#include "MyGrabCutEnv.h"

#include "grabcut/MyGrabCut.h"

#include <opencv2/core/core.hpp>
#include <opencv2/highgui/highgui.hpp>
#include "opencv2/highgui/highgui_c.h"
#include <opencv2/imgproc/imgproc.hpp>
#include "opencv2/imgproc/types_c.h"
#include "CuboidUtil.h"
#include <iostream>
#include "cuboid_approx/util/AMUtil2.h"
#include "IGMM.h"

using namespace std;

namespace am
{
    // init static fields
    const std::string MyGrabCutEnv::win_name_rgb = "annotate rgb";
    const std::string MyGrabCutEnv::win_name_dep = "annotate dep";

    inline cv::Point AsPoint2( cv::Point3i const& pnt )
    {
        return cv::Point( pnt.x, pnt.y );
    }

    // global, tmp structure used by annotate, to store mouse clicks before returning them
    struct
    {
            t_snakes            pnts;
            std::vector<uchar>  labels_of_pnts;
            uchar               label_current    = 0;
            uchar               max_filled_label = 0;
            bool                changed          = false;

            void clear()
            {
                pnts          .clear();
                labels_of_pnts.clear();
                label_current       = 0;
                max_filled_label    = 0;
                changed             = false;
            }

            void new_snake( bool new_label = false )
            {
                pnts.push_back( std::vector<cv::Point3i>() );

                if ( !labels_of_pnts.size() ) // if no previous label
                {
                    labels_of_pnts.push_back( label_current );
                }
                else if ( new_label )         // if new snake needs new label
                {
                    labels_of_pnts.push_back( ++label_current );
                }
                else                            // if new snake needs no new label
                {
                    labels_of_pnts.push_back( label_current );
                }
            }

            void inc_label()
            {
                if ( label_current <= max_filled_label )
                    ++label_current;
            }

            void dec_label()
            {
                if ( label_current > 0 )
                    --label_current;
            }

            void printState()
            {
                std::cout << "Snakes: \n";
                for ( int i = 0; i < pnts.size(); ++i )
                {
                    std::cout << "\tSnake with " << pnts[i].size() << " and label " << (int)labels_of_pnts[i] << std::endl;
                }
                std::cout << std::endl;
            }

            cv::Scalar get_label_colour( int label )
            {
                return cv::Scalar( ((label+1) * 32) % 255,
                                   ((label+5) * 48) % 255,
                                   ((label-8) * 24) % 255 );
            }
    } ctx;

    void mouseEvent( int evt, int x, int y, int flags, void* param )
    {
        using namespace cv;

        Mat* rgb = (Mat*) param;
        if (    ( (evt == CV_EVENT_LBUTTONDOWN)                                   )
             || ( (evt == CV_EVENT_MOUSEMOVE  ) && (flags & CV_EVENT_FLAG_LBUTTON)) )
        {
            // if right button: start new snake
            if ( (evt == CV_EVENT_LBUTTONDOWN) )
            {
                ctx.new_snake( false );
            }

            // debug print
            /*printf("%d %d: %d, %d, %d\n",
                   x, y,
                   (int)(*rgb).at<Vec3b>(y, x)[0],
                   (int)(*rgb).at<Vec3b>(y, x)[1],
                   (int)(*rgb).at<Vec3b>(y, x)[2] );
            fflush(stdout);*/

            // init
            if ( !ctx.pnts.size() )
                ctx.new_snake( true );

            if ( ctx.max_filled_label < ctx.label_current )
            {
                ctx.max_filled_label = ctx.label_current;
            }
            // add
            ctx.pnts.back().push_back( Point3i(x,y,ctx.label_current) );
            // notify
            ctx.changed = true;
        }
        else if ( evt == CV_EVENT_RBUTTONDOWN )
        {
            if ( !(flags & CV_EVENT_FLAG_SHIFTKEY) )
            {
                ctx.inc_label();
            }
            if ( flags & CV_EVENT_FLAG_SHIFTKEY )
            {
                ctx.dec_label();
            }
            std::cout << "painting with label: " << (int)ctx.label_current << std::endl;
            //ctx.printState();
        }
        else if ( evt == CV_EVENT_LBUTTONUP )
        {
            ctx.printState();
        }
    }

    int annotate( t_snakes &points, cv::Mat &mask )
    {
        // [391, 210; 510, 106]
        points.clear();
        points.push_back( std::vector<cv::Point3i>() );
        points.back().push_back( cv::Point3i(391, 210,cv::GC_PR_FGD) );
        points.back().push_back( cv::Point3i(510, 106,cv::GC_PR_FGD) );
        // [230, 371; 232, 409; 280, 409; 342, 409; 395, 389; 517, 377; 548, 376; 549, 333; 501, 302; 452, 311; 387, 316; 347, 343; 274, 359; 254, 371; 335, 382; 376, 350; 376, 361; 429, 338; 464, 365; 490, 343; 516, 349]

        for ( auto const& snake : points )
            for ( int i = 1; i < snake.size(); ++i )
            {
                std::cout << "drawing line " << snake[i-1] << "->" << snake[i] << std::endl;
                // draw on mask
                cv::line( mask, AsPoint2(snake[i-1]), AsPoint2(snake[i]), cv::Scalar(am::Labels::probably_label(1)), 8 );
                cv::line( mask,
                          cv::Point( snake[i-1].y, snake[i-1].x ),
                          cv::Point( snake[i  ].y, snake[i  ].x ),
                          cv::Scalar(cv::GC_PR_FGD),
                        8 );
            }
        cv::imshow( "mask", mask * 32 );
        cv::waitKey();
    }

    int
    MyGrabCutEnv::_annotate( t_snakes &points, cv::Mat &mask, cv::Mat const& img, cv::Mat const& dep )
    {
        using namespace cv;

        cv::Mat tmp( img.clone() );
        cv::Mat tmp_dep( dep.clone() );

        // prepare global context
        ctx.clear();

        mask = cv::Mat::ones( img.size(), CV_8UC1 ) * GC_PR_BGD;

        // create windows
        cv::destroyWindow( win_name_rgb );
        cv::namedWindow  ( win_name_rgb );
        cv::moveWindow   ( win_name_rgb, 10, 10 );
        cvSetMouseCallback( win_name_rgb.c_str(), mouseEvent, &tmp );

        if ( !tmp_dep.empty() )
        {
            namedWindow       ( win_name_dep );
            cv::moveWindow    ( win_name_dep, 10, 530 );
            cvSetMouseCallback( win_name_dep.c_str(), mouseEvent, &tmp_dep );
        }

        char key = 0;
        while ( !(   ((int)key == 27)
                  || ((int)key == 10) ) )
        {
            if ( ctx.changed )
            {
                if ( ctx.pnts.back().size() > 1 )
                {
                    // draw lines
                    int label = ctx.labels_of_pnts.back();
                    cv::Scalar img_color = ctx.get_label_colour( label );
                    // draw on image
                    cv::line( tmp,
                              AsPoint2( ctx.pnts.back().back() ),
                              AsPoint2( ctx.pnts.back()[ctx.pnts.back().size()-2] ),
                            img_color, 4 );
                    // draw on depth image
                    if ( !tmp_dep.empty() )
                        cv::line( tmp_dep,
                                  AsPoint2( ctx.pnts.back().back() ),
                                  AsPoint2( ctx.pnts.back()[ctx.pnts.back().size()-2] ),
                                img_color, 4 );

                    // draw on mask
                    cv::line( mask,
                              AsPoint2( ctx.pnts.back().back() ),
                              AsPoint2( ctx.pnts.back()[ctx.pnts.back().size()-2] ),
                              Scalar(ctx.label_current), 4 );
                }

                // draw circles
                //if ( ctx.pnts.back().size() > 0 )
                //{
                //    cv::circle( tmp, ctx.pnts.back().back(), 2, Scalar(255,0,200), 2 );
                //    if ( !tmp_dep.empty() ) cv::circle( tmp_dep, ctx.pnts.back().back(), 2, Scalar(255,0,200), 2 );
                //}

                // update model
                ctx.changed = false;
            }

            // show
            imshow( win_name_rgb.c_str(), tmp );
            if ( !tmp_dep.empty() ) imshow( win_name_dep.c_str(), tmp_dep );

            key = waitKey(10);
        }

        points = ctx.pnts;
        return key;
    }

    // utils
    void bbox( cv::Rect &rect, t_snakes const& snakes )
    {
        using namespace cv;

        Point tl( INT_MAX, INT_MAX );
        Point br( 0, 0 );
        for ( auto const& snake : snakes )
        {
            for ( auto const& pnt : snake )
            {
                if ( pnt.x < tl.x ) tl.x = pnt.x;
                if ( pnt.x > br.x ) br.x = pnt.x;
                if ( pnt.y < tl.y ) tl.y = pnt.y;
                if ( pnt.y > br.y ) br.y = pnt.y;
            }
        }
        std::cout << "tl: " << tl.x << "," << tl.y << "\t br: " << br.x << "," << br.y << std::endl;
        rect = cv::Rect( tl.x, tl.y, br.x-tl.x, br.y-tl.y );
    }

    void MyGrabCutEnv::showMask( const cv::Mat & mask, cv::Mat const& img, std::string const title, int x, int y )
    {
        if ( mask.empty() )
        {
            std::cerr << "MyGrabCutEnv::showMask(): mask is empty, exiting" << std::endl;
            return;
        }


        // Generate output img
        cv::Mat foreground( img.size(),CV_8UC3,cv::Scalar(64,64,64));

        // Get the pixels marked as likely foreground
        cv::Mat local_mask = mask.clone();

        cv::compare( mask, cv::GC_PR_FGD, local_mask, cv::CMP_EQ );
        img.copyTo( foreground, local_mask ); // bg pixels not copied
        cv::compare( mask, cv::GC_FGD, local_mask, cv::CMP_EQ );
        img.copyTo( foreground, local_mask ); // bg pixels not copied

        // display result
        cv::namedWindow( title );
        cv::imshow( title, foreground );
        if ( x > -1 && y > -1 )
        {
            cv::moveWindow( title, x, y );
        }
    }

    // logic
    int MyGrabCutEnv::loadAndAnnotate( t_snakes &snakes, cv::Mat &mask, cv::Mat &img, cv::Mat &dep,
                                       const std::string &path, const std::string &depth_path,
                                       cv::Mat *p_dep8, bool cheat )
    {
        // read RGB24
        {
            img = cv::imread( path, cv::IMREAD_UNCHANGED );
            cout << img.cols << " " << img.rows << " " << img.channels() << endl;
            if ( img.channels() < 3 )
            {
                cv::Mat tmp = cv::Mat::zeros( img.size(), img.type() );
                cv::merge( std::vector<cv::Mat>({img,img,img}), tmp );
                tmp.copyTo( img );
            }
            else if ( img.channels() > 3 )
            {
                cv::Mat tmp;
                cv::cvtColor( img, tmp, CV_RGBA2RGB );
                tmp.copyTo( img );
            }
        }

        // read DEP
        cv::Mat dep8;
        {
            dep = am::CuboidUtil::loadDepth( depth_path );

            std::cout << "MyGrabCutEnv:run with \n\t"
                      << path << "(" << img.cols << "x" << img.rows << "@" << img.channels() << ")"
                      << " and \n\t"
                      << depth_path << "(" << dep.cols << "x" << dep.rows << "@" << dep.channels() << "," << dep.depth() << ")" << std::endl;

            if ( dep.channels() > 1 )
            {
                cv::Mat tmp;
                cv::cvtColor( dep, tmp, CV_RGB2GRAY );
                dep = tmp;
            }

            double minVal, maxVal;
            cv::minMaxLoc( dep, &minVal, &maxVal );
            std::cout << "depminmax: " << minVal << "," << maxVal << std::endl;
            if ( dep.type() != CV_8UC1 ) { dep.convertTo( dep8, CV_8UC1, 255.f/maxVal ); }
            else                           dep.copyTo( dep8 );

            if ( p_dep8 ) dep8.copyTo( *p_dep8 );
        }

        // annotate image
        {
            if ( cheat )
            {
                //mask = cv::Mat::ones( img.size(), CV_8UC1 ) * Labels::probably_background();
                mask = cv::Mat::ones( img.size(), CV_8UC1 ) * cv::GC_PR_BGD;
                annotate( snakes, mask );
                return 10;
            }
            else
            {
                int key = _annotate( snakes, mask, img, dep8 );
                if (    (key == 27)
                     || (key == 10) )
                    return key;
            }
        }
    }

    int MyGrabCutEnv::run( GrabcutResult &result, string const& path, string const depth_path, bool cheat, Params *p_params )
    {
        using namespace cv;
        using namespace std;

        int key = 0;
        key = MyGrabCutEnv::loadAndAnnotate( m_snakes, m_mask, m_img, m_dep, path, depth_path, &m_dep8, cheat );

        if ( key != 27 )
        {
            t_annotated_input input;
            input.snakes = m_snakes;
            input.mask   = m_mask;
            input.img    = m_img;
            input.dep    = m_dep;
            input.dep8   = m_dep8;
            if ( p_params )
            {
                input.betas.resize(2);
                input.betas[0] = p_params->getDouble( "betas[0]" );
                input.betas[1] = p_params->getDouble( "betas[1]" );
            }

            this->run( result, input );
        }

        return (key == 27) ? EXIT_FAILURE : EXIT_SUCCESS;
    }

    int MyGrabCutEnv::run( GrabcutResult &result, cv::Mat const& img, cv::Mat const& dep, bool cheat, Params *p_params )
    {
        using namespace cv;
        using namespace std;
        CV_Assert( img.channels() == 3 );
        CV_Assert( dep.channels() == 1 );
        m_img = img.clone();
        m_dep = dep.clone();

        // cast to dep8 for display and annotation
        cv::Mat dep8;
        {
            double minVal, maxVal;
            cv::minMaxLoc( dep, &minVal, &maxVal );
            if ( dep.type() != CV_8UC1 ) { dep.convertTo( dep8, CV_8UC1, 255.f/maxVal ); }
            else                           dep.copyTo( dep8 );
        }

        int key = 0;
        //key = MyGrabCutEnv::loadAndAnnotate( m_snakes, m_mask, m_img, m_dep, path, depth_path, &m_dep8, cheat );
        key = _annotate( m_snakes, m_mask, img, dep8 );

        if ( key != 27 )
        {
            t_annotated_input input;
            input.snakes = m_snakes;
            input.mask   = m_mask;
            input.img    = m_img;
            input.dep    = m_dep;
            input.dep8   = m_dep8;
            if ( p_params )
            {
                input.betas.resize(2);
                input.betas[0] = p_params->getDouble( "betas[0]" );
                input.betas[1] = p_params->getDouble( "betas[1]" );
            }

            this->run( result, input );
        }

        return (key == 27) ? EXIT_FAILURE : EXIT_SUCCESS;
    }

    void MyGrabCutEnv::run( GrabcutResult &result, t_annotated_input const& input )
    {
        std::cerr << "Cannot call run of MyGrabCut class..." << std::endl;
        //run( result, input.snakes, input.mask, input.img, input.dep, input.dep8 );
    }

    //void MyGrabCutEnv::run( GrabcutResult &result, t_snakes const& snakes, cv::Mat const& mask, cv::Mat const& img, cv::Mat const& dep, cv::Mat const& dep8 )
    //{
    //    std::cerr << "Cannot call run of MyGrabCut class..." << std::endl;
    //}

    int
    MyGrabCutEnv::destroyWindows()
    {
        cv::destroyWindow( win_name_rgb );
        cv::destroyWindow( win_name_dep );
    }

    /// SET, GET
    t_annotated_input MyGrabCutEnv::getAnnotatedInput()
    {
        t_annotated_input out;
        out.dep     = this->getDep();
        out.dep8    = this->getDep8();
        out.img     = this->getImg();
        out.mask    = this->getMask();
        out.snakes  = this->getSnakes();

        return out;
    }

    t_snakes const& MyGrabCutEnv::getSnakes()
    {
        return m_snakes;
    }

    cv::Mat const& MyGrabCutEnv::getMask()
    {
        return m_mask;
    }

    cv::Mat const& MyGrabCutEnv::getImg()
    {
        return m_img;
    }

    cv::Mat const& MyGrabCutEnv::getDep()
    {
        return m_dep;
    }

    cv::Mat const& MyGrabCutEnv::getDep8()
    {
        return m_dep8;
    }

} // ns am
