#ifndef __CA_ANNOTATIONCONTROLLER_H__
#define __CA_ANNOTATIONCONTROLLER_H__

#include <map>
#include <string>
#include <iostream>
#include <memory>

#include "vtkRendererCollection.h"
#include "vtkSmartPointer.h"
#include "pcl/visualization/pcl_visualizer.h"
#include <opencv2/core.hpp>

#include <QObject>

#include "cuboid_approx/common.h"
#include "Params.h"
#include "cuboid_approx/util/MyIntrinsics.h"
#include "cuboid_approx/ui/Annot3DInteractorStyle.h"

#include "cuboid_approx/primitives/Contact.h"
#include "cuboid_approx/ui/DisplayedPrimitive.h"
#include "cuboid_approx/AnnotationState.h"

class QVTKWidget;

namespace am
{
    class Plane;
    class Cuboid;
    class Primitive;
    //class Contact;

    class Cuboid;
    class Cuboids;

    class AnnotationController : public QObject
    {
            Q_OBJECT
        // SIGNALs, SLOTs
        public Q_SLOTS:
            void pointPickedSlot( int point_id );
        Q_SIGNALS:
            void pointPickedSignal( int point_id );

        public:
            typedef float Scalar;

            /// STATIC METHODS

            /// CONSTR
            AnnotationController();
            int initViewer(); // some of the view stuff can't be done at construction time

            /// METHODS
            // events
            int
            annotate_button_clicked( Params const& params );
            int
            loadNewCloudFromImagePaths(  std::string const img_path
                                       , std::string const depth_path
                                       , float spatial_sigma    = 2.f
                                       , float range_sigma      = 1.f
                                       , int kernel_radius      = 2
                                       , int iteration_count    = 1 );
            static int
            loadNewCloudFromImagePaths( PCloud::Ptr                 cloud
                                        , std::string        const  img_path
                                        , std::string        const  depth_path
                                        , MyIntrinsics       const &depth_intrinsics
                                        , float              const  spatial_sigma   = 2.f
                                        , float              const  range_sigma     = 1.f
                                        , int                const  kernel_radius   = 2
                                        , int                const  iteration_count = 1
                                        , cv::Mat                  *p_img           = nullptr
                                        , cv::Mat                  *p_dep           = nullptr
                                        , PCloud::Ptr              *p_aux_cloud     = nullptr
                                        );



            int
            removePrimitive( std::vector<int> id );
            int
            selectPrimitive(std::vector<int> ids );
            int
            toggleHidePrimitive( int id );
            int
            setPrimitiveIsFloor( int id, bool is_floor, bool is_selected = false );
            bool
            getPrimitiveIsFloor( int id );
            int
            setPrimitiveIsWall(int id, bool is_wall , bool is_selected = false );
            bool
            getPrimitiveIsWall( int id );
            int
            setPrimitiveIsTarget( int id, bool is_target );
            int
            setPrimitiveIsSphere( int id, bool is_sphere, bool is_selected = false );
            int
            setPrimitiveIsCylinder( int id, bool is_cylinder, bool is_selected = false );

            std::vector<Contact>
            getLatestContacts() const;

            int
            redrawPrimitive( int id, bool is_selected );
            int
            redrawAllPrimitives();

            ca::DisplayedPrimitive::Map const&
            getPrimitives();
            ca::DisplayedPrimitive const&
            getPrimitive( int id );
            int
            fillPrimitive( int id );

            std::string
            getRotInfo();

            int
            save(std::string directory, Params const& params);
            int
            saveCuboidsToPLY( std::string path );
            int
            savePointsToPLY( std::string path ) const;

            int
            loadCuboids( std::string path, Params const* params, std::vector<am::Contact> *contacts_arg = NULL );

            static int
            loadRegions( std::vector< pcl::PointIndices::Ptr >      & region_indices
                         , std::string                         const& path
                         , Params                              const& params
                         , PCloud::Ptr                                cloud );
            int
            loadRegions( std::string const& path, Params const& params);

            int
            region_growing();

            inline void
            _mouseCallbacks( const pcl::visualization::MouseEvent event );

            int
            resetCamera();
            int
            centerCamera();

            int
            saveCameraPose( std::string path );


            int
            showAuxCloud( int id, bool doShow );

            int
            removeAllPoints();
            int
            addModel( std::string path );

            std::pair<int,std::string>
            rotateCamera( bool rotate, QVTKWidget *display );

            ///
            /// \brief segment
            /// \param min_cluster_size
            /// \param clusterTolerance
            /// \param voxelgrid_resolution
            /// \param slab_size_threshold_percentage
            /// \return
            /// \dependency _state._current_indices_ptr So, hidden objects are excluded
            /// \dependency _state._objects             Requires at least one floor, floors and walls are excluded locally
            /// \sideeffect _state._objects             Cuboids are added to the current objects
            ///
            int
            segment( int min_cluster_size                    = 5000
                     , double clusterTolerance               = 0.05
                     , float  voxelgrid_resolution           = .125f
                     , float  slab_size_threshold_percentage = .01f
                     , int    max_slab_count                 = 25
                     , Eigen::Vector3f const* up = NULL );

            int
            cluster( int min_cluster_size
                     , double cluster_tolerance
                     );
            int
            drawIntersections( bool                                     do_show
                               , std::vector<Contact>            const* contacts = nullptr
                               , int                                    viewport = 0
                               , pcl::visualization::PCLVisualizer::Ptr viewer   = nullptr );

            int
            saveScreenshot( std::string path );

            /// SET, GET
            vtkSmartPointer<vtkRenderWindow>       getRenderWindow();
            vtkSmartPointer<vtkRendererCollection> getRendererCollection();

            std::string
            getOutputPath();

            int
            setPointSize( double size );

            int
            setColourCodeCuboids( bool checked );
            int
            setColourEdges( bool checked );

            int
            setCuboidColour( int id, Eigen::Vector4f colour );
            Eigen::Vector4f
            getCuboidColour( int id );
            int
            setCuboidOpacity( int id, double opacity );

            int
            setShowCuboidIDs( bool show );

            std::vector<Eigen::Vector3f>
            getCameraPosFocalUp();
            int
            setCameraPosFocalUp( std::vector<Eigen::Vector3f> pos_focal_up );

            int
            setSphereSize( double value );

            template <typename ScalarTypeInFile, class CloudPtrT> static inline int
            readAdditionalPoints( CloudPtrT &cloud, std::string path );

            template <class CloudPtrT> inline int
            addPoints( CloudPtrT cloud );

            std::vector< std::shared_ptr<am::Cuboid> >
            getFloors();
            std::vector< std::shared_ptr<am::Cuboid> >
            getRotationTargets();
            pcl::visualization::PCLVisualizer::Ptr  getViewer();

            Eigen::Matrix4d getProjMatrix();
            Eigen::Matrix4d getViewMatrix();

            int
            mapCurrentCloud();

            int
            createCuboid( am::Params const* params = NULL );

            int
            clearCurrentPick();

            enum COLOUR_CLOUD_MODE { NONE, DISTANCE_BASED };
            int
            colourCloud( COLOUR_CLOUD_MODE mode, void* param = NULL );

            // public constants
            static const Eigen::Vector4f CUBOID_COLOUR;
            static const Eigen::Vector4f CUBOID_FLOOR_COLOUR;
            static const Eigen::Vector4f CUBOID_WALL_COLOUR;
            static const double          CUBOID_SELECTED_LINE_WIDTH;
            static const double          CUBOID_UNSELECTED_LINE_WIDTH;
            static const float           INTERSECT_OFFSET;

            int
            addCuboid( am::Cuboid const& cuboid, std::string name );

        protected:
            int
            _addCuboids( Cuboids const& cuboids, std::string tag = "" );
            int
            _addCuboids(am::Cuboids const& cuboids, std::vector<std::string> tags );
            int
            _addCuboid(am::Cuboid const& cuboid , int cube_id, std::string tag = "" );
            int
            _removePoints( pcl::PointIndices::ConstPtr indices_ptr );
            int
            _addPoints( pcl::PointIndices::ConstPtr indices_ptr );

            std::vector<am::Contact>
            _calculateIntersections();

            static Eigen::Matrix4d
            _getProjMatrix( pcl::visualization::PCLVisualizer::Ptr viewer, int viewport = 0 );
            static Eigen::Matrix4d
            _getViewMatrix( pcl::visualization::PCLVisualizer::Ptr viewer, int viewport = 0 );

            // FIELDS

            MyIntrinsics                                _depth_intrinsics;

            pcl::visualization::PCLVisualizer::Ptr      _viewer;
            vtkSmartPointer<ca::Annot3DInteractorStyle> _interactor_style;
            enum MOUSE_STATE { RIGHT_DOWN = 2 };
            int                                         _mouse_state;

            ca::AnnotationState                         _state;

            std::vector<cv::Point3f>                _colors;
            int                                     _colorHead; // what's the next color to use from colors
            const std::string                       _output_path;

            cv::Mat                                 _img, _dep; // stores the last loaded images

            bool                                    _colour_code_cuboids;
            float                                   _default_opacity;


    };

} // ns am

#ifndef __CA_ANNOTATIONCONTROLLER_HPP__
#define __CA_ANNOTATIONCONTROLLER_HPP__

#include "CloudViewer.h"

template <typename ScalarTypeInFile, class CloudPtrT> inline int
am::AnnotationController::readAdditionalPoints( CloudPtrT &cloud, std::string path )
{
    typedef typename CloudPtrT::element_type CloudT;
    typedef typename CloudT::PointType       PointT;

    int err = EXIT_SUCCESS;

    // init cloud
    if ( !cloud )   cloud.reset( new CloudT );
    cloud->clear();

    typedef struct tag_float3
    {
        ScalarTypeInFile _v[3];
    } float3;

    FILE *fp = fopen( path.c_str(), "rb" );
    if (!fp)
    {
        std::cerr << __func__ << ": could not open " << path << ", exiting..." << std::endl;
        return EXIT_FAILURE;
    }

    if ( path.find("pts2") != std::string::npos )
    {
        ScalarTypeInFile up[3];
        int read_count = fread( up, sizeof(ScalarTypeInFile), 3, fp );
        if ( read_count != 3 ) { std::cerr << "fread error " << read_count << "/3 read" << std::endl; err += 1; }
        std::cout << "up: " << up[0] << ", " << up[1] << ", " << up[2] << std::endl;
    }

    int n_vertices;
    int read_count = fread( &n_vertices, sizeof(int), 1, fp );
    if ( n_vertices <= 0 || (read_count != 1) )
    {
        std::cerr << __func__ << ": no points in file, count is zero, or read_count " << read_count << "!= 1..." << std::endl;
        return EXIT_FAILURE;
    }

    int n_fields = 9;
    if ( path.find("pts2") != std::string::npos )   n_fields = 3;

    float3 *vertices = NULL, *normals = NULL, *colours = NULL;

    if ( n_vertices > 0 )
    {
        vertices = new float3[n_vertices];
        read_count = fread( vertices, sizeof(float3), n_vertices, fp );
        if (read_count != n_vertices) { std::cerr << "fread vertices error: " << read_count << "/" << n_vertices << " read..." << std::endl; err += 2; return EXIT_FAILURE; }

        if ( n_fields > 3 )
        {
            normals = new float3[n_vertices];
            read_count = fread( normals, sizeof(float3), n_vertices, fp );
            if (read_count != n_vertices) { std::cerr << "fread normals error: " << read_count << "/" << n_vertices << " read..." << std::endl; err += 4; }
        }

        if ( n_fields > 6 )
        {
            colours = new float3[n_vertices];
            read_count = fread( colours, sizeof(float3), n_vertices, fp );
            if (read_count != n_vertices) { std::cerr << "fread colours error: " << read_count << "/" << n_vertices << " read..." << std::endl; err += 8; }
        }
    }

    // scale to be 0..10
    float max_val = 0.f, min_val = FLT_MAX;
    for ( int i = 0; i < n_vertices; ++i )
    {
        /*for ( int d = 0; d != 3; ++d )*/
        if ( vertices[i]._v[2] > max_val )     max_val = vertices[i]._v[2];
        if ( vertices[i]._v[2] < min_val )     min_val = vertices[i]._v[2];
//        if ( max_val > 1000.f )         break;
    }

    float                           scale = 1.f;
    if      ( max_val > 1000.f  )   scale =  .001f;
    else if ( max_val > 100.f   )   scale =  .01f;
    else if ( max_val > 10.f    )   scale =  .1f;

    // debug
    if ( scale != 1.f )     std::cout << "max_val: " << max_val << " so multiplying every point by " << scale
                                      << ", min_val: " <<  min_val
                                      << std::endl;

    // copy to cloud
    cloud->reserve( n_vertices );
    for ( int i = 0; i < n_vertices; ++i )
    {
        PointT pnt;
        pnt.x = vertices[i]._v[0] * scale;
        pnt.y = vertices[i]._v[1] * scale;
        pnt.z = vertices[i]._v[2] * scale;

        uint32_t rgb;
        if ( colours )
        {
            rgb = (uint32_t)(colours[i]._v[0] * 255.f) << 16 | (uint32_t)(colours[i]._v[1] * 255.f) << 8 | (uint32_t)(colours[i]._v[2]*255.f);
            pnt.rgb = *reinterpret_cast<float*>(&rgb);
        }
        else
        {
            rgb = 0;
            pnt.rgb = *reinterpret_cast<float*>(&rgb);
        }

        cloud->push_back( pnt );
    }

    if ( vertices ) { delete [] vertices; vertices = NULL; }
    if ( normals  ) { delete [] normals;  normals  = NULL; }
    if ( colours  ) { delete [] colours;  colours  = NULL; }

    return err;
}

template <class CloudPtrT> inline int
am::AnnotationController::addPoints( CloudPtrT cloud )
{
    typedef typename CloudPtrT::element_type CloudT;

    int err = EXIT_SUCCESS;

    if ( !_state.     MainCloud() ) _state.     MainCloudRef(ca::AnnotationState::FULL).reset( new CloudT()             );
    if ( !_state.  CurrentCloud() ) _state.  CurrentCloudRef(ca::AnnotationState::FULL).reset( new CloudT()             );
    if ( !_state.CurrentIndices() ) _state.CurrentIndicesRef(                         ).reset( new pcl::PointIndices()  );

    // reserve sizes
    const int fullN = _state.MainCloud()->size() + cloud->size();
    _state.     MainCloudRef(ca::AnnotationState::NONE)->        reserve( fullN );
    _state.  CurrentCloudRef(ca::AnnotationState::NONE)->        reserve( fullN );
    _state.CurrentIndicesRef(                         )->indices.reserve( fullN );

    // copy points to main cloud
    for ( size_t pid = 0; pid != cloud->size(); ++pid )
    {
        _state.     MainCloudRef(ca::AnnotationState::NONE)->        push_back( cloud->at(pid) );
        _state.  CurrentCloudRef(ca::AnnotationState::NONE)->        push_back( cloud->at(pid) );
        _state.CurrentIndicesRef(                         )->indices.push_back( _state.CurrentCloud()->size() - 1 );
    }

    if ( _viewer->getCloudActorMap()->find("big_cloud") != _viewer->getCloudActorMap()->end() )
    {
        if ( !am::CloudViewer::updatePointCloud( _viewer, _state.CurrentCloud(), "big_cloud", _state.CurrentIndices() ) )
            err += (am::CloudViewer::addPointCloud( _viewer, _state.CurrentCloud(), "big_cloud", _state.CurrentIndices() ) ? 0 : 1);
    }
    else
    {
        err += (am::CloudViewer::addPointCloud( _viewer, _state.CurrentCloud(), "big_cloud", _state.CurrentIndices() ) ? 0 : 1);
    }

    return err;
}

#endif // __CA_ANNOTATIONCONTROLLER_HPP__



#endif // __CA_ANNOTATIONCONTROLLER_H__
