#ifndef __CA_NYUINIT_H__
#define __CA_NYUINIT_H__

#include <pcl/point_types.h> // pcl::PointXYZRGBA -> MyPointT

namespace ca
{
    typedef pcl::PointXYZRGBL MyPointT;

    class NYUInit
    {
        public:
            template <typename Scalar, typename PointT> inline static int
            run( int argc, char** argv );

            template <typename Scalar, typename PointT> inline static int
            filterContacts( int argc, char** argv );
    };
} // ... ns ca

// HPP
#include <pcl/console/parse.h>
#include "primitives/Cuboids.h"
#include "CloudViewer.h"
#include "AnnotationController.h"
#include "cuboid_approx/io/walls_io.h"
#include "CuboidUtil.h"
#include "cuboid_approx/proc3d/ContactsFiltering.h"

namespace ca
{
    template <typename Scalar, typename PointT> int
    NYUInit::filterContacts( int argc, char** argv )
    {
        using namespace am;

        std::string                             cuboids_path;
        std::string                             walls_path;
        std::string                             contacts_path;
        Cuboids                                 cuboids;
        am::Cuboid::TypeMap                     walls;
        am::Contact::Vec                        contacts, gt_contacts;
        am::Contact::Vec                        filtered_contacts;
        boost::shared_ptr<CloudViewer>          vp                  = NULL;
        bool                                    verbose             = false;
        int err = EXIT_SUCCESS;

        // READ
        {
            // parse
            if ( EXIT_SUCCESS == err )
            {
                if ( pcl::console::find_switch(argc,argv,"-v") || pcl::console::find_switch(argc,argv,"--verbose") )
                    verbose = true;

                pcl::console::parse_argument( argc, argv, "--cuboids" ,  cuboids_path );

                if ( verbose )
                    std::cout << "run with paths: "
                              << "\n\t" << cuboids_path
                              << std::endl;
                if ( cuboids_path.empty() )
                {
                    std::cerr << "[" << __func__ << "]: " << "empty cuboids_path, exiting" << std::endl;
                    err = EXIT_FAILURE;
                }
            } // ... parse

            // read boxes
            if ( EXIT_SUCCESS == err )
            {
                err = am::Cuboids::readFromFile( /*   [out]               cuboids: */ cuboids
                                                 , /* [in]                   path: */ cuboids_path
                                                 , /* [in]           inliers_path: */ NULL          // only used for clouds from images
                                                 , /* [in]                  cloud: */ NULL          // no cloud
                                                 , /* [in]                indices: */ NULL          // no cloud
                                                 , /* [out]           GT_contacts: */ &gt_contacts  // need to save it
                                                 , /* [out]            GT_statics: */ NULL          // returned in cuboids anyway
                                                 , /* [in] plane inlier_threshold: */ NULL          // no cloud
                                                 , /* [in]                verbose: */ false
                                                 );
                if ( !cuboids.size() )
                {
                    std::cerr << "[" << __func__ << "]: " << "0 cuboids, exiting" << std::endl;
                    err = EXIT_FAILURE;
                }

                if ( (EXIT_SUCCESS == err) && gt_contacts.size() )
                {
                    err = am::util::saveGTContacts( gt_contacts, cuboids_path );
                    if ( EXIT_SUCCESS != err ) std::cout << "[" << __func__ << "]: " << "did NOT save " << gt_contacts.size() << " gt_contacts..." << std::endl;
                    else                       std::cout << "[" << __func__ << "]: " << "did save " << gt_contacts.size() << " gt_contacts..." << std::endl;
                }

                if ( verbose )  std::cout << "[" << __func__ << "]: " << "finished read of " << cuboids.size() << " cuboids" << std::endl; fflush(stdout);
            } // ... read boxes

            // read walls
            if ( EXIT_SUCCESS == err )
            {
                if ( pcl::console::parse_argument(argc,argv,"--walls",walls_path) <= 0 )
                {
                    // guess
                    boost::filesystem::path cuboids_boost_path( cuboids_path );

                    std::vector<std::string> wall_files;
                    am::util::get_files_by_ext( /*    out: */ wall_files
                                                , /* dirs: */ boost::filesystem::path(cuboids_boost_path.parent_path())
                                                , /*  ext: */ ".walls" );
                    if ( !wall_files.size() )
                    {
                        std::cerr << "[" << __func__ << "]: " << "tried to guess wall file, but no walls...exiting\n"; fflush(stderr);
                        err = EXIT_FAILURE;
                    }
                    else if ( wall_files.size() == 1 )
                    {
                        std::cout << "setting walls path to " << wall_files[0] << std::endl;
                        walls_path = wall_files[0];
                    }
                    else
                    {
                        std::string suggested = am::util::cuboidPathToWallsPath( cuboids_path );
                        if ( boost::filesystem::exists( suggested ) )
                        {
                               walls_path = suggested;
                        }
                        else
                        {
                            std::cerr << "[" << __func__ << "]: " << "many wall files, could not guess which to use, exiting \n";
                            err = EXIT_FAILURE;
                        }
                    }
                }

                if ( (EXIT_SUCCESS == err) && !walls_path.empty() )
                {
                    const int cuboids_size = cuboids.size();
                    err = ca::io::readWalls( walls, walls_path, &cuboids_size );
                }
                else
                    std::cout << "did not read walls, since path: " << walls_path << std::endl;

                std::cout << "[" << __func__ << "]: " << "read " << walls.size() << " walls\n";
            } // ... read walls

            // read contacts
            if ( EXIT_SUCCESS == err )
            {
                if ( pcl::console::parse_argument( argc, argv, "--contacts", contacts_path ) <= 0 )
                {
                    std::cerr << "no --contacts argument provided" << std::endl;
                    err = EXIT_FAILURE;
                } // ... if parse

                if ( EXIT_SUCCESS == err )
                {
                    int max_proxy_id = 0;

                    err = am::Contact::read( contacts, contacts_path, &max_proxy_id );
                    if (  max_proxy_id >= cuboids.size() )
                    {
                        std::cerr << "[" << __func__ << "]: " << "Warning: More proxies in contacts (" << max_proxy_id << "), than cuboids (" << cuboids.size()-1 << ")...filtering, continuing\n"; fflush(stderr);

                        int orig_size = contacts.size();
                        am::Contact::filterByMaxCuboidID( contacts, contacts, cuboids.size() - 1 );
                        std::cout << "[" << __func__ << "]: " << "removed " << orig_size - int(contacts.size()) << " contacts." << std::endl;
                    }
                }

                std::cout << "[" << __func__ << "]: " << "read " << contacts.size() << " contacts\n";
            } // ... read contacts
        } // ...READ

        // filter
        if ( EXIT_SUCCESS == err )
        {
            err = ca::ContactsFiltering::filterContacts( filtered_contacts
                                                         , contacts
                                                         , cuboids
                                                         , walls );
            if ( EXIT_SUCCESS != err )
            {
                std::cerr << "[" << __func__ << "]: " << "filterContacts returned error " << err << std::endl; fflush(stderr);
            }
            else
            {
                boost::filesystem::path filtered_boost_path( contacts_path );
                std::string             filtered_path( filtered_boost_path.parent_path().string() + "/" + filtered_boost_path.stem().string() + ".filtered.txt" );

                if ( boost::filesystem::exists(filtered_path) )
                {
                    if ( boost::filesystem::exists(filtered_path + ".bak") )
                        boost::filesystem::remove( filtered_path + ".bak" );
                    boost::filesystem::rename( filtered_path, filtered_path + ".bak" );
                }

                err = am::Contact::write( filtered_contacts, filtered_path );
                if ( EXIT_SUCCESS != err )  std::cerr << "[" << __func__ << "]: " << "could not write to " << filtered_boost_path << std::endl;
                else                        std::cout << "[" << __func__ << "]: " << "wrote " << filtered_contacts.size() << " contacts to " << filtered_boost_path << std::endl;
            } // ... if filter success
        } // ... filter

        return err;
    } // ... function filterContacts

    template <typename Scalar, typename PointT> int
    NYUInit::run( int argc, char** argv )
    {
        using namespace am;

        std::string                             cloud_path, cuboids_path, initials_path;
        typename pcl::PointCloud<PointT>::Ptr   cloud               = NULL;
        Cuboids                                 gt_cuboids, cuboids, initials;
        std::vector<Primitive::Ptr>             statics;                // holds non-cuboids from gt
        boost::shared_ptr<CloudViewer>          vp                  = NULL;
        Eigen::Matrix<Scalar,4,1>               floor_plane_coeffs  = Eigen::Matrix<Scalar,4,1>::Zero();
        bool                                    verbose             = false;
        int                                     cuboids_skipped     = 0;

        // READ
        {
            // parse
            {
                if ( pcl::console::find_switch(argc,argv,"-v") || pcl::console::find_switch(argc,argv,"--verbose") )      verbose = true;

                pcl::console::parse_argument( argc, argv, "--cloud"   ,    cloud_path );
                pcl::console::parse_argument( argc, argv, "--cuboids" ,  cuboids_path );
                pcl::console::parse_argument( argc, argv, "--initials", initials_path );

                if ( verbose )
                    std::cout << "run with paths: " << cloud_path << ", " << cuboids_path << ", " << initials_path << std::endl;
            } // ... parse

            // read cloud
            {
                if ( cloud_path.find("pts2") != std::string::npos )
                    AnnotationController::readAdditionalPoints<double,typename pcl::PointCloud<PointT>::Ptr>( cloud, cloud_path ); // *.pts  - Tianjia's format
                else if ( cloud_path.find("pts") != std::string::npos )
                    AnnotationController::readAdditionalPoints<float,typename pcl::PointCloud<PointT>::Ptr>( cloud, cloud_path ); // *.pts  - Tianjia's format
            } // ... read cloud

            // read boxes
            {
                am::Cuboids::readFromFile( gt_cuboids, cuboids_path, NULL, cloud, nullptr, NULL, &statics, NULL );
                if ( !statics.size() )
                {
                    std::cerr << "[" << __func__ << "]: " << "no floor, exiting" << std::endl;
                    return EXIT_FAILURE;
                }

                if ( !initials_path.empty() )
                {
                    Cuboids::readFromFile( initials, initials_path, NULL, cloud );
                    std::cout << "gt_cuboids.size(): " << gt_cuboids.size() << ", " << initials.size() << ", " << statics.size() << ", " << std::endl;
                }

                if ( verbose )  std::cout << "finished read" << std::endl; fflush(stdout);

                Plane::Ptr plane        = std::static_pointer_cast<Plane>( statics[0] );
                floor_plane_coeffs      = plane->Coeffs();

                if ( verbose )  std::cout << "floor: " << floor_plane_coeffs.transpose() << std::endl;
            } // ... read boxes

            // show boxes and cloud
            if ( verbose )
            {
                vp = CloudViewer::showCloud( cloud, "cloud", nullptr, 2., false );
                // draw cuboids
                for ( int i = 0; i != gt_cuboids.size(); ++i )
                {
                    char cub_name[64];
                    sprintf( cub_name, "cuboid%d", i );
                    gt_cuboids[i].draw( vp->VP(), cub_name );
                }

                vp->VP()->addArrow( pcl::PointXYZ(0, 0, 0),
                                    pcl::PointXYZ(floor_plane_coeffs(0), floor_plane_coeffs(1),floor_plane_coeffs(2)),
                                    1., 0., 0., false, "up", 0 );

                vp->VP()->spin();
            } // ... show
        } // ... READ

        // WORK
        {
            for ( int cuboid_id = 0; cuboid_id != gt_cuboids.size(); ++cuboid_id )
            {
                Cuboid const& gt_cuboid = gt_cuboids.at( cuboid_id );
                if ( cuboid_id >= gt_cuboids.size() - statics.size() )
                {
                    cuboids.add( gt_cuboid );
                    continue;
                }

                pcl::PointIndices::Ptr inside_indices;

                const float step_distance_threshold = 0.05f;
                float distance_threshold = -FLT_EPSILON + step_distance_threshold;
                do
                {
                    distance_threshold -= step_distance_threshold;
                    gt_cuboid.getInlierIndices( inside_indices, cloud, NULL, distance_threshold );
                }
                while ( (inside_indices->indices.size() < 10) && (distance_threshold > -1.f) );
                if ( (distance_threshold > -1.f) && (distance_threshold != -FLT_EPSILON) )
                    std::cout << "[" << __func__ << "]: " << "found >10 points at distance_threshold " << distance_threshold << std::endl;
                else
                {
                    if ( verbose )
                    {
                        pcl::visualization::PCLVisualizer::Ptr vptr( new pcl::visualization::PCLVisualizer() );
                        vptr->setBackgroundColor( .5 ,.6, .6 );
                        CloudViewer::addPointCloud( vptr, cloud );
                        char name[1024];
                        sprintf( name, "cuboid%02d", cuboid_id );
                        gt_cuboid.draw( vptr, name );
                        std::cout << "[" << __func__ << "]: " << "did not find >10 points at " << distance_threshold << std::endl;
                        vptr->spin();
                    }
                }

                if ( !inside_indices->indices.size() )
                {
                    std::cerr << "[" << __func__ << "]: " << "skipping cuboid " << cuboid_id << ", since inlier count is " << inside_indices->indices.size() << std::endl;
                    ++cuboids_skipped;
                    continue;
                    //return EXIT_FAILURE;
                }
                std::cout << "[" << __func__ << "]: " << "cuboid" << cuboid_id << std::endl;

                if ( verbose )
                {
                    vp->VP()->removePointCloud( "cloud" );
                    vp->addCloud( cloud, "cloud", inside_indices, 0, 2. );
                }

                {
                    // calc frame using upvector
                    Eigen::Matrix4f frame;
                    {
                        DepthProcessor::PCA_withPlane( /*             out_frame: */ frame
                                                       , /*              coeffs: */ floor_plane_coeffs
                                                       , /*               cloud: */ cloud
                                                       , /*             inliers: */ inside_indices
                                                       , /*          downsample: */ false
                                                       , /* out_projected_cloud: */ NULL
                                                       , /*               debug: */ false );

                        if ( verbose )
                        {
                            std::cout << "1dot2: " << frame.block<3,1>( 0,2).dot( frame.block<3,1>( 0,1) ) << std::endl;
                            std::cout << "0dot2: " << frame.block<3,1>( 0,2).dot( frame.block<3,1>( 0,0) ) << std::endl;
                            std::cout << "frame: " << frame << std::endl;
                        }
                    }

                    // Cloud2Cuboid
                    Cuboid cluster_cuboid;
                    Cuboid::fromFrame( cluster_cuboid, frame, cloud, inside_indices );

                    cuboids.add( cluster_cuboid );
                }

                if ( verbose && vp )
                    vp->VP()->spinOnce();
            } // ... for gt_cuboids

            // add from initials
            if ( initials.size() )
            {
                if ( initials.size() != gt_cuboids.size() + statics.size() )
                {
                    std::cerr << "[" << __func__ << "]: " << "initials.size() != gt_cuboids.size() + statics.size()...exiting\n";
                    return EXIT_FAILURE;
                }

                for ( int i = initials.size() - statics.size(); i != initials.size(); ++i )
                {
                    std::cout << "adding " << i << " from statics to " << cuboids.size() << "th cuboid\n";
                    cuboids.add( initials.at(i) );
                }

                if ( cuboids.size() != gt_cuboids.size() + statics.size() )
                    std::cerr << "initials.size(): " << initials.size() << ", cuboids.size(): " << cuboids.size() << std::endl;
            }

            if ( verbose )
            {
                boost::shared_ptr<CloudViewer> vp2 = CloudViewer::showCloud( cloud, "init cuboids", NULL, 2., false );
                for ( int cuboid_id = 0; cuboid_id != cuboids.size(); ++cuboid_id )
                {
                    Cuboid const& cuboid = cuboids.at( cuboid_id );
                    char tmp[1024];
                    sprintf( tmp, "cubiod%d", cuboid_id );
                    cuboid.draw( vp2->VP(), tmp );
                }

                vp2->VP()->spin();
            }

        } // ... WORK

        // WRITE
        {
            std::string unsuccess_path( "" );
            if ( (cuboids.size() != gt_cuboids.size()) || cuboids_skipped  )
                unsuccess_path = "/unsuccessful";

            boost::filesystem::path boost_path( cuboids_path );

            std::string stem     = boost_path.stem().string();
            //std::string out_path = (boost_path.parent_path() / boost_path.stem()).string() + ".dat";
            std::string out_path = boost_path.parent_path().string() + unsuccess_path + "/initial2_" + stem.substr( stem.find("_")+1 ) + ".dat";

            Cuboids::writeToFile( cuboids, out_path, 0, 0 );
        }

        return EXIT_SUCCESS;
    }
}

#endif // __CA_NYUINIT_H__
