#ifndef __CA_CLOUDCOLOURING_H__
#define __CA_CLOUDCOLOURING_H__

#include "Eigen/Dense"
#include <limits>

namespace ca
{

    class CloudColouring
    {
        public:
            template <class CloudPtrT, typename Scalar = float> static inline int
            distanceBased( CloudPtrT                    cloud
                           , Eigen::Matrix<Scalar,3,1>  origin      = Eigen::Matrix<Scalar,3,1>::Zero()
                           , Scalar                     max_dist    = Scalar(-1)
                           , Scalar                     min_dist    = Scalar(-1)
                           , Scalar                     scale       = Scalar(255)                      );

            template <typename Scalar = float> static inline Eigen::Matrix<int,3,1>
            getColourForDistance( Scalar distance, Scalar max_dist, Scalar min_dist );
    };

} // ... ns ca

// HPP

namespace ca
{
    template <class CloudPtrT, typename Scalar> inline int
    CloudColouring::distanceBased( CloudPtrT cloud, Eigen::Matrix<Scalar,3,1> origin, Scalar max_dist, Scalar min_dist, Scalar scale )
    {
        if ( scale != 255 )
        {
            std::cerr << "[" << __func__ << "]: " << "not implemented for scale!=255" << std::endl;
            return EXIT_FAILURE;
        }

        bool calc_max_dist = (max_dist < Scalar(0)),
             calc_min_dist = (min_dist < Scalar(0));
        min_dist = std::numeric_limits<Scalar>::max();

        std::vector<Scalar> distances( cloud->size() );
        for ( size_t pid = 0; pid != cloud->size(); ++pid )
        {
            distances[ pid ] = (cloud->at( pid ).getVector3fMap() - origin).norm();

            if ( calc_max_dist && (max_dist < distances[pid]) )
                max_dist = distances[pid];
            if ( calc_min_dist && (min_dist > distances[pid]) )
                min_dist = distances[pid];
        }

        std::cout << "[" << __func__ << "]: " << "max_dist: " << max_dist << ", min_dist: " << min_dist << std::endl;

        // colour
        for ( size_t pid = 0; pid != cloud->size(); ++pid )
        {
            Eigen::Matrix<int,3,1>   colour  = getColourForDistance( distances[pid], max_dist, min_dist );

            cloud->at(pid).r = colour(0)/* * scale*/;
            cloud->at(pid).g = colour(1)/* * scale*/;
            cloud->at(pid).b = colour(2)/* * scale*/;
        }

        return EXIT_SUCCESS;
    }

    template <typename Scalar> Eigen::Matrix<int,3,1> inline
    CloudColouring::getColourForDistance( Scalar distance, Scalar max_dist, Scalar min_dist )
    {
        Eigen::Matrix<int,3,1> ret_colour( Eigen::Matrix<int,3,1>::Zero() );

        Scalar spread   = (max_dist - min_dist);            // calculate domain width for normalization
        Scalar val      = (distance - min_dist) / spread;   // unbias, normalize

        if ( (val > Scalar(1)) || (val < Scalar(0)) )
        {
            std::cerr << "distance: " << distance
                      << ", max_dist: " << max_dist
                      << ", min_dist: " << min_dist
                      << ", spread: " << spread
//                      << ", dist2: " << dist2
                      << ", val: " << val << std::endl;
        }
#if 0
             if ( val < Scalar(.33) )   ret_colour(0) = val;
        else if ( val < Scalar(.66) )   ret_colour(1) = val;
        else                            ret_colour(2) = val;
#else
//        for( unsigned int i = 0 ; i < 2048 ; i++) {
//            float v = i/2048.0;
//            v = pow(v, 3)* 6;
//            m_gamma[i] = v*6*256;
//        }

        // borrowed from openkinect
        val = 0.22 + val * 0.323996; // map spread of visualization...
              // 0.046007 + val * 0.498993;
        int pval = val * val * val * 36 * 256; //m_gamma[depth[i]];
        int lb   = pval & 0xff;
        switch ( pval>>8 )
        {
            case 0:
                ret_colour(0) = 255;
                ret_colour(1) = 255-lb;
                ret_colour(2) = 255-lb;
                break;
            case 1:
                ret_colour(0) = 255;
                ret_colour(1) = lb;
                ret_colour(2) = 0;
                break;
            case 2:
                ret_colour(0) = 255-lb;
                ret_colour(1) = 255;
                ret_colour(2) = 0;
                break;
            case 3:
                ret_colour(0) = 0;
                ret_colour(1) = 255;
                ret_colour(2) = lb;
                break;
            case 4:
                ret_colour(0) = 0;
                ret_colour(1) = 255-lb;
                ret_colour(2) = 255;
                break;
            case 5:
                ret_colour(0) = 0;
                ret_colour(1) = 0;
                ret_colour(2) = 255-lb;
                break;
            default:
                ret_colour(0) = 0;
                ret_colour(1) = 0;
                ret_colour(2) = 0;
                break;
        }

#endif

        return ret_colour;
    }
}

#endif // __CA_CLOUDCOLOURING_H__
