#ifndef AMUTIL_H
#define AMUTIL_H

#include <string>
#include <algorithm>

#include <eigen3/Eigen/Dense>
#include <boost/shared_ptr.hpp>

#include <opencv2/core/core.hpp>
#include <opencv2/highgui/highgui.hpp>

#include "pcl/io/openni_camera/openni_depth_image.h"
#include "pcl/io/openni_camera/openni_image_rgb24.h"
#include "pcl/point_types.h"
#include "pcl/ModelCoefficients.h"
#include "pcl/sample_consensus/sac_model_plane.h"
#include "pcl/visualization/pcl_visualizer.h"
#include <pcl/PointIndices.h>

#include "amDefines.h"

namespace am
{
    namespace util
    {
        namespace eigen
        {
            int sort_eigen_vectors_desc( ::Eigen::Vector3f const& eigen_values, ::Eigen::Matrix3f & eigen_vectors );
        }

        namespace cv
        {
            extern void
            unsignedIntToFloat( ::cv::Mat & img32FC1, ::cv::Mat const& img32UC1 );

            extern int
            imread( /* out: */ ::cv::Mat &mat, /*  in: */ std::string const& path );

            extern int
            writePNG( std::string title, ::cv::Mat const& mat, bool silent = false );

            extern int
            blend( ::cv::Mat &blendedUC3, ::cv::Mat const& dep, float depMax, ::cv::Mat const& rgb, float rgbMax = 255.f );

            template <typename T>
            int
            histogram( std::vector<uint> &hist, ::cv::Mat const& img, int bins = 0, T the_max = 255.f )
            {
                T bin_width = 1.f;
                if ( !bins )
                {
                    double minVal, maxVal;
                    ::cv::minMaxLoc( img, &minVal, &maxVal );
                    bins = static_cast<int>( maxVal ) + 1;
                }
                else
                {
                    bin_width = static_cast<T>( ceil(the_max / (float)bins) );
                }

                hist.resize( bins );

                for ( unsigned y = 0; y < img.rows; ++y )
                {
                    for ( unsigned x = 0; x < img.cols; ++x )
                    {
                        uchar pixel = img.at<uchar>( y, x );
                        int bin = pixel / bin_width;
                        if ( bin != 1 )
                        ++(hist[ std::min(bins-1,bin) ]);
                    }
                }

                return EXIT_SUCCESS;
            }

            template <typename T>
            int
            printHistogram( std::string title, ::cv::Mat const& img, int bins = 0, T the_max = 255, std::vector<unsigned> *p_hist = NULL )
            {
                std::vector<uint> histogram;
                am::util::cv::histogram<T>( histogram, img, bins, the_max );

                std::cout << "histogram(" << title << "): ";
                int id = 0;

                for ( std::vector<uint>::const_iterator it = histogram.begin();
                      it != histogram.end();
                      ++it )
                {
                    if ( *it > 0 )
                        std::cout << "(" << id << "," << (int)*it << "),";
                    ++id;
                }
                std::cout << std::endl;

                if ( p_hist )
                {
                    *p_hist = histogram;
                }

                return EXIT_SUCCESS;
            }

            int
            replaceValue( ::cv::Mat &img, ::cv::InputArray value_to_compare, ::cv::InputArray value_to_write );

            template <typename Timg, typename Tout>
            int
            saveDepthAsBinary( ::cv::Mat const& dep, std::string const& dat_path, float alpha = 1.f )
            {
                FILE *fp = fopen( dat_path.c_str(), "wb" );
                if ( !fp )
                {
                    printf("save Depthmap with size HFirst failed when opening %s\n", dat_path.c_str() );
                    return false;
                }

                fwrite( &dep.rows, sizeof(int), 1, fp );
                fwrite( &dep.cols, sizeof(int), 1, fp );
                // save as linear
                Tout *p_depth = new Tout[dep.cols * dep.rows];
                int p_depth_index = 0;
                for ( unsigned y = 0; y < dep.rows; ++y  )
                {
                    for ( unsigned x = 0; x < dep.cols; ++x, ++p_depth_index )
                    {
                        p_depth[p_depth_index] = (Tout) dep.at<Timg>( y, x ) * alpha;
                    }
                }

                fwrite( p_depth, sizeof(Tout), dep.rows*dep.cols, fp );

                fclose( fp );

                if ( p_depth )
                    delete [] p_depth;

                return EXIT_SUCCESS;
            }

            // hfirst
            // Tout is the format of the data file, Timg is the format of cv::Mat
            template <typename Timg, typename Tout>
            int
            loadDepthAsBinary( ::cv::Mat &img, std::string path, float alpha = 1.f )
            {
                FILE *fp = fopen( path.c_str(), "rb" );
                if (!fp)
                {
                    printf( "loadDepthAsBinary failed, when opening %s\n", path.c_str() );
                    return false;
                }

                int w,h, read_size = 0;
                read_size += fread( &h,sizeof(int),1,fp );
                read_size += fread( &w,sizeof(int),1,fp );
                if ( read_size != 2 )
                {
                    std::cerr << "[" << __func__ << "]: " << "w or h not read properly: " << w << ", " << h << std::endl;
                }

                Tout *p_depth = new Tout[ h * w ];

                read_size = fread( p_depth, sizeof(Tout), w*h, fp );
                if ( read_size != w*h )
                    std::cerr << "[" << __func__ << "]: " << "read size != " << w*h << std::endl;

                img.create( h, w, CV_16UC1 );
                int p_depth_index = 0;
                for ( unsigned y = 0; y < img.rows; ++y  )
                {
                    for ( unsigned x = 0; x < img.cols; ++x, ++p_depth_index )
                    {
                        img.at<Timg>(y,x) = p_depth[ p_depth_index ] * alpha;
                    }
                }

                fclose( fp );
                if ( p_depth )
                    delete [] p_depth;

                return EXIT_SUCCESS;
            }

            template <typename Timg, typename Tout>
            int
            saveDepthAsBinary( std::string const& depth_path, std::string const& dat_path, float alpha = 1.f )
            {
                ::cv::Mat dep = ::cv::imread( depth_path, ::cv::IMREAD_UNCHANGED );
                return saveDepthAsBinary<Timg,Tout>( dep, dat_path, alpha );
            }
        }

        namespace pcl
        {
            Eigen::Vector3f
            point2To3D( Eigen::Vector2f const& pnt2, Eigen::Matrix3f const& intrinsics );

            Eigen::Vector2f
            point3To2D( Eigen::Vector3f const& pnt3, Eigen::Matrix3f const& intrinsics );

            template <typename MyPointT>
            MyPointT
            projectPointToPlane( MyPointT point, ::pcl::ModelCoefficients::ConstPtr coeffs )
            {
                MyPointT retPoint;
                Eigen::Vector4f coeffs_v4;
                coeffs_v4 << coeffs->values[0],
                        coeffs->values[1],
                        coeffs->values[2],
                        coeffs->values[3];
                ::pcl::projectPoint( point, coeffs_v4, retPoint );
                return retPoint;

            }

            ::pcl::PointXYZ
            asPointXYZ( Eigen::Vector3f const& vector3f );

            ::pcl::PointXYZ
            v4asPointXYZ(Eigen::Vector4f const& vector4f, bool dont_divide = false );

            template <typename MyPointT>
            Eigen::Vector3f
            asVector3f( MyPointT const& v )
            {
                return { v.x, v.y, v.z };
            }

            template <typename MyPointT>
            int
            setPointColor( MyPointT &pnt,
                           int r,
                           int g,
                           int b )
            {
                uint32_t rgb = ( static_cast<uint32_t>(r) << 16 |
                                 static_cast<uint32_t>(g) << 8  |
                                 static_cast<uint32_t>(b)         );
                pnt.rgb = *reinterpret_cast<float*>( &rgb );

                return EXIT_SUCCESS;
            }

            ::pcl::PointXYZRGB
            PointXYZRGB( float x, float y, float z, uint8_t r, uint8_t g, uint8_t b );

            std::vector< std::vector<Eigen::Vector3f> >
            getCamerasPosFocalUp( ::pcl::visualization::PCLVisualizer::Ptr viewer );
            int
            setCameraPosFocalUp( ::pcl::visualization::PCLVisualizer::Ptr viewer
                                 , std::vector<Eigen::Vector3f>           pos_focal_up
                                 , int                                    viewport_id  = 0 );

            template <typename inCloudT>
            ::pcl::PointCloud<::pcl::PointXYZRGB>::Ptr toXYZRGBCloud( inCloudT cloud_ptr )
            {
                ::pcl::PointCloud<::pcl::PointXYZRGB>::Ptr rgb_cloud_ptr( new ::pcl::PointCloud<::pcl::PointXYZRGB>() );

                rgb_cloud_ptr->reserve( cloud_ptr->size() );
                for ( auto it  = cloud_ptr->begin();
                           it != cloud_ptr->end();
                         ++it )
                {
                    ::pcl::PointXYZRGB pnt;
                    pnt.x = it->x;
                    pnt.y = it->y;
                    pnt.z = it->z;
                    pnt.rgb = it->rgb;

                    rgb_cloud_ptr->push_back( pnt );
                }
                return rgb_cloud_ptr;
            }


            template <typename inCloudT>
            ::pcl::PointIndices::Ptr allIndicesOf( inCloudT cloud_ptr )
            {
                ::pcl::PointIndices::Ptr indices_ptr( new ::pcl::PointIndices() );
                indices_ptr->indices.reserve( cloud_ptr->size() );

                for ( int i = 0; i != cloud_ptr->size(); ++i )
                    indices_ptr->indices.push_back( i );

                return indices_ptr;
            }

            template <class VectorType>
            ::pcl::ModelCoefficients::Ptr coeffsFrom( VectorType v )
            {
                ::pcl::ModelCoefficients::Ptr model(new ::pcl::ModelCoefficients() );

                model->values.resize( v.rows() );
                for ( int i = 0; i != model->values.size(); ++i )
                    model->values[i] = v(i);

                return model;
            }
        } // ns pcl

        namespace algo
        {
            template <typename T>
            int sort3_desc( T &a, T &b, T &c, std::vector<int> *p_indices = nullptr )
            {
                if ( a < b )
                {
                    std::swap( a, b );
                    if ( p_indices ) std::swap ( p_indices->at(0), p_indices->at(1) );
                }
                if ( b < c )
                {
                    std::swap( b, c );
                    if ( p_indices ) std::swap ( p_indices->at(1), p_indices->at(2) );
                }
                if ( a < b )
                {
                    std::swap( a, b );
                    if ( p_indices ) std::swap ( p_indices->at(0), p_indices->at(1) );
                }

                return EXIT_SUCCESS;
            }

            template <typename T>
            int sort3_desc( T a, T b, T c, std::vector<int> &indices )
            {
                if ( a < b )
                {
                    std::swap( a, b );
                    std::swap ( indices.at(0), indices[1] );
                }
                if ( b < c )
                {
                    std::swap( b, c );
                    std::swap ( indices.at(1), indices[2] );
                }
                if ( a < b )
                {
                    std::swap( a, b );
                    std::swap ( indices.at(0), indices[1] );
                }

                return EXIT_SUCCESS;
            }

            // implements vector1 - vector2
            template <typename T>
            int
            vectorDifference( std::vector<T> &difference, std::vector<T> vector1, std::vector<T> vector2 )
            {
                difference.clear();
                difference.reserve( vector1.size() );
                std::sort(vector1.begin(), vector1.end());
                std::sort(vector2.begin(), vector2.end());
                std::set_difference(
                            vector1.begin(), vector1.end(),
                            vector2.begin(), vector2.end(),
                            std::back_inserter( difference )
                            );

                return EXIT_SUCCESS;
            }

            // implements vector1 - vector2
            template <typename T>
            int
            vectorIntersection( std::vector<T> &intersection, std::vector<T> vector1, std::vector<T> vector2 )
            {
                intersection.clear();
                intersection.reserve( vector1.size() );

                std::sort(vector1.begin(), vector1.end());
                std::sort(vector2.begin(), vector2.end());

                std::set_intersection(
                            vector1.begin(), vector1.end(),
                            vector2.begin(), vector2.end(),
                            std::back_inserter( intersection )
                            );

                return EXIT_SUCCESS;
            }

            // implements vector1 U vector2
            template <typename T>
            int
            vectorUnion( std::vector<T> &out_union, std::vector<T> vector1, std::vector<T> vector2 )
            {
                out_union.clear();
                out_union.reserve( vector1.size() );

                std::sort(vector1.begin(), vector1.end());
                std::sort(vector2.begin(), vector2.end());

                std::set_union(
                            vector1.begin(), vector1.end(),
                            vector2.begin(), vector2.end(),
                            std::back_inserter( out_union )
                            );

                return EXIT_SUCCESS;
            }

            template<typename T>
            T median( std::vector<T> values )
            {
                  T median;
                  size_t size = values.size();

                  std::sort( values.begin(), values.end() );

                  if ( size % 2 )
                      median = values[size / 2];
                  else
                      median = (values[size / 2 - 1] + values[size / 2]) / (T)2;

                  return median;
            }

            template <typename T>
            std::vector<size_t> sort_indices( const std::vector<T> &v )
            {
              // initialize original index locations
              std::vector<size_t> indices( v.size() );
              for ( size_t i = 0; i != indices.size(); ++i )
                  indices[ i ] = i;

              // sort indexes based on comparing values in v
              std::sort( indices.begin(), indices.end(),
                    [&v]( size_t i1, size_t i2 ) { return v[i1] < v[i2]; } );

              return indices;
            }
        }

        template <typename T>
        std::string
        sprintf( char const* format, T val )
        {
            char chr[1024];
            std::sprintf( chr, format, val );

            return std::string(chr);
        }

        template <typename T>
        std::string
        sprintf( char const* format, T val, T val2 )
        {
            char chr[1024];
            std::sprintf( chr, format, val, val2 );

            return std::string(chr);
        }

        extern int
        savePFM( ::cv::Mat const& imgF, std::string path, float scale = -1.f, bool silent = false );

        extern int
        loadPFM( ::cv::Mat & imgF, std::string path );

        int
        PclDepthImageToCvMat( ::cv::Mat* pCvDepth16, boost::shared_ptr<openni_wrapper::DepthImage> const& depth_img, ::cv::Mat* pCvDepth8 = NULL );
        int
        PclRgbImageToCvMat  (::cv::Mat *pImgRGB, const boost::shared_ptr<openni_wrapper::Image> &img );

        template <typename T>
        int
        printVector( std::vector<T> & vec, std::string title, int limit = 0 )
        {
            int counter = 0;
            std::cout << "Vector(" << title << "):" << std::endl << "\t";
            for ( T & elem : vec )
            {
                std::cout << elem;
                if ( ++counter < vec.size() )
                    std::cout << ", ";

                if ( (limit > 0) && (counter > limit) )
                    break;
            }
            std::cout << std::endl;

            return EXIT_SUCCESS;
        }

        template <typename T>
        bool
        in( T elem, std::vector<T> elements )
        {
            return ( std::find( elements.begin(), elements.end(), elem ) != elements.end() );
        }

        std::string
        timestamp2Str();

        // generates n different colours
        ::cv::Point3f hsv2rgb( ::cv::Point3f const& in );
        std::vector<::cv::Point3f    > nColoursCv   ( int n, float scale = 255.f, bool random_shuffle = true );
        // brightness_limit: needs at least one channel, to be below this
        std::vector<::cv::Point3f    > nColoursCv2  ( int n, float scale = 255.f, float value_random_part = .3f, double start = .8, float brightness_limit = .7f );
        std::vector<::Eigen::Vector3f> nColoursEigen( int n, float scale = 255.f, bool random_shuffle = true );

        std::string
        toHexString( int number );
    } // end ns util

    namespace geom
    {
        extern
        Eigen::Vector3f
        lineToPlaneIntersection( Eigen::VectorXf line, Eigen::Vector4f plane );
    }
} // end ns am

#endif // AMUTIL_H
