#ifndef VIEWPOINTMAPPER_H
#define VIEWPOINTMAPPER_H

#include <eigen3/Eigen/Dense>
#include "MyIntrinsics.h"
#include <iostream>

namespace smartgeometry
{

    class ViewpointMapper
    {
        public:
            // returns a two floats in a vector, camera space 2D coordinates
           static Eigen::Vector2f world2Cam( const Eigen::Vector3f xw, // world coordinates of point
                                             const float *F,  // focal distance
                                             const float *C,  // center of projection
                                             const float k1, // K1
                                             const float k2, // K2
                                             const float p1, // K3
                                             const float p2, // K4
                                             const float k3, // K5
                                             const float alpha = 0.f // skew parameter
                                           );
            // returns a three floats in a vector, homogeneous world coordinates
            static Eigen::Vector3f cam2World( const Eigen::Vector2f xd, // x_distorted
                                              const float *F, // focal distance
                                              const float *C, // center of projection
                                              const float k1, // K1
                                              const float k2, // K2
                                              const float p1, // K3
                                              const float p2, // K4
                                              const float k3, // K5
                                              const float alpha = 0.f // skew parameter
            );

            static float kduToMM( const float depthInKDU,
                                  const float dc1,
                                  const float dc2 ); // converts KDU to MM


            template <typename T>
            static void mapViewpoint( const T *in,          // input depth
                                      /* */ T *out,                // mapped depth
                                      const int w,                 // width (of both)
                                      const int h,                 // height (of both)
                                      const std::size_t in_pitch,  // row_step of in
                                      const std::size_t out_pitch, // row_step of out
                                      const MyIntrinsics dep_intr, // intrinsics of depth camera
                                      const MyIntrinsics rgb_intr, // intrinsics of rgb camera
                                      const Eigen::Matrix3f& R,    // rotation part of extrinsic mx
                                      const Eigen::Vector3f& t,    // translation part of extrinsic mx
                                      bool isInMM                  // data given in MM already 
                                      );


    };

    template <typename T>
    void ViewpointMapper::mapViewpoint( const T *in,                 // input depth
                                        /* */ T *out,                // mapped depth
                                        const int w,                 // width (of both)
                                        const int h,                 // height (of both)
                                        const std::size_t in_pitch,  // row_step of in
                                        const std::size_t out_pitch, // row_step of out
                                        const MyIntrinsics dep_intr, // intrinsics of depth camera
                                        const MyIntrinsics rgb_intr, // intrinsics of rgb camera
                                        const Eigen::Matrix3f& R,    // rotation part of extrinsic mx
                                        const Eigen::Vector3f& t,    // translation part of extrinsic mx
                                        bool isInMM                  // data given in MM already
                                        )
    {
        const float f_dep[2] = { dep_intr.fx, dep_intr.fy }; // focal distances of depth camera
        const float c_dep[2] = { dep_intr.cx, dep_intr.cy }; // centre of projection of depth camera
        const float f_rgb[2] = { rgb_intr.fx, rgb_intr.fy }; // focal distances of rgb camera
        const float c_rgb[2] = { rgb_intr.cx, rgb_intr.cy }; // centre of projection of rgb camera

        // for every pixel in image
        for ( int y = 0; y < h; ++y )
            for ( int x = 0; x < w; ++x )
            {
                // read depth value
                float d = in[ y * in_pitch + x ];

                // convert to MM if necessary
                if (!isInMM) {
                    d = kduToMM(d, dep_intr.dc1, dep_intr.dc2);
                }

                // create point
                Eigen::Vector2f p = { x, y };

                // from depth camera camera-space to depth camera world-space
                Eigen::Vector3f P_world_left = d * cam2World( p,
                                                              f_dep,
                                                              c_dep,
                                                              dep_intr.k1, dep_intr.k2, dep_intr.p1, dep_intr.p2, dep_intr.k3,
                                                              dep_intr.alpha );


                // from depth camera world-space to rgb camera world-space
                // R * [X; Y; Z] + T
                Eigen::Vector3f P_world_right = R * P_world_left + t;

                // from rgb camera world-space to rgb camera camera-space
                Eigen::Vector2f p2_right = world2Cam( P_world_right,
                                                      f_rgb,
                                                      c_rgb,
                                                      rgb_intr.k1, rgb_intr.k2, rgb_intr.p1, rgb_intr.p2, rgb_intr.k3,
                                                      rgb_intr.alpha );
                p2_right(0) = round(p2_right(0));
                p2_right(1) = round(p2_right(1));

                // round to integer coordinates
                int int_p_right[2] = { static_cast<int>(p2_right(0)),
                                       static_cast<int>(p2_right(1)) };

                // check boundaries
                if (    (int_p_right[0] >= w)
                        || (int_p_right[1] >= h)
                        || (int_p_right[0] <  0)
                        || (int_p_right[1] <  0) )
                    continue;

                // put minimum depth to output
                int linear_index_in_out = int_p_right[1] * out_pitch + int_p_right[0];

                //if (in[y * in_pitch + x] > 0)
                    //std::cout << in[y * in_pitch + x] << std::endl;
                //if ((y * in_pitch + x) > 301000) {
                    //std::cout << w << " " << h << std::endl;
                    //std::cout << out_pitch << " " << int_p_right[0] << " " << int_p_right[1] << std::endl;
                    //std::cout << "Y * IN_PITCH + X: " << y * in_pitch + x << std::endl;
                    //std::cout << "LINEAR_INDEX_IN_OUT: " << linear_index_in_out << std::endl;
                //}

                if (in[y * in_pitch + x] > 2046)
                    continue;

                out[ linear_index_in_out ] = in[ y * in_pitch + x ];

                // float old = atomicCAS ( (int*)&(out[ int_p_right[1] * out_pitch + int_p_right[0] ]), 0, *pZ );
                // if ( old != 0.f )
                // {
                //      atomicMin( (int*)&(out[ int_p_right[1] * out_pitch + int_p_right[0] ]), *pZ );
                // }
            }
    }

} // namespace smartgeometry

#endif // VIEWPOINTMAPPER_H
