#include "MultiLabelGrabCut.h"

#include <opencv2/core/core.hpp>
#include "opencv2/core/types_c.h"
#include <opencv2/highgui/highgui.hpp>

#include "GMM.h"
#include "GMM2.h"
#include "grabcut/gcgraph.hpp"
#include <memory>
#include "gc_types.h"
#include "AMUtil2.h"
#include "GCoptimization.h"

typedef double t_colour;

namespace am
{
    using namespace cv;

    //extern void checkMask( const cv::Mat& img, const cv::Mat& mask );
    //extern void estimateSegmentation( GCGraph<double>& graph, cv::Mat& mask );
    //extern void initMaskWithRect( cv::Mat& mask, cv::Size imgSize, cv::Rect rect );

    // Calculate beta - parameter of GrabCut algorithm.
    // beta = 1/(2*avg(sqr(||color[i] - color[j]||)))
    template <int cn, typename t_img>
    static double calcBeta( const cv::Mat& img )
    {
        double beta = 0;
        for( int y = 0; y < img.rows; y++ )
        {
            for( int x = 0; x < img.cols; x++ )
            {
                cv::Vec<double,cn> diff, color = (cv::Vec<double,cn>)img.at<t_img>(y,x);
                if( x>0 ) // left
                {
                    diff = color - (cv::Vec<double,cn>)img.at<t_img>(y,x-1);
                    beta += diff.dot(diff);

                    if( y>0 ) // upleft
                    {
                        diff = color - (cv::Vec<double,cn>)img.at<t_img>(y-1,x-1);
                        beta += diff.dot(diff);
                    }
                }

                if( y>0 ) // up
                {
                    diff = color - (cv::Vec<double,cn>)img.at<t_img>(y-1,x);
                    beta += diff.dot(diff);

                    if ( x < img.cols-1 ) // upright
                    {
                        diff = color - (cv::Vec<double,cn>)img.at<t_img>(y-1,x+1);
                        beta += diff.dot(diff);
                    }
                }

            }
        }
        if ( beta <= std::numeric_limits<double>::epsilon() ) beta = 0;
        else                                                  beta = 1. / (2. * beta/(4.*img.cols*img.rows - 3.*img.cols - 3.*img.rows + 2.) );

        return beta;
    }

    // Calculate weights of noterminal vertices of graph.
    // beta and gamma - parameters of GrabCut algorithm.
    template <int cn, typename t_img>
    static void calcNWeights( Mat& leftW, Mat& upleftW, Mat& upW, Mat& uprightW, const Mat& img, double beta, double gamma )
    {
        const double gammaDivSqrt2 = gamma / std::sqrt(2.0f);
        leftW   .create( img.rows, img.cols, CV_64FC1 );
        upleftW .create( img.rows, img.cols, CV_64FC1 );
        upW     .create( img.rows, img.cols, CV_64FC1 );
        uprightW.create( img.rows, img.cols, CV_64FC1 );
        for ( int y = 0; y < img.rows; ++y )
        {
            for ( int x = 0; x < img.cols; ++x )
            {
                cv::Vec<double,cn> color = (cv::Vec<double,cn>)img.at<t_img>(y,x);
                if( x-1>=0 ) // left
                {
                    cv::Vec<double,cn> diff = color - (cv::Vec<double,cn>)img.at<t_img>(y,x-1);
                    leftW.at<double>(y,x) = gamma * exp(-beta*diff.dot(diff));
                }
                else
                    leftW.at<double>(y,x) = 0;
                if( x-1>=0 && y-1>=0 ) // upleft
                {
                    cv::Vec<double,cn> diff = color - (cv::Vec<double,cn>)img.at<t_img>(y-1,x-1);
                    upleftW.at<double>(y,x) = gammaDivSqrt2 * exp(-beta*diff.dot(diff));
                }
                else
                    upleftW.at<double>(y,x) = 0;
                if( y-1>=0 ) // up
                {
                    cv::Vec<double,cn> diff = color - (cv::Vec<double,cn>)img.at<t_img>(y-1,x);
                    upW.at<double>(y,x) = gamma * exp(-beta*diff.dot(diff));
                }
                else
                    upW.at<double>(y,x) = 0;
                if( x+1<img.cols && y-1>=0 ) // upright
                {
                    cv::Vec<double,cn> diff = color - (cv::Vec<double,cn>)img.at<t_img>(y-1,x+1);
                    uprightW.at<double>(y,x) = gammaDivSqrt2 * exp(-beta*diff.dot(diff));
                }
                else
                    uprightW.at<double>(y,x) = 0;
            }
        }
    }

    template <typename t_img>
    cv::Mat readPixelToMat( cv::Mat const& img, cv::Point const& p );
    template <>
    cv::Mat readPixelToMat<cv::Vec3b>( cv::Mat const& img, cv::Point const& p )
    {
        cv::Mat out( 1, 3, CV_64FC1 );
        cv::Vec3b val = img.at<cv::Vec3b>( p );
        out.at<double>( 0, 0 ) = static_cast<double>( val[0] );
        out.at<double>( 0, 1 ) = static_cast<double>( val[1] );
        out.at<double>( 0, 2 ) = static_cast<double>( val[2] );
        return out;
    }
    template <>
    cv::Mat readPixelToMat<uchar>( cv::Mat const& img, cv::Point const& p )
    {
        cv::Mat out( 1, 1, CV_64FC1 );
        uchar val = img.at<uchar>( p );
        out.at<double>( 0, 0 ) = static_cast<double>( val );
        return out;
    }
    template <>
    cv::Mat readPixelToMat<ushort>( cv::Mat const& img, cv::Point const& p )
    {
        cv::Mat out( 1, 1, CV_64FC1 );
        ushort val = img.at<ushort>( p );
        out.at<double>( 0, 0 ) = static_cast<double>( val );
        return out;
    }

    // Initialize GMM background and foreground models using kmeans algorithm.
    // \param mask Contains labels: l0 = 0, l0_probably  = 1, l1 = 2, l1_probably = 3, etc...
    template <typename t_img>
    static void initGMM2( const Mat& img, const Mat& mask, uchar label, GMM2& gmm )
    {
        const int kMeansItCount = 10;
        const int kMeansType    = KMEANS_PP_CENTERS;

        CV_Assert( mask.type() == CV_8UC1 );

        Mat labels;
        std::vector< cv::Mat > samples_vector;
        cv::Mat value;
        Point p;
        // debug hist of mask
        {
            std::vector<uint> histogram;
            am::util::cv::histogram<uchar>( histogram, mask );
            std::cout << "histogram: ";
            for ( auto elem : histogram )
            {
                std::cout << (int)elem << ", ";
            }
            std::cout << std::endl;
        }
        for ( p.y = 0; p.y < img.rows; ++p.y )
        {
            for ( p.x = 0; p.x < img.cols; ++p.x )
            {
                if ( (mask.at<uchar>(p) == label) || (mask.at<uchar>(p) == label+1) )
                {
                    samples_vector.push_back( readPixelToMat<t_img>( img, p ) );
                }
            }
        }

        CV_Assert( !samples_vector.empty() );
        Mat samples_mat( (int)samples_vector.size(), gmm.get_channels(), CV_32FC1 );
        for ( int sample_id = 0; sample_id < samples_vector.size(); ++sample_id )
        {
            for (int column_id = 0; column_id < samples_vector[sample_id].cols; ++column_id)
            {
                samples_mat.at<t_colour>( sample_id, column_id ) = samples_vector[sample_id].at<t_colour>(column_id);
            }

        }
        kmeans( samples_mat,
                gmm.get_components_count(),
                labels,
                TermCriteria( CV_TERMCRIT_ITER, kMeansItCount, 0.0 ),
                0,
                kMeansType );

        gmm.initLearning();
        for( int i = 0; i < (int)samples_vector.size(); ++i )
            gmm.addSample( labels.at<int>(i,0), samples_vector[i] );
        gmm.endLearning();
    }

    // Assign GMMs components for each pixel.
    template <int cn, typename t_img>
    static void assignGMMsComponents( Mat& compIdxs, const Mat& img, const Mat& mask, uchar label, const GMM2& gmm )
    {
        Point p;
        for ( p.y = 0; p.y < img.rows; ++p.y )
        {
            for ( p.x = 0; p.x < img.cols; ++p.x )
            {
                cv::Vec<double,cn> color = img.at<t_img>(p);
                if ( (mask.at<uchar>(p) == label) || (mask.at<uchar>(p) == label+1) )
                    compIdxs.at<int>(p) = gmm.whichComponent( color );
            }
        }
    }

    // Learn GMMs parameters.
    template <int cn, typename t_img>
    static void learnGMMs( const Mat& img, const Mat& mask, const Mat& compIdxs, GMM<cn>& bgdGMM, GMM<cn>& fgdGMM )
    {
        bgdGMM.initLearning();
        fgdGMM.initLearning();
        Point p;
        for( int ci = 0; ci < GMM<cn>::kComponentsCount; ci++ )
        {
            for( p.y = 0; p.y < img.rows; p.y++ )
            {
                for( p.x = 0; p.x < img.cols; p.x++ )
                {
                    if( compIdxs.at<int>(p) == ci )
                    {
                        if (    (mask.at<uchar>(p) == GC_BGD   )
                             || (mask.at<uchar>(p) == GC_PR_BGD) )  bgdGMM.addSample( ci, img.at<t_img>(p) );
                        else                                        fgdGMM.addSample( ci, img.at<t_img>(p) );
                    }
                }
            }
        }
        bgdGMM.endLearning();
        fgdGMM.endLearning();
    }

    // Construct GCGraph
    template <int cn, typename t_img>
    static void constructGCGraph( const Mat& img, const Mat& mask, const GMM<cn>& bgdGMM, const GMM<cn>& fgdGMM, double lambda,
                                  const Mat& leftW, const Mat& upleftW, const Mat& upW, const Mat& uprightW,
                                  GCGraph<double>& graph )
    {
        int vtxCount = img.cols*img.rows,
            edgeCount = 2*(4*img.cols*img.rows - 3*(img.cols + img.rows) + 2);
        graph.create(vtxCount, edgeCount);
        Point p;
        for( p.y = 0; p.y < img.rows; p.y++ )
        {
            for( p.x = 0; p.x < img.cols; p.x++)
            {
                // add node
                int vtxIdx = graph.addVtx();
                t_img color = img.at<t_img>(p);

                // set t-weights
                double fromSource, toSink;
                if( mask.at<uchar>(p) == GC_PR_BGD || mask.at<uchar>(p) == GC_PR_FGD )
                {
                    fromSource  = -log( bgdGMM(color) );
                    toSink      = -log( fgdGMM(color) );
                }
                else if( mask.at<uchar>(p) == GC_BGD )
                {
                    fromSource = 0;
                    toSink = lambda;
                }
                else // GC_FGD
                {
                    fromSource = lambda;
                    toSink = 0;
                }
                graph.addTermWeights( vtxIdx, fromSource, toSink );

                // set n-weights
                if( p.x>0 )
                {
                    double w = leftW.at<double>(p);
                    graph.addEdges( vtxIdx, vtxIdx-1, w, w );
                }
                if( p.x>0 && p.y>0 )
                {
                    double w = upleftW.at<double>(p);
                    graph.addEdges( vtxIdx, vtxIdx-img.cols-1, w, w );
                }
                if( p.y>0 )
                {
                    double w = upW.at<double>(p);
                    graph.addEdges( vtxIdx, vtxIdx-img.cols, w, w );
                }
                if( p.x<img.cols-1 && p.y>0 )
                {
                    double w = uprightW.at<double>(p);
                    graph.addEdges( vtxIdx, vtxIdx-img.cols+1, w, w );
                }
            }
        }
    }


    int MultiLabelGrabCut::grabCut( cv::InputArray const& images_arg,
                                     cv::InputArray masks_arg,
                                     cv::InputOutputArray models,
                                     cv::OutputArray results,
                                     int iterCount,
                                     int mode )
    {
        std::cout << "MultiLabelGrabCut::grabCut:"
                  << "\n\t images.size: " << images_arg.size()
                  << " isMat: "     << ((images_arg.kind() == cv::_InputArray::MAT) ? "MAT" : "NO")
                  << " isVector: "  << ((images_arg.kind() == cv::_InputArray::STD_VECTOR_MAT) ? "STD_VECTOR_MAT" : "NO")
                  << "\n\t masks.size: " << masks_arg.size() << " kind: " << masks_arg.kind()
                  << " isMat: " << ((masks_arg.kind() == cv::_InputArray::MAT) ? "MAT" : "NO")
                  << " isVector: " << ((masks_arg.kind() == cv::_InputArray::STD_VECTOR_MAT) ? "STD_VECTOR_MAT" : "NO")
                  << std::endl;

        // read masks
        std::vector<cv::Mat> masks;
        {
            if ( masks_arg.kind() == cv::_InputArray::STD_VECTOR_MAT )
            {
                masks_arg.getMatVector( masks );
            }
            else if ( masks_arg.kind() == cv::_InputArray::MAT )
            {
                masks.push_back( masks_arg.getMat() );
            }
            else
            {
                std::cerr << "MultiLabelGrabCut::grabCut(): masks_arg kind not recognized!!!" << std::endl;
                return EXIT_FAILURE;
            }
        }

        // read images
        std::vector<cv::Mat> images;
        {
            if ( images_arg.kind() == cv::_InputArray::STD_VECTOR_MAT)
            {
                images_arg.getMatVector( images );
            }
            else if ( images_arg.kind() == cv::_InputArray::MAT )
            {
                images.push_back( images_arg.getMat() );
            }
            else
            {
                std::cerr << "MultiLabelGrabCut::grabCut(): images_arg kind not recognized!!!" << std::endl;
                return EXIT_FAILURE;
            }
        }

        CV_Assert( masks.size() == images.size() );

        // check masks
        uchar max_label = Labels::probably_background();
        {
            double minVal, maxVal;
            for ( int img_id = 0; img_id < masks.size(); ++img_id )
            {
                cv::minMaxLoc( masks[img_id], &minVal, &maxVal );
                if ( maxVal > max_label ) max_label = maxVal;
            }
        }

        // check images
        {
            for ( int img_id = 0; img_id < images.size(); ++img_id )
            {
                CV_Assert( images[0].size() == images[img_id].size() );
            }
        }

        // show input
        {
            char win_title[256];
            for ( int img_id = 0; img_id < images.size(); ++img_id )
            {
                CV_Assert( images[0].size() == images[img_id].size() );
                sprintf( win_title,"images[%d]", img_id );
                cv::imshow( win_title, images[img_id] );
                sprintf( win_title,"masks[%d]", img_id );
                cv::imshow( win_title, masks[img_id] );
            }
        }

        // GMM, betas
        const uint   width      = images[0].cols,
                     height     = images[0].rows,
                     num_pixels = width * height,
                     num_labels = max_label;
        const double gamma      = 50.;
        const double lambda     = 9. * gamma;
        double       *betas     = new double[ images.size() ];
        std::vector< std::vector<cv::Mat> > neighbour_weights(4);
        std::vector< std::shared_ptr<GMM2> > gmms; // one gmm for each image? //TODO: for each label AND each image

        // prepare calculation
        for ( int img_id = 0; img_id < images.size(); ++img_id )
        {
            cv::Mat const& img_ref = images[img_id];

            auto p_gmm = std::shared_ptr<GMM2>( new GMM2( models.getMatRef(img_id), img_ref.channels()) );

            if ( img_ref.type() == CV_8UC1 )
            {
                initGMM2<uchar>( img_ref, masks[img_id], img_id * 2, *p_gmm );
                betas[img_id] = calcBeta<1,uchar>( img_ref );
                calcNWeights<1,uchar>(
                            /* leftWeights: */ neighbour_weights[img_id][0],
                        /*   upLeftWeights: */ neighbour_weights[img_id][1],
                        /*       upWeights: */ neighbour_weights[img_id][2],
                        /*  upRightWeights: */ neighbour_weights[img_id][3],
                        /*     input image: */ images           [img_id]   ,
                        /*            beta: */ betas            [img_id]   ,
                        /*           gamma: */ gamma );
            }
            else if ( img_ref.type() == CV_8UC3 )
            {
                initGMM2<cv::Vec3b>( images[img_id], masks[img_id], img_id * 2, *p_gmm );
                betas[img_id] = calcBeta<3,uchar>( img_ref );
                calcNWeights<3,uchar>(
                            /* leftWeights: */ neighbour_weights[img_id][0],
                        /*   upLeftWeights: */ neighbour_weights[img_id][1],
                        /*       upWeights: */ neighbour_weights[img_id][2],
                        /*  upRightWeights: */ neighbour_weights[img_id][3],
                        /*     input image: */ images           [img_id]   ,
                        /*            beta: */ betas            [img_id]   ,
                        /*           gamma: */ gamma );
            }
            else if ( img_ref.type() == CV_16UC1 )
            {
                initGMM2<ushort>( images[img_id], masks[img_id], img_id * 2, *p_gmm );
                betas[img_id] = calcBeta<1,ushort>( img_ref );
                calcNWeights<1,ushort>(
                            /* leftWeights: */ neighbour_weights[img_id][0],
                        /*   upLeftWeights: */ neighbour_weights[img_id][1],
                        /*       upWeights: */ neighbour_weights[img_id][2],
                        /*  upRightWeights: */ neighbour_weights[img_id][3],
                        /*     input image: */ images           [img_id]   ,
                        /*            beta: */ betas            [img_id]   ,
                        /*           gamma: */ gamma );
            }
            else
            {
                std::cerr << "MultiLabelGrabCut::grabCut(): unknown image format!!! exiting..." << std::endl;
                return EXIT_FAILURE;
            }

            gmms.push_back( p_gmm );
        }


#if 0
        {
            cv::Mat compIdxs( height, width, CV_32SC1 );
            for( int i = 0; i < iterCount; i++ )
            {
                GCGraph<double> graph;

                for ( int img_id = 0; img_id < images.size(); ++img_id )
                {
                    cv::Mat const& img_ref = images[img_id];
                    std::shared_ptr<GMM2> p_gmm = gmms[ img_id ];

                    if ( img_ref.type() == CV_8UC1 )
                    {// TODO: unfinshed
                        assignGMMsComponents<1,uchar>( compIdxs, img_ref, masks[mask_id],  );
                    }
                }

                assignGMMsComponents<cn,t_img>( img, mask, bgdGMM, compIdxs );
                learnGMMs           <cn,t_img>( img, mask, compIdxs, bgdGMM, fgdGMM );
                constructGCGraph    <cn,t_img>( img, mask, bgdGMM, fgdGMM, lambda, leftW, upleftW, upW, uprightW, graph );
                estimateSegmentation( graph, mask );
            }

        }
#endif

#if 0 // GCOpt
        // work
        try
        {
            GCoptimizationGridGraph *gc = new GCoptimizationGridGraph( width, height, num_labels );

            for ( int l1 = 0; l1 < num_labels; ++l1 )
                for ( int l2 = 0; l2 < num_labels; ++l2 )
                {
                    gc->setSmoothCost( l1, l2, cost );
                }

            // first set up data costs individually
            for ( int i = 0; i < num_pixels; i++ )
                for (int l = 0; l < num_labels; ++l )
                    if (i < 25 )
                    {
                        if(  l == 0 ) gc->setDataCost(i,l,0);
                        else gc->setDataCost(i,l,10);
                    }
                    else {
                        if(  l == 5 ) gc->setDataCost(i,l,0);
                        else gc->setDataCost(i,l,10);
                    }

            SAFE_DELETE( gc );
        }
        catch ( GCException e )
        {
            e.Report();
        }
#endif
        SAFE_DELETE( betas );

        cv::waitKey();
        return EXIT_SUCCESS;

#if 0
        using namespace cv;

        Mat  img      = _img.getMat();
        Mat& mask     = _mask.getMatRef();
        Mat& bgdModel = _bgdModel.getMatRef();
        Mat& fgdModel = _fgdModel.getMatRef();

        if ( img.empty()           ) CV_Error( CV_StsBadArg, "image is empty" );
        //if ( img.type() != CV_8UC3 ) CV_Error( CV_StsBadArg, "image mush have CV_8UC3 type" );

        GMM<cn> bgdGMM( bgdModel ),
                fgdGMM( fgdModel );
        Mat compIdxs( img.size(), CV_32SC1 );

        if(    (mode == GC_INIT_WITH_RECT)
            || (mode == GC_INIT_WITH_MASK) )
        {
            if    ( mode == GC_INIT_WITH_RECT )  initMaskWithRect( mask, img.size(), rect );
            else /* mode == GC_INIT_WITH_MASK */ checkMask( img, mask );

            initGMMs<cn,t_img>( img, mask, bgdGMM, fgdGMM );
        }

        if ( iterCount <= 0) return;

        if ( mode == GC_EVAL ) checkMask( img, mask );

        const double gamma  = 50;
        const double lambda = 9*gamma;
        const double beta = calcBeta<cn,t_img>( img );

        Mat leftW, upleftW, upW, uprightW;
        calcNWeights<cn,t_img>( img, leftW, upleftW, upW, uprightW, beta, gamma );

        for( int i = 0; i < iterCount; i++ )
        {
            GCGraph<double> graph;

            assignGMMsComponents<cn,t_img>( img, mask, bgdGMM, fgdGMM, compIdxs );
            learnGMMs           <cn,t_img>( img, mask, compIdxs, bgdGMM, fgdGMM );
            constructGCGraph    <cn,t_img>( img, mask, bgdGMM, fgdGMM, lambda, leftW, upleftW, upW, uprightW, graph );
            estimateSegmentation( graph, mask );
        }
    #endif
    }

} // ns am
