#ifndef GRABCUT_HPP
#define GRABCUT_HPP

#include "MyGrabCut.h"

#include <opencv2/core/core.hpp>

#include "grabcut/GMM.h"
#include "grabcut/GMM.hpp"
#include "grabcut/gcgraph.hpp"
#include "cuboid_approx/common.h"

namespace am
{
    extern void checkMask( const cv::Mat& img, const cv::Mat& mask );
    extern void estimateSegmentation( GCGraph<double>& graph, cv::Mat& mask );
    extern void initMaskWithRect( cv::Mat& mask, cv::Size imgSize, cv::Rect rect );

    // Calculate beta - parameter of GrabCut algorithm.
    // beta = 1/(2*avg(sqr(||color[i] - color[j]||)))
    template <int cn, typename t_img>
    static double calcBeta( const cv::Mat& img )
    {
        TIC;
        double beta = 0;
        for( int y = 0; y < img.rows; y++ )
        {
            for( int x = 0; x < img.cols; x++ )
            {
                cv::Vec<double,cn> color = (cv::Vec<double,cn>)img.at<t_img>(y,x);
                if( x>0 ) // left
                {
                    cv::Vec<double,cn> diff = color - (cv::Vec<double,cn>)img.at<t_img>(y,x-1);
                    beta += diff.dot(diff);
                }
                if( y>0 && x>0 ) // upleft
                {
                    cv::Vec<double,cn> diff = color - (cv::Vec<double,cn>)img.at<t_img>(y-1,x-1);
                    beta += diff.dot(diff);
                }
                if( y>0 ) // up
                {
                    cv::Vec<double,cn> diff = color - (cv::Vec<double,cn>)img.at<t_img>(y-1,x);
                    beta += diff.dot(diff);
                }
                if( y>0 && x<img.cols-1) // upright
                {
                    cv::Vec<double,cn> diff = color - (cv::Vec<double,cn>)img.at<t_img>(y-1,x+1);
                    beta += diff.dot(diff);
                }
            }
        }
        if ( beta <= std::numeric_limits<double>::epsilon() ) beta = 0;
        else                                                  beta = 1.f / (2 * beta/(4*img.cols*img.rows - 3*img.cols - 3*img.rows + 2) );

        TOC("calcBeta");
        return beta;
    }

    // Calculate weights of noterminal vertices of graph.
    // beta and gamma - parameters of GrabCut algorithm.
    template <int cn, typename t_img>
    static void calcNWeights( const cv::Mat& img, cv::Mat& leftW, cv::Mat& upleftW, cv::Mat& upW, cv::Mat& uprightW,
                              double beta, double gamma )
    {
        TIC;

        const double gammaDivSqrt2 = gamma / std::sqrt(2.0f);
        leftW   .create( img.rows, img.cols, CV_64FC1 );
        upleftW .create( img.rows, img.cols, CV_64FC1 );
        upW     .create( img.rows, img.cols, CV_64FC1 );
        uprightW.create( img.rows, img.cols, CV_64FC1 );
        for ( int y = 0; y < img.rows; y++ )
        {
            for ( int x = 0; x < img.cols; x++ )
            {
                cv::Vec<double,cn> color = (cv::Vec<double,cn>)img.at<t_img>(y,x);
                if( x-1>=0 ) // left
                {
                    cv::Vec<double,cn> diff = color - (cv::Vec<double,cn>)img.at<t_img>(y,x-1);
                    leftW.at<double>(y,x) = gamma * exp(-beta*diff.dot(diff));
                }
                else
                    leftW.at<double>(y,x) = 0;
                if( x-1>=0 && y-1>=0 ) // upleft
                {
                    cv::Vec<double,cn> diff = color - (cv::Vec<double,cn>)img.at<t_img>(y-1,x-1);
                    upleftW.at<double>(y,x) = gammaDivSqrt2 * exp(-beta*diff.dot(diff));
                }
                else
                    upleftW.at<double>(y,x) = 0;
                if( y-1>=0 ) // up
                {
                    cv::Vec<double,cn> diff = color - (cv::Vec<double,cn>)img.at<t_img>(y-1,x);
                    upW.at<double>(y,x) = gamma * exp(-beta*diff.dot(diff));
                }
                else
                    upW.at<double>(y,x) = 0;
                if( x+1<img.cols && y-1>=0 ) // upright
                {
                    cv::Vec<double,cn> diff = color - (cv::Vec<double,cn>)img.at<t_img>(y-1,x+1);
                    uprightW.at<double>(y,x) = gammaDivSqrt2 * exp(-beta*diff.dot(diff));
                }
                else
                    uprightW.at<double>(y,x) = 0;
            }
        }

        TOC("calcNWeights")
    }

    // Initialize GMM background and foreground models using kmeans algorithm.
    template <int cn, typename t_img>
    void MyGrabCut::initGMMs( const cv::Mat& img, const cv::Mat& mask, GMM<cn>& bgdGMM, GMM<cn>& fgdGMM )
    {
        TIC;

        const int kMeansItCount = 10;
        const int kMeansType    = cv::KMEANS_PP_CENTERS;

        cv::Mat bgd_labels, fgd_labels;
        std::vector< cv::Vec<double,cn> > bgd_samples_vector, fgd_samples_vector;
        cv::Point p;
        for( p.y = 0; p.y < img.rows; ++p.y )
        {
            for( p.x = 0; p.x < img.cols; ++p.x )
            {
                cv::Vec<double,cn> px = (cv::Vec<double,cn>)img.at<t_img>(p);
                if ( ((cn == 1) && (px[0] == 0))
                     || ((cn == 4) && (px[3] < DBL_EPSILON))) continue;

                if (   (mask.at<uchar>(p) == cv::GC_BGD   )
                       || (mask.at<uchar>(p) == cv::GC_PR_BGD) )
                {
                    bgd_samples_vector.push_back( px);
                }
                else // GC_FGD | GC_PR_FGD
                {
                    fgd_samples_vector.push_back( px );
                }

            }
        }

        CV_Assert( !bgd_samples_vector.empty() && !fgd_samples_vector.empty() );

        auto bgd_future = std::async( std::launch::async, [&bgd_labels, &bgdGMM, &bgd_samples_vector]()
        {
            cv::Mat bgd_samples_mat( (int)bgd_samples_vector.size(), cn, CV_32FC1, &bgd_samples_vector[0][0] );
            cv::kmeans( bgd_samples_mat,
                        GMM<cn>::kComponentsCount,
                        bgd_labels,
                        cv::TermCriteria( CV_TERMCRIT_ITER, kMeansItCount, 0.0),
                        0,
                        kMeansType );

            bgdGMM.initLearning();
            for( int i = 0; i < (int)bgd_samples_vector.size(); i++ )
                bgdGMM.addSample( bgd_labels.at<int>(i,0), bgd_samples_vector[i] );
            bgdGMM.endLearning();
        });


        auto fgd_future = std::async( std::launch::async, [&fgd_labels, &fgdGMM, &fgd_samples_vector]()
        {
            cv::Mat fgd_samples_mat( (int)fgd_samples_vector.size(), cn, CV_32FC1, &fgd_samples_vector[0][0] );
            cv::kmeans( fgd_samples_mat,
                        GMM<cn>::kComponentsCount,
                        fgd_labels,
                        cv::TermCriteria( CV_TERMCRIT_ITER, kMeansItCount, 0.0),
                        0,
                        kMeansType );

            fgdGMM.initLearning();
            for( int i = 0; i < (int)fgd_samples_vector.size(); i++ )
                fgdGMM.addSample( fgd_labels.at<int>(i,0), fgd_samples_vector[i] );
            fgdGMM.endLearning();
        });

        // gather
        {
            bgd_future.get();
            fgd_future.get();
        }

        TOC("initGMMs")
    }

    // Assign GMMs components for each pixel.
    template <int cn, typename t_img>
    void MyGrabCut::assignGMMsComponents( cv::Mat& compIdxs, const cv::Mat& img, const cv::Mat& mask, const GMM<cn>& bgdGMM, const GMM<cn>& fgdGMM )
    {
        MyGrabCut::assignGMMsComponents<cn,t_img>( img, mask, bgdGMM, fgdGMM, compIdxs );
    }

    template <int cn, typename t_img>
    void MyGrabCut::assignGMMsComponents( const cv::Mat& img, const cv::Mat& mask, const GMM<cn>& bgdGMM, const GMM<cn>& fgdGMM, cv::Mat& compIdxs )
    {
        //cv::Point p;

        TIC;
        //std::vector<std::future<void>> futures;
        //futures.reserve(img.rows);

        for( int y = 0; y < img.rows; ++y )
        {
            //futures.push_back( std::async( std::launch::async, [y,&img,&mask,&bgdGMM,&fgdGMM,&compIdxs]()
            //{
                for( int x = 0; x < img.cols; ++x )
                {
                    cv::Vec<double,cn> color = img.at<t_img>(y,x);
                    compIdxs.at<int>(y,x) = mask.at<uchar>(y,x) == cv::GC_BGD || mask.at<uchar>(y,x) == cv::GC_PR_BGD ?
                                              bgdGMM.whichComponent(color) : fgdGMM.whichComponent(color);
                }
            //}));
        }

        //for( int y = 0; y < img.rows; ++y )
        //    futures[y].get();

        TOC("assignGMMComps");
    }

    // Learn GMMs parameters.
    template <int cn, typename t_img>
    void MyGrabCut::learnGMMs( GMM<cn>& bgdGMM, GMM<cn>& fgdGMM, const cv::Mat& img, const cv::Mat& mask, const cv::Mat& compIdxs )
    {
        MyGrabCut::learnGMMs<cn,t_img>( img, mask, compIdxs, bgdGMM, fgdGMM );
    }

    template <int cn, typename t_img>
    void MyGrabCut::learnGMMs( const cv::Mat& img, const cv::Mat& mask, const cv::Mat& compIdxs, GMM<cn>& bgdGMM, GMM<cn>& fgdGMM )
    {
        TIC;

                bgdGMM.initLearning();
        fgdGMM.initLearning();
        cv::Point p;

        for( int ci = 0; ci < GMM<cn>::kComponentsCount; ci++ )
        {
            for( p.y = 0; p.y < img.rows; p.y++ )
            {
                for( p.x = 0; p.x < img.cols; p.x++ )
                {
                    if ( compIdxs.at<int>(p) == ci )
                    {
                        cv::Vec<double,cn> px = img.at<t_img>(p);
                        if ( cn==1 && px[0] == 0 ) continue;

                        if (    (mask.at<uchar>(p) == cv::GC_BGD   )
                                || (mask.at<uchar>(p) == cv::GC_PR_BGD) )
                            bgdGMM.addSample( ci, img.at<t_img>(p) );
                        else
                            fgdGMM.addSample( ci, img.at<t_img>(p) );
                    }
                }
            }
        }
        bgdGMM.endLearning();
        fgdGMM.endLearning();

        TOC("learnGMMs")
    }

    // Construct GCGraph
    template <int cn, typename t_img>
    static void constructGCGraph( const cv::Mat& img, const cv::Mat& mask, const GMM<cn>& bgdGMM, const GMM<cn>& fgdGMM, double lambda,
                                  const cv::Mat& leftW, const cv::Mat& upleftW, const cv::Mat& upW, const cv::Mat& uprightW,
                                  GCGraph<double>& graph )
    {
        TIC;

        int vtxCount = img.cols*img.rows,
                edgeCount = 2*(4*img.cols*img.rows - 3*(img.cols + img.rows) + 2);
        graph.create( vtxCount, edgeCount );
        cv::Point p;
        for( p.y = 0; p.y < img.rows; p.y++ )
        {
            for( p.x = 0; p.x < img.cols; p.x++)
            {
                // add node
                int vtxIdx = graph.addVtx();
                t_img color = img.at<t_img>(p);

                // set t-weights
                double fromSource = 0., toSink = 0.;
                if (    (mask.at<uchar>(p) == cv::GC_PR_BGD)
                        || (mask.at<uchar>(p) == cv::GC_PR_FGD) )
                {
                    fromSource  = -log( bgdGMM(color) );
                    toSink      = -log( fgdGMM(color) );
                }
                else if( mask.at<uchar>(p) == cv::GC_BGD ) toSink = lambda;
                else                                       fromSource = lambda;

                graph.addTermWeights( vtxIdx, fromSource, toSink );

                // set n-weights
                double w;
                if ( p.x > 0 )
                {
                    w = leftW.at<double>(p);
                    graph.addEdges( vtxIdx, vtxIdx-1, w, w );

                    if ( p.y > 0 )
                    {
                        w = upleftW.at<double>(p);
                        graph.addEdges( vtxIdx, vtxIdx-img.cols-1, w, w );
                    }
                }
                if ( p.y>0 )
                {
                    w = upW.at<double>(p);
                    graph.addEdges( vtxIdx, vtxIdx-img.cols, w, w );

                    if( p.x < img.cols-1 )
                    {
                        w = uprightW.at<double>(p);
                        graph.addEdges( vtxIdx, vtxIdx-img.cols+1, w, w );
                    }
                }
            }
        }

        TOC("constructGCGraph");
    }


    template <int cn, typename t_img>
    void MyGrabCut::grabCut( cv::InputArray _img, cv::InputOutputArray _mask, cv::Rect rect,
                             cv::InputOutputArray _bgdModel, cv::InputOutputArray _fgdModel,
                             int iterCount, int mode )
    {
        using namespace cv;

        Mat  img      = _img.getMat();
        Mat& mask     = _mask.getMatRef();
        Mat& bgdModel = _bgdModel.getMatRef();
        Mat& fgdModel = _fgdModel.getMatRef();

        if ( img.empty()           ) CV_Error( CV_StsBadArg, "image is empty" );
        //if ( img.type() != CV_8UC3 ) CV_Error( CV_StsBadArg, "image mush have CV_8UC3 type" );

        GMM<cn> bgdGMM( bgdModel ),
                fgdGMM( fgdModel );
        Mat compIdxs( img.size(), CV_32SC1 );

        if(    (mode == GC_INIT_WITH_RECT)
               || (mode == GC_INIT_WITH_MASK) )
        {
            if    ( mode == GC_INIT_WITH_RECT )  initMaskWithRect( mask, img.size(), rect );
            else /* mode == GC_INIT_WITH_MASK */ checkMask( img, mask );

            initGMMs<cn,t_img>( img, mask, bgdGMM, fgdGMM );
        }

        if ( iterCount <= 0) return;

        if ( mode == GC_EVAL ) checkMask( img, mask );

        const double gamma  = 50;
        const double lambda = 9*gamma;
        const double beta = calcBeta<cn,t_img>( img );

        Mat leftW, upleftW, upW, uprightW;
        calcNWeights<cn,t_img>( img, leftW, upleftW, upW, uprightW, beta, gamma );

        for( int i = 0; i < iterCount; i++ )
        {
            GCGraph<double> graph;

            assignGMMsComponents<cn,t_img>( img, mask, bgdGMM, fgdGMM, compIdxs );
            learnGMMs           <cn,t_img>( img, mask, compIdxs, bgdGMM, fgdGMM );
            constructGCGraph    <cn,t_img>( img, mask, bgdGMM, fgdGMM, lambda, leftW, upleftW, upW, uprightW, graph );
            estimateSegmentation( graph, mask );
        }
    }


} // ns am

#endif // GRABCUT_HPP
