#include "MyGrabCut.h"
#include "MyGrabCut.hpp"

#include "cuboid_approx/util/AMUtil2.h"

namespace am
{
    using namespace my_grabcut;

    // Calculate weights of noterminal vertices of graph.
    // beta and gamma - parameters of GrabCut algorithm.
    template <typename t_img, typename t_dep>
    void MyGrabCutD::calcNWeights( std::vector<cv::Mat> const& images,
                                   cv::Mat& leftW, cv::Mat& upleftW, cv::Mat& upW, cv::Mat& uprightW,
                                   double *betas, double gamma )
    {
        cv::Mat const& img = images[IMG_RGB];
        cv::Mat const& dep = images[IMG_DEP];
        const double gammaDivSqrt2 = gamma / std::sqrt(2.);
        leftW   .create( img.rows, img.cols, CV_64FC1 ); leftW   .setTo( 0. );
        upleftW .create( img.rows, img.cols, CV_64FC1 ); upleftW .setTo( 0. );
        upW     .create( img.rows, img.cols, CV_64FC1 ); upW     .setTo( 0. );
        uprightW.create( img.rows, img.cols, CV_64FC1 ); uprightW.setTo( 0. );

        cv::Vec<double,3> diff_color;
        cv::Vec<double,1> diff_depth;
        for ( int y = 0; y < img.rows; ++y )
        {
            for ( int x = 0; x < img.cols; ++x )
            {
                cv::Vec<double,3> color = static_cast<cv::Vec<double,3>>( img.at<t_img>(y,x) );
                cv::Vec<double,1> depth = static_cast<cv::Vec<double,1>>( dep.at<t_dep>(y,x) );

                if ( x-1 >= 0 ) // left
                {
                    diff_color = color - static_cast<cv::Vec<double,3>>( img.at<t_img>(y,x-1) );
                    diff_depth = depth - static_cast<cv::Vec<double,1>>( dep.at<t_dep>(y,x-1) );

                    leftW.at<double>(y,x) = gamma * exp( -betas[IMG_RGB] * diff_color.dot(diff_color) - betas[IMG_DEP] * diff_depth.dot(diff_depth) );

                    if ( y-1 >= 0 ) // upleft
                    {
                        diff_color = color - static_cast<cv::Vec<double,3>>( img.at<t_img>(y-1,x-1) );
                        diff_depth = depth - static_cast<cv::Vec<double,1>>( dep.at<t_dep>(y-1,x-1) );

                        upleftW.at<double>(y,x) = gammaDivSqrt2 * exp( -betas[IMG_RGB] * diff_color.dot(diff_color) - betas[IMG_DEP] * diff_depth.dot(diff_depth) );
                    }
                }

                if ( y-1 >= 0 ) // up
                {
                    diff_color = color - static_cast<cv::Vec<double,3>>( img.at<t_img>(y-1,x) );
                    diff_depth = depth - static_cast<cv::Vec<double,1>>( dep.at<t_dep>(y-1,x) );

                    upW.at<double>(y,x) = gamma * exp( -betas[IMG_RGB] * diff_color.dot(diff_color) - betas[IMG_DEP] * diff_depth.dot(diff_depth) );

                    if ( x+1 < img.cols ) // upright
                    {
                        diff_color = color - static_cast<cv::Vec<double,3>>( img.at<t_img>(y-1,x+1) );
                        diff_depth = depth - static_cast<cv::Vec<double,1>>( dep.at<t_dep>(y-1,x+1) );

                        uprightW.at<double>(y,x) = gammaDivSqrt2 * exp( -betas[IMG_RGB] * diff_color.dot(diff_color) - betas[IMG_DEP] * diff_depth.dot(diff_depth) );
                    }
                }
            }
        }
    }

    // Construct GCGraph for img + depth
    template <typename t_img, typename t_dep>
    void MyGrabCutD::constructGCGraph(const std::vector<cv::Mat>                  &images,
                                       const cv::Mat                               &mask,
                                       const std::vector<IGMM*>                    &gmms,
                                       double                                       lambda,
                                       const std::vector< std::vector< cv::Mat > > &weights,
                                       GCGraph<double>                             &graph,
                                       params_struct *params )
    {
        TIC;

        if ( weights.size() != 1 )
        {
            std::cerr << "#warning MyGrabCut::constructGCGraph(): currently expecting weights to be aggregated in calcNWeights, so only weights[0][0..4] will be used\n";
        }

        cv::Mat const& img = images[IMG_RGB];
        int vtxCount = img.cols * img.rows,
            edgeCount = 2*(4*img.cols*img.rows - 3*(img.cols + img.rows) + 2);
        graph.create( vtxCount, edgeCount );

        // betas_coeffs
        //if ( )
        //std::vector<std::future<void>> futures;
        //futures.reserve(images[IMG_RGB].rows);

        cv::Point p;
        for( p.y = 0; p.y < images[IMG_RGB].rows; p.y++ )
        {
            //futures.push_back( std::async( std::launch::async, [y,&images,&graph,&mask,&gmms,&lambda,&weights,&img]() {
            //cv::Point p(0,y);
            for( p.x = 0; p.x < images[IMG_RGB].cols; p.x++)
            {
                // add node
                int vtxIdx = graph.addVtx();
                t_img color = images[IMG_RGB].at<t_img>( p );
                t_dep depth = images[IMG_DEP].at<t_dep>( p );

                // set t-weights
                double fromSource = 0., toSink = 0.;
                if (   (mask.at<uchar>(p) == cv::GC_PR_BGD)
                    || (mask.at<uchar>(p) == cv::GC_PR_FGD) )
                {
                    fromSource  = -log( gmms[GMM_RGB_BG]->AsGMM3Pointer()->operator()(color) ) - log( gmms[GMM_DEP_BG]->AsGMM1Pointer()->operator()(depth) );
                    toSink      = -log( gmms[GMM_RGB_FG]->AsGMM3Pointer()->operator()(color) ) - log( gmms[GMM_DEP_FG]->AsGMM1Pointer()->operator()(depth) );
                    //fromSource  = std::max( -log( gmms[GMM_RGB_BG]->AsGMM3Pointer()->operator()(color) ), - log( gmms[GMM_DEP_BG]->AsGMM1Pointer()->operator()(depth) ) );
                    //toSink      = std::max( -log( gmms[GMM_RGB_FG]->AsGMM3Pointer()->operator()(color) ), - log( gmms[GMM_DEP_FG]->AsGMM1Pointer()->operator()(depth) ) );
                }
                else if( mask.at<uchar>(p) == cv::GC_BGD ) toSink     = lambda;
                else                                       fromSource = lambda;

                graph.addTermWeights( vtxIdx, fromSource, toSink );

                // set n-weights
                double w;
                if ( p.x > 0 )
                {
                    w = weights[IMG_RGB][WEIGHTS_LEFT].at<double>(p);
                    graph.addEdges( vtxIdx, vtxIdx-1, w, w );

                    if ( p.y > 0 )
                    {
                        w = weights[IMG_RGB][WEIGHTS_UPLEFT].at<double>(p);
                        graph.addEdges( vtxIdx, vtxIdx-img.cols-1, w, w );
                    }
                }
                if ( p.y > 0 )
                {
                    w = weights[IMG_RGB][WEIGHTS_UP].at<double>(p);
                    graph.addEdges( vtxIdx, vtxIdx-img.cols, w, w );

                    if( p.x < img.cols-1 )
                    {
                        w = weights[IMG_RGB][WEIGHTS_UPRIGHT].at<double>(p);
                        graph.addEdges( vtxIdx, vtxIdx-img.cols+1, w, w );
                    }
                }
            }
            //})); // futures.push_back
        } // for y

        // gather
        //for( int y = 0; y < images[IMG_RGB].rows; y++ )
        //{
        //    futures[y].get();
        //} // for y

        TOC("constructGCGraph");
    }

    template <typename t_img, typename t_dep>
    void MyGrabCutD::grabCutD( cv::InputArray       const& images_arg,
                               cv::InputOutputArray const& mask_arg,
                               cv::Rect                    rect,
                               cv::InputOutputArray        &models_arg,
                               int iterCount, int mode,
                               params_struct *params )
    {
        using namespace cv;

        if ( images_arg.total() != 2 )
        {
            std::cerr << "MyGrabCut::grabCutD(): expects images_arg to contain two images (rgb, d)...exiting" << std::cout;
            return;
        }
        if ( models_arg.total() != 4 )
        {
            std::cerr << "MyGrabCut::grabCutD(): expects models_arg to contain { BG_RGB_MDL, FG_RGB_MDL, BG_DEP_GMM, FG_DEP_GMM }...exiting...\n";
            return;
        }

        // parse input
        std::vector<cv::Mat> images;
        images_arg.getMatVector( images );
        std::vector<cv::Mat> models;
        models_arg.getMatVector( models );
        Mat const&  img      = images[IMG_RGB];
        Mat const&  dep      = images[IMG_DEP];
        Mat         mask_out = mask_arg.getMatRef().clone();

        if ( img.empty() || dep.empty() ) CV_Error( CV_StsBadArg, "image or dep is empty" );
        if ( img.type () != CV_8UC3     ) CV_Error( CV_StsBadArg, "image must have CV_8UC3 type" );

        double beta_coeffs[2] = { 1., 1. };
        if ( params )
        {
            beta_coeffs[0] = params->beta_coeffs[0];
            beta_coeffs[1] = params->beta_coeffs[1];
        }


        std::vector<IGMM*> gmms = { new GMM<3>(models[GMM_RGB_BG]), new GMM<3>(models[GMM_RGB_FG]),
                                    new GMM<1>(models[GMM_DEP_BG]), new GMM<1>(models[GMM_DEP_FG]) };
        std::vector<cv::Mat> component_indices = { cv::Mat(img.size(), CV_32SC1), cv::Mat(dep.size(), CV_32SC1) };

        if(    (mode == GC_INIT_WITH_RECT)
            || (mode == GC_INIT_WITH_MASK) )
        {
            if    ( mode == GC_INIT_WITH_RECT )  initMaskWithRect( mask_out, img.size(), rect );
            else /* mode == GC_INIT_WITH_MASK */ checkMask( img, mask_out );

            // map
            auto init3_future = std::async( std::launch::async, [&images,&mask_out,&gmms]() {
                MyGrabCut::initGMMs<3,t_img>( images[IMG_RGB], mask_out, *reinterpret_cast<GMM<3>*>(gmms[GMM_RGB_BG]), *reinterpret_cast<GMM<3>*>(gmms[GMM_RGB_FG]) );
            });
            auto init1_future = std::async( std::launch::async, [&images,&mask_out,&gmms]() {
                MyGrabCut::initGMMs<1,t_dep>( images[IMG_DEP], mask_out, *reinterpret_cast<GMM<1>*>(gmms[GMM_DEP_BG]), *reinterpret_cast<GMM<1>*>(gmms[GMM_DEP_FG]) );
            });

            // gather
            init3_future.get();
            init1_future.get();
        }

        if ( iterCount <= 0) return;

        if ( mode == GC_EVAL ) checkMask( img, mask_out );

        const double gamma  = 50;
        const double lambda = 9*gamma;
        double betas[2] = { calcBeta<3,t_img>(images[IMG_RGB]) * beta_coeffs[0],
                            calcBeta<1,t_dep>(images[IMG_DEP]) * beta_coeffs[1] };

        std::vector< std::vector< cv::Mat > > weights(1);
        weights[IMG_RGB].resize(4); // img weights
        //weights[IMG_DEP].resize(4); // dep weights
        MyGrabCutD::calcNWeights<t_img, t_dep>( images,
                                                weights[IMG_RGB][WEIGHTS_LEFT],
                                                weights[IMG_RGB][WEIGHTS_UPLEFT],
                                                weights[IMG_RGB][WEIGHTS_UP],
                                                weights[IMG_RGB][WEIGHTS_UPRIGHT],
                                                betas, gamma );
        //calcNWeights<1,t_dep>( images[IMG_DEP], weights[WEIGHTS_LEFT], weights[WEIGHTS_UPLEFT], weights[WEIGHTS_UP], weights[WEIGHTS_UPRIGHT], betas[IMG_RGB], gamma );

        for ( int i = 0; i < iterCount; ++i )
        {
            // map
            auto assign3_future = std::async( std::launch::async, [&component_indices,&img,&mask_out,&gmms]() {
                MyGrabCut::assignGMMsComponents<3,t_img>( component_indices[IMG_RGB],
                                                          img, mask_out,
                                                          *reinterpret_cast<GMM<3>*>(gmms[GMM_RGB_BG]),
                                                          *reinterpret_cast<GMM<3>*>(gmms[GMM_RGB_FG]) );
            });
            auto assign1_future = std::async( std::launch::async, [&component_indices,&dep,&mask_out,&gmms]() {
                MyGrabCut::assignGMMsComponents<1,t_dep>( component_indices[IMG_DEP],
                                                          dep, mask_out,
                                                          *reinterpret_cast<GMM<1>*>(gmms[GMM_DEP_BG]),
                                                          *reinterpret_cast<GMM<1>*>(gmms[GMM_DEP_FG]) );
            });

            // gather
            {
                assign3_future.get();
                assign1_future.get();
            }

            // map independent tasks to async exec
            auto learnGMMs3_future = std::async( std::launch::async, [&component_indices,&img,&dep,&mask_out,&gmms]() {
                MyGrabCut::learnGMMs<3,t_img>( *reinterpret_cast<GMM<3>*>(gmms[GMM_RGB_BG]),
                                               *reinterpret_cast<GMM<3>*>(gmms[GMM_RGB_FG]),
                                               img, mask_out, component_indices[IMG_RGB] );
            });

            auto learnGMMs1_future = std::async( std::launch::async, [&component_indices,&dep,&mask_out,&gmms]() {
                MyGrabCut::learnGMMs<1,t_dep>( *reinterpret_cast<GMM<1>*>(gmms[GMM_DEP_BG]),
                                               *reinterpret_cast<GMM<1>*>(gmms[GMM_DEP_FG]),
                                               dep, mask_out, component_indices[IMG_DEP] );
            });

            // gather
            {
                learnGMMs3_future.get();
                learnGMMs1_future.get();
            }

            GCGraph<double> graph;
            MyGrabCutD::constructGCGraph<t_img,t_dep>( images, mask_out, gmms, lambda, weights, graph, params );
            estimateSegmentation( graph, mask_out );
        }

        // floodfill ouptut
        {
            // cache
            const cv::Mat& mask_ref = mask_arg.getMatRef();

            // find a point labelled foreground (by hand)
            cv::Point seed(-1,-1);
            {
                for ( unsigned y = 0; (y < mask_ref.rows) && (seed.x < 0); ++y  )
                {
                    for ( unsigned x = 0; (x < mask_ref.cols) && (seed.x < 0); ++x )
                    {
                        if ( mask_ref.at<uchar>(y,x) == GC_FGD )
                        {
                            seed.x = x;
                            seed.y = y;
                        }
                    }
                }
            }

            // floodfill input
            cv::Mat mask_copy = mask_out.clone();
            {
                // remap to FG, PRFG = as_is, BG, PR_BG = 200
                am::util::cv::replaceValue( mask_copy, GC_PR_BGD, 200 );
                am::util::cv::replaceValue( mask_copy, GC_BGD,    200 );
            }

            // prepare floodfill output
            cv::Mat floodfill_out_mask( mask_copy.rows + 2, mask_copy.cols + 2, CV_8UC1 );
            {
                floodfill_out_mask.setTo( 0 );
            }

            // floodfill
            {
                cv::floodFill( mask_copy, floodfill_out_mask, seed, cv::Scalar(GC_PR_FGD), 0, cv::Scalar(2), cv::Scalar(2), 8 );
            }

            // return output
            {
                cv::Mat cropped_mask = floodfill_out_mask(cv::Range(1,mask_out.rows), cv::Range(1,mask_out.cols));
                mask_arg.getMatRef().setTo( GC_PR_FGD, floodfill_out_mask(cv::Range(1,mask_out.rows+1), cv::Range(1,mask_out.cols+1)) );
                //mask.copyTo( mask_arg.getMatRef(), floodfill_out_mask );
            }
        }

        for ( IGMM*& gmm_ptr : gmms )
        {
            SAFE_DELETE( gmm_ptr );
        }
    }

    template <typename t_img, typename t_dep>
    void MyGrabCutD::grabCutD4( cv::InputArray       const& images_arg,
                                cv::InputOutputArray const& mask_arg,
                                cv::Rect                    rect,
                                cv::InputOutputArray        &models_arg,
                                int iterCount, int mode,
                                params_struct *params )
    {
        using namespace cv;

        if ( images_arg.total() != 2 )
        {
            std::cerr << "MyGrabCut::grabCutD4(): expects images_arg to contain two images (rgb, d)...exiting" << std::cout;
            return;
        }
        if ( models_arg.total() != 2 )
        {
            std::cerr << "MyGrabCut::grabCutD4(): expects models_arg to contain { BG_MDL, FG_MDL }...exiting...\n";
            return;
        }

        // parse input
        std::vector<cv::Mat> images;
        images_arg.getMatVector( images );
        std::vector<cv::Mat> models;
        models_arg.getMatVector( models );
        Mat const&  img      = images[IMG_RGB];
        Mat const&  dep      = images[IMG_DEP];
        Mat         mask_out = mask_arg.getMatRef().clone();

        if ( img.empty() || dep.empty() ) CV_Error( CV_StsBadArg, "image or dep is empty" );
        if ( img.type () != CV_8UC3     ) CV_Error( CV_StsBadArg, "image must have CV_8UC3 type" );

        cv::Mat img_normed, dep_normed;
        cv::normalize( img, img_normed, 0.f, 1.f, NORM_MINMAX, CV_32F );
        cv::imshow( "img_normed", img_normed );
        cv::normalize( dep, dep_normed, 0.f, 1.f, NORM_MINMAX, CV_32F );
        cv::imshow( "dep_normed", dep_normed );
        //cv::waitKey(10);

        Mat img4( img.rows, img.cols, CV_32FC4 );
        for ( unsigned y = 0; y < img4.rows; ++y  )
        {
            for ( unsigned x = 0; x < img4.cols; ++x )
            {
                cv::Vec4f& px = img4.at<cv::Vec4f>( y, x );
                cv::Vec3f rgb = img_normed.at<cv::Vec3f>(y, x);

//                std::cout << "img.at<vec3b>(" << y << ", " << x << "): "
//                          << (int)img.at<t_img>(y,x)[0] << ", "
//                          << (int)img.at<t_img>(y,x)[1] << ", "
//                          << (int)img.at<t_img>(y,x)[2] << ", "
//                          << (int)dep.at<t_dep>(y,x) << "\tvs\t";

                for ( int chan = 0; chan < rgb.rows; ++chan )
                {
                    px[ chan ] = static_cast<float>( rgb[chan] );
//                    std::cout << px[chan] << ", ";
                }
                px[3] = static_cast<float>( dep_normed.at<float>(y, x) );
//                std::cout << px[3] << std::endl;
            }
        }


        double beta_coeffs[2] = { 1., 1. };
        if ( params )
        {
            beta_coeffs[0] = params->beta_coeffs[0];
            beta_coeffs[1] = params->beta_coeffs[1];
        }


        std::vector<IGMM*> gmms = { new GMM<4>(models[GMM_RGB_BG]),
                                    new GMM<4>(models[GMM_RGB_FG])  };
        // TODO: this far
        std::vector<cv::Mat> component_indices = { cv::Mat(img.size(), CV_32SC1) };

        if(    (mode == GC_INIT_WITH_RECT)
            || (mode == GC_INIT_WITH_MASK) )
        {
            if    ( mode == GC_INIT_WITH_RECT )  initMaskWithRect( mask_out, img.size(), rect );
            else /* mode == GC_INIT_WITH_MASK */ checkMask( img, mask_out );

            MyGrabCut::initGMMs<4,cv::Vec4f>( img4, mask_out,
                                              *reinterpret_cast<GMM<4>*>(gmms[GMM_RGB_BG]),
                                              *reinterpret_cast<GMM<4>*>(gmms[GMM_RGB_FG]) );
        }

        if ( iterCount <= 0) return;

        if ( mode == GC_EVAL ) checkMask( img, mask_out );

        const double gamma  = 50;
        const double lambda = 9*gamma;
        double betas[2] = { calcBeta<3,float>(img_normed) * beta_coeffs[0],
                            calcBeta<1,float>(dep_normed) * beta_coeffs[1] };

        std::vector< std::vector< cv::Mat > > weights(1);
        weights[IMG_RGB].resize(4); // img weights
        //weights[IMG_DEP].resize(4); // dep weights
        MyGrabCutD::calcNWeights<float, float>( {img_normed, dep_normed},
                                                weights[IMG_RGB][WEIGHTS_LEFT],
                                                weights[IMG_RGB][WEIGHTS_UPLEFT],
                                                weights[IMG_RGB][WEIGHTS_UP],
                                                weights[IMG_RGB][WEIGHTS_UPRIGHT],
                                                betas, gamma );
        //calcNWeights<1,t_dep>( images[IMG_DEP], weights[WEIGHTS_LEFT], weights[WEIGHTS_UPLEFT], weights[WEIGHTS_UP], weights[WEIGHTS_UPRIGHT], betas[IMG_RGB], gamma );

        for ( int i = 0; i < iterCount; ++i )
        {
                MyGrabCut::assignGMMsComponents<4,cv::Vec4f>( component_indices[IMG_RGB],
                                                          img4, mask_out,
                                                          *reinterpret_cast<GMM<4>*>(gmms[GMM_RGB_BG]),
                                                          *reinterpret_cast<GMM<4>*>(gmms[GMM_RGB_FG]) );

            // map independent tasks to async exec
                MyGrabCut::learnGMMs<4,cv::Vec4f>( *reinterpret_cast<GMM<4>*>(gmms[GMM_RGB_BG]),
                                                   *reinterpret_cast<GMM<4>*>(gmms[GMM_RGB_FG]),
                                                   img4, mask_out, component_indices[IMG_RGB] );

            GCGraph<double> graph;
            MyGrabCutD::constructGCGraph4<cv::Vec4f>( {img4}, mask_out, gmms, lambda, weights, graph, params );
            estimateSegmentation( graph, mask_out );
        }

        // floodfill ouptut
        {
            // cache
            const cv::Mat& mask_ref = mask_arg.getMatRef();

            // find a point labelled foreground (by hand)
            cv::Point seed(-1,-1);
            {
                for ( unsigned y = 0; (y < mask_ref.rows) && (seed.x < 0); ++y  )
                {
                    for ( unsigned x = 0; (x < mask_ref.cols) && (seed.x < 0); ++x )
                    {
                        if ( mask_ref.at<uchar>(y,x) == GC_FGD )
                        {
                            seed.x = x;
                            seed.y = y;
                        }
                    }
                }
            }

            // floodfill input
            cv::Mat mask_copy = mask_out.clone();
            {
                // remap to FG, PRFG = as_is, BG, PR_BG = 200
                am::util::cv::replaceValue( mask_copy, GC_PR_BGD, 200 );
                am::util::cv::replaceValue( mask_copy, GC_BGD,    200 );
            }

            // prepare floodfill output
            cv::Mat floodfill_out_mask( mask_copy.rows + 2, mask_copy.cols + 2, CV_8UC1 );
            {
                floodfill_out_mask.setTo( 0 );
            }

            // floodfill
            {
                cv::floodFill( mask_copy, floodfill_out_mask, seed, cv::Scalar(GC_PR_FGD), 0, cv::Scalar(2), cv::Scalar(2), 8 );
            }

            // return output
            {
                cv::Mat cropped_mask = floodfill_out_mask(cv::Range(1,mask_out.rows), cv::Range(1,mask_out.cols));
                mask_arg.getMatRef().setTo( GC_PR_FGD, floodfill_out_mask(cv::Range(1,mask_out.rows+1), cv::Range(1,mask_out.cols+1)) );
                //mask.copyTo( mask_arg.getMatRef(), floodfill_out_mask );
            }
        }

        for ( IGMM*& gmm_ptr : gmms )
        {
            SAFE_DELETE( gmm_ptr );
        }
    }

    // Construct GCGraph for img + depth
    template <typename t_img>
    void MyGrabCutD::constructGCGraph4(const std::vector<cv::Mat>                  &images,
                                       const cv::Mat                               &mask,
                                       const std::vector<IGMM*>                    &gmms,
                                       double                                       lambda,
                                       const std::vector< std::vector< cv::Mat > > &weights,
                                       GCGraph<double>                             &graph,
                                       params_struct *params )
    {
        TIC;

        if ( weights.size() != 1 )
        {
            std::cerr << "#warning MyGrabCut::constructGCGraph(): currently expecting weights to be aggregated in calcNWeights, so only weights[0][0..4] will be used\n";
        }

        cv::Mat const& img = images[IMG_RGB];
        int vtxCount = img.cols * img.rows,
            edgeCount = 2*(4*img.cols*img.rows - 3*(img.cols + img.rows) + 2);
        graph.create( vtxCount, edgeCount );

        // betas_coeffs
        //if ( )
        //std::vector<std::future<void>> futures;
        //futures.reserve(images[IMG_RGB].rows);

        cv::Point p;
        for( p.y = 0; p.y < images[IMG_RGB].rows; ++p.y )
        {
            //futures.push_back( std::async( std::launch::async, [y,&images,&graph,&mask,&gmms,&lambda,&weights,&img]() {
            //cv::Point p(0,y);
            for( p.x = 0; p.x < images[IMG_RGB].cols; ++p.x )
            {
                // add node
                int vtxIdx = graph.addVtx();
                t_img color = images[IMG_RGB].at<t_img>( p );

                // set t-weights
                double fromSource = 0., toSink = 0.;
                if (   (mask.at<uchar>(p) == cv::GC_PR_BGD)
                    || (mask.at<uchar>(p) == cv::GC_PR_FGD) )
                {
                    fromSource  = -log( gmms[GMM_RGB_BG]->AsGMM4Pointer()->operator()(color) );
                    toSink      = -log( gmms[GMM_RGB_FG]->AsGMM4Pointer()->operator()(color) );
                    //fromSource  = std::max( -log( gmms[GMM_RGB_BG]->AsGMM3Pointer()->operator()(color) ), - log( gmms[GMM_DEP_BG]->AsGMM1Pointer()->operator()(depth) ) );
                    //toSink      = std::max( -log( gmms[GMM_RGB_FG]->AsGMM3Pointer()->operator()(color) ), - log( gmms[GMM_DEP_FG]->AsGMM1Pointer()->operator()(depth) ) );
                }
                else if( mask.at<uchar>(p) == cv::GC_BGD ) toSink     = lambda;
                else                                       fromSource = lambda;

                graph.addTermWeights( vtxIdx, fromSource, toSink );

                // set n-weights
                double w;
                if ( p.x > 0 )
                {
                    w = weights[IMG_RGB][WEIGHTS_LEFT].at<double>(p);
                    graph.addEdges( vtxIdx, vtxIdx-1, w, w );

                    if ( p.y > 0 )
                    {
                        w = weights[IMG_RGB][WEIGHTS_UPLEFT].at<double>(p);
                        graph.addEdges( vtxIdx, vtxIdx-img.cols-1, w, w );
                    }
                }
                if ( p.y > 0 )
                {
                    w = weights[IMG_RGB][WEIGHTS_UP].at<double>(p);
                    graph.addEdges( vtxIdx, vtxIdx-img.cols, w, w );

                    if( p.x < img.cols-1 )
                    {
                        w = weights[IMG_RGB][WEIGHTS_UPRIGHT].at<double>(p);
                        graph.addEdges( vtxIdx, vtxIdx-img.cols+1, w, w );
                    }
                }
            }
            //})); // futures.push_back
        } // for y

        // gather
        //for( int y = 0; y < images[IMG_RGB].rows; y++ )
        //{
        //    futures[y].get();
        //} // for y

        TOC("constructGCGraph4");
    }


} // ns am
