#ifndef __CA_CUBOID_H__
#define __CA_CUBOID_H__

#include "Plane.h"
#include "Planes.h"

#include "PrimitiveContainer.h"
#include "Primitive.h"

#include "pcl/visualization/pcl_visualizer.h"
#include <Eigen/Dense>
#include <vector>

#define USE_WM5 1

#if USE_WM5
//class BOX3_F;
namespace Wm5
{
    template <typename T>
    class Box3;
}
#endif

namespace am
{

    class Cuboid : public Primitive
    {
        public:
            enum TYPE { CUBOID, FLOOR, WALL, SPHERE, CYLINDER };
            typedef std::map<int,am::Cuboid::TYPE> TypeMap;
            typedef float Scalar;

            Cuboid operator=( const Cuboid& other );

            /// STATIC METHODS
            static int
            from_cloud(Cuboid                        &cuboid,
                        PCloud::ConstPtr             cloud_ptr,
                        pcl::PointIndices::Ptr const &indices_ptr,
                        double                       distanceThreshold = .02,
                        bool                         debug             = false );

            template <typename PointT> static int
            fromCloudWithPCA( am::Cuboid &cuboid_out
                              , boost::shared_ptr<pcl::PointCloud<PointT> > cloud
                              , boost::shared_ptr<pcl::PointIndices > indices_arg = nullptr
                              , int   const K                                     = 15
                              , float const radius                                = -1.f
                              , int   const filter_nmeans                         = 50
                              , float const filter_stddev                         = .1
                              , float const density_cap_median_multiplier         = 1.5f
                              , bool  const debug                                 = false );

            template <typename CloudPtrT> static int
            fromCloudWithPCA2( Cuboid                             &cuboid_out
                               , CloudPtrT                         cloud
                               , pcl::PointIndices::ConstPtr       selection
                               , pcl::PointIndices::ConstPtr       indices_arg
                               , float                       const offset = -FLT_EPSILON );

            static int
            fromCorners(am::Cuboid                   &cuboid,
                         std::vector<float>          corner_coords,
                         PCloud::ConstPtr            cloud_ptr,
                         pcl::PointIndices::ConstPtr indices_ptr
                        , bool debug = false );

            static int
            fromFrame( am::Cuboid &cuboid
                       , Eigen::Matrix4f const& frame
                       , PCloud::ConstPtr cloud
                       , pcl::PointIndices::ConstPtr indices );

            static int
            from1Plane( am::Cuboid &cuboid
                        , am::Plane const& plane
                        , float assumed_depth = .05f);

            static int
            calculateCorners( std::vector<pcl::PointXYZ> &corners,
                              MyPoint                    &min_pt,
                              MyPoint                    &max_pt,
                              Cuboid               const &cuboid );

            static int
            writeToFile( Cuboid      const &cuboid,
                         std::string        path,
                         bool               append = false );
            static int
            writeToTextFile( Cuboid      const &cuboid,
                             std::string        path,
                             bool               append = false );

            static int
            reassignPlanes( am::Planes &planes,
                            PCloud::ConstPtr cloud_ptr,
                            pcl::PointIndices::ConstPtr indices_ptr,
                            double distanceThreshold );

            static int
            clampPointsToOrientedPlanes( Cuboid          & out_cuboid,
                                         Cuboid     const& in_cuboid,
                                         am::Planes const& planes,
                                         double     const  distanceThreshold = -FLT_EPSILON );

            static int
            calculateCentroidFromCorners( am::Cuboid &cuboid );

            static int
            calculateFrameFromPlanes( am::Cuboid &cuboid, bool debug = false );

            static std::string
            cuboidName( int cube_id );

            static std::string
            cuboidLineName( int cube_id, int line_id );

            static int
            verify( am::Cuboid const& cuboid );

            virtual int
            unDraw( pcl::visualization::PCLVisualizer::Ptr v,
                    std::string                            cloud_name ) override;

            static int
            draw( Cuboid                                 const &cuboid,
                  pcl::visualization::PCLVisualizer::Ptr        visualizer,
                  std::string                                   cloud_name,
                  Eigen::Vector3f                               colour          = {.9,.9,.0},
                  double                                        line_width      = 8.
                  , bool                                        colour_edges    = false
                  , bool                                        draw_name       = false       );

            static int
            draw4(Cuboid                                 const &cuboid,
                  pcl::visualization::PCLVisualizer::Ptr       v,
                  std::string                                  cloud_name,
                  Eigen::Vector4f                              colour       = { .9f, .9f, .0f, .15f }
                  , double                                     line_width   = 8.
                  , bool                                       colour_edges = false
                  , bool                                       draw_name    = false
                  , TYPE                                 const shape        = Cuboid::CUBOID );

            /// CONSTR
            int
            cloneTo( Cuboid & cuboid ) const;

            /// METHODS
            int
            draw( pcl::visualization::PCLVisualizer::Ptr visualizer
                  //std::string                            cloud_name = "Cuboid",
                  , std::string                          cloud_name
                  , Eigen::Vector3f                      colour       = {.9, .9, 0.}
                  , double                               line_width   = 8.
                  , bool                                 colour_edges = false
                  , bool                                 draw_name    = false       ) const;

            int
            draw4( pcl::visualization::PCLVisualizer::Ptr visualizer,
                   std::string                            cloud_name,
                   Eigen::Vector4f                        colour       = {.9f, .9f, 0.f, .15f},
                   double                                 line_width   = 8.
                   , bool                                 colour_edges = false
                   , bool                                 draw_name    = false
                   , Cuboid::TYPE                         shape        = Cuboid::CUBOID ) const;

            int
            drawPlanes                      ( pcl::visualization::PCLVisualizer::Ptr visualizer
                                            , std::string                            cloud_name ) const;

            int
            getInlierIndices                ( pcl::PointIndices::Ptr               &indices_ptr
                                            , PCloud::ConstPtr                      cloud_ptr
                                            , pcl::PointIndices::ConstPtr           in_indices_ptr
                                            , float                                 outside_distance_threshold    = .0f
                                            , bool                          const   debug                         = false ) const;

            std::vector<double>
            getSideLengths                  () const;

            Eigen::Vector3f
            getExtent                       () const; // returns 3 dim size using _center, _corners and _R

            int
            moveAlongFrame                  ( Eigen::Vector3f offset );
            int
            resizeAlongFrame                ( Eigen::Vector3f offset );
            int
            rotateAroundFrameAxis           ( Eigen::Vector3f angles );

            int
            getMinMaxCorner                 ( pcl::PointXYZ &mn, pcl::PointXYZ &mx );

            Eigen::Matrix<Scalar,3,1>
            getOutwardsNormalOfLargestPlane () const;

            int
            writeLabelsToFile               ( std::string dat_path
                                            , int         width   ) const;

#if USE_WM5
            static int
            toBox3                          ( std::shared_ptr<Wm5::Box3<float>>       &box
                                            , am::Cuboid                        const &cub
                                            , float                             const  offset = 0.f );

            static bool
            intersect(am::Cuboid const& a, am::Cuboid const& b , float offset = 0.f);
#endif
            /// SET, GET
            // _center
            Eigen::Vector3f const&            Center() const;
            Eigen::Vector3f      &            CenterRef();

            // _R
            Eigen::Matrix3f const&            R() const;
            Eigen::Matrix3f      &            RRef();

            // _planes
            am::Planes      const&            Planes() const;
            am::Planes           &            PlanesRef();

            // _corners
            std::vector<pcl::PointXYZ> const& Corners()       const;
            std::vector<pcl::PointXYZ>      & Corners();


            static inline std::string typeToString( TYPE type )
            {
                switch ( type )
                {
                    case TYPE::SPHERE:
                        return "sphere"; break;
                    case TYPE::CYLINDER:
                        return "cylinder"; break;
                    case TYPE::CUBOID:
                    default:
                        return "cuboid"; break;
                }

                return "cuboid";
            }

#if 0
            static int
            from_vertex_list( Cuboid                 &cuboid,
                              Eigen::Matrix3Xf const &vertex_list );

            static int
            from_cloud_old( Cuboid                       &cuboid,
                        PCloud::ConstPtr             cloud_ptr,
                        pcl::PointIndices::Ptr const &indices_ptr );
#endif
        protected:
            Eigen::Vector3f            _center;  // center of mass
            Eigen::Matrix3f            _R;       // column-major eigen vectors spanning the cuboid
            am::Planes                 _planes;  // planes of the cuboid's ROI cloud
            std::vector<pcl::PointXYZ> _corners;

    };
} // ns am

#ifndef __CA_CUBOID_HPP_INC__
#   define __CA_CUBOID_HPP_INC__
#   include "primitives/Cuboid.hpp"
#endif

#endif // __CA_CUBOID_H__
