#ifndef PLANE_H
#define PLANE_H

#include <ostream>

#include <Eigen/Dense>

#include "pcl/visualization/pcl_visualizer.h"

#include "cuboid_approx/common.h"

//#include "PrimitiveContainer.h"
#include "primitives/Primitive.h"

namespace am
{
    class Cuboid;

    struct Histogram
    {
            Histogram(int N)
                : histogram(N)
                , mins(N)
                , maxs(N)
                //, members(N)
            {
                std::fill( histogram.begin(), histogram.end(), 0 );
                std::fill( mins.begin(), mins.end(), DBL_MAX );
                std::fill( maxs.begin(), maxs.end(), -DBL_MAX );
            }

            int
            getMax( int *val, int *pos = nullptr ) const
            {
                auto it_max_pos = std::max_element( histogram.begin(), histogram.end() );
                if ( val )
                    *val = *it_max_pos;
                if ( pos )
                    *pos = std::distance( histogram.begin(), it_max_pos );
            }

            std::vector<int> histogram;
            std::vector<double> mins;
            std::vector<double> maxs;
            //std::vector<std::vector<int> > members;
    };

    class Plane : public Primitive
    {
        public:
            typedef std::shared_ptr<      Plane>      Ptr;
            typedef std::shared_ptr<const Plane> ConstPtr;

            /// STATIC METHODS
            static int
            fromCloud( Plane                        &plane,
                       PCloud::ConstPtr             cloud_ptr,
                       pcl::PointIndices::Ptr const indices_ptr,
                       double                       distanceThreshold = .1 );

            static int
            draw(am::Plane const& plane
                  , pcl::visualization::PCLVisualizer::Ptr v
                  , std::string plane_name = "plane0"
                  , double r        = .2
                  , double g        = .0
                  , double b        = .0
                  , bool drawPlane  = true
                  , bool drawNormal = false
                  , bool drawFrame  = false
                  , int viewport_id = 0
                  );

            static int
            draw( am::Plane const& plane
                  , pcl::visualization::PCLVisualizer::Ptr v
                  , int plane_id    = 0
                  , double r        = .2
                  , double g        = .0
                  , double b        = .0
                  , bool drawPlane  = true
                  , bool drawNormal = false
                  , bool drawFrame  = false
                  , int viewport_id = 0
                  );

            // writes plane as 4 floats
            static int
            writeToFile( Plane const& plane, std::string path, bool append );

            static int
            checkNormalFlip(Plane &plane,
                             PCloud::ConstPtr cloud_ptr,
                             pcl::PointIndices::ConstPtr indices_ptr,
                             double distanceThreshold = .1 );

            static int
            processHistogram( std::vector<double>              &offsets,
                              std::vector<am::Histogram> const &histograms,
                              double                            threshold = .2 );

            static int
            fromCorners( am::Plane                        &plane,
                         std::vector<pcl::PointXYZ> const &corners,
                         std::vector<int>           const &corner_ids,
                         PCloud::ConstPtr                 cloud_ptr,
                         pcl::PointIndices::ConstPtr      indices_ptr,
                         double                           distanceThreshold = .1 );


            /// CONSTR
            Plane();
            Plane( pcl::ModelCoefficients::Ptr coeffs, PCloud::ConstPtr cloud_ptr );

            Plane( am::Plane const& theOther );
            Plane& operator=( Plane const& other );

            Plane clone()
            {
                Plane plane(*this);
                plane.CoeffsPtrRef() = pcl::ModelCoefficients::Ptr( new pcl::ModelCoefficients() );
                *plane.CoeffsPtrRef() = *this->_coeffs_ptr;
                return plane;
            }

            /// METHODS
            int
            getInlierIndices(pcl::PointIndices::Ptr &inliers,
                              PCloud::ConstPtr       cloud_ptr,
                              pcl::PointIndices::ConstPtr indices_ptr = nullptr,
                              double                 threshold   = .1 ) const;
            int
            getInlierPoints ( PCloud::Ptr            &inlier_points,
                              PCloud::ConstPtr       cloud_ptr,
                              double                 threshold = .1 ) const;

            int
            getOutlierIndices( pcl::PointIndices::Ptr &outlier_indices,
                               PCloud::ConstPtr       cloud_ptr,
                               double                 threshold = .1 ) const;

            int
            draw( pcl::visualization::PCLVisualizer::Ptr v
                 , int plane_id
                 , double r        = .2
                 , double g        = .0
                 , double b        = .0
                 , bool drawPlane  = true
                 , bool drawNormal = false
                 , bool drawFrame  = false
                 , int viewport_id = 0) const;

            int
            draw( pcl::visualization::PCLVisualizer::Ptr v
                 , std::string plane_name
                 , double r        = .2
                 , double g        = .0
                 , double b        = .0
                 , bool drawPlane  = true
                 , bool drawNormal = false
                 , bool drawFrame  = false
                 , int viewport_id = 0) const;

            int
            countFinitePoints() const;

            Eigen::Vector3f
            getCentroid( pcl::PointIndices::ConstPtr indices_ptr = nullptr ) const;

            Eigen::Vector3f
            getCentroidFromCorners() const;

            Eigen::Vector3f
            getNormalFromCorners() const;

            std::vector<Eigen::Vector3f>
            getSides() const;

            static int
            calculateCorners( am::Plane &plane,
                              Eigen::Matrix4f const& frame);

            static int
            makePerpendicular( am::Plane &plane,
                               Eigen::Vector3f normal,
                               PCloud::ConstPtr cloud_ptr,
                               pcl::PointIndices::ConstPtr indices_ptr,
                               double distanceThreshold
                               );

            static int
            projectPointsToNormal( Histogram                  &histogram,
                                   am::Plane            const &plane,
                                   PCloud::ConstPtr            cloud_ptr,
                                   pcl::PointIndices::ConstPtr indices_ptr,
                                   int                         N_bins
                                   , double distanceThreshold
                                   , bool                        debug = false );
            int
            flipNormal();
            int
            offsetAlongNormal( double dist );

            float
            getArea() const;
            // SET, GET
            pcl::ModelCoefficients::ConstPtr  CoeffsPtr()     const;
            pcl::ModelCoefficients::Ptr&      CoeffsPtrRef();
            Eigen::Vector4f                   Coeffs()        const;
            Eigen::Vector3f                   Normal()        const;
            int                               setNormal( Eigen::Vector3f const& normal );
            float                             D()             const;
            std::vector<pcl::PointXYZ> const& Corners()       const;
            std::vector<pcl::PointXYZ>      & Corners();
            Eigen::Matrix4f            const& Frame()         const;
            Eigen::Matrix4f                 & Frame();

        protected:
            pcl::ModelCoefficients::Ptr _coeffs_ptr;
            std::vector<pcl::PointXYZ>  _corners;
            Eigen::Matrix4f             _frame;
    };

    inline std::ostream& operator << (std::ostream& s, Plane const& plane )
    {
        s << "Plane(" << plane.Coeffs().transpose() << ")" << std::endl;
        return (s);
    }
} // ns am

#endif // PLANE_H
