#ifndef CA_DEPTHPROCESSOR_H
#define CA_DEPTHPROCESSOR_H

#include <pcl/visualization/pcl_visualizer.h>
#include <pcl/point_cloud.h>

#include <opencv2/core/core.hpp>

#include "cuboid_approx/util/AMUtil2.h"
#include "cuboid_approx/util/MyIntrinsics.h"

#include "cuboid_approx/common.h"

#include "Params.h"

//#include "Cuboid.h"
//#include "Plane.h"

namespace am
{
    class Cuboid;
    class Plane;

    class GrabcutResult;
    extern double disparity_to_depth(double disparity );

    inline pcl::PointXYZ
    toWorld( pcl::PointXYZ const& pnt, Eigen::Matrix4f const& frame )
    {
        Eigen::Vector4f pnt_eigen = pnt.getVector4fMap();
        return am::util::pcl::v4asPointXYZ( frame * pnt_eigen );
    }

    inline pcl::PointXYZ
    toWorld( MyPoint const& pnt, Eigen::Matrix4f const& frame )
    {
        Eigen::Vector4f pnt_eigen = pnt.getVector4fMap();
        return am::util::pcl::v4asPointXYZ( frame * pnt_eigen );
    }

    class DepthProcessor
    {
        public:
            template <typename depT>
            static int
            cloud_from_images( PCloud::Ptr         &cloud_ptr,
                               cv::Mat       const &img,
                               cv::Mat       const &depth,
                               MyIntrinsics  const &intrinsics,
                               PCloud::Ptr         *p_invalid_cloud_ptr = nullptr );

            static int
            cloud_2_vertex_list( Eigen::Matrix3Xf& vertex_list, PCloud::ConstPtr cloud );

            template <typename depT>
            static int
            mats_2_3d( PCloud::Ptr                                 &cloudPtr,
                       cv::Mat                               const &dep,
                       cv::Mat                               const &img,
                       Eigen::Matrix3f                       const &intrinsics,
                       /* depth scale: */ double             const alpha,
                       Eigen::Affine3f                       const *pose        = nullptr,
                       std::vector< ::pcl::Vertices>               *pFaces      = nullptr,
                       PCloud::Ptr                                 *p_invalid_cloud_ptr = nullptr,
                       double(*transform_distance)(double)                      = &disparity_to_depth );

            template <typename My_PointT>
            static int
            statistical_clean( boost::shared_ptr<      pcl::PointIndices          >       &indices_ptr,
                               boost::shared_ptr<const pcl::PointCloud<My_PointT> > const  cloud_ptr,
                               boost::shared_ptr<      pcl::PointIndices          > const *p_selected_indices_ptr = nullptr,
                               boost::shared_ptr<      pcl::PointIndices          >       *p_removed_indices_ptr  = nullptr,
                               double                                               const  meanK                  = 50.,
                               double                                               const  stddevMulThresh        = .1
                               , bool const silent = true );

            template <typename My_PointT>
            static int
            statistical_clean( boost::shared_ptr<      pcl::PointCloud<My_PointT> >       &filtered_cloud_ptr,
                               boost::shared_ptr<const pcl::PointCloud<My_PointT> > const  cloud_ptr          ,
                               boost::shared_ptr<      pcl::PointIndices          > const *p_selected_indices_ptr = nullptr,
                               boost::shared_ptr<      pcl::PointIndices          >       *p_removed_indices_ptr  = nullptr,
                               double                                               const  meanK                  = 50.,
                               double                                               const  stddevMulThresh        = .1
                               , bool const silent = true );



            template <typename My_PointT>
            static int
            extractCloud( boost::shared_ptr<      pcl::PointCloud<My_PointT>> &out_cloud_ptr,
                          boost::shared_ptr<const pcl::PointCloud<My_PointT>> cloud_ptr,
                          boost::shared_ptr<const pcl::PointIndices         > indices_ptr );


            static int
            projectPlaneInliers( PCloud::Ptr                      &inliers_cloud_ptr,
                                 PCloud::ConstPtr                  cloud_ptr,
                                 pcl::PointIndices::ConstPtr       indices_ptr,
                                 pcl::ModelCoefficients::ConstPtr  coeffs_ptr );

            static  int
            cloud2Local( PCloud::Ptr                 &cloud_ptr,
                         Eigen::Matrix4f    const    &frame,
                         PCloud::ConstPtr             in_cloud_ptr,
                         pcl::PointIndices::ConstPtr  indices_ptr = nullptr );
            int
            local2World( PCloud::Ptr                 &cloud_ptr,
                         Eigen::Matrix4f       const &frame,
                         PCloud::ConstPtr             in_cloud_ptr,
                         pcl::PointIndices::ConstPtr  indices_ptr = nullptr );

            template <typename MyPointT>
            static MyPointT
            local2World( MyPointT          const& pnt
                         , Eigen::Matrix4f const& transform );

            static int
            PCA( Eigen::Matrix4f             &frame,
                 //MyPoint                     &min_pt,
                 //MyPoint                     &max_pt,
                 PCloud::ConstPtr             cloud_ptr,
                 pcl::PointIndices::ConstPtr  indices_ptr );

            static int
            PCA_withPlane( Eigen::Matrix4f               & frame
                          , Eigen::Vector4f         const& plane_coords
                          , PCloud::ConstPtr               cloud_ptr
                          , pcl::PointIndices::ConstPtr    indices_ptr
                          , bool                           downsample       = false
                          , PCloud::Ptr                  * projected_cloud  = NULL
                          , bool                    const  debug            = false );


            static int
            calculateExtremaInFrame( MyPoint                       &min_pt
                                     , MyPoint                     &max_pt
                                     , Eigen::Matrix4f       const &frame
                                     , PCloud::ConstPtr             cloud_ptr
                                     , pcl::PointIndices::ConstPtr  indices_ptr
                                     );

            static bool
            isValidDepth( float d );

            template <typename depT>
            static int
            getMaskedIndices( pcl::PointIndices::Ptr       &masked_indices_ptr,
                               cv::Mat                const &masked_depth,
                               cv::Mat                const &full_depth );
    };
} // ns am

#ifndef INC_DEPTHPROCESSOR_HPP
#   define INC_DEPTHPROCESSOR_HPP
#   include "DepthProcessor.hpp"
#endif

#endif // CA_DEPTHPROCESSOR_H
