#ifndef DEPTHPROCESSOR_HPP
#define DEPTHPROCESSOR_HPP

#include "DepthProcessor.h"

#include "pcl/io/pcd_io.h"
#include "pcl/io/ply_io.h"
#include "pcl/filters/statistical_outlier_removal.h"
#include "pcl/common/common.h"

#include <opencv2/imgproc/imgproc.hpp>
#include <opencv2/highgui/highgui.hpp>

//#include "cuboid_approx/util/ViewpointMapper.h"
#include "CloudViewer.h"
#include "GrabcutResult.h"

#include "Segmentation.h"

namespace am
{
    template <typename depT>
    int
    DepthProcessor::cloud_from_images( PCloud::Ptr         &cloud_ptr,
                                       cv::Mat       const &img,
                                       cv::Mat       const &depth,
                                       MyIntrinsics  const &intrinsics,
                                       PCloud::Ptr *p_invalid_cloud_ptr )
    {
        int err = EXIT_SUCCESS;

        // to cloud
        {
            err = DepthProcessor::mats_2_3d<depT>( cloud_ptr,
                                                   depth,
                                                   img,
                                                   intrinsics.asMatrix3f(),
                                                   1. / 1000.,
                                                   nullptr,
                                                   nullptr,
                                                   p_invalid_cloud_ptr );
            if ( err != EXIT_SUCCESS )
            {
                std::cerr << "DepthProcessor::mats_2_3d<depT>(): Returned " << (int)err << std::endl;
                return err;
            }
            if ( cloud_ptr->empty() )
            {
                std::cerr << "DepthProcessor::cloud_from_images(): cloud empty! exiting...\n";
                return EXIT_FAILURE;
            }
        }

#if 0
        // Filter cloud
        pcl::PointCloud<pcl::PointXYZRGB>::Ptr filtered_cloud_ptr;
        {
            err = DepthProcessor::statistical_clean<pcl::PointXYZRGB>( filtered_cloud_ptr, cloud_ptr );
            if ( err != EXIT_SUCCESS )
            {
                std::cerr << "DepthProcessor::statistical_clean(): exited with error " << (int)err << std::endl;
                return err;
            }
            if ( filtered_cloud_ptr->empty() )
            {
                std::cerr << "DepthProcessor::calculate_cuboids_from_masks(): cloud empty! exiting...\n";
                return EXIT_FAILURE;
            }
        }

        // output
        {
            cloud_ptr.swap( filtered_cloud_ptr );
        }
#endif

        return err;
    }

    template <typename depT>
    int
    DepthProcessor::getMaskedIndices( pcl::PointIndices::Ptr       &masked_indices_ptr,
                                      cv::Mat                const &masked_depth,
                                      cv::Mat                const &full_depth )
    {
        // select point indices from mask
        if ( !masked_indices_ptr )
        {
            std::cout << "masked_indices_ptr initing..." << std::endl;
            masked_indices_ptr = boost::make_shared<pcl::PointIndices>();
        }
        else
        {
            masked_indices_ptr->indices.clear();
        }

        int cloud_point_index = 0;
        for ( unsigned y = 0; y < masked_depth.rows; ++y  )
        {
            for ( unsigned x = 0; x < masked_depth.cols; ++x )
            {
                // record index, if belongs to mask
                if ( masked_depth.at<uchar>(y,x) > 0 ) // !!! mask is always CV_8UC1
                {
                    masked_indices_ptr->indices.push_back( cloud_point_index );
                }

                if ( DepthProcessor::isValidDepth(full_depth.at<depT>(y,x)) ) // keep track of valid point cloud ids
                {
                    ++cloud_point_index;
                }
            }
        }

        return EXIT_SUCCESS;
    }

    template <typename depT>
    int
    DepthProcessor::mats_2_3d( PCloud::Ptr                                  &cloudPtr,
                               cv::Mat                                const &dep,
                               cv::Mat                                const &img,
                               Eigen::Matrix3f                        const &intrinsics,
                               double                                 const alpha,
                               Eigen::Affine3f                        const *pose,
                               std::vector<::pcl::Vertices>                 *pFaces,
                               PCloud::Ptr                                  *p_invalid_cloud_ptr,
                               double (*transform_distance)(double) )
    {
        // check input
        if ( !img.empty() )
        {
            if ( dep.size() != img.size() )
            {
                std::cerr << "matsTo3D(): dep and rgb need the same size! "
                          << dep.rows << "x" << dep.cols << ", "
                          << img.rows << "x" << img.cols << std::endl;
                return EXIT_FAILURE;
            }
        }

        if ( (!img.empty()) && (img.channels() != 3) )
        {
            std::cerr << "matsTo3D(): rgb should be 3 channels! " << img.channels() << std::endl;
            return EXIT_FAILURE;
        }

        // allocate ouptut
        cloudPtr = PCloud::Ptr( new PCloud );
        if ( p_invalid_cloud_ptr )
        {
            *p_invalid_cloud_ptr = PCloud::Ptr( new PCloud );
        }

        Eigen::Vector3f pnt3D;
        uint32_t rgb = 128 << 16 | 128 << 8 | 128;

        Eigen::Matrix3f rotation;
        Eigen::Vector3f translation;
        if ( pose )
        {
            rotation    = pose->rotation();
            translation = pose->translation();
        }

        // prepare faces output
        if ( pFaces )
        {
            pFaces->clear();
            pFaces->reserve( dep.rows * dep.cols * 2 );
        }

        // copy inputs
        for ( int y = 0; y < dep.rows; ++y )
        {
            for ( int x = 0; x < dep.cols; ++x )
            {
                //  read depth
                float depth = (float)dep.at<depT>( y,x ) * alpha;
                // check validity
                bool isValid = DepthProcessor::isValidDepth(depth);
                if ( !isValid )
                    if ( p_invalid_cloud_ptr )  depth = 10.f;
                    else                        continue;

                // calculate point
                pnt3D = am::util::pcl::point2To3D( (Eigen::Vector2f){x,y},
                                                   intrinsics             )
                        * depth;

                // apply pose
                if ( pose )
                {
                    pnt3D = rotation * pnt3D + translation;
                }

                // convert
                MyPoint point;
                point.x = pnt3D(0);
                point.y = pnt3D(1);
                point.z = pnt3D(2);

                // add colour
                if ( !img.empty() )
                {
                    rgb = (static_cast<uint32_t>(img.at<uchar>(y, x * img.channels() + 2)) << 16 |
                           static_cast<uint32_t>(img.at<uchar>(y, x * img.channels() + 1)) << 8  |
                           static_cast<uint32_t>(img.at<uchar>(y, x * img.channels()    ))        );
                }
                point.rgb = *reinterpret_cast<float*>( &rgb );

                // add label
                point.label = y * dep.cols + x;

                // add point
                if ( p_invalid_cloud_ptr && !isValid )
                  (*p_invalid_cloud_ptr)->push_back( point );
                else
                    cloudPtr->points.push_back( point );

                // faces
                const float dlimit = 300.f; // 30 cm
                if ( (pFaces)                                         &&
                     (y+1 < dep.rows)                                 &&
                     (static_cast<float>(dep.at<depT>(y,x)) > dlimit)    )
                {
                    // left face
                    if ( (x > 0)                                              &&
                         (static_cast<float>(dep.at<depT>(y+1,x-1)) > dlimit) &&
                         (static_cast<float>(dep.at<depT>(y+1,x  )) > dlimit)    )
                    {
                        pFaces->push_back( ::pcl::Vertices() );
                        pFaces->back().vertices.resize(3);
                        pFaces->back().vertices[0] =  y    * dep.cols + x    ; // y,x
                        pFaces->back().vertices[1] = (y+1) * dep.cols + x - 1; // y+1,x-1
                        pFaces->back().vertices[2] = (y+1) * dep.cols + x    ; // y+1,x

                        //++fid;
                    }

                    // right face
                    if ( (x+1 < dep.cols)                                     &&
                         (static_cast<float>(dep.at<depT>(y+1,x  )) > dlimit) &&
                         (static_cast<float>(dep.at<depT>(y  ,x+1)) > dlimit)    )
                    {
                        pFaces->push_back( ::pcl::Vertices() );
                        pFaces->back().vertices.resize(3);
                        pFaces->back().vertices[0] =  y    * dep.cols + x    ; // y,x
                        pFaces->back().vertices[1] = (y+1) * dep.cols + x    ; // y+1,x
                        pFaces->back().vertices[2] =  y    * dep.cols + x + 1; // y+1,x+1
                    }
                }
            }
        }
        cloudPtr->width = (int)cloudPtr->points.size ();
        cloudPtr->height = 1;

        return EXIT_SUCCESS;

    } // matsTo3D

    // \brief Statistical clean
    // \param[out] indices_ptr           Indices of inliers in cloud_ptr
    // \param[in]  cloud_ptr             Pointcloud to clean
    // \param[in]  selected_indices_ptr  Indices of interest in cloud_ptr
    // \param[out] p_removed_indices_ptr Indices of removed points, selected_indices = indices + removed_indices
    template <typename My_PointT>
    int
    DepthProcessor::statistical_clean( boost::shared_ptr<      pcl::PointIndices          >       &indices_ptr,
                                       boost::shared_ptr<const pcl::PointCloud<My_PointT> > const  cloud_ptr,
                                       boost::shared_ptr<      pcl::PointIndices          > const *p_selected_indices_ptr,
                                       boost::shared_ptr<      pcl::PointIndices          >       *p_removed_indices_ptr,
                                       double                                               const  meanK,
                                       double                                               const  stddevMulThresh
                                       , bool const silent )
    {
        // Create the filtering object
        pcl::StatisticalOutlierRemoval<My_PointT> sor( /* extract_removed_indices: */ true );
        sor.setInputCloud      ( cloud_ptr            );
        if ( p_selected_indices_ptr )
            sor.setIndices         ( *p_selected_indices_ptr );
        sor.setMeanK           ( meanK                ); // 50
        sor.setStddevMulThresh ( stddevMulThresh      ); // 1.
        sor.filter             ( indices_ptr->indices );

        if ( p_removed_indices_ptr )
            sor.getRemovedIndices( **p_removed_indices_ptr );

        if ( !silent )
            std::cout << "statistical_clean left "
                  << indices_ptr->indices.size() << " out of "
                  << ( p_selected_indices_ptr ? (*p_selected_indices_ptr)->indices.size() : cloud_ptr->size() ) << " points" << std::endl;

        return EXIT_SUCCESS;
    } // statistical_removal

    // \brief Statistical clean
    // \param[out] filtered_cloud_ptr    Inliers point cloud from cloud_ptr
    // \param[in]  cloud_ptr             Pointcloud to clean
    // \param[in]  selected_indices_ptr  Indices of interest in cloud_ptr
    // \param[out] p_removed_indices_ptr Indices of removed points, selected_indices = indices + removed_indices
    template <typename My_PointT>
    int
    DepthProcessor::statistical_clean( boost::shared_ptr<      pcl::PointCloud<My_PointT> >       &filtered_cloud_ptr,
                                       boost::shared_ptr<const pcl::PointCloud<My_PointT> > const  cloud_ptr          ,
                                       boost::shared_ptr<      pcl::PointIndices          > const *p_selected_indices_ptr,
                                       boost::shared_ptr<      pcl::PointIndices          >       *p_removed_indices_ptr,
                                       double                                               const  meanK,
                                       double                                               const  stddevMulThresh
                                       , bool const silent                                                              )
    {
        if ( !filtered_cloud_ptr ) filtered_cloud_ptr = PCloud::Ptr( new PCloud() );

        // Create the filtering object
        pcl::StatisticalOutlierRemoval<My_PointT> sor( /* extract_removed_indices: */ true );
        sor.setInputCloud      ( cloud_ptr           );
        if ( p_selected_indices_ptr )
            sor.setIndices( *p_selected_indices_ptr );
        sor.setMeanK           ( meanK               ); // 50.
        sor.setStddevMulThresh ( stddevMulThresh     ); // 1.
        sor.filter             ( *filtered_cloud_ptr );

        if ( p_removed_indices_ptr )
            sor.getRemovedIndices( **p_removed_indices_ptr );

        if ( !silent )
            std::cout << "statistical_clean left "
                  << filtered_cloud_ptr->size() << " out of "
                  << ( p_selected_indices_ptr ? (*p_selected_indices_ptr)->indices.size() : cloud_ptr->size() ) << " points" << std::endl;

        return EXIT_SUCCESS;
    } // statistical_removal

    template <typename My_PointT>
    int
    DepthProcessor::extractCloud( boost::shared_ptr<      pcl::PointCloud<My_PointT>> &out_cloud_ptr,
                                  boost::shared_ptr<const pcl::PointCloud<My_PointT>> cloud_ptr,
                                  boost::shared_ptr<const pcl::PointIndices         > indices_ptr )
    {
        if ( !out_cloud_ptr ) out_cloud_ptr = PCloud::Ptr( new PCloud );

        pcl::ExtractIndices<MyPoint> extraction;
        extraction.setInputCloud( cloud_ptr      );
        extraction.setIndices   ( indices_ptr    );
        extraction.setNegative  ( false          );
        extraction.filter       ( *out_cloud_ptr );

        return EXIT_SUCCESS;
    } // extract_cloud


    template <typename MyPointT>
    MyPointT
    DepthProcessor::local2World( MyPointT          const& pnt
                                 , Eigen::Matrix4f const& transform )
    {
        MyPointT out = pnt;

        Eigen::Vector3f pt( pnt.x, pnt.y, pnt.z );
        out.x = static_cast<float>( transform (0, 0) * pt.coeffRef (0) + transform (0, 1) * pt.coeffRef (1) + transform (0, 2) * pt.coeffRef (2) + transform (0, 3) );
        out.y = static_cast<float>( transform (1, 0) * pt.coeffRef (0) + transform (1, 1) * pt.coeffRef (1) + transform (1, 2) * pt.coeffRef (2) + transform (1, 3) );
        out.z = static_cast<float>( transform (2, 0) * pt.coeffRef (0) + transform (2, 1) * pt.coeffRef (1) + transform (2, 2) * pt.coeffRef (2) + transform (2, 3) );

        return out;
    }

} // ns am

#endif // DEPTHPROCESSOR_HPP
