#ifndef __CA_MYVOXELGRID_H__
#define __CA_MYVOXELGRID_H__

#include "primitives/Cuboids.h"

#include "cuboid_approx/common.h"

#include <pcl/visualization/pcl_visualizer.h>
#include <pcl/filters/voxel_grid.h>

#include <Eigen/Dense>
#include <iostream>

namespace am
{
    struct CompactnessStruct
    {
            CompactnessStruct( int dir, int width, int start, int sum, float compactness )
                : dir(dir), width(width), start(start), sum(sum), compactness(compactness)
                , outlierscore(0.f)
            {
            }

            bool
            operator< ( CompactnessStruct const& rhs ) const
            {
                return this->compactness < rhs.compactness;
            }

            bool
            operator> ( CompactnessStruct const& rhs ) const
            {
                if ( this->outlierscore == rhs.outlierscore )
                {
                    return this->compactness > rhs.compactness;
                }
                return this->outlierscore > rhs.outlierscore;
            }

            void print()
            {
                std::cout << "projections"
                          << "[dimension:" << dir << "]"
                          << "[starts:" << start << "]"
                          << "[width:" << width << "]"
                          << ": "
                          << "\tscore: " << outlierscore
                          << "\tcompact: " << compactness
                          << "\t(sum:" << sum << ")"
                          << std::endl;
            }

            int dir, width, start, sum;
            float compactness;
            float outlierscore;
    };

    struct MaxPlanes
    {
            MaxPlanes();

            int
            getMaxDimCoord() const;
            int
            getDim1Coord() const;
            int
            getDim2Coord() const;

            // highest pointcount in a plane along x,y,z axis
            Eigen::Vector3i max_point_count;
            // x,y and z coordinate of highest pointcount plane along x,y and z axis
            Eigen::Vector3i max_coords;

            int max_dim_id;       // which dimension has biggest plane
            int dim1_id, dim2_id; // the other two dimensions
    };

    class LeafIndexer
    {
        public:
            LeafIndexer( int max_dim_id, int dim1_id, int dim2_id, Eigen::Vector3i div_mul, int max_dim_coord );

            int
            leafIndex( int i, int j, int k ) const;

            int
            leafIndex( int j, int k ) const;

        private:
            Eigen::Vector3i _dim_ids;
            Eigen::Vector3i _div_mul;
            int             _max_dim_coord;
    };

    class MyVoxelGrid
    {
        public:
            typedef Eigen::MatrixX2i Slab;

            ///
            /// \brief segment
            /// \param[out] cuboids
            /// \param[in]  cloud_ptr
            /// \param[in]  frame
            /// \param[in]  resolution                       Voxelgrid resolution
            /// \param[in]  slab_size_threshold_percentage   Stops, when slab.pointcount < cloud.pointcount * threshold.
            /// \param[in]  max_slab_count                   How many slabs to return at maximum
            /// \return
            ///
            static int
            segment(am::Cuboids            &cuboids
                     , am::PCloud::ConstPtr cloud_ptr // TODO: change to template
                     , Eigen::Matrix4f              const& frame
                     , pcl::PointIndices::ConstPtr indices_ptr  = nullptr
                     , float resolution                         = .15f
                     , float slab_size_threshold_percentage     = .1f
                     , int   max_slab_count                     = 25
                     , bool  debug                              = false );

            static int
            getVolume( Slab slab );

            static int
            cuboidFromSlab( am::Cuboid &cuboid
                            , Eigen::MatrixX2i              const& slab
                            , MaxPlanes                     const& mp
                            , pcl::VoxelGrid<pcl::PointXYZRGB>     *sor
                            );

            static int
            getLeafId( int c0, int c1, int c2, int d1, int d2, Eigen::Vector3i const& div_mul );

            static std::string
            getCellGridId( Eigen::Vector3i ijk );

            static int
            drawSlab( pcl::visualization::PCLVisualizer::Ptr v_ptr
                      , Eigen::MatrixX2i              const& minmax
                      , MaxPlanes                     const& mp
                      , pcl::VoxelGrid<pcl::PointXYZRGB>     *sor
                      , std::string                          title  = "cuboid"
                      , Eigen::Vector3f                      colour = { .9, .9, .0 }
                      //, int                                  viewport = 0
                      );

            static int
            project2D( std::vector<std::vector<int> >       & projections
                       , std::vector<int>              const& points_in_voxels
                       , Eigen::MatrixX2i              const& slab
                       , MaxPlanes                     const& mp
                       , pcl::VoxelGrid<pcl::PointXYZRGB>   * sor
                       , bool                                 print = false );
            static bool
            compareSlabsByVolume( Eigen::MatrixX2i const& a, Eigen::MatrixX2i const& b );


            static int
            splitSlabs( std::vector<Eigen::MatrixX2i>        & out_slabs
                        , std::vector<Eigen::MatrixX2i> const& slabs
                        , float                                threshold
                        , std::vector<int>                   & leafs
                        , std::vector<int>              const& points_in_voxels
                        , MaxPlanes                     const& mp
                        , pcl::VoxelGrid<pcl::PointXYZRGB>   * sor       );

            static int
            splitSlabs2(std::vector<Eigen::MatrixX2i>        & out_slabs
                        , std::vector<Eigen::MatrixX2i> const& slabs
                        , float                                threshold
                        , std::vector<int>                     leafs
                        , std::vector<int>              const& points_in_voxels
                        , MaxPlanes                     const& mp
                        , pcl::VoxelGrid<pcl::PointXYZRGB>   * sor
                        , pcl::visualization::PCLVisualizer::Ptr vptr           = nullptr );

            static int
            mergeSlabs( std::vector<std::pair<Slab, MaxPlanes> >        & out_slabs
                        , std::vector<std::pair<Slab, MaxPlanes> > const& slabs
                        , float                                           threshold
                        , std::vector<int>                                leafs // need the local copy
                        , pcl::VoxelGrid<pcl::PointXYZRGB>              * sor
                        , pcl::visualization::PCLVisualizer::Ptr          vptr = nullptr );

            static int
            shrinkSlab( Eigen::MatrixX2i                           & out_slab // contains slab min,max coords in [ maxdim, dim1, dim2 ] order
                        , Eigen::MatrixX2i                    const& slab    // contains limits in [ maxdim,dim1,dim2] order, inclusive!
                        , MaxPlanes                           const& mp
                        , std::vector<int>                    const& leafs
                        , pcl::VoxelGrid<pcl::PointXYZRGB>         *const sor
                        , /*debug*/ pcl::visualization::PCLVisualizer::Ptr v_ptr = nullptr
                        , int level = 0 );
            template <typename PointT>
            static int
            shrinkSlabInsideGrid( am::Cuboid                                 & out_cuboid
                                  , Eigen::MatrixX2i                    const& slab    // contains limits in [ maxdim,dim1,dim2] order, inclusive!
                                  , MaxPlanes                           const& mp
                                  , std::vector<int>                    const& leafs
                                  , boost::shared_ptr<const pcl::PointCloud<PointT> >       cloud_ptr
                                  , pcl::PointIndices::ConstPtr                             indices_ptr
                                  , pcl::VoxelGrid<pcl::PointXYZRGB>         *const sor
                                  , /*debug*/ pcl::visualization::PCLVisualizer::Ptr v_ptr = nullptr
                                  , int level = 0 );

            static Eigen::Vector3f
            getCentroidCoord( Eigen::Vector3i ijk, pcl::VoxelGrid<pcl::PointXYZRGB> *sor );
            static Eigen::Vector3f
            getCentroidCoord( int i, int j, int k, Eigen::Vector3i const& min_box_coords, Eigen::Vector3f const& leaf_size );

            template <typename PointT>
            static int
            drawGrid( pcl::visualization::PCLVisualizer::Ptr        v_ptr
                      , std::vector<int>                     const& leafs
                      , pcl::VoxelGrid<PointT>                    * sor  );

            template <typename PointT>
            static int
            drawRealGrid( pcl::visualization::PCLVisualizer::Ptr        v_ptr
                          , std::vector<int>                     const& leafs
                          , pcl::VoxelGrid<PointT>                    * sor );

            template <typename PointT>
            static int
            drawCentroids( pcl::visualization::PCLVisualizer::Ptr v_ptr
                           , pcl::VoxelGrid<PointT>               *sor
                           , std::string                          title    = "voxel_grid"
                           , int                                  viewport = 0
                           );
    };

} // ns am

#ifndef __CA_INC_MYVOXELGRID_HPP__
#define __CA_INC_MYVOXELGRID_HPP__
#   include "MyVoxelGrid.hpp"
#endif

#endif // __CA_MYVOXELGRID_H__
