#ifndef SEGMENTATION_HPP
#define SEGMENTATION_HPP

#include "Segmentation.h"

#include "boost/smart_ptr.hpp"
#include "pcl/segmentation/sac_segmentation.h"
#include "pcl/features/normal_3d.h"
#include "pcl/filters/extract_indices.h"
#include "pcl/segmentation/region_growing.h"
#include "pcl/common/angles.h"

#include "cuboid_approx/common.h"
#include "proc3d/CloudViewer.h"
#include "proc3d/DepthProcessor.h"

namespace am
{
    template <typename MyPointT>
    int Segmentation::extract_cylinder( boost::shared_ptr<pcl::PointCloud<MyPointT>>       &cylinder_cloud,
                                        pcl::ModelCoefficients::Ptr                        &cylinder_coefficients,
                                        boost::shared_ptr<const pcl::PointCloud<MyPointT>> in_cloud )
    {
        // prepare output
        if ( cylinder_cloud )        { cylinder_cloud->clear(); }
        else                         { cylinder_cloud = boost::make_shared<pcl::PointCloud<MyPointT>>(); }
        //        if ( !cylinder_coefficients ) { cylinder_coefficients = boost::make_shared<pcl::ModelCoefficients>(); }
        //        else                          { cylinder_coefficients->clear(); }

        // var
        pcl::NormalEstimation<MyPointT, pcl::Normal>           ne;
        pcl::SACSegmentationFromNormals<MyPointT, pcl::Normal> seg;
        typename pcl::search::KdTree<MyPointT>::Ptr            tree          (new pcl::search::KdTree<MyPointT> );
        pcl::PointCloud<pcl::Normal>::Ptr                      cloud_normals (new pcl::PointCloud<pcl::Normal>);
        pcl::ExtractIndices<MyPointT>                          extract_point_indices;

        // Normals
        {
            ne.setSearchMethod( tree );
            ne.setInputCloud ( in_cloud );
            ne.setKSearch (50);
            ne.compute (*cloud_normals);
        }

        // Prepare segmentation
        {
            seg.setOptimizeCoefficients (true);
            seg.setModelType (pcl::SACMODEL_CYLINDER);
            seg.setMethodType (pcl::SAC_RANSAC);
            seg.setNormalDistanceWeight (0.1);
            seg.setMaxIterations (10000);
            seg.setDistanceThreshold (0.05);
            seg.setRadiusLimits (0, 0.1);
            seg.setInputCloud (in_cloud);
            seg.setInputNormals (cloud_normals);
        }

        // Perform segmentation
        pcl::PointIndices::Ptr inliers_cylinder( new pcl::PointIndices );
        {
            seg.segment ( *inliers_cylinder, *cylinder_coefficients );
            std::cerr << "Cylinder coefficients: " << *cylinder_coefficients << std::endl;
        }

        // Extract results
        {
            // in
            extract_point_indices.setInputCloud ( in_cloud );
            extract_point_indices.setIndices    ( inliers_cylinder );
            // prop
            extract_point_indices.setNegative   ( false );
            // out
            extract_point_indices.filter        ( *cylinder_cloud );
        }

        return EXIT_SUCCESS;
    }

    template <typename MyPointT>
    int Segmentation::extract_plane( pcl::ModelCoefficients::Ptr                        &out_coefficients,
                                     boost::shared_ptr<const pcl::PointCloud<MyPointT>> in_cloud_ptr,
                                     pcl::PointIndices::ConstPtr                        in_indices_ptr,
                                     double                                             distanceThreshold,
                                     int                                                sac_model,
                                     Eigen::Vector3f                                    *p_normal            )
    {
        // prepare output
        if ( !out_coefficients )
            out_coefficients = boost::make_shared< pcl::ModelCoefficients >();

        // Normals
        pcl::PointCloud<pcl::Normal>::Ptr roi_cloud_normals_ptr (new pcl::PointCloud<pcl::Normal>);
        {
            pcl::NormalEstimation<MyPointT, pcl::Normal> normal_estimation;
            typename pcl::search::KdTree<MyPointT>::Ptr  tree( new pcl::search::KdTree<MyPointT> );

            normal_estimation.setSearchMethod( tree           );
            normal_estimation.setInputCloud  ( in_cloud_ptr   );
            normal_estimation.setKSearch     ( 20            );
            if ( in_indices_ptr )
                normal_estimation.setIndices( in_indices_ptr );

            normal_estimation.compute        ( *roi_cloud_normals_ptr );
        }

        // resize normals cloud to match cloud_ptr, copy estimated ones to right positions, leave others as 0
        pcl::PointCloud<pcl::Normal>::Ptr cloud_normals_ptr ( new pcl::PointCloud<pcl::Normal> );
        if ( in_indices_ptr )
        {
            _expandNormals( /*   out_cloud: */ cloud_normals_ptr,
                            /*   in_indice: */ in_indices_ptr,
                            /*  in_normals: */ roi_cloud_normals_ptr,
                            /* output_size: */ in_cloud_ptr->size() );
        }

        // Segmentation
        pcl::PointIndices::Ptr inliers_indices( new pcl::PointIndices );
        pcl::SACSegmentationFromNormals<MyPointT, pcl::Normal> seg;
        // in
        seg.setInputCloud           ( in_cloud_ptr                   );
        seg.setInputNormals         ( in_indices_ptr ? cloud_normals_ptr : roi_cloud_normals_ptr );
        if ( in_indices_ptr ) seg.setIndices( in_indices_ptr );
        // props
        seg.setOptimizeCoefficients ( true                       );
        seg.setMethodType           ( pcl::SAC_RANSAC            );
        seg.setDistanceThreshold    ( distanceThreshold          ); // .1, .05
        seg.setMaxIterations        ( 500                        );

        switch ( sac_model )
        {
            case pcl::SACMODEL_PERPENDICULAR_PLANE:
                seg.setModelType ( pcl::SACMODEL_PARALLEL_PLANE );
                seg.setAxis      ( *p_normal                    );
                seg.setEpsAngle  ( pcl::deg2rad(20.f)           );
                break;
            case pcl::SACMODEL_NORMAL_PLANE:
                seg.setModelType ( pcl::SACMODEL_NORMAL_PLANE );
                seg.setNormalDistanceWeight ( 0.1                        );
                break;
            case pcl::SACMODEL_NORMAL_PARALLEL_PLANE:
                seg.setModelType ( pcl::SACMODEL_NORMAL_PARALLEL_PLANE );
                seg.setAxis      ( *p_normal                    );
                seg.setEpsAngle  ( pcl::deg2rad(5.f)           );
                break;
            case pcl::SACMODEL_PARALLEL_PLANE:
                seg.setModelType ( pcl::SACMODEL_PARALLEL_PLANE );
                seg.setAxis      ( *p_normal                    );
                seg.setEpsAngle  ( pcl::deg2rad(5.f)           );
                break;
            default:
                std::cerr << "sac_model_plane unknown" << std::endl;
                break;
        }

        // out
        seg.segment ( *inliers_indices, *out_coefficients );
        return (out_coefficients->values.size() > 0) ? EXIT_SUCCESS : EXIT_FAILURE;
    }


    template <typename MyPointT>
    int
    Segmentation::region_grow( std::vector <pcl::PointIndices::Ptr>                &cluster_ptrs,
                               boost::shared_ptr<const pcl::PointCloud<MyPointT> > in_cloud_ptr,
                               pcl::PointIndices::ConstPtr                         in_indices_ptr,
                               int                                                 normalK )
    {
        am::CloudViewer v( "region_grow_input" );

        PCloud::Ptr roi_cloud_ptr( new PCloud() );
        am::DepthProcessor::extractCloud( roi_cloud_ptr,
                                          in_cloud_ptr,
                                          in_indices_ptr );


        typename pcl::search::Search<MyPointT>::Ptr tree( new pcl::search::KdTree<MyPointT> );
        pcl::PointCloud <pcl::Normal>::Ptr normals (new pcl::PointCloud <pcl::Normal>);
        {
            pcl::NormalEstimation<MyPointT, pcl::Normal> normal_estimator;
            normal_estimator.setSearchMethod( tree         );
            normal_estimator.setInputCloud  ( in_cloud_ptr );
            if ( in_indices_ptr )
                normal_estimator.setIndices     ( in_indices_ptr );
            normal_estimator.setKSearch     ( normalK      ); // 50
            normal_estimator.compute        ( *normals     );
        }

        v.ViewerPtr()->addPointCloud<MyPointT>( roi_cloud_ptr, "cloud" );
        v.ViewerPtr()->addPointCloudNormals<MyPointT,pcl::Normal>( roi_cloud_ptr, normals, 10, 0.1, "normals" );
        v.ViewerPtr()->spin();
        //v.ViewerPtr()->addPointCloudNormals( normals, 100, 0.02f, "normals" );

//        pcl::IndicesPtr indices (new std::vector <int>);
//        pcl::PassThrough<MyPointT> pass;
//        pass.setInputCloud (cloud);
//        pass.setFilterFieldName ("z");
//        pass.setFilterLimits (0.0, 1.0);
//        pass.filter (*indices);

        pcl::RegionGrowing<MyPointT, pcl::Normal> reg;
        {
            reg.setInputCloud           ( roi_cloud_ptr       );
            //reg.setIndices              ( in_indices_ptr     );
            reg.setSearchMethod         ( tree               );
            reg.setMinClusterSize       ( 50                 );
            reg.setMaxClusterSize       ( 10000              );
            reg.setNumberOfNeighbours   ( 30                  );

            //reg.setIndices (indices);
            reg.setInputNormals         ( normals            );
            reg.setSmoothnessThreshold  ( pcl::deg2rad(7.0) );
            reg.setCurvatureThreshold   ( 1.                   );
        }

        std::vector <pcl::PointIndices> clusters;
        reg.extract( clusters );

        std::cout << "Number of clusters is equal to " << clusters.size () << std::endl;
        if ( !clusters.size() )
            return EXIT_FAILURE;

        std::cout << "First cluster has " << clusters[0].indices.size () << " points." << std::endl;
        std::cout << "These are the indices of the points of the initial" <<
                     std::endl << "cloud that belong to the first cluster:" << std::endl;
        int counter = 0;
        while (counter < 5 || counter > clusters[0].indices.size ())
        {
            std::cout << clusters[0].indices[counter] << std::endl;
            counter++;
        }

        pcl::PointCloud<pcl::PointXYZRGB>::Ptr colored_cloud_ptr = reg.getColoredCloud ();
        am::CloudViewer viewer("Cluster viewer");

        //pcl::visualization::PointCloudColorHandlerRGBField<MyPoint>  color_handler( colored_cloud );
        //pcl::visualization::PointCloudGeometryHandlerCustom<MyPoint> geom_handler ( colored_cloud, "x","y","z" );
        //viewer.VP()->addPointCloud<MyPoint>   ( colored_cloud, geom_handler, "cloud" );
        //viewer.VP()->updatePointCloud<MyPoint>( colored_cloud, color_handler, "cloud" );
        viewer.VP()->addPointCloud( colored_cloud_ptr );
        viewer.VP()->spin();

        return (0);
    }
} // ns am


#endif // SEGMENTATION_HPP
