#include "AnnotationController.h"

#include <boost/filesystem/operations.hpp>

#include "pcl/io/ply_io.h"
#include "pcl/io/vtk_lib_io.h"
#include "CuboidUtil.h"

#include "GrabcutResult.h"
#include "cuboid_approx/util/MyIntrinsics.h"
#include "DepthProcessor.h"
#include "MyGrabCutEnv.h"
#include "GrabCutTest.h"
#include "Cuboids.h"
#include "MyVoxelGrid.h"
#include "cuboid_approx/primitives/Contact.h"

#if CA_WITH_BFCUDA
    #include "BilateralFilterCuda.hpp"
#endif
#include "QVTKWidget.h" // for ui update
#include "cuboid_approx/mapBack.h"
#include "cuboid_approx/ui/DisplayedPrimitive.h"
#include "cuboid_approx/AnnotationState.h"

// for screenshot
#include "vtkRenderWindow.h"
#include "vtkWindowToImageFilter.h"
#include "vtkPNGWriter.h"
#include "vtkBMPWriter.h"
#include "vtkCamera.h"
#include "vtkImageActor.h"

// clustering
#include "pcl/segmentation/extract_clusters.h"

#include "cuboid_approx/proc3d/CloudColouring.h"

namespace am
{
    using ca::AnnotationState;
    using ca::DisplayedPrimitive;

    /// ------------------------------------------------------------------------------------------------ ///
    /// CONSTs
    const Eigen::Vector4f AnnotationController::CUBOID_COLOUR               ( .9f, .9f, .0f, .15f );
//    const Eigen::Vector4f AnnotationController::CUBOID_FLOOR_COLOUR         ( 214.f/255.f, 210.f/255.f, 196.f/255.f, .15f );
//    const Eigen::Vector4f AnnotationController::CUBOID_WALL_COLOUR          ( 214.f/255.f, 210.f/255.f, 196.f/255.f, .15f );
    const Eigen::Vector4f AnnotationController::CUBOID_FLOOR_COLOUR         ( 173.f/255.f, 173.f/255.f, 193.f/255.f, .15f );
    const Eigen::Vector4f AnnotationController::CUBOID_WALL_COLOUR          ( 173.f/255.f, 173.f/255.f, 193.f/255.f, .15f );
    const double          AnnotationController::CUBOID_SELECTED_LINE_WIDTH  ( 8.         );
    const double          AnnotationController::CUBOID_UNSELECTED_LINE_WIDTH( 2.         );
    const float           AnnotationController::INTERSECT_OFFSET            ( 0.005f );
//    const Eigen::Vector4f AnnotationController::CONTACT_TOUCHING_COLOUR     ( 80.f/255.f, 7.f/255.f, 120.f/255.f, .8f );
//    const Eigen::Vector4f AnnotationController::CONTACT_FIXED_COLOUR        ( 0.f, 40.f/255.f, 255.f/255.f, .8f );


    /// ------------------------------------------------------------------------------------------------ ///
    /// LOCAL
    inline pcl::PointXYZRGB
    PointXYZRGB( float x, float y, float z, int r = 255, int g = 255, int b = 255 )
    {
        pcl::PointXYZRGB pnt;
        pnt.x = x;
        pnt.y = y;
        pnt.z = z;
        pnt.r = r;
        pnt.g = g;
        pnt.b = b;

        return pnt;
    }

    std::string
    _auxCloudName( int id )
    {
        char title[255];
        sprintf( title, "aux_cloud%d", id );
        return title;
    }

    // should be moved to amutilPCL later
    std::vector< std::vector<Eigen::Vector3f> >
    getCamerasPosFocalUp( ::pcl::visualization::PCLVisualizer::Ptr viewer )
    {
        std::vector< std::vector<Eigen::Vector3f> > ret;

        auto rens = viewer->getRendererCollection();
        rens->InitTraversal ();
        vtkRenderer* renderer = NULL;
        while ((renderer = rens->GetNextItem ()) != NULL)
        {
            ret.push_back( std::vector<Eigen::Vector3f>() );

            auto pos         = renderer->GetActiveCamera()->GetPosition();
            ret.back().push_back( (Eigen::Vector3f() <<         pos[0],         pos[1],         pos[2]).finished() );

            auto focal_point = renderer->GetActiveCamera()->GetFocalPoint();
            ret.back().push_back( (Eigen::Vector3f() << focal_point[0], focal_point[1], focal_point[2]).finished() );

            auto view_up     = renderer->GetActiveCamera()->GetViewUp();
            ret.back().push_back( (Eigen::Vector3f() <<     view_up[0],     view_up[1],     view_up[2]).finished() );

            /*std::cout << renderer->GetActiveCamera ()->GetPosition ()[0] << ", ";
            std::cout << renderer->GetActiveCamera ()->GetPosition ()[1]<< ", ";
            std::cout << renderer->GetActiveCamera ()->GetPosition ()[2]<< "\n";
            std::cout << renderer->GetActiveCamera ()->GetFocalPoint ()[0]<< ", ";
            std::cout << renderer->GetActiveCamera ()->GetFocalPoint ()[1]<< ", ";
            std::cout << renderer->GetActiveCamera ()->GetFocalPoint ()[2]<<"\n";
            std::cout << renderer->GetActiveCamera ()->GetViewUp ()[0]<< ", ";
            std::cout << renderer->GetActiveCamera ()->GetViewUp ()[1]<< ", ";
            std::cout << renderer->GetActiveCamera ()->GetViewUp ()[2]<<"\n";*/
        }

        return ret;
    }

    // should be moved to amutilPCL later
    int
    setCameraPosFocalUp( ::pcl::visualization::PCLVisualizer::Ptr viewer
                         , std::vector<Eigen::Vector3f>           pos_focal_up
                         , int                                    viewport_id  = 0 )
    {
        int vid = 0;

        auto rens = viewer->getRendererCollection();
        rens->InitTraversal ();
        vtkRenderer* renderer = NULL;
        while ((renderer = rens->GetNextItem ()) != NULL)
        {
            if ( (viewport_id == 0) || (vid == viewport_id) )
            {
                auto active_cam = renderer->GetActiveCamera();

                if ( active_cam )
                {
                    renderer->GetActiveCamera()->SetPosition  ( pos_focal_up[0][0], pos_focal_up[0][1], pos_focal_up[0][2] );
                    renderer->GetActiveCamera()->SetFocalPoint( pos_focal_up[1][0], pos_focal_up[1][1], pos_focal_up[1][2] );
                    renderer->GetActiveCamera()->SetViewUp    ( pos_focal_up[2][0], pos_focal_up[2][1], pos_focal_up[2][2] );
                }
                else
                    std::cerr << __func__ << ": active_cam nullptr..." << std::endl;

                /*std::cout << renderer->GetActiveCamera ()->GetPosition ()[0] << ", ";
                std::cout << renderer->GetActiveCamera ()->GetPosition ()[1]<< ", ";
                std::cout << renderer->GetActiveCamera ()->GetPosition ()[2]<< "\n";
                std::cout << renderer->GetActiveCamera ()->GetFocalPoint ()[0]<< ", ";
                std::cout << renderer->GetActiveCamera ()->GetFocalPoint ()[1]<< ", ";
                std::cout << renderer->GetActiveCamera ()->GetFocalPoint ()[2]<<"\n";
                std::cout << renderer->GetActiveCamera ()->GetViewUp ()[0]<< ", ";
                std::cout << renderer->GetActiveCamera ()->GetViewUp ()[1]<< ", ";
                std::cout << renderer->GetActiveCamera ()->GetViewUp ()[2]<<"\n";*/
            }

            ++vid;
        }

        return EXIT_SUCCESS;
    }

    int
    _printCamInfo( pcl::visualization::PCLVisualizer::Ptr viewer )
    {
        std::vector<pcl::visualization::Camera> cameras;
        viewer->getCameras( cameras );
        auto camera = cameras[0];

        Eigen::Matrix4d view_mat;
        camera.computeViewMatrix( view_mat );
        std::cout << "view_mat:\n" <<  view_mat << std::endl;

        Eigen::Matrix4d proj;
        camera.computeProjectionMatrix( proj );
        std::cout << "projmatrix:\n " << proj << std::endl;

        std::cout << "camera.window_size: " << camera.window_size[0] << ", " << camera.window_size[1] << std::endl;
        std::cout << "camera.fovy: " << camera.fovy << std::endl;
        std::cout << "camera.clip: " << camera.clip[0] << ", " << camera.clip[1] << std::endl;
        std::cout << "camera.window_pos: " << camera.window_pos[0] << ", " << camera.window_pos[1] << std::endl;

        auto vis_cameras = getCamerasPosFocalUp( viewer );
        for ( int cam_id = 0; cam_id != vis_cameras.size(); ++cam_id )
        {
            std::cout << "camera" << cam_id << std::endl;
            for ( int vec_id = 0; vec_id != vis_cameras[cam_id].size(); ++vec_id )
            {
                std::cout << "vec" << vec_id << vis_cameras[cam_id][vec_id].transpose() << std::endl;
            }
        }
    }

    Eigen::Matrix4d
    AnnotationController::_getProjMatrix( pcl::visualization::PCLVisualizer::Ptr viewer, int viewport )
    {
        std::vector<pcl::visualization::Camera> cameras;
        viewer->getCameras( cameras );
        auto camera = cameras[viewport];

        Eigen::Matrix4d proj;
        camera.computeProjectionMatrix( proj );
        return proj;
    }

    Eigen::Matrix4d
    AnnotationController::_getViewMatrix( pcl::visualization::PCLVisualizer::Ptr viewer, int viewport )
    {
        std::vector<pcl::visualization::Camera> cameras;
        viewer->getCameras( cameras );
        auto camera = cameras[viewport];

        Eigen::Matrix4d view_mat;
        camera.computeViewMatrix( view_mat );
        return view_mat;
    }

    /// ------------------------------------------------------------------------------------------------ ///
    /// CONSTR
    AnnotationController::AnnotationController()
        : _viewer( new pcl::visualization::PCLVisualizer("Viewer", false) )
        //, _depth_intrinsics( 521.7401f, 522.1379f, 323.4402f, 258.1387f, 0.f, 0.f, 0.f, 0.f, 0.f, 0.f ) // arons
        , _depth_intrinsics( 5.1930334103339817e+02f, 5.1816401430246583e+02f, 3.2850951551345941e+02f, 2.5282555217253503e+02f )
        , _colors( am::util::nColoursCv2(/* n: */ 50, /* scale: */ 1.f) )
        , _colorHead( 0 )
        , _output_path( am::util::timestamp2Str() )
        , _colour_code_cuboids( false )
        , _default_opacity( .15f )
        , _mouse_state( 0 )
    {
//        std::cout<<"_colors:";for(size_t vi=0;vi!=11;++vi)std::cout<<_colors[vi]<<"\n";std::cout << "\n";
//        std::vector<std::vector<float>> tmp_colours =
//        {{1.000000,1.000000,0.000000},
//        {0.710820,0.192403,0.049229},
//        {.900000,1.000000,0.682353},
//        {0.326217,0.781641,0.061963},
//        {0.967051,0.030074,0.647523},
//        {0.400000,1.000000,0.960784}};
//        for ( int i = 0; i < tmp_colours.size(); ++i )
//        {
//            _colors[i+1] = cv::Point3f( tmp_colours[i][0], tmp_colours[i][1], tmp_colours[i][2] );
//        }
//        std::cout<<"_colors:";for(size_t vi=0;vi!=11;++vi)std::cout<<_colors[vi]<<"\n";std::cout << "\n";

        _viewer->setShowFPS( false );
        //_viewer->setBackgroundColor( .6, .6, .7, 0 );
        _viewer->setBackgroundColor( 1, 1, 1, 0 );
        //_viewer->addCoordinateSystem( 0.1, "coord_sys", 0 );

//        for ( auto &colour : _colors)
//        {
//            if ( (colour.x > 0.9f) && (colour.y > 0.9f) && (colour.z > 0.9f) )
//                std::cerr << colour.x << ", " << colour.y << ", " << colour.z << std::endl;
//            else
//                std::cout << colour.x << ", " << colour.y << ", " << colour.z << std::endl;
//        }

        connect( this, SIGNAL(pointPickedSignal(int)), this, SLOT(pointPickedSlot(int)) );
    }

    int
    AnnotationController::initViewer()
    {
        // interactor_style
        {
            _interactor_style = vtkSmartPointer<ca::Annot3DInteractorStyle>( ca::Annot3DInteractorStyle::New() );
            _interactor_style->Initialize();
            //_interactor_style->setKeyboardModifier( pcl::visualization::INTERACTOR_KB_MOD_SHIFT );
            vtkSmartPointer<vtkRendererCollection> rens =  _viewer->getRendererCollection();
            _interactor_style->setRendererCollection( rens );
            //        _interactor_style->setCloudActorMap( _vis->getCloudActorMap() );

            // setup mouse callbacks
            boost::function<void (const pcl::visualization::MouseEvent)> fMouse = boost::bind( &am::AnnotationController::_mouseCallbacks, this, _1 );
            boost::signals2::connection cMouse = _interactor_style->registerMouseCallback( fMouse );

            _viewer->setupInteractor( _viewer->getRenderWindow()->GetInteractor(), _viewer->getRenderWindow(), _interactor_style );
        }

        return EXIT_SUCCESS;
    }

    /// ------------------------------------------------------------------------------------------------ ///
    /// Events

    int
    AnnotationController::loadNewCloudFromImagePaths(   PCloud::Ptr               cloud
                                                      , std::string        const  img_path
                                                      , std::string        const  depth_path
                                                      , MyIntrinsics       const &depth_intrinsics
                                                      , float              const  spatial_sigma
                                                      , float              const  range_sigma
                                                      , int                const  kernel_radius
                                                      , int                const  iteration_count
                                                      , cv::Mat                  *p_img
                                                      , cv::Mat                  *p_dep
                                                      , PCloud::Ptr              *p_aux_cloud
                                                      )
    {
        int err = EXIT_SUCCESS;

        // load image
        cv::Mat img = cv::imread( img_path, cv::IMREAD_UNCHANGED );
        if      ( img.empty()            ) { /*std::cerr << "img empty"               << std::endl;*/ return EXIT_FAILURE; }
        else if ( img.type() != CV_8UC3  ) { std::cerr << "img not CV_8UC3"         << std::endl; return EXIT_FAILURE; }

        // load depth
        cv::Mat dep = am::CuboidUtil::loadDepth( depth_path );
        if      ( dep.empty()            ) { std::cerr << "dep empty"               << std::endl; return EXIT_FAILURE; }
        else if ( dep.type() != CV_16UC1 ) { std::cerr << "depth type not CV_16UC1" << std::endl; return EXIT_FAILURE; }

#if CA_WITH_BFCUDA
        // filter depth
        if ( (spatial_sigma > .0f) && (range_sigma > .0f) )
        {
            cv::Mat dep2;
            std::cout << __func__ << ": spatial_sigma, range_sigma, kernel_radius, iteration_count"
                      << spatial_sigma << ", " << range_sigma << ", " << kernel_radius << ", " << iteration_count << std::endl;
            BilateralFilterCuda<float> bfc;
            bfc.setIterations( iteration_count );
            bfc.setFillMode( FILL_NON_ZEROS | SKIP_ZEROS );
            bfc.runBilateralFiltering( dep, img, dep2, spatial_sigma, range_sigma, kernel_radius );
            dep = dep2;
        }
#endif // CA_WITH_BFCUDA

        // cloud - create coloured cloud
        {
            err += am::DepthProcessor::cloud_from_images<unsigned short>( /*        output cloud: */ cloud,
                                                                          /*                 rgb: */ img,
                                                                          /*               depth: */ dep,
                                                                          /*          intrinsics: */ depth_intrinsics,
                                                                          /* p_invalid_cloud_ptr: */ p_aux_cloud );
        }

        // return
        if ( p_img ) *p_img = img;
        if ( p_dep ) *p_dep = dep;

        return err;
    }
    int
    AnnotationController::loadNewCloudFromImagePaths(   std::string const img_path
                                                      , std::string const depth_path
                                                      , float             spatial_sigma
                                                      , float             range_sigma
                                                      , int               kernel_radius
                                                      , int               iteration_count )
    {
        // load image
        //cv::Mat _img;
        _img = cv::imread( img_path, cv::IMREAD_UNCHANGED );
        if      ( _img.empty()            ) {    return EXIT_FAILURE; }
        else if ( _img.type() != CV_8UC3  ) { std::cerr << "img not CV_8UC3"         << std::endl; return EXIT_FAILURE; }

        // load depth
        _dep = am::CuboidUtil::loadDepth( depth_path );
        if      ( _dep.empty()            ) { std::cerr << "dep empty"               << std::endl; return EXIT_FAILURE; }
        else if ( _dep.type() != CV_16UC1 ) { std::cerr << "depth type not CV_16UC1" << std::endl; return EXIT_FAILURE; }

#if CA_WITH_BFCUDA
        // filter depth
        if ( (spatial_sigma > .0f) && (range_sigma > .0f) )
        {
            cv::Mat dep2;
            std::cout << __func__ << ": spatial_sigma, range_sigma, kernel_radius, iteration_count"
                      << spatial_sigma << ", " << range_sigma << ", " << kernel_radius << ", " << iteration_count << std::endl;
            BilateralFilterCuda<float> bfc;
            //bfc.setGaussianParameters( spatial_sigma, kernel_radius );
            bfc.setIterations( iteration_count );
            bfc.setFillMode( FILL_NON_ZEROS | SKIP_ZEROS );
            bfc.runBilateralFiltering( _dep, _img, dep2, spatial_sigma, range_sigma, kernel_radius );
            _dep = dep2;
        }
#endif

        // cloud - create coloured cloud
        {
            _state._aux_clouds.clear();
            _state._aux_clouds.push_back( nullptr );
            am::DepthProcessor::cloud_from_images<unsigned short>( /*        output cloud: */ _state.MainCloudRef( AnnotationState::FULL ), //_state._main_cloud_ptr,
                                                                   /*                 rgb: */ _img,
                                                                   /*               depth: */ _dep,
                                                                   /*          intrinsics: */ _depth_intrinsics,
                                                                   /* p_invalid_cloud_ptr: */ &(_state._aux_clouds[0]) );
        }

        // display
        {
            _viewer->removeAllPointClouds();
            _viewer->removeAllShapes();
            //_viewer->removePolygonMesh( )
            _state._drawn_models.clear();

            //std::cout << "adding point cloud: big cloud" << std::endl;
            am::CloudViewer::addCloud( _viewer, _state.MainCloud(), "big_cloud" );

            _state._objects.clear();
            _state._next_primitive_id = 0;
        }

        // copy to "current cloud"
        {
            if ( !_state.CurrentCloud() )
                _state.CurrentCloudRef( AnnotationState::FULL ).reset( new PCloud() );
            pcl::copyPointCloud( *_state.MainCloud(),
                                 *_state.CurrentCloudRef(AnnotationState::FULL) );
        }
        // create indices
        {
            _state.CurrentIndicesRef() = pcl::PointIndices::Ptr( new pcl::PointIndices );
            _state.CurrentIndicesRef()->indices.reserve( _state.MainCloud()->size() );
            for ( int point_id = 0; point_id < _state.MainCloud()->size(); ++point_id )
            {
                _state.CurrentIndicesRef()->indices.push_back( point_id );
            }
        }

        return EXIT_SUCCESS;
    }

    // \param img_path
    // \param dep_path
    // \param plane_distance_threshold
    int
    AnnotationController::annotate_button_clicked( Params const& params )
    {
        // add main cloud, if needed
        if ( !_state.MainCloud() )
        {
            loadNewCloudFromImagePaths( params.getString("img_path"),
                                        params.getString("dep_path") );
        }

        // route task_id
        int task_id = params.getInt( "task_selected" );
        switch ( task_id )
        {
            // PLANES
            case 0:
            {
                // grabcut annotation
                am::GrabcutResult                 gc_results;
                std::shared_ptr<am::MyGrabCutEnv> p_gc3( new am::GrabCutDTest );
                {
                    // grabcut
                    am::Params gc_params;
                    {
                        gc_params.setDouble( "betas[0]", params.getDouble("betas[rgb]") );
                        gc_params.setDouble( "betas[1]", params.getDouble("betas[depth]") );
                    }

                    int gcres = p_gc3->run( /*     output: */ gc_results,
                                            ///*   img_path: */ params.getString("img_path"),
                                            ///* depth_path: */ params.getString("dep_path"),
                                            /*        img: */ _img,
                                            /*      depth: */ _dep,
                                            /*      cheat: */ false,
                                /*     params: */ &gc_params );
                    if ( gcres )
                    {
                        p_gc3->destroyWindows();
                        break;
                    }
                }

                // estimate planes from segmentation
                am::Planes planes;
                {
                    cv::Mat tmp;
                    gc_results.get_depth().convertTo( tmp, CV_32FC1, 1.f / 1000.f );
                    gc_results.set_depth( tmp );

                    Planes::from_masks<float>( planes,
                                               gc_results,
                                               _state.MainCloud(),
                                               _state.CurrentIndices(),
                                               params.getDouble("plane_distance_threshold") );

                }

                // apply planes
                {
                    // for every plane
                    int plane_id = 0;
                    for ( Plane & plane : planes.PrimitivesRef() )
                    {
#if 0
                        // select inliers
                        pcl::PointIndices::Ptr inlier_indices_ptr( new pcl::PointIndices );
                        {
                            plane.getInlierIndices( /* out_indices: */ inlier_indices_ptr,
                                                    /*   cloud_ptr: */ _state._main_cloud_ptr,
                                                    /* indices_ptr: */ _state._current_indices_ptr,
                                                    /*   threshold: */ params.getDouble("plane_distance_threshold") );
                        }

                        // exclude plane's points from active cloud
                        {
                            am::util::algo::vectorDifference( /* A-B: */ _state._current_indices_ptr->indices,
                                                              /*   A: */ _state._current_indices_ptr->indices,
                                                              /*   B: */  inlier_indices_ptr->indices );
                        }
#endif

                        am::Cuboid cuboid;
                        {
                            // make it extend backwards
                            if ( plane.Normal().dot( (Eigen::Vector3f()<<0.f,0.f,1.f).finished() ) > 0.f )
                            {
                                plane.flipNormal();
                            }
                            Cuboid::from1Plane( cuboid, plane, params.getDouble("plane_distance_threshold") );
                            // comment this out, if you don't want all inliers
                            {
                                plane.getInlierIndices( /* out_indices: */ cuboid.PlanesRef()[0].IndicesPtrRef(),
                                        /*   cloud_ptr: */ _state.MainCloud(),
                                        /* indices_ptr: */ _state.CurrentIndices(),
                                        /*   threshold: */ params.getDouble("plane_distance_threshold") );
                                cuboid.IndicesPtrRef()->indices = cuboid.Planes()[0].IndicesPtr()->indices;
                            }

                            Cuboid::verify( cuboid );
                            _addCuboid( cuboid, _state._next_primitive_id++ );
                        }

                        // name
                        char plane_name[255];
                        {
                            sprintf( plane_name, "plane%02d", plane_id );
                        }

                        // show plane
                        {
                            // colour inliers
                            {
                                cv::Point3f c = _colors[ _colorHead++ % _colors.size() ];
                                for ( int point_id = 0; point_id < plane.IndicesPtr()->indices.size(); ++point_id )
                                {
                                    int index    = plane.IndicesPtr()->indices[ point_id ];
                                    MyPoint &pnt = _state.CurrentCloud()->at( index );

                                    am::util::pcl::setPointColor( pnt, c.x, c.y, c.z ); // needs 255
                                }
                                am::CloudViewer::updatePointCloud( _viewer, _state.CurrentCloud(), "big_cloud", nullptr, 0, _state._display_point_size );
                            }
                        }

                        // save plane
                        {
                            std::cout << "saving " << plane_name << std::endl;
                            pcl::io::savePCDFile( /*    path: */  std::string( plane_name ) + ".pcd",
                                                  /*   cloud: */ *plane.CloudPtr(),
                                                  /* indices: */  plane.IndicesPtr()->indices );
                        }

                        ++plane_id;
                    }
                }

                // destroy cv windows
                {
                    p_gc3->destroyWindows();
                }

                // save planes to file
                {
                    if ( !boost::filesystem::create_directory( _output_path )
                         && !boost::filesystem::exists(_output_path) )
                    {
                        std::cerr << "WARNING: could not create and does not exist: " << _output_path << std::endl;
                        //throw std::runtime_error("could not create path");
                    }
                    Planes::writeToFile( planes, _output_path + "/planes.dat" );
                }

                break;
            } // case 0

                // CUBOIDS
            case 1:
            {
                am::GrabcutResult                 gc_results;
                std::shared_ptr<am::MyGrabCutEnv> p_gc3( new am::GrabCutDTest );
                {
                    // grabcut
                    am::Params gc_params;
                    gc_params.setDouble( "betas[0]", params.getDouble("betas[rgb]") );
                    gc_params.setDouble( "betas[1]", params.getDouble("betas[depth]") );

                    int gcres = p_gc3->run( /*     output: */ gc_results,
                                            ///*   img_path: */ params.getString("img_path"),
                                            ///* depth_path: */ params.getString("dep_path"),
                                            /*        img: */ _img,
                                            /*      depth: */ _dep,
                                            /*      cheat: */ false,
                                            /*     params: */ &gc_params );

                    if ( gcres != EXIT_SUCCESS )
                    {
                        p_gc3->destroyWindows();
                        break;
                    }
                }

                // Cuboid estimation
                am::Cuboids cuboids;
                {

                    {
                        double minVal, maxVal;
                        cv::minMaxLoc( gc_results.get_depth(), &minVal, &maxVal );
                        std::cout << "gc_results.get_depth() before: "
                                  << gc_results.get_depth().rows << ", "
                                  << gc_results.get_depth().cols << ", "
                                  << gc_results.get_depth().channels() << ", " << maxVal << std::endl;
                    }

                    cv::Mat tmp;
                    gc_results.get_depth().convertTo( tmp, CV_32FC1, 1.f / 1000.f );
                    gc_results.set_depth( tmp );

                    Cuboids::from_masks<float>( /*            output: */ cuboids,
                                                /*       input masks: */ gc_results,
                                                /*        main cloud: */ _state.MainCloud(),
                                                /*    active indices: */ _state.CurrentIndices(),
                                                /* distanceThreshold: */ params.getDouble("plane_distance_threshold") );
                }

                // save cuboid to disk
                for ( int cuboid_id = 0; cuboid_id < cuboids.size(); ++cuboid_id )
                {
                    char tmp[1024];
                    sprintf( tmp, "cuboid%02d.pcd", cuboid_id );
                    std::cout << "saving " << tmp << std::endl;
                    pcl::io::savePCDFile( std::string(tmp),
                                          *cuboids[cuboid_id].CloudPtr(),
                                          cuboids[cuboid_id].IndicesPtr()->indices );
                    std::cerr << "saved " << cuboids[cuboid_id].IndicesPtr()->indices.size() << " points" << std::endl;

                    //                    cuboid.Planes().colourCloud( _state._current_cloud_ptr,
                    //                                                 &_colors );
                }

                // destroy cv windows
                {
                    p_gc3->destroyWindows();
                }

                _addCuboids( cuboids );

                break;
            } // case 1

            default:
            {
                std::cerr << "AnnotationController::annotate_button_clicked(): uncrecognized task id..." << std::endl;
                return EXIT_FAILURE;
                break;
            }
        }

        fflush( stdout );

        return EXIT_SUCCESS;
    }

    int
    AnnotationController::selectPrimitive( std::vector<int> ids )
    {
        auto it = _state._objects.begin();
        for ( int pos = 0; pos < _state._objects.size(); ++pos, ++it )
        {
            if ( it->second.hidden )    continue;

            if ( it->second.primitive_ptr )
            {
                it->second.reDraw( _viewer
                                   , /* name: */ it->first
                                   , /* selected: */std::find( ids.begin(), ids.end(), pos )!=ids.end()
                                                                                              , /* do_colour_code_cuboids: */ _colour_code_cuboids
                                   , /* colour_edges: */ _state._colour_edges
                                   , /* showName: */ _state._show_IDs
                                   //, it->second.shape
                                   );
                continue;
            }

            //            it->second.primitive_ptr->unDraw( _viewer, it->first );

            //            // redraw cuboid
            //            std::shared_ptr<am::Cuboid> cuboid_ptr = std::static_pointer_cast< am::Cuboid,am::Primitive>( it->second.primitive_ptr );
            //            if ( cuboid_ptr != nullptr )
            //            {
            //                auto colour = it->second.getDrawingColour( _colour_code_cuboids );

            //                cuboid_ptr->draw4( _viewer
            //                                  , it->first
            //                                  , /*     colour: */ colour
            //                                  , /* line_width: */ (std::find( ids.begin(), ids.end(), pos )!=ids.end()) ? CUBOID_SELECTED_LINE_WIDTH : CUBOID_UNSELECTED_LINE_WIDTH
            //                                  , _state._colour_edges
            //                                  , /* showName: */ _state._show_IDs );
            //                continue;
            //            }

            // redraw plane // TODO
            std::cerr << "planes are not redrawn yet" << std::endl;
        }
    }

    int
    AnnotationController::toggleHidePrimitive( int id )
    {
        auto it = _state._objects.begin();
        std::advance( it, id );
        std::shared_ptr<am::Cuboid> cuboid_ptr = std::static_pointer_cast< am::Cuboid,am::Primitive>( it->second.primitive_ptr );

        if ( it->second.hidden = !it->second.hidden )
        {
            cuboid_ptr->unDraw( _viewer, it->first );

            // TODO: remove this, it is read from file...
            /*
             *pcl::PointIndices::Ptr cuboid_inlier_indices_ptr;
            cuboid_ptr->getInlierIndices( cuboid_inlier_indices_ptr,
                                          _state._main_cloud_ptr,
                                          nullptr );
            _removePoints( cuboid_inlier_indices_ptr );
            */

            _removePoints( cuboid_ptr->IndicesPtr() );
        }
        else
        {
            //cuboid_ptr->draw( _viewer, it->first,  );
            //it->second.draw( _viewer, it->first )
            it->second.reDraw( _viewer
                               , /*                name: */ it->first
                               , /*            selected: */ true
                               , /* do_colour_code_cubs: */ _colour_code_cuboids
                               , /*        colour_edges: */ _state._colour_edges
                               , /*            showName: */ _state._show_IDs );
                               //, it->second.shape );

            /*pcl::PointIndices::Ptr cuboid_inlier_indices_ptr;
            cuboid_ptr->getInlierIndices( cuboid_inlier_indices_ptr,
                                          _state._main_cloud_ptr,
                                          nullptr );*/
            _addPoints( cuboid_ptr->IndicesPtr() );
        }

        return EXIT_SUCCESS;
    }

    int
    AnnotationController::saveCuboidsToPLY( std::string path )
    {
        typedef pcl::PointXYZRGB PointT;
        pcl::PolygonMesh       ::Ptr mesh ( new pcl::PolygonMesh       ()  );
        pcl::PointCloud<PointT>::Ptr cloud( new pcl::PointCloud<PointT>()  );

        int polygon_id = 0;
        for ( auto object_pair : _state._objects )
        {
            std::shared_ptr<am::Cuboid> cuboid_ptr = std::static_pointer_cast< am::Cuboid,am::Primitive>( object_pair.second.primitive_ptr );
            if ( !cuboid_ptr ) continue;

            Eigen::Vector4f colour = object_pair.second.getDrawingColour( true );

            mesh->polygons.resize( mesh->polygons.size() + 12 );
            for ( int plane_id = 0; plane_id != cuboid_ptr->Planes().size(); ++plane_id )
            {
                Plane const& plane = cuboid_ptr->Planes().at( plane_id );

                for ( int corner_id = 0; corner_id != 4; ++corner_id )
                {
                    // pnt
                    pcl::PointXYZRGB pnt;
                    pnt.x = plane.Corners()[corner_id].x;
                    pnt.y = plane.Corners()[corner_id].y;
                    pnt.z = plane.Corners()[corner_id].z;
                    pnt.r = colour(0) * 255.f;
                    pnt.g = colour(1) * 255.f;
                    pnt.b = colour(2) * 255.f;
                    cloud->push_back( pnt );

                    // polygon
                    if ( corner_id != 3 )
                        mesh->polygons[ polygon_id ].vertices.push_back( cloud->size() - 1 );
                    if ( corner_id != 1 )
                        mesh->polygons[ polygon_id+1 ].vertices.push_back( cloud->size() - 1 );
                }

                polygon_id += 2;
            }
        }

        // grab_060514h/1744
        cloud->reserve( cloud->size() + _state.MainCloud()->size() );
        for ( int pid = 0; pid != _state.MainCloud()->size(); ++pid )
        {
            auto pnt = _state.MainCloud()->at( pid );
            cloud->push_back( PointXYZRGB(pnt.x,pnt.y,pnt.z,pnt.r,pnt.g,pnt.b) );
        }

        // convert cloud to pclpointcloud2
        pcl::PCLPointCloud2::Ptr cloud2( new pcl::PCLPointCloud2() );
        pcl::toPCLPointCloud2( *cloud, *cloud2 );
        mesh->cloud = *cloud2;

        return pcl::io::savePLYFile( path, *mesh );
    }

    int
    AnnotationController::save( std::string directory, const Params &params )
    {
        if (    !boost::filesystem::create_directory( directory )
             && !boost::filesystem::exists(directory)            )
        {
            std::cerr << "WARNING: could not create and does not exist: " << directory << std::endl;
        }

        am::Planes planes;
        am::Cuboids cuboids;
        for ( auto object_pair : _state._objects )
        {
            std::shared_ptr<am::Cuboid> cuboid_ptr = std::static_pointer_cast< am::Cuboid,am::Primitive>( object_pair.second.primitive_ptr );
            if ( cuboid_ptr != nullptr )
            {
                cuboids.add( *cuboid_ptr );
                //cuboids.back().CloudPtrRef() = _state._main_cloud_ptr;
//                cuboids.back().getInlierIndices( cuboids.back().IndicesPtrRef(),
//                                                 _state._main_cloud_ptr,
//                                                 nullptr );
                continue;
            }

            std::shared_ptr<am::Plane> plane_ptr   = std::static_pointer_cast< am::Plane, am::Primitive>( object_pair.second.primitive_ptr );
            if ( plane_ptr != nullptr )
            {
                planes.add( *plane_ptr );
                continue;
            }

            std::cout << "damn..." << std::endl;
        }

        //cv::Mat depth = am::CuboidUtil::loadDepth( params.getString("dep_path") );
        cv::Mat const& depth = _dep;
        Cuboids::writeToFile( cuboids, directory + "/cuboids.dat", depth.rows, depth.cols );
        Planes::writeToFile( planes, directory + "/planes.dat" );

        // save depth dat
        boost::filesystem::path dep_path( params.getString("dep_path") );
        if ( boost::filesystem::exists( dep_path ) )
        {
            //cv::Mat depth = am::CuboidUtil::loadDepth( dep_path.string() ); //cv::imread( dep_path.string(), cv::IMREAD_UNCHANGED );
            am::util::cv::saveDepthAsBinary<ushort,float>( depth, directory + "/" + dep_path.stem().string() + ".dat", 1.f / 1000.f );
            std::cout << "saved " << _output_path + "/" + dep_path.stem().string() + ".dat" << std::endl;
        }

        // copy depth
        {
            std::string command = "cp " + dep_path.string() + " " + directory + "/" + dep_path.filename().string();
            int err = system( command.c_str() );
            if ( err != EXIT_SUCCESS ) { std::cerr << "[" << __func__ << "]: " << "\"system(" << command << "\") failed with code " << err << std::endl; }
        }

        // copy rgb
        {
            boost::filesystem::path rgb_path( params.getString("img_path") );
            if ( boost::filesystem::exists( rgb_path ) )
            {
                std::string command = "cp " + rgb_path.string() + " " + directory + "/" + rgb_path.filename().string();
                int err = system( command.c_str() );
                if ( err != EXIT_SUCCESS ) { std::cerr << "[" << __func__ << "]: " << "\"system(" << command << "\") failed with code " << err << std::endl; }
            }
        }

        // screenshot
        {
            //_viewer->saveScreenshot( directory + "/screenshot.png" );
            vtkWindowToImageFilter *windowToImage = vtkWindowToImageFilter::New();
            //QVTKWidget *windowVTK = qobject_cast<QVTKWidget *>(activeWindow);

            ///Capture Window
            auto p_render_window = _viewer->getRenderWindow();
            windowToImage->SetInput( p_render_window.GetPointer() );
            windowToImage->Update();

            ///Write/Save as PNG
            vtkPNGWriter *writer = vtkPNGWriter::New();
            writer->SetInputConnection(windowToImage->GetOutputPort());
            writer->SetFileName( (directory + "/screenshot.png").c_str() );
            _viewer->getRenderWindow()->Render();
            writer->Write();
            writer->Delete();
        }

        // zip
        {
            char command[1024];
            sprintf( command, "find %s -path '*/.*' -prune -o -type f -print | zip %s.zip -@ -j", directory.c_str(), directory.c_str() );
            std::cout << "executing " << command << std::endl;
            {
                int err = system( command );
                if ( err != EXIT_SUCCESS ) { std::cerr << "[" << __func__ << "]: " << "\"system(" << command << "\") failed with code " << err << std::endl; }
            }
        }
    }

    int
    AnnotationController::saveScreenshot( std::string path )
    {
        vtkWindowToImageFilter *windowToImage = vtkWindowToImageFilter::New();

        ///Capture Window
        auto p_render_window = _viewer->getRenderWindow();
        windowToImage->SetInput( p_render_window.GetPointer() );
        windowToImage->Update();

        vtkImageWriter* writer = NULL;
        if ( path.find( "png") != std::string::npos )
            writer = vtkPNGWriter::New();
        else if ( path.find("bmp") != std::string::npos )
            writer = vtkBMPWriter::New();
        else
        {
            std::cerr << "[" << __func__ << "]: " << "extension not recognized...exiting" << std::endl;
            return EXIT_FAILURE;
        }

        ///Write/Save as PNG
        //vtkPNGWriter *writer = vtkPNGWriter::New();
        //vtkBMPWriter *writer = vtkBMPWriter::New();

        writer->SetInputConnection(windowToImage->GetOutputPort());
        writer->SetFileName( (path.substr(0,path.rfind(".")) + "2.png").c_str() );
        _viewer->getRenderWindow()->Render();
        std::cout << __func__ << ": writing screenshot to " << path << std::endl;
        writer->Write();
        writer->Delete();

        // two pass rendering
        std::cout << "[" << __func__ << "]: " << "starting experiment..." << std::endl;
        {
            // create tmp window
            vtkSmartPointer<vtkRenderWindow> renderWindow = vtkSmartPointer<vtkRenderWindow>::New();
            {
                renderWindow->SetSize( _viewer->getRenderWindow()->GetSize() );
            }

            // create tmp renderer
            vtkSmartPointer<vtkRenderer> renderer = vtkSmartPointer<vtkRenderer>::New();
            {
                // add to window
                renderWindow->AddRenderer( renderer );
                // borrow camera
                renderer->SetActiveCamera( _viewer->getRendererCollection()->GetFirstRenderer()->GetActiveCamera() );
            }

            // add front content = contacts
            {
                auto actors = _viewer->getShapeActorMap(); // get all participating shapes
                for ( auto it = actors->begin(); it != actors->end(); ++it )
                {
                    if ( (it->first.find("contact") != std::string::npos) )  // render contacts only
                        renderer->AddActor( it->second );
                }
            }

            // background content
            {
                // get background image from _viewer window content
                vtkSmartPointer<vtkWindowToImageFilter> windowToImageFilter  = vtkSmartPointer<vtkWindowToImageFilter>::New();
                windowToImageFilter->SetInput( _viewer->getRenderWindow() );
                windowToImageFilter->SetInputBufferTypeToRGBA(); //also record the alpha (transparency) channel
                windowToImageFilter->Update();

                // create background image
                vtkSmartPointer<vtkImageActor> backgroundImageActor = vtkSmartPointer<vtkImageActor>::New();
                vtkSmartPointer<vtkImageData>  backgroundImageData = windowToImageFilter->GetOutput();
                backgroundImageActor->SetInput( backgroundImageData );

                // create background renderer
                vtkSmartPointer<vtkRenderer> backgroundRenderer = vtkSmartPointer<vtkRenderer>::New();
                backgroundRenderer->SetLayer( 0 );
                backgroundRenderer->InteractiveOff();
                // add background to renderer
                backgroundRenderer->AddActor( backgroundImageActor );

                // add as background to tmp output window
                renderWindow->SetNumberOfLayers( 2 );
                renderWindow->GetRenderers()->GetFirstRenderer()->SetLayer( 1 );
                renderWindow->AddRenderer( backgroundRenderer );

                // render once to figure out where the background camera will be
                renderWindow->Render();
//                renderWindow()->SetSize(640, 480);

                // set up the background camera to fill the renderer with the image
                vtkCamera* camera = backgroundRenderer->GetActiveCamera();
                {
                    double origin[3];
                    double spacing[3];
                    int extent[6];
                    backgroundImageData->GetOrigin( origin );
                    backgroundImageData->GetSpacing( spacing );
                    backgroundImageData->GetExtent( extent );

                    camera->ParallelProjectionOn();

                    double xc = origin[0] + 0.5*(extent[0] + extent[1])*spacing[0];
                    double yc = origin[1] + 0.5*(extent[2] + extent[3])*spacing[1];
                    //double xd = (extent[1] - extent[0] + 1)*spacing[0];
                    double yd = (extent[3] - extent[2] + 1)*spacing[1];
                    double d = camera->GetDistance();
                    camera->SetParallelScale(0.5*yd);
                    camera->SetFocalPoint(xc,yc,0.0);
                    camera->SetPosition(xc,yc,d);
                }
            }

            // render
            {
                renderer->Render();
            }

            // save
            {
                vtkSmartPointer<vtkWindowToImageFilter> windowToImageFilter = vtkSmartPointer<vtkWindowToImageFilter>::New();
                windowToImageFilter->SetInput(renderWindow);
                windowToImageFilter->SetInputBufferTypeToRGBA(); //also record the alpha (transparency) channel
                windowToImageFilter->Update();

                vtkSmartPointer<vtkPNGWriter> writer = vtkSmartPointer<vtkPNGWriter>::New();
                writer->SetFileName( path.c_str() );
                writer->SetInputConnection( windowToImageFilter->GetOutputPort() );
                writer->Write();
            }
        }

        return EXIT_SUCCESS;
    }

    std::string
    AnnotationController::getRotInfo()
    {
        // get floor
        auto floor_ptrs = getFloors();
        if ( floor_ptrs.empty() )
        {
            return  "Need a floor for upvector" ;
        }

        // get cuboids centroid
        auto target_ptrs = getRotationTargets();
        if ( target_ptrs.empty() )
        {
            return "Need a target for center" ;
        }

        Eigen::Vector3f centroid(0,0,0);
        // rotationcentroid
        for ( std::shared_ptr<am::Cuboid> const cuboid_ptr : target_ptrs )
        {
            centroid += cuboid_ptr->Center();
        }
        centroid /= (float)target_ptrs.size();

        // draw rotation axis
        //Eigen::Vecto3f rot_axis = floor_ptrs[0]->Planes()[2].Normal().normalized();
        std::stringstream ss;
        ss << floor_ptrs[0]->Planes()[2].Normal().normalized().transpose() << "; " << centroid << std::endl;
        return ss.str();

    }

    int
    AnnotationController::saveCameraPose( std::string path )
    {
        auto pos_focal_up = getCameraPosFocalUp();

        std::stringstream ss;
        for ( int i = 0; i != pos_focal_up.size(); ++i )
            ss << pos_focal_up[i](0) << ","
               << pos_focal_up[i](1) << ","
               << pos_focal_up[i](2) << (i < pos_focal_up.size()-1 ? "," : "");

        std::ofstream myfile;
        myfile.open ( path );
        if ( !myfile.is_open() )
        {
            std::cerr << "could not open file " << path << std::endl;
            return EXIT_FAILURE;
        }

        std::cout << "writing camera pose to " << path << std::endl;

        myfile << ss.str();
        myfile.close();

        return EXIT_SUCCESS;
    }

    int
    AnnotationController::loadCuboids( std::string path, Params const* params, std::vector<am::Contact> *contacts_arg )
    {
        //std::cout << "AnnotationController::loadCuboids()...BEGIN\n\n";

        am::Cuboids                     cuboids;
        std::vector<am::Contact>        contacts;
        std::vector<am::Primitive::Ptr> statics;
        double                          inlier_thresh = params->getDouble( "plane_distance_threshold", .15 );
        am::Cuboids::readFromFile( cuboids, path, NULL, _state.MainCloud(), nullptr, contacts_arg ? contacts_arg : &contacts, &statics, &inlier_thresh );

        boost::filesystem::path cuboids_path( path );
        _addCuboids( cuboids, /* tag: */ cuboids_path.stem().string() );

        //std::cout << "AnnotationController::loadCuboids()...END\n";

        return EXIT_SUCCESS;
    }

    int
    AnnotationController::fillPrimitive( int id )
    {
        // get cuboid
        auto it = _state._objects.begin();
        std::advance( it, id );
        std::shared_ptr<am::Cuboid> cuboid_ptr = std::static_pointer_cast< am::Cuboid,am::Primitive>( it->second.primitive_ptr );

        // get inliers (maybe unncessary)
        cuboid_ptr->getInlierIndices( cuboid_ptr->IndicesPtrRef(),
                                      _state.MainCloud(),
                                      _state.CurrentIndices() );

#if 1
        // cache vars
        am::Planes const& planes = cuboid_ptr->Planes();

        // init sac model on roi cloud
        pcl::SampleConsensusModelPlane<MyPoint> sac_model( cuboid_ptr->CloudPtr() );
        sac_model.setIndices( cuboid_ptr->IndicesPtr()->indices );
        // allocate point2plane distances
        std::vector< std::vector<double> > distances( planes.size() );
        // allocate assignment ids
        std::vector<int> closest_plane_ids( cuboid_ptr->IndicesPtr()->indices.size() );
        std::fill( closest_plane_ids.begin(), closest_plane_ids.end(), 0 );


        // assign points to planes
        for ( int plane_id = 0; plane_id < planes.size(); ++plane_id )
        {
            // calc distances
            sac_model.getDistancesToModel( planes[plane_id].Coeffs(), distances[plane_id] );
            if ( !plane_id ) continue;

            // compare to previous distances
            for ( int point_id = 0; point_id < distances[plane_id].size(); ++point_id )
            {
                if ( distances[plane_id][point_id] < distances[closest_plane_ids[point_id]][point_id] )
                    closest_plane_ids[point_id] = plane_id;
            }
        }

        // prepare output
        PCloud::Ptr filled_cloud_ptr( new PCloud() );
        // copy all points
        pcl::copyPointCloud( *cuboid_ptr->CloudPtr(), cuboid_ptr->IndicesPtr()->indices, *filled_cloud_ptr );

        MyPoint new_point;
        for ( int pid = 0; pid < closest_plane_ids.size(); ++pid )
        {
            int     point_id          = cuboid_ptr->IndicesPtr()->indices[pid];
            MyPoint pnt               = cuboid_ptr->CloudPtr()->at( point_id );
            int     plane_id          = closest_plane_ids[pid];
            int     opposite_plane_id = ((plane_id + 3) > 5) ? (plane_id - 3) : (plane_id + 3);
            float   distance          = distances[plane_id][pid];


            // project
            pcl::projectPoint( pnt, planes[opposite_plane_id].Coeffs(), new_point );
            // offset along normal
            Eigen::Vector3f pnt3f = new_point.getVector3fMap() + planes[opposite_plane_id].Normal() * distance;
            new_point.x = pnt3f(0); new_point.y = pnt3f(1); new_point.z = pnt3f(2);
            new_point.rgb = pnt.rgb;
            // store
            filled_cloud_ptr->push_back( new_point );
        }
        am::CloudViewer::showCloud( filled_cloud_ptr, "filled_cloud", nullptr, 5. );
#endif
        return EXIT_SUCCESS;
    }

    int
    AnnotationController::removePrimitive( std::vector<int> ids )
    {
        TIC;

        std::sort( ids.begin(), ids.end() );
        int id = 0;
        for ( auto it = _state._objects.begin(); it != _state._objects.end(); ++id )
        {
            if ( std::find( ids.begin(), ids.end(), id ) == ids.end() )
            {
                //std::cout << "keeping " << id << std::endl; fflush( stdout );
                ++it;
                continue;
            }

            std::cout << "removing " << id << "by unDraw( _viewer, " << it->first << ")\t"; fflush(stdout);
            it->second.primitive_ptr->unDraw( _viewer, it->first );
            std::cout << " removed"; fflush(stdout);
            _state._objects.erase( it++ ); // it will point to next valid it, so ++it not necessary
            std::cout << " erased"; fflush(stdout);
            std::cout << "std::distance(_state._objects.begin(), it): " << std::distance(_state._objects.begin(), it )
                      << ", size: " << _state._objects.size()
                      << std::endl; fflush( stdout );
        }

        if ( !_state._objects.size() ) _state._next_primitive_id = 0;

        TOC(__func__);
    }

    int
    AnnotationController::showAuxCloud( int id, bool doShow )
    {
        if ( id >= _state._aux_clouds.size() )
        {
            std::cerr << "AnnotationController::showAuxCloud(): no such aux_cloud id: " << id << std::endl;
            return EXIT_FAILURE;
        }
        if ( doShow )
        {
            pcl::visualization::CloudActorMap::const_iterator found = _viewer->getCloudActorMap()->find(_auxCloudName(id) );
            if ( found != _viewer->getCloudActorMap()->end() )
            {
                if ( !am::CloudViewer::updatePointCloud(_viewer, _state._aux_clouds[id], _auxCloudName(id), NULL, 0, _state._display_point_size) )
                {
                    std::cout << "please..." << std::endl;
                    am::CloudViewer::addPointCloud( _viewer, _state._aux_clouds[id], _auxCloudName(id), NULL, 0, _state._display_point_size );
                }
            }
            else
            {
                am::CloudViewer::addPointCloud( _viewer, _state._aux_clouds[id], _auxCloudName(id), NULL, 0, _state._display_point_size );
            }
        }
        else
        {
            _viewer->removePointCloud( _auxCloudName(id) );
        }
    }

    int
    AnnotationController::centerCamera()
    {
        Eigen::Vector3f center( Eigen::Vector3f::Zero() );
        int avg_n = 0;

        DisplayedPrimitive::Map::const_iterator end_it = _state._objects.end();
        for ( DisplayedPrimitive::Map::const_iterator it = _state._objects.begin(); it != end_it; ++it )
        {
            if ( std::shared_ptr<Cuboid> cub = it->second.CuboidPtr() )
            {
                Eigen::Vector3f cuboid_center = cub->Center();
                center += cuboid_center;
                ++avg_n;

            }
            else if ( std::shared_ptr<Plane> plane = it->second.PlanePtr() )
            {
                Eigen::Vector3f plane_center = plane->getCentroidFromCorners();
                center += plane_center;
                ++avg_n;
            }
        }

        if ( avg_n )
        {
            center /= float( avg_n );
            std::vector<Eigen::Vector3f> pos_focal_up = getCameraPosFocalUp();
            pos_focal_up[1] = center;
            setCameraPosFocalUp( pos_focal_up );
            // std::cout << "[" << __func__ << "]: " << "focal set to " << center.transpose() << std::endl;
        }

        return EXIT_SUCCESS;
    }

    int
    AnnotationController::resetCamera()
    {
        //printCamInfo( _viewer );

        // save window size
        std::vector<pcl::visualization::Camera> cameras;
        _viewer->getCameras( cameras );
        auto camera = cameras[ 0 ];

        _viewer->setCameraPosition( 0, 0, 0, 0, -1, 0 );
        Eigen::Matrix4f extrinsic = Eigen::Matrix4f::Identity();
        extrinsic(1,1) = -1.f;
        auto intr = _depth_intrinsics.asMatrix3f();
        intr(0,2) *= 2.f;
        intr(1,2) *= 2.f;
        _viewer->setCameraParameters( intr, extrinsic );
        _viewer->setSize( camera.window_size[0], camera.window_size[1] );

        //_printCamInfo( _viewer );
    }

    std::vector< std::shared_ptr<am::Cuboid> >
    AnnotationController::getFloors()
    {
        // get a floor
        std::vector<std::shared_ptr<am::Cuboid>> floor_cuboid_ptrs;

        // check for at least one floor element
        for ( auto const& pair: _state._objects )
        {
            if ( pair.second.floor )
            {
                floor_cuboid_ptrs.push_back( pair.second.CuboidPtr() );
            }
        }

        return floor_cuboid_ptrs;
    }


    std::vector< std::shared_ptr<am::Cuboid> >
    AnnotationController::getRotationTargets()
    {
        // get a floor
        std::vector<std::shared_ptr<am::Cuboid>> rot_target_cuboid_ptrs;

        // check for at least one floor element
        for ( auto const& pair: _state._objects )
        {
            if ( pair.second.rotation_target )
            {
                rot_target_cuboid_ptrs.push_back( pair.second.CuboidPtr() );
            }
        }

        return rot_target_cuboid_ptrs;
    }

    int
    AnnotationController::segment( int      min_cluster_size
                                   , double cluster_tolerance
                                   , float  voxelgrid_resolution
                                   , float  slab_size_threshold_percentage
                                   , int    max_slab_count
                                   , Eigen::Vector3f const* up )
    {
        PCloud::ConstPtr const& cloud_ptr = _state.MainCloud();
        Eigen::Vector4f floor_plane_coeffs;

        // UP dir
        if ( up )
        {
            floor_plane_coeffs << *up, 0.f;
        }
        else
        {
            // get a floor
            std::shared_ptr<am::Cuboid> floor_cuboid_ptr( nullptr );
            {
                // check for at least one floor element
                for ( auto const& pair: _state._objects )
                {
                    if ( pair.second.floor )
                    {
                        floor_cuboid_ptr = pair.second.CuboidPtr();
                        break;
                    }
                }
                // if not found, quit
                if ( !floor_cuboid_ptr )
                {
                    std::cerr << "Annotationcontroller::segment() needs at least one floor cuboid to be annotated...exiting\n";
                    return EXIT_FAILURE;
                }
            }
            floor_plane_coeffs = floor_cuboid_ptr->Planes()[2].Coeffs();
        }
        std::cout << "floor_plane_coeffs: " << floor_plane_coeffs.transpose() << std::endl;

        // create full cloud indices
        //pcl::PointIndices::Ptr indices_ptr( am::util::pcl::allIndicesOf(cloud_ptr) );
        pcl::PointIndices::Ptr indices_ptr( new pcl::PointIndices() );
        *indices_ptr = *(_state.CurrentIndices());

        // remove floors and walls
        for ( auto const& pair: _state._objects )
        {
            if ( pair.second.floor || pair.second.wall )
            {
                // select inliers
//                pcl::PointIndices::Ptr inlier_indices_ptr( new pcl::PointIndices );
//                {
//                    pair.second.CuboidPtr()->getInlierIndices( /*           out_indices: */ inlier_indices_ptr
//                                                               , /*           cloud_ptr: */ cloud_ptr
//                                                               , /*         indices_ptr: */ indices_ptr
//                                                               , /* outside_dist_thresh: */ 1.f
//                                                               , /*               debug: */ true
//                                                               );
//                }

                // exclude plane's points from active cloud
                {
                    am::util::algo::vectorDifference( /*   A-B: */ indices_ptr->indices
                                                      , /*   A: */ indices_ptr->indices
                                                      , /*   B: */ pair.second.CuboidPtr()->IndicesPtr()->indices );
                }
            }
        }

        // do clustering
        std::vector<pcl::PointIndices::Ptr> cluster_indices_ptrs;
        {
            std::vector<pcl::PointIndices>              clusterIndices;
            pcl::search::KdTree<am::MyPoint>::Ptr tree (new pcl::search::KdTree<am::MyPoint>);
            tree->setInputCloud( cloud_ptr );

            am::CloudViewer::showCloud( cloud_ptr, "segmentinput", indices_ptr );
            pcl::EuclideanClusterExtraction<am::MyPoint> ec;
            ec.setIndices         ( indices_ptr );
            ec.setClusterTolerance( cluster_tolerance );
            ec.setMinClusterSize  ( min_cluster_size );
            ec.setMaxClusterSize  ( cloud_ptr->size() );
            ec.setSearchMethod    ( tree );
            ec.setInputCloud      ( cloud_ptr );
            std::cout << "[" << __func__ << "]: " << "extraction starting...";
            ec.extract            ( clusterIndices );
            std::cout << "extraction finished...clustercount: " << clusterIndices.size() << std::endl;

            int cluster_id = 0;
            for ( auto const& indices : clusterIndices )
            {
                // store
                cluster_indices_ptrs.emplace_back( pcl::PointIndices::Ptr(new pcl::PointIndices(indices)) );

                // save to disk
                char title[255];
                sprintf( title, "cluster%02d.pcd", cluster_id++ );
                std::cout << "saving " << title << std::endl;
                pcl::io::savePCDFile( /*    path: */ std::string(title),
                                      /*   cloud: */ *cloud_ptr,
                                      /* indices: */ indices.indices );
            }
        }

        am::Cuboids              cuboids;
        std::vector<std::string> cluster_tags;

        // prepare show
        am::CloudViewer v;
        v.VP()->setCameraPosition( 0, 0, 0, 0, -1, 0 );

        int cluster_id = 0;
        for ( auto const& cluster_indices_ptr : cluster_indices_ptrs )
        {
            // debug
            std::cout << "showing cluster with " << cluster_indices_ptr->indices.size() << std::endl;

            // calc frame using upvector
            Eigen::Matrix4f frame;
            {
                DepthProcessor::PCA_withPlane( frame
                                               , floor_plane_coeffs
                                               , cloud_ptr
                                               , cluster_indices_ptr );
#if 0
                PCloud::Ptr tmp_cloud_ptr;
                am::DepthProcessor::projectPlaneInliers( tmp_cloud_ptr, cloud_ptr, indices_ptr, floor_plane.CoeffsPtr() );
                am::DepthProcessor::PCA( frame, tmp_cloud_ptr, nullptr );
                frame.block<3,1>(0,2) = floor_plane.Normal();
                frame.block<3,1>(0,1) = frame.block<3,1>(0,2).cross(frame.block<3,1>(0,0));
#endif
                std::cout << "1dot2: " << frame.block<3,1>( 0,2).dot( frame.block<3,1>( 0,1) ) << std::endl;
                std::cout << "0dot2: " << frame.block<3,1>( 0,2).dot( frame.block<3,1>( 0,0) ) << std::endl;
                std::cout << "frame: " << frame << std::endl;
            }

            // Cloud2Cuboid
            am::Cuboid cluster_cuboid;
            {
                // calc bounding box
                MyPoint min_pt, max_pt;
                am::DepthProcessor::calculateExtremaInFrame( min_pt, max_pt, frame, cloud_ptr, cluster_indices_ptr );
                pcl::PointXYZ min_pt_world = toWorld(min_pt,frame);
                pcl::PointXYZ max_pt_world = toWorld(max_pt,frame);
                std::vector<float> extrema = { min_pt.x, min_pt.y, min_pt.z,
                                               max_pt.x, max_pt.y, max_pt.z };

                // debug
                std::cout << "min_pt_world: " << min_pt_world << std::endl;
                std::cout << "max_pt_world: " << max_pt_world << std::endl;

                std::vector<float> corner_coords(24);
                for ( int x = 0; x < 2; ++x )
                    for ( int y = 0; y < 2; ++y )
                        for ( int z = 0; z < 2; ++z )
                        {
                            int corner_id = (x*4+y*2+z)*3;
                            pcl::PointXYZ local_pnt( extrema[ x*3 ], extrema[ y*3+1 ], extrema[ z*3+2 ] );
                            pcl::PointXYZ world_pnt = toWorld( local_pnt, frame );

                            corner_coords[ corner_id   ] = world_pnt.x;
                            corner_coords[ corner_id+1 ] = world_pnt.y;
                            corner_coords[ corner_id+2 ] = world_pnt.z;
                        }

                am::Cuboid::fromCorners( cluster_cuboid, corner_coords, cloud_ptr, cluster_indices_ptr );
            }

            // auto_cuboids
            {
                am::Cuboids tmp_cuboids;
                MyVoxelGrid::segment( /*                      out_cuboids: */ tmp_cuboids
                                      , /*                       in_cloud: */ cloud_ptr
                                      , /*                    local_frame: */ frame
                                      , /*                    indices_ptr: */ cluster_indices_ptr
                                      , /*           voxelgrid_resolution: */ voxelgrid_resolution
                                      , /* slab_size_threshold_percentage: */ slab_size_threshold_percentage
                                      , /*                 max_slab_count: */ max_slab_count
                                      , /*                          debug: */ false );
                // add cluster tag
                char chr[255];
                sprintf( chr, "cluster%d", cluster_id );
                // copy to output
                for ( auto const& cuboid : tmp_cuboids.Primitives() )
                {
                    cluster_tags.push_back( chr );
                    cuboids.add( cuboid );
                }
            }

            // draw
            {
                // name
                char chr[255];
                sprintf( chr, "cluster%d", cluster_id );
                // cuboid
                cluster_cuboid.draw( v.VP(), std::string(chr) + "_cuboid" );
                // cluster cloud
                v.addPointCloud( cloud_ptr, chr, cluster_indices_ptr, 0, 2. );

                // floor normal
                v.VP()->addArrow( am::util::pcl::asPointXYZ( cluster_cuboid.Center() + floor_plane_coeffs.head<3>() )
                        , am::util::pcl::asPointXYZ( cluster_cuboid.Center() )
                        , .9, .0, .0
                        , false
                        , std::string(chr)+"_upvector" );
            }

            ++cluster_id;
        }

        //store
        _addCuboids( cuboids, cluster_tags );

        // show
        if ( !v.VP()->wasStopped() )
            v.VP()->spin();

        return EXIT_SUCCESS;
    }

    int
    AnnotationController::cluster( int min_cluster_size
                                   , double cluster_tolerance
                                   )
    {
        PCloud::ConstPtr const& cloud_ptr = _state.MainCloud();

        // get a floor
        std::shared_ptr<am::Cuboid> floor_cuboid_ptr( nullptr );
        {
            // check for at least one floor element
            for ( auto const& pair: _state._objects )
            {
                if ( pair.second.floor )
                {
                    floor_cuboid_ptr = pair.second.CuboidPtr();
                    break;
                }
            }
            // if not found, quit
            if ( !floor_cuboid_ptr )
            {
                std::cerr << "Annotationcontroller::segment() needs at least one floor cuboid to be annotated...exiting\n";
                return EXIT_FAILURE;
            }
        }

        // copy cloud indices
        pcl::PointIndices::Ptr indices_ptr( new pcl::PointIndices() );
        *indices_ptr = *(_state.CurrentIndices());

        // remove floors and walls
        for ( auto const& pair: _state._objects )
        {
            if ( pair.second.floor || pair.second.wall )
            {
                am::util::algo::vectorDifference( /*   A-B: */ indices_ptr->indices
                                                  , /*   A: */ indices_ptr->indices
                                                  , /*   B: */ pair.second.CuboidPtr()->IndicesPtr()->indices );
            }
        }

        // do clustering
        std::vector<pcl::PointIndices::Ptr> cluster_indices_ptrs;
        {
            std::vector<pcl::PointIndices>              clusterIndices;
            pcl::search::KdTree<am::MyPoint>::Ptr tree (new pcl::search::KdTree<am::MyPoint>);
            tree->setInputCloud( cloud_ptr );

            pcl::EuclideanClusterExtraction<am::MyPoint> ec;
            ec.setIndices         ( indices_ptr );
            ec.setClusterTolerance( cluster_tolerance );
            ec.setMinClusterSize  ( min_cluster_size );
            ec.setMaxClusterSize  ( cloud_ptr->size() );
            ec.setSearchMethod    ( tree );
            ec.setInputCloud      ( cloud_ptr );
            ec.extract            ( clusterIndices );

            int cluster_id = 0;
            for ( auto const& indices : clusterIndices )
            {
                // store
                cluster_indices_ptrs.emplace_back( pcl::PointIndices::Ptr(new pcl::PointIndices(indices)) );

                // save to disk
                char title[255];
                sprintf( title, "cluster%02d.pcd", cluster_id++ );
                std::cout << "saving " << title << std::endl;
                pcl::io::savePCDFile( /*    path: */ std::string(title),
                                      /*   cloud: */ *cloud_ptr,
                                      /* indices: */ indices.indices );
            }
        }

        am::Cuboids              cuboids;
        std::vector<std::string> cluster_tags;

        // prepare show
        am::CloudViewer v;
        v.VP()->setCameraPosition( 0, 0, 0, 0, -1, 0 );

        am::Plane const& floor_plane = floor_cuboid_ptr->Planes()[2];

        int cluster_id = 0;
        for ( auto const& cluster_indices_ptr : cluster_indices_ptrs )
        {
            // debug
            std::cout << "showing cluster with " << cluster_indices_ptr->indices.size() << std::endl;

            // calc frame using upvector
            Eigen::Matrix4f frame;
            {
                DepthProcessor::PCA_withPlane( frame
                                               , floor_plane.Coeffs()
                                               , cloud_ptr
                                               , cluster_indices_ptr );
            }

            // Cloud2Cuboid
            am::Cuboid cluster_cuboid;
            {
                // calc bounding box
                MyPoint min_pt, max_pt;
                am::DepthProcessor::calculateExtremaInFrame( min_pt, max_pt, frame, cloud_ptr, cluster_indices_ptr );
                std::vector<float> extrema = { min_pt.x, min_pt.y, min_pt.z,
                                               max_pt.x, max_pt.y, max_pt.z };

                std::vector<float> corner_coords(24);
                for ( int x = 0; x < 2; ++x )
                    for ( int y = 0; y < 2; ++y )
                        for ( int z = 0; z != 2; ++z )
                        {
                            int corner_id = (x*4+y*2+z)*3;
                            pcl::PointXYZ local_pnt( extrema[ x*3 ], extrema[ y*3+1 ], extrema[ z*3+2 ] );
                            pcl::PointXYZ world_pnt = toWorld( local_pnt, frame );

                            corner_coords[ corner_id   ] = world_pnt.x;
                            corner_coords[ corner_id+1 ] = world_pnt.y;
                            corner_coords[ corner_id+2 ] = world_pnt.z;
                        }

                am::Cuboid::fromCorners( cluster_cuboid,
                                         corner_coords,
                                         cloud_ptr,
                                         cluster_indices_ptr );
            }
            cuboids.add( cluster_cuboid );
            cluster_tags.push_back( am::util::sprintf("cluster%d", cluster_id) );

            // draw
            {
                // name
                char chr[255];
                sprintf( chr, "cluster%d", cluster_id );
                // cuboid
                cluster_cuboid.draw( v.VP(), std::string(chr) + "_cuboid" );
                // cluster cloud
                v.addPointCloud( cloud_ptr, chr, cluster_indices_ptr, 0, 2. );

                // floor normal
                v.VP()->addArrow( am::util::pcl::asPointXYZ( cluster_cuboid.Center() + floor_plane.Normal() )
                        , am::util::pcl::asPointXYZ( cluster_cuboid.Center() )
                        , .9, .0, .0
                        , false
                        , std::string(chr)+"_upvector" );
            }

            ++cluster_id;
        }

        //store
        _addCuboids( cuboids, cluster_tags );

        // show
        if ( !v.VP()->wasStopped() )
            v.VP()->spin();

        return EXIT_SUCCESS;
    }

    std::vector<am::Contact>
    AnnotationController::_calculateIntersections()
    {
        std::vector<am::Contact> contacts;

        int i = 0, j = 0;
        for ( auto it = _state._objects.begin(); it != _state._objects.end(); ++it, ++i )
        {
            j = i+1;
            for ( auto it2 = std::next(it); it2 != _state._objects.end(); ++it2, ++j )
            {
                am::Cuboid &cubA = *(it ->second.CuboidPtr());
                am::Cuboid &cubB = *(it2->second.CuboidPtr());

                if ( am::Cuboid::intersect(cubA, cubB, INTERSECT_OFFSET) )
                {
                    contacts.push_back( (am::Contact){(std::vector<int>){i, j}, am::Contact::TOUCHING} );
                }
            }
        }

        return contacts;
    }

    int
    AnnotationController::drawIntersections( bool                                       do_show
                                             , std::vector<Contact>              const* contacts_arg
                                             , int                                      viewport
                                             , pcl::visualization::PCLVisualizer::Ptr   viewer_arg )
    {
        pcl::visualization::PCLVisualizer::Ptr viewer = viewer_arg ? viewer_arg : _viewer;
        std::vector<Contact> contacts = contacts_arg ? *contacts_arg : _calculateIntersections();

        if ( do_show )
        {
            if ( _state._drawn_contacts.size() )
            {
                std::cerr << __func__ << ": not empty contacts to draw on!!!" << std::endl;
                return EXIT_FAILURE;
            }

            for ( Contact const& contact : contacts )
            {
                auto it = _state._objects.begin();
                std::advance( it, contact.ids[0] );
                auto it2 = _state._objects.begin();
                std::advance( it2, contact.ids[1] );

                am::Cuboid &cubA = *(it ->second.CuboidPtr());
                am::Cuboid &cubB = *(it2->second.CuboidPtr());

                std::vector<std::string> drawn_shapes = contact.draw( cubA
                                                                      , it->second.getDrawingColour( _colour_code_cuboids )
                                                                      , cubB
                                                                      , it2->second.getDrawingColour( _colour_code_cuboids )
                                                                      , viewer
                                                                      , viewport
                                                                      , Contact::CONTACT_LINE_WIDTH // AnnotationController::CUBOID_SELECTED_LINE_WIDTH
                                                                      , _state._max_sphere_size
                                                                      , _state._min_sphere_size );
                _state._drawn_contacts.insert( _state._drawn_contacts.end(), drawn_shapes.begin(), drawn_shapes.end() );
            }

            _state._latest_contacts = contacts;
        }
        else // !do_show
        {
            for ( std::string &shape_name : _state._drawn_contacts )
            {
                viewer->removeShape( shape_name, viewport );
            }

//            if ( !_state._drawn_contacts.size() )
//                std::cerr << __func__ << ": nothing to remove" << std::endl;
            _state._drawn_contacts.clear();
        }
    }

    int
    AnnotationController::loadRegions( std::vector< pcl::PointIndices::Ptr >      & region_indices
                                       , std::string                         const& path
                                       , Params                              const& params
                                       , PCloud::Ptr                                cloud )
    {
        cv::Mat regions = imread( path, cv::IMREAD_UNCHANGED );
        if ( regions.empty() || (regions.type() != CV_16UC1) )
        {
            std::cerr << "[" << __func__ << "]: " << "regions.empty() || (regions.type != CV_16UC1)...exiting\n";
            return EXIT_FAILURE;
        }
        if ( cloud->size() != regions.rows * regions.cols )
        {
            std::cerr << "cloud.size() " << cloud->size()
                      << " != " << regions.rows * regions.cols << " regions.rows * regions.cols\n";
        }

        for ( unsigned y = 0; y < regions.rows; ++y  )
        {
            for ( unsigned x = 0; x < regions.cols; ++x )
            {
                int region_id = regions.at<ushort>(y,x);
                int point_id  = y * regions.cols + x;

                /// store point index
                // resize storage
                while ( region_indices.size() <= region_id )
                    region_indices.emplace_back( pcl::PointIndices::Ptr(new pcl::PointIndices()) );
                // store
                region_indices[ region_id ]->indices.push_back( point_id );
            }
        }
    }

    int
    AnnotationController::loadRegions( std::string const& path, Params const& params )
    {
        int err = EXIT_SUCCESS;

        std::vector< pcl::PointIndices::Ptr > region_indices;
        err = AnnotationController::loadRegions( region_indices
                                                 , path
                                                 , params
                                                 , _state.CurrentCloud() );

        // save to disk
        if ( 1 )
        {
            char fname[512];
            pcl::PointCloud<MyPoint>::Ptr tmp_cloud( new pcl::PointCloud<MyPoint>() );
            for ( int rid = 1; rid != region_indices.size(); ++rid ) // rid==0 is "unlabelled"
            {
                pcl::copyPointCloud( *(_state.CurrentCloud()), region_indices[rid]->indices, *tmp_cloud );
                sprintf( fname, "region_%02d.pcd", rid );
                pcl::io::savePCDFileBinary( fname, *tmp_cloud );
                std::cout << "sved " << fname << std::endl;
            }
        }

        // colour
        {
            for ( size_t rid = 0; rid != region_indices.size(); ++rid )
            {
                for ( size_t iid = 0; iid != region_indices[rid]->indices.size(); ++iid )
                {
                    int point_id  = region_indices[rid]->indices[iid];

                    //  colour point
                    MyPoint &p = _state.CurrentCloudRef(AnnotationState::COLOUR)->at( point_id );
                    cv::Point3f colour = rid ? _colors[ rid % _colors.size() ]
                                             : cv::Point3f(0.f,0.f,0.f);
                    am::util::pcl::setPointColor( p, colour.x * 255.f, colour.y * 255.f, colour.z * 255.f );
                }
            }
        }

//        return err;

        // extract cuboid
        for ( int rid = 1; rid != region_indices.size(); ++rid ) // rid==0 is "unlabelled"
        {
            //std::cout << "region size: " << region_indices[rid]->indices.size() << std::endl;
            Cuboid cuboid;
//            err = Cuboid::from_cloud( cuboid
//                                      , _state._current_cloud_ptr
//                                      , region_indices[rid]
//                                      , params.getDouble("plane_distance_threshold")
//                                      , false );
            err = Cuboid::fromCloudWithPCA( /*                  cuboid_out: */ cuboid
                                            , /*                     cloud: */ _state.CurrentCloud()
                                            , /*                   indices: */ region_indices[rid]
                                            , /*         max_neighbours(K): */ 15
                                            , /*                    radius: */ -1.f
                                            , /*             clean_n_means: */ 10
                                            , /*              clean_stddev: */ .01
                                            , /* density_median_multiplier: */ 1.2f
                                            , /*                     debug: */ false
                                           );
            if ( err == EXIT_SUCCESS )
                _addCuboid( cuboid, _state._next_primitive_id++ );
            else
                std::cerr << "[" << __func__ << "]: " << "region " << rid << "did not produce valid cuboid" << std::endl;
        }

        return err + am::CloudViewer::updatePointCloud( _viewer, _state.CurrentCloud(), "big_cloud", nullptr, 0, _state._display_point_size );
    }

    int
    AnnotationController::region_growing()
    {
        std::vector <pcl::PointIndices::Ptr> cluster_indices;

        return am::Segmentation::region_grow( cluster_indices
                                              , boost::const_pointer_cast<const PCloud>( _state.CurrentCloud() ) // _state._current_cloud_ptr
                                              , _state.CurrentIndices()
                                              );
    }

    inline void
    AnnotationController::_mouseCallbacks( const pcl::visualization::MouseEvent event )
    {
        // toggle state
        if ( /**/ event.getType()              == pcl::visualization::MouseEvent::MouseButtonPress
             &&   event.getButton()            == pcl::visualization::MouseEvent::RightButton
             &&   event.getKeyboardModifiers() &  pcl::visualization::KeyboardEvent::Shift )
        {
            _mouse_state |= RIGHT_DOWN;
            std::cout << "[" << __func__ << "]: " << "setting mouse to right_down " << _mouse_state << std::endl; fflush(stdout);

            // disable camera navigation
            _interactor_style->setPropagateMouseMove( false );

            // start new region
            //_indices_vec.push_back( std::vector<int>() );
        }
        else if ( /**/ event.getType()   == pcl::visualization::MouseEvent::MouseButtonRelease
                  &&   event.getButton() == pcl::visualization::MouseEvent::RightButton          )
        {
            _mouse_state &= (~RIGHT_DOWN);
            std::cout << "[" << __func__ << "]: " << "releasing mouse from right_down " << _mouse_state << std::endl; fflush(stdout);

            // reenable camera navigation
            _interactor_style->setPropagateMouseMove( true );

        } // state change handling

        // left press
        if ( (_mouse_state & RIGHT_DOWN)
             && (event.getKeyboardModifiers() &  pcl::visualization::KeyboardEvent::Shift) )
        {
            std::cout << "[" << __func__ << "]: " << "picking " << _mouse_state << std::endl; fflush(stdout);

            vtkSmartPointer<vtkPointPicker> point_picker = vtkPointPicker::New();
            vtkSmartPointer<vtkRenderWindow> win = _viewer->getRenderWindow();
            if ( !win->GetInteractor() )
            {
                std::cout << "[" << __func__ << "]: " << "setting interactor" << std::endl; fflush(stdout);
                win->GetInteractor()->SetPicker( point_picker );
            }
            else
            {
                std::cout << "[" << __func__ << "]: " << "interactor set already" << std::endl; fflush(stdout);
            }

            // pick point based on XY position
            point_picker->Pick(event.getX(), event.getY(), 0.0, _viewer->getRendererCollection()->GetFirstRenderer() );

            // if a point is under the cursor and it belongs to a surface, highlight the surface
            if (point_picker->GetDataSet() != NULL)
            {
                int idx = static_cast<int>( point_picker->GetPointId() );
                std::cout << "idx: " << idx << std::endl;

                // broadcast signal
                Q_EMIT pointPickedSignal( idx );
            }
        }
    }

    void
    AnnotationController::pointPickedSlot( int point_id )
    {
        std::cout << "[" << __func__ << "]: " << "point_id: " << point_id << std::endl;

        bool update = false;
        if ( _state.hasPick(point_id) >= 0 )
        {
            _state.removePick( point_id );

            if ( _state.CurrentCloud() )
            {
                _state.CurrentCloud()->at(point_id).rgb = _state.MainCloud()->at(point_id).rgb;
                update = true;
            }
            _viewer->removeShape( "pick_sphere", 0 );
        }
        else
        {
            float dist = _state.getClosestPickedTo( point_id );
            std::cout << "[" << __func__ << "]: " << "dist: " << dist << std::endl;

            if ( (dist < .3f) || !_state.pickedIndices().size() )
            {
                _state.addPick( point_id );

                if ( _state.CurrentCloud() )
                {
                    _state.CurrentCloud()->at( point_id ).r = 255;
                    update = true;
                }
                _viewer->removeShape( "pick_sphere", 0 );
#if CA_PCL_MODIFIED
                _viewer->addSphere( _state.CurrentCloud()->at(point_id), .01f, 255, 30, 30, "pick_sphere", 0, 20 );
#else
                _viewer->addSphere( _state.CurrentCloud()->at(point_id), .01f, 255, 30, 30, "pick_sphere", 0 );
#endif
                _viewer->setShapeRenderingProperties( pcl::visualization::PCL_VISUALIZER_OPACITY, _default_opacity, "pick_sphere", 0 );
            }
        }

        if ( update )
        {
            am::CloudViewer::updatePointCloud( _viewer, _state.CurrentCloud(), "big_cloud", nullptr, 0, _state._display_point_size );
            _viewer->getRenderWindow()->Render();
        }
    }

    /// ------------------------------------------------------------------------------------------------ ///
    /// METHODS

    int
    AnnotationController::redrawPrimitive( int id, bool is_selected )
    {
        auto it = _state._objects.begin();
        std::advance( it, id );
        // redraw
        it->second.reDraw( /*        viewer: */ _viewer,
                           /* name_from_map: */ it->first,
                           /*   is_selected: */ is_selected
                           , _colour_code_cuboids
                           , _state._colour_edges
                           , _state._show_IDs );
    }

    int
    AnnotationController::redrawAllPrimitives()
    {
        int ret = EXIT_SUCCESS;

        for ( auto it = _state._objects.begin();
              it != _state._objects.end();
              ++it )
        {
            //std::advance( it, id );

            // redraw
            ret += it->second.reDraw( /*        viewer: */ _viewer,
                                      /* name_from_map: */ it->first,
                                      /*   is_selected: */ false // TODO: store this in object
                                      , _colour_code_cuboids
                                      , _state._colour_edges
                                      , _state._show_IDs );
        }

        return ret;
    }

    int
    AnnotationController::_addCuboids( Cuboids const& cuboids, std::string tag )
    {
        std::vector<std::string> tags = { tag };
        return _addCuboids( cuboids, tags );
    }

    int
    AnnotationController::_addCuboids( Cuboids const& cuboids, std::vector<std::string> tags )
    {
        // display cuboids
        int tag_id = 0;
        for ( auto const& cuboid : cuboids.Primitives() )
        {
            _addCuboid( cuboid, _state._next_primitive_id++,
                        tags[std::min(tag_id,(int)tags.size()-1)] );
            ++tag_id;
        }
    }

    int
    AnnotationController::_addCuboid( Cuboid const& cuboid, int cube_id, std::string tag )
    {
        // name
        std::string cuboid_name = Cuboid::cuboidName( cube_id );
        // cuboid
        _state._objects[ cuboid_name ] = DisplayedPrimitive( cuboid );
        _state._objects[ cuboid_name ].tag = tag;
        auto tmp_colour = _colors[ cube_id % _colors.size() ];
        _state._objects[ cuboid_name ].colour = { tmp_colour.x, tmp_colour.y, tmp_colour.z, _default_opacity };

        _state._objects[ cuboid_name ].reDraw( _viewer, cuboid_name, false, _colour_code_cuboids );
        //am::Cuboid::draw( cuboid, _viewer, cuboid_name, _color_code_cuboids?CUBOID_COLOUR, CUBOID_UNSELECTED_LINE_WIDTH );
    }

    int
    AnnotationController::_removePoints( pcl::PointIndices::ConstPtr indices_ptr )
    {
        // debug
        std::cout << __func__ << "subtracting " << indices_ptr->indices.size()
                  << " from " << _state.CurrentIndices()->indices.size();

        // diff
        util::algo::vectorDifference( _state.CurrentIndicesRef()->indices,
                                      _state.CurrentIndices()->indices,
                                      indices_ptr->indices );
        std::cout << " = " << _state.CurrentIndices()->indices.size() << std::endl;

        // copy cloud
        pcl::copyPointCloud( *_state.MainCloud(), _state.CurrentIndices()->indices, *_state.CurrentCloudRef(AnnotationState::FULL) );

        // show
        am::CloudViewer::updatePointCloud( _viewer, _state.CurrentCloud(), "big_cloud", nullptr, 0, _state._display_point_size );

        return EXIT_SUCCESS;
    }

    int
    AnnotationController::_addPoints( pcl::PointIndices::ConstPtr indices_ptr )
    {
        std::cout << __PRETTY_FUNCTION__ << "adding " << indices_ptr->indices.size()
                  << " to " << _state.CurrentIndices()->indices.size();

        util::algo::vectorUnion( _state.CurrentIndicesRef()->indices,
                                 _state.CurrentIndices()->indices,
                                 indices_ptr->indices );

        std::cout << " = " << _state.CurrentIndices()->indices.size() << std::endl;

        // TODO: preserve colour
        pcl::copyPointCloud( *_state.MainCloud(), _state.CurrentIndices()->indices, *_state.CurrentCloudRef(AnnotationState::FULL) );

        am::CloudViewer::updatePointCloud( _viewer, _state.CurrentCloud(), "big_cloud", nullptr, 0, _state._display_point_size );
    }

    int
    AnnotationController::mapCurrentCloud()
    {
        typedef unsigned char uchar;

        ca::MapBack mapBack;
        cv::Mat img( mapBack.ViewportMat()(2), mapBack.ViewportMat()(2), CV_8UC1 );
        img.setTo(255);
        std::cout << "img.size(): " << img.rows << "x" << img.cols << std::endl;
        Eigen::Vector3f min_p, max_p;
        min_p.setConstant( 3, 1, FLT_MAX );
        max_p.setZero( 3 );
        std::cout << "starting with " << min_p.transpose() << ", " << max_p.transpose() << std::endl;

        Eigen::Matrix<Scalar,3,1> pnt2D;
        for ( size_t pid_id = 0; pid_id != _state.CurrentIndices()->indices.size(); ++pid_id )
        {
            const int pid = _state.CurrentIndices()->indices[ pid_id ];

            mapBack.project( pnt2D, _state.CurrentCloud()->at( pid ).getVector3fMap() );
            //pnt2D.template head<2>() = am::util::pcl::point3To2D( _state._current_cloud_ptr->at( pid ).getVector3fMap(), _depth_intrinsics.asMatrix3f() );
            //if ( pnt2D(2) != 0.f )
            //    pnt2D.template head<2>() /= pnt2D(2);

            // stats
            for ( int d = 0; d != 3; ++d )
            {
                if ( pnt2D(d) < min_p(d) ) min_p(d) = pnt2D(d);
                if ( pnt2D(d) > max_p(d) ) max_p(d) = pnt2D(d);
            }

            //pnt2D(2) = (pnt2D(2) + 1.f) / 2.f * 255.f;

            if (    (pnt2D(0) < img.cols) && (pnt2D(0) >= 0)
                 && (pnt2D(1) < img.rows) && (pnt2D(1) >= 0) )
            {
                //std::cout << "setting img.at<uchar>( " << pnt2D(1) << ", " << pnt2D(0) << ") = " << pnt2D(2) << "~=" << (int)static_cast<uchar>(pnt2D(2)) << std::endl;
                img.at<uchar>( pnt2D(1), pnt2D(0) ) = std::min( static_cast<uchar>(pnt2D(2)), img.at<uchar>(pnt2D(1),pnt2D(0)) );
            }
//            else
//                std::cout << "skipping " << pnt2D.transpose() << std::endl;
        }

        std::cout << "min_p: " << min_p.transpose() << ", max_p: " << max_p.transpose() << std::endl;

        cv::imshow( "backproj", img );
        cv::waitKey();

        return EXIT_SUCCESS;
    }

    int
    AnnotationController::createCuboid( am::Params const* params )
    {
        pcl::PointIndices::Ptr point_indices( new pcl::PointIndices() );

        //std::vector<int> indices = _state.pickedIndices();
        point_indices->indices = _state.pickedIndices();

        if ( !point_indices->indices.size() )
        {
            std::cerr << "[" << __func__ << "]: " << "!point_indices->indices.size()...exiting\n";
            return EXIT_FAILURE;
        }
        else
            std::cout << "[" << __func__ << "]: " << "point_indices->indices.size(): " << point_indices->indices.size() << std::endl;

        std::cout<<"indices:"; fflush(stdout);
        for ( size_t vi = 0; vi != point_indices->indices.size(); ++vi )
        {
            std::cout << point_indices->indices.at(vi) << " ";  fflush(stdout);
        }
        std::cout << "\n"; fflush(stdout);

        double inlier_thresh = .02;
        if ( params )
            inlier_thresh = params->getDouble( "plane_distance_threshold", .02 );

        Cuboid cuboid;
//        Cuboid::from_cloud( cuboid
//                            , _state.CurrentCloud()
//                            , point_indices
//                            , inlier_thresh
//                            , false );
        Cuboid::fromCloudWithPCA2( cuboid
                                   , boost::const_pointer_cast<const ca::PCloud       >( _state.CurrentCloud() )
                                   , boost::const_pointer_cast<const pcl::PointIndices>( point_indices )
                                   , boost::const_pointer_cast<const pcl::PointIndices>( _state.CurrentIndices() )
                                   );

        std::stringstream ss; ss << "fromPick" << _state._objects.size();
        _addCuboid( cuboid, _state._next_primitive_id++, ss.str() );

        // TODO:
        std::cout << "[" << __func__ << "]: " << "finshed" << std::endl;


        return EXIT_SUCCESS;
    }

    int
    AnnotationController::addCuboid( am::Cuboid const& cuboid, std::string name )
    {
        std::stringstream ss; ss << "name" << _state._objects.size();
        _addCuboid( cuboid, _state._next_primitive_id++, ss.str() );
    }

    int
    AnnotationController::clearCurrentPick()
    {
        std::vector<int> picked = _state.pickedIndices();
        for ( int pid_id = 0; pid_id != picked.size(); ++pid_id )
        {
            int pid = picked[pid_id];
            if ( pid < _state.CurrentCloud()->size() )
                _state.CurrentCloudRef( ca::AnnotationState::COLOUR )->at( pid ).r = _state.MainCloud()->at( pid ).r;
        }
        am::CloudViewer::updatePointCloud( _viewer, _state.CurrentCloud(), "big_cloud", _state.CurrentIndices(), 0, _state._display_point_size );
        return _state.clearPick();
    }

    int
    AnnotationController::colourCloud( COLOUR_CLOUD_MODE mode, void* param )
    {
        switch ( mode )
        {
            case COLOUR_CLOUD_MODE::NONE:
                pcl::copyPointCloud( *_state.MainCloud(), *_state.CurrentCloudRef( am::AnnotationState::COLOUR ) );
                CloudViewer::updatePointCloud( _viewer, _state.CurrentCloud(), "big_cloud", _state.CurrentIndices(), 0, _state._display_point_size );
                break;

            case COLOUR_CLOUD_MODE::DISTANCE_BASED:
            {
                Eigen::Vector3f origin( Eigen::Vector3f::Zero() );
                if ( param )
                {
                    Eigen::Vector3f *camera = static_cast<Eigen::Vector3f *>( param );
                    if ( camera )   origin = *camera;
                }
                ca::CloudColouring::distanceBased( _state.CurrentCloudRef(am::AnnotationState::COLOUR), origin );
                CloudViewer::updatePointCloud( _viewer, _state.CurrentCloud(), "big_cloud", _state.CurrentIndices(), 0, _state._display_point_size );
            }
                break;

            default:
                std::cout << "[" << __func__ << "]: " << "unrecognized mode" << std::endl;
                break;
        }

        return EXIT_SUCCESS;
    }

    /// ------------------------------------------------------------------------------------------------ ///
    /// SET, GET
    vtkSmartPointer<vtkRenderWindow>
    AnnotationController::getRenderWindow()
    {
        return _viewer->getRenderWindow();
    }

    vtkSmartPointer<vtkRendererCollection>
    AnnotationController::getRendererCollection()
    {
        return _viewer->getRendererCollection();
    }

    int
    AnnotationController::setPointSize( double size )
    {
        if ( _viewer->getCloudActorMap()->size() )
            _viewer->setPointCloudRenderingProperties( pcl::visualization::PCL_VISUALIZER_POINT_SIZE, size, "big_cloud" );

        for ( int aux_cloud_id = 0; aux_cloud_id < _state._aux_clouds.size(); ++aux_cloud_id )
        {
            std::string cloud_name = _auxCloudName(aux_cloud_id);
            if ( _viewer->getCloudActorMap()->find(cloud_name) != _viewer->getCloudActorMap()->end() )
                _viewer->setPointCloudRenderingProperties( pcl::visualization::PCL_VISUALIZER_POINT_SIZE, size, cloud_name );
        }

        _state._display_point_size = size;
    }

    int
    AnnotationController::setColourCodeCuboids( bool checked )
    {
        _colour_code_cuboids = checked;
        return this->redrawAllPrimitives();
    }

    int
    AnnotationController::setColourEdges( bool checked )
    {
        _state._colour_edges = checked;
        return this->redrawAllPrimitives();
    }

    int
    AnnotationController::setPrimitiveIsTarget( int id, bool is_target )
    {
        auto it = _state._objects.begin();
        std::advance( it, id );
        it->second.rotation_target = is_target;
    }

    int
    AnnotationController::setPrimitiveIsWall( int id, bool is_wall, bool is_selected )
    {
        auto it = _state._objects.begin();
        std::advance( it, id );
        it->second.wall = is_wall;

        // redraw
        it->second.reDraw( /*        viewer: */ _viewer,
                           /* name_from_map: */ it->first,
                           /*   is_selected: */ is_selected
                           , _colour_code_cuboids
                           , _state._colour_edges );

        /*it->second.CuboidPtr()->unDraw( _viewer, it->first );
        it->second.CuboidPtr()->draw( _viewer
                                      , it->first
                                      , is_wall ? CUBOID_WALL_COLOUR : CUBOID_COLOUR
                                      , CUBOID_SELECTED_LINE_WIDTH );*/

//        std::cout << "primitive " << id << " is now wall: " << (it->second.wall?"YES":"NO") << std::endl;

        return EXIT_SUCCESS;
    }

    bool
    AnnotationController::getPrimitiveIsWall( int id )
    {
        auto it = _state._objects.begin();
        std::advance( it, id );
        return it->second.wall;
    }
    bool
    AnnotationController::getPrimitiveIsFloor( int id )
    {
        auto it = _state._objects.begin();
        std::advance( it, id );
        return it->second.floor;
    }

    int
    AnnotationController::setPrimitiveIsFloor( int id, bool is_floor, bool is_selected )
    {
        auto it = _state._objects.begin();
        std::advance( it, id );
        it->second.floor = is_floor;
        // redraw
        it->second.reDraw( /*        viewer: */ _viewer,
                           /* name_from_map: */ it->first,
                           /*   is_selected: */ is_selected
                           , _colour_code_cuboids
                           , _state._colour_edges
                           );

        /*it->second.CuboidPtr()->unDraw( _viewer, it->first );
        it->second.CuboidPtr()->draw( _viewer
                                      , it->first
                                      , is_floor ? CUBOID_FLOOR_COLOUR : CUBOID_COLOUR
                                      , CUBOID_SELECTED_LINE_WIDTH );*/

        //std::cout << "primitive " << id << " is now floor: " << (it->second.floor?"YES":"NO") << std::endl;

        return EXIT_SUCCESS;
    }

    int
    AnnotationController::setPrimitiveIsSphere( int id, bool is_sphere, bool is_selected )
    {
        auto it = _state._objects.begin();
        std::advance( it, id );
        it->second.shape = is_sphere ? am::Cuboid::TYPE::SPHERE : am::Cuboid::TYPE::CUBOID;

//        if ( is_sphere )
//            if ( it->second.primitive_ptr )
//            {
//                std::shared_ptr<am::Cuboid> cuboid_ptr = std::static_pointer_cast< am::Cuboid,am::Primitive>( it->second.primitive_ptr );
//                if ( cuboid_ptr )
//                {
//                    Eigen::Vector3f extent = cuboid_ptr->getExtent();
//                    Eigen::Vector3f::Index min_extent_id;
//                    float min_extent = extent.minCoeff( &min_extent_id );
//                    std::cout << "[" << __func__ << "]: " << "min_extent_id: " << min_extent_id << std::endl;
//                    Eigen::Vector3f resize_offset( Eigen::Vector3f::Zero() );
//                    for ( int d = 0; d != 3; ++d )
//                    {
//                        if ( d == min_extent_id ) continue;
//                        resize_offset( d ) = min_extent - extent(d);
//                    }
//                    std::cout << "resizing by : " << resize_offset.transpose() << std::endl;
//                    cuboid_ptr->resizeAlongFrame( resize_offset );
//                }
//                else
//                    std::cerr << "[" << __func__ << "]: " << "cuboid NULL" << std::endl;
//            }
//            else
//            {
//                std::cerr << "[" << __func__ << "]: " << "primitive NULL" << std::endl;
//            }
        // redraw
        it->second.reDraw( /*        viewer: */ _viewer,
                           /* name_from_map: */ it->first,
                           /*   is_selected: */ is_selected
                           , _colour_code_cuboids
                           , _state._colour_edges
                           );

        /*it->second.CuboidPtr()->unDraw( _viewer, it->first );
        it->second.CuboidPtr()->draw( _viewer
                                      , it->first
                                      , is_floor ? CUBOID_FLOOR_COLOUR : CUBOID_COLOUR
                                      , CUBOID_SELECTED_LINE_WIDTH );*/

        //std::cout << "primitive " << id << " is now floor: " << (it->second.floor?"YES":"NO") << std::endl;

        return EXIT_SUCCESS;
    }

    int
    AnnotationController::setPrimitiveIsCylinder( int id, bool is_cylinder, bool is_selected )
    {
        auto it = _state._objects.begin();
        std::advance( it, id );
        it->second.shape = is_cylinder ? am::Cuboid::TYPE::CYLINDER : am::Cuboid::TYPE::CUBOID;

//        if ( is_cylinder )
//            if ( it->second.primitive_ptr )
//            {
//                std::shared_ptr<am::Cuboid> cuboid_ptr = std::static_pointer_cast< am::Cuboid,am::Primitive>( it->second.primitive_ptr );
//                if ( cuboid_ptr )
//                {
//                    Eigen::Vector3f extent = cuboid_ptr->getExtent();
//                    Eigen::Vector3f::Index min_extent_id;
//                    float min_extent = extent.minCoeff( &min_extent_id );
//                    std::cout << "[" << __func__ << "]: " << "min_extent_id: " << min_extent_id << std::endl;
//                    Eigen::Vector3f resize_offset( Eigen::Vector3f::Zero() );
//                    for ( int d = 0; d != 3; ++d )
//                    {
//                        if ( d == min_extent_id ) continue;
//                        resize_offset( d ) = min_extent - extent(d);
//                    }
//                    std::cout << "resizing by : " << resize_offset.transpose() << std::endl;
//                    cuboid_ptr->resizeAlongFrame( resize_offset );
//                }
//                else
//                    std::cerr << "[" << __func__ << "]: " << "cuboid NULL" << std::endl;
//            }
//            else
//            {
//                std::cerr << "[" << __func__ << "]: " << "primitive NULL" << std::endl;
//            }
        // redraw
        it->second.reDraw( /*        viewer: */ _viewer,
                           /* name_from_map: */ it->first,
                           /*   is_selected: */ is_selected
                           , _colour_code_cuboids
                           , _state._colour_edges
                           );

        /*it->second.CuboidPtr()->unDraw( _viewer, it->first );
        it->second.CuboidPtr()->draw( _viewer
                                      , it->first
                                      , is_floor ? CUBOID_FLOOR_COLOUR : CUBOID_COLOUR
                                      , CUBOID_SELECTED_LINE_WIDTH );*/

        //std::cout << "primitive " << id << " is now floor: " << (it->second.floor?"YES":"NO") << std::endl;

        return EXIT_SUCCESS;
    }

    std::vector<Contact>
    AnnotationController::getLatestContacts() const
    {
        return _state._latest_contacts;
    }

    DisplayedPrimitive const&
    AnnotationController::getPrimitive( int id )
    {
        auto it = _state._objects.begin();
        std::advance( it, id );
        return it->second;
    }

    DisplayedPrimitive::Map const&
    AnnotationController::getPrimitives()
    {
        return _state._objects;
    }

    std::string
    AnnotationController::getOutputPath()
    {
        return _output_path;
    }

    int
    AnnotationController::setCuboidColour( int id, Eigen::Vector4f colour )
    {
        auto it = _state._objects.begin();
        std::advance( it, id );
        it->second.colour = colour;
        redrawPrimitive( id, true );
    }

    Eigen::Vector4f
    AnnotationController::getCuboidColour( int id )
    {
        auto it = _state._objects.begin();
        std::advance( it, id );
        return it->second.colour;
    }

    int
    AnnotationController::setCuboidOpacity( int id, double opacity )
    {
        for ( auto it = _state._objects.begin();
                   it != _state._objects.end();
                 ++it )
        {
            // if ID is specifiied
            if ( id > -1 )
            {
                std::advance( it, id );
                it->second.colour(3) = opacity;
                return redrawPrimitive( id, true );
            }

            // otherwise, change all
            it->second.colour(3) = opacity;
        }

        // redraw, if all were changed
        if ( id == -1 )
        {
            _default_opacity = opacity;
            return redrawAllPrimitives();
        }

        // shouldn't reach this
        return EXIT_FAILURE;
    }

    int
    AnnotationController::setShowCuboidIDs( bool show )
    {
        _state._show_IDs = show;
        return redrawAllPrimitives();
    }

    std::vector<Eigen::Vector3f>
    AnnotationController::getCameraPosFocalUp()
    {
        return getCamerasPosFocalUp( _viewer )[0];
    }

    int
    AnnotationController::setCameraPosFocalUp( std::vector<Eigen::Vector3f> pos_focal_up )
    {
        return am::setCameraPosFocalUp( _viewer, pos_focal_up );
    }

    int
    AnnotationController::setSphereSize( double value )
    {
        //std::cout << "[" << __func__ << "]: " << "sphere_size: " << value << std::endl;
        _state._max_sphere_size = value;

        if ( _state._drawn_contacts.size() )
        {
            drawIntersections( false, &_state._latest_contacts ); // ! provide contacts argument, otherwise info is lost...
            drawIntersections( true , &_state._latest_contacts );
        }
    }

    int
    AnnotationController::removeAllPoints()
    {
        _state.CurrentCloudRef(AnnotationState::FULL)->clear();
        _state.CurrentIndicesRef()->indices.clear();
        _state.MainCloudRef(AnnotationState::FULL)->clear(); // recently added
        _viewer->removePointCloud( "big_cloud", 0 );
    }

    int
    AnnotationController::addModel( std::string path )
    {
        pcl::PolygonMesh mesh;
        pcl::io::loadPolygonFile( path, mesh );
        std::string poly_name = am::util::sprintf( "polygonmeshmodel_%d", _state._drawn_models.size() );
        _viewer->addPolygonMesh( mesh, poly_name.c_str() );
        _state._drawn_models.push_back( poly_name );
    }

    std::pair<int,std::string>
    AnnotationController::rotateCamera( bool rotate, QVTKWidget *display )
    {
        std::string message;
        if ( !rotate )
        {
            // TODO: stop rotating
            return std::pair<int,std::string>( EXIT_SUCCESS, "" );
        }

        /// start rotating
        // get floor
        auto floor_ptrs = getFloors();
        if ( floor_ptrs.empty() )
        {
            return std::pair<int,std::string>( EXIT_FAILURE, "Need a floor for upvector" );
        }
        // get cuboids centroid
        auto target_ptrs = getRotationTargets();
        if ( target_ptrs.empty() )
        {
            return std::pair<int,std::string>( EXIT_FAILURE, "Need a target for center" );
        }

        // rotationcentroid
        Eigen::Vector3f centroid;
        for ( std::shared_ptr<am::Cuboid> const cuboid_ptr : target_ptrs )
        {
            centroid += cuboid_ptr->Center();
        }
        centroid /= target_ptrs.size();

        // draw rotation axis
        Eigen::Vector3f rot_axis = floor_ptrs[0]->Planes()[2].Normal().normalized();
        //Eigen::AngleAxisf rot( ang, tmp_frame.block<3,1>(0,2) );
        //tmp_frame.block<3,1>(0,0) = rot * tmp_frame.block<3,1>(0,0);

        _viewer->addArrow( am::util::pcl::asPointXYZ(centroid-rot_axis*1.5f) //upvector is always from plane_id=2
                ,am::util::pcl::asPointXYZ(centroid)
                ,0, 0, 0
                , false
                , "rotuparrow", 0 );
        display->update();

        std::vector<Eigen::Vector3f> pos_focal_up = getCameraPosFocalUp(), tmp_pos_focal_up;
        tmp_pos_focal_up = pos_focal_up;
        std::cout << "firstpos is: " << pos_focal_up[0] << std::endl;

        for ( float deg = 0.f; deg < 360.f; ++deg )
        {
            Eigen::AngleAxisf rot( pcl::deg2rad(deg), rot_axis );
            tmp_pos_focal_up[0] = rot * pos_focal_up[0];

            setCameraPosFocalUp( tmp_pos_focal_up );
            std::cout << "new pos is: " << tmp_pos_focal_up[0].transpose() << std::endl;
            display->update();
            usleep(100000);
        }

        // start thread to rotate

        return { EXIT_SUCCESS, "Ok" };
    }

    pcl::visualization::PCLVisualizer::Ptr
    AnnotationController::getViewer()
    {
        return _viewer;
    }

    Eigen::Matrix4d
    AnnotationController::getProjMatrix()
    {
        return _getProjMatrix( _viewer, 0 );
    }

    Eigen::Matrix4d
    AnnotationController::getViewMatrix()
    {
        return _getViewMatrix( _viewer, 0 );
    }

    int
    AnnotationController::savePointsToPLY( std::string path ) const
    {
        return pcl::io::savePLYFileASCII( path, *_state.CurrentCloud() );
    }
} // ns am
