#include "cuboid_approx/AnnotationState.h"

#include "cuboid_approx/primitives/Contact.h"

namespace ca
{
    /// ------------------------------------------------------------------------------------------------ ///
    /// CLASS AnnotationState
    /// ------------------------------------------------------------------------------------------------ ///
    AnnotationState::AnnotationState()
        : _main_cloud_ptr( nullptr )
        , _current_cloud_ptr( nullptr )
        , _current_indices_ptr( nullptr )
        , _next_primitive_id( 0 )
        , _max_sphere_size( .04f )
        , _min_sphere_size( .02f )
        , _colour_edges( false )
        , _show_IDs( false )
        , _display_point_size( 1. )
        , _main_cloud_dirty( true )
        , _current_cloud_dirty( true )
        , _current_indices_dirty( true )
    {
        _objects.clear();

        // make space for picking at least one object
        _picked.resize( 1 );
    }

    int
    AnnotationState::removePick( int point_id, int obj_id )
    {
        if ( obj_id >= static_cast<int>(_picked.size()) )
        {
            std::cerr << "[" << __func__ << "]: " << "obj_id " << obj_id << ">= " << _picked.size() << " picked.size()...exiting\\n";
            return EXIT_FAILURE;
        }

        if ( obj_id < 0 ) _picked.back() .erase( point_id );
        else              _picked[obj_id].erase( point_id );

        return EXIT_SUCCESS;
    }

    int
    AnnotationState::hasPick( int point_id, int obj_id )
    {
        if ( obj_id >= static_cast<int>(_picked.size()) )
        {
            std::cerr << "[" << __func__ << "]: " << "obj_id " << obj_id << ">= " << _picked.size() << " picked.size()...exiting\\n";
            return -1;
        }

        std::set<int>::const_iterator it;
        if ( obj_id < 0 )
        {
            it = _picked.back() .find( point_id );
            if ( it != _picked.back().end() )
                return *it;
        }
        else
        {
            it = _picked[obj_id].find( point_id );
            if ( it != _picked[obj_id].end() )
                return *it;
        }

        return -1;
    }

    int
    AnnotationState::addPick( int point_id, int obj_id )
    {
        if ( obj_id >= static_cast<int>(_picked.size()) )
        {
            std::cerr << "[" << __func__ << "]: " << "obj_id " << obj_id << ">= " << _picked.size() << " picked.size()...exiting\\n";
            return EXIT_FAILURE;
        }

        if ( obj_id < 0 )   _picked.back() .insert( point_id );
        else                _picked[obj_id].insert( point_id );

        return EXIT_SUCCESS;
    }

    float
    AnnotationState::getClosestPickedTo( int point_id, int obj_id )
    {
        if ( obj_id >= static_cast<int>(_picked.size()) )
        {
            std::cerr << "[" << __func__ << "]: " << "obj_id " << obj_id << ">= " << _picked.size() << " picked.size()...exiting\\n";
            return -1.f;
        }

        std::set<int> &picked = ( (obj_id < 0) ? _picked.back() : _picked[obj_id] );

        float min_dist = FLT_MAX;
        for ( auto it = picked.begin(); it != picked.end(); ++it )
            min_dist = std::min( (CurrentCloud()->at(*it).getVector3fMap() - CurrentCloud()->at(point_id).getVector3fMap()).norm(), min_dist );

        return min_dist;
    }

    int
    AnnotationState::clearPick( int obj_id )
    {
        if ( obj_id >= static_cast<int>(_picked.size()) )
        {
            std::cerr << "[" << __func__ << "]: " << "obj_id >= picked.size...exiting\\n";
            return EXIT_FAILURE;
        }

        if ( obj_id < 0 )   _picked.back()   .clear();
        else                _picked[ obj_id ].clear();
    }


    std::vector<int>
    AnnotationState::pickedIndices( int obj_id )
    {
        if ( obj_id >= static_cast<int>(_picked.size()) )
        {
            std::cerr << "[" << __func__ << "]: " << "obj_id >= picked.size...exiting\\n";
            return std::vector<int>();
        }

        if ( obj_id < 0 )   return std::vector<int>( _picked.back().begin(), _picked.back().end() );
        else                return std::vector<int>( _picked[obj_id].begin(), _picked[obj_id].end() );
    }

    PCloud::Ptr&
    AnnotationState::MainCloudRef( int change_level )
    {
        if ( change_level == FULL )
            _main_cloud_dirty = true;

        return _main_cloud_ptr;
    }

    PCloud::Ptr const&
    AnnotationState::MainCloud() const
    {
        return _main_cloud_ptr;
    }

    PCloud::Ptr&
    AnnotationState::CurrentCloudRef( int change_level )
    {
        if ( change_level == FULL )
            _current_cloud_dirty = true;

        return _current_cloud_ptr;
    }

    PCloud::Ptr const&
    AnnotationState::CurrentCloud() const
    {
        return _current_cloud_ptr;
    }

    pcl::PointIndices::Ptr&
    AnnotationState::CurrentIndicesRef()
    {
        _current_indices_dirty = true;
        return _current_indices_ptr;
    }

    pcl::PointIndices::Ptr const&
    AnnotationState::CurrentIndices() const
    {
        return _current_indices_ptr;
    }

} // ... ns ca
