#include <iostream>

#include <pcl/console/parse.h>

#include <map>
#include <Eigen/Dense>
#include <vtkObjectFactory.h>
#include <vtkSmartPointer.h>
#include <vtkPointPicker.h>
#include <vtkRenderWindow.h>
#include <vtkRenderWindowInteractor.h>
#include <pcl/visualization/pcl_visualizer.h>
#include <pcl/io/ply_io.h>
#include "my_screenshot_manager.h"
#include "primitives/Cuboid.h"

#include "cuboid_approx/ui/Annot3DInteractorStyle.h"

namespace ca
{
    class KinfuAnnotator
    {
        public:
            typedef pcl::PointXYZRGBA PointT;
            enum MOUSE_STATE { LEFT_DOWN = 1 };

            KinfuAnnotator( pcl::visualization::PCLVisualizer::Ptr vptr );

            inline int
            initialize( pcl::visualization::PCLVisualizer::Ptr vptr );

            static int
            run( std::string const& path, std::string const& poses_path, int argc, char** argv );

            void _mouseCallbacks( const pcl::visualization::MouseEvent event );

        protected:
            pcl::visualization::PCLVisualizer::Ptr  _vis;
            vtkSmartPointer<Annot3DInteractorStyle>   _interactor_style;
            int                                     _mouse_state;
            //pcl::PointCloud<PointT>::Ptr            _cloud;

            std::vector<std::vector<int> >          _indices_vec;

            KinfuAnnotator();

    };

    KinfuAnnotator::KinfuAnnotator( pcl::visualization::PCLVisualizer::Ptr vptr )
        : KinfuAnnotator()
    {
        this->initialize( vptr );
    }

    KinfuAnnotator::KinfuAnnotator()
        : _mouse_state( 0 )
    {}

    inline void
    KinfuAnnotator::_mouseCallbacks( const pcl::visualization::MouseEvent event )
    {
        // toggle state
        if ( /**/ event.getType()   == pcl::visualization::MouseEvent::MouseButtonPress
             &&   event.getButton() == pcl::visualization::MouseEvent::LeftButton
             &&   event.getKeyboardModifiers() & pcl::visualization::KeyboardEvent::Shift )
        {
            _mouse_state |= LEFT_DOWN;

            // disable camera navigation
            _interactor_style->setPropagateMouseMove( false );

            // start new region
            _indices_vec.push_back( std::vector<int>() );
        }
        else if ( /**/ event.getType()   == pcl::visualization::MouseEvent::MouseButtonRelease
                  &&   event.getButton() == pcl::visualization::MouseEvent::LeftButton          )
        {
            _mouse_state &= !LEFT_DOWN;

            // reenable camera navigation
            _interactor_style->setPropagateMouseMove( true );

            // add cuboid
            if ( _indices_vec.size() && _indices_vec.back().size() )
            {
                using namespace am;

#if 0
                // copy indices
                pcl::PointIndices::Ptr tmp_indices( new pcl::PointIndices() );
                {
                    tmp_indices->indices = _indices_vec.back();
                }

                // copy pointcloud
                pcl::PointCloud<am::MyPoint>::Ptr tmp_cloud( new pcl::PointCloud<am::MyPoint>() );
                {
                    tmp_cloud->reserve( _cloud->size() );
                    for ( size_t pid = 0; pid != _cloud->size(); ++pid )
                    {
                        am::MyPoint pnt;
                        pnt.x = _cloud->at( pid ).x;
                        pnt.y = _cloud->at( pid ).y;
                        pnt.z = _cloud->at( pid ).z;
                        pnt.r = _cloud->at( pid ).r;
                        pnt.g = _cloud->at( pid ).g;
                        pnt.b = _cloud->at( pid ).b;
                        pnt.label = 0;
                        tmp_cloud->push_back( pnt );
                    }
                }

                Cuboid cuboid;
                int err = Cuboid::fromCloudWithPCA( /*                  cuboid_out: */ cuboid
                                                    , /*                     cloud: */ tmp_cloud //boost::const_pointer_cast<const pcl::PointCloud<PointT> >(_cloud)
                                                    , /*                   indices: */ tmp_indices
                                                    , /*         max_neighbours(K): */ 15
                                                    , /*                    radius: */ -1.f
                                                    , /*             clean_n_means: */ 10
                                                    , /*              clean_stddev: */ .01
                                                    , /* density_median_multiplier: */ 1.2f
                                                    , /*                     debug: */ false
                                                    );

                {
                    char tmp[1024];
                    sprintf( tmp, "cuboid%d", static_cast<int>(_indices_vec.size()) );

                    cuboid.draw4( _vis, tmp );
                }
#endif
            }
        }

        // left press
        if ( _mouse_state & LEFT_DOWN )
        {
            vtkSmartPointer<vtkPointPicker> point_picker = vtkPointPicker::New();
            vtkSmartPointer<vtkRenderWindow> win = _vis->getRenderWindow();
            win->GetInteractor()->SetPicker( point_picker );

            // pick point based on XY position
            point_picker->Pick(event.getX(), event.getY(), 0.0, _vis->getRendererCollection()->GetFirstRenderer() );

            // if a point is under the cursor and it belongs to a surface, highlight the surface
            if (point_picker->GetDataSet() != NULL)
            {
                int idx = static_cast<int>( point_picker->GetPointId() );
                std::cout << "idx: " << idx << std::endl;
                _indices_vec.back().push_back( idx );

//                _cloud->at(idx).r = 255;
//                _vis->updatePointCloud( _cloud, "cloud" );
//                _vis->getRenderWindow()->Render();
            }
        }
    }
    
    inline int
    KinfuAnnotator::initialize( pcl::visualization::PCLVisualizer::Ptr vptr )
    {
        _interactor_style = vtkSmartPointer<Annot3DInteractorStyle>( Annot3DInteractorStyle::New() );
        _interactor_style->Initialize();
        _interactor_style->setKeyboardModifier( pcl::visualization::INTERACTOR_KB_MOD_SHIFT );
        vtkSmartPointer<vtkRendererCollection> rens = vptr->getRendererCollection();
        _interactor_style->setRendererCollection( rens );
        //        _interactor_style->setCloudActorMap( _vis->getCloudActorMap() );

        // setup mouse callbacks
        boost::function<void (const pcl::visualization::MouseEvent)> fMouse = boost::bind( &ca::KinfuAnnotator::_mouseCallbacks, this, _1 );
        boost::signals2::connection cMouse = _interactor_style->registerMouseCallback( fMouse );

        vptr->setupInteractor( vptr->getRenderWindow()->GetInteractor(), vptr->getRenderWindow(), _interactor_style );
    }

    inline int
    KinfuAnnotator::run( const std::string &path, std::string const& poses_path, int argc, char** argv )
    {
        // debug paths
        {
            std::cout << "[" << __func__ << "]: "
                      << "path: " << path
                      << ", pose_path: " << poses_path
                      << std::endl;
        }

        // read paths
        {
            std::map<int,Eigen::Affine3f> poses;
            Eigen::Matrix3f intrinsics;
            am::MyScreenshotManager::readPoses( poses_path, poses, &intrinsics );

            for ( auto it = poses.begin(); it != poses.end(); ++it )
                std::cout << "pose" << std::distance(poses.begin(), it) << "\n" << it->second.matrix() << std::endl;
            std::cout << "intrinsics:\n" << intrinsics << std::endl;
        }

        // read cloud
        pcl::PointCloud<PointT>::Ptr cloud;
        if ( !cloud) cloud.reset( new pcl::PointCloud<PointT>() );
        {
            pcl::io::loadPLYFile( path, *cloud );
        }

        // init viewer
        pcl::visualization::PCLVisualizer::Ptr vis( new pcl::visualization::PCLVisualizer() );
        vis->setBackgroundColor( .5, .5, .5 );

        // display cloud
        {
            vis->addPointCloud( cloud, "cloud" );
            vis->setPointCloudRenderingProperties( pcl::visualization::PCL_VISUALIZER_POINT_SIZE, 3.0 );
        }

        KinfuAnnotator kinfu_annot( vis );

        // loop
        vis->spin();

        return 0;
    }
}

int main( int argc, char** argv )
{
    std::cout << "Hello kinfuAnnotator" << std::endl;

    std::string cloud_path, poses_path;
    if ( !pcl::console::parse_argument(argc, argv, "--points", cloud_path) )
    {
        std::cerr << "[" << __func__ << "]: " << "please add --points argument" << std::endl;
        return 1;
    }

    if ( !pcl::console::parse_argument(argc, argv, "--poses", poses_path) )
    {
        std::cerr << "[" << __func__ << "]: " << "please add --poses argument" << std::endl;
        return 1;
    }

    ca::KinfuAnnotator::run( cloud_path, poses_path, argc, argv );

    return 0;
}
