/*
 * =====================================================================================
 *
 *       Filename:  furniture.cpp
 *
 *    Description:  
 *
 *        Version:  1.0
 *        Created:  10/02/2013 10:01:32 AM
 *       Revision:  none
 *       Compiler:  gcc
 *
 *         Author:  Moos Hueting (mh), mooshueting@gmail.com
 *   Organization:  
 *
 * =====================================================================================
 */
#include <stdlib.h>
#include "furniture.h"

FurnitureSuite::FurnitureSuite()
    : cloud(new PointCloud), viewer(new pcl::visualization::PCLVisualizer("3D Viewer"))
{
    // default constructor
}

FurnitureSuite::FurnitureSuite(std::string filename)
    : cloud(new PointCloud), viewer(new pcl::visualization::PCLVisualizer("3D Viewer"))
{
    pcl::PCDReader reader;
    reader.read(filename, *cloud);
    std::cout << "Read PointCloud in " << filename << " with " << 
        cloud->points.size() << " points." << std::endl;
}

void FurnitureSuite::removeStatisticalNoise()
{
    // create temporary point cloud to store filtered result
    PointCloud::Ptr tmpCloud (new PointCloud);

    // setup statistical outlier removal
    pcl::StatisticalOutlierRemoval<pcl::PointXYZRGB> sor;
    sor.setInputCloud(cloud);
    sor.setMeanK(50);
    sor.setStddevMulThresh(1.0);
    sor.filter(*tmpCloud);
    *cloud = *tmpCloud;
    std::cout << "PointCloud after statistical outlier removal has " << cloud->points.size() << " points." << std::endl;

}

void FurnitureSuite::decimatePointCloud()
{
    // create temporary point cloud to store filtered result
    PointCloud::Ptr tmpCloud (new PointCloud);

    // setup voxel grid
    pcl::VoxelGrid<pcl::PointXYZRGB> vg;
    vg.setInputCloud(cloud);
    vg.setLeafSize(0.01f, 0.01f, 0.01f); // voxels of 1cm^3
    vg.filter(*tmpCloud);
    *cloud = *tmpCloud;
    std::cout << "PointCloud after decimation has " << cloud->points.size() << " points."
        << std::endl;

}

void FurnitureSuite::detectPlanes()
{
    // setup Sample COnsensus segmentation for planar models in scene
    pcl::SACSegmentation<pcl::PointXYZRGB> seg;
    // indices of points that lie in found plane
    pcl::PointIndices::Ptr inliers (new pcl::PointIndices);
    // coefficients of found plane
    pcl::ModelCoefficients::Ptr coefficients (new pcl::ModelCoefficients);
    pcl::PointCloud<pcl::PointXYZRGB>::Ptr cloud_plane (new PointCloud ()), tmpCloud (new PointCloud ());

    seg.setOptimizeCoefficients(true);
    seg.setModelType(pcl::SACMODEL_PLANE);
    seg.setMethodType(pcl::SAC_RANSAC);
    seg.setMaxIterations(100);
    seg.setDistanceThreshold(0.05);

    int cloud_size = (int) cloud->points.size();
    while (cloud->points.size() > 0.3 * cloud_size) {
        // find plane with largest number of points in cloud
        seg.setInputCloud(cloud);
        seg.segment(*inliers, *coefficients);
        if (inliers->indices.size() < 10000) {
            std::cout << "Plane with highest number of points too small, stopping." << std::endl;
            break;
        }

        // Extract plane from cloud
        pcl::ExtractIndices<pcl::PointXYZRGB> extract;
        extract.setInputCloud(cloud);
        extract.setIndices(inliers);
        extract.setNegative(false);
        extract.filter(*cloud_plane);
        std::cout << "PointCloud representing the planar component: " << cloud_plane->points.size () << " data points." << std::endl;
        
        // Extract remaining cloud
        extract.setNegative(true);
        extract.filter(*tmpCloud);
        *cloud = *tmpCloud;
        
        // Add plane to set
        planeCoefficients.push_back(*coefficients);

    }

}

void FurnitureSuite::segment()
{
    // create kd-tree
    pcl::search::KdTree<pcl::PointXYZRGB>::Ptr tree (new pcl::search::KdTree<pcl::PointXYZRGB>);
    tree->setInputCloud(cloud);

    pcl::EuclideanClusterExtraction<pcl::PointXYZRGB> ec;
    ec.setClusterTolerance(0.02); // 2cm
    ec.setMinClusterSize(250);
    ec.setMaxClusterSize(25000);
    ec.setSearchMethod(tree);
    ec.setInputCloud(cloud);
    ec.extract(clusterIndices);
}

// UI
void FurnitureSuite::showPointCloud() {
    // colour in cloud according to RGB data
    pcl::visualization::PointCloudColorHandlerRGBField<pcl::PointXYZRGB> rgb(cloud);

    // add point cloud with colour to viewer
    viewer->addPointCloud<pcl::PointXYZRGB>(cloud, rgb, "Main Cloud");
}

void FurnitureSuite::showPlanes() {
    // initialize iterator and counter
    std::vector<pcl::ModelCoefficients>::const_iterator planeIterator;
    int i;

    // iterate over plane coefficients, and add them to the viewer
    for (planeIterator = planeCoefficients.begin(), i = 0; planeIterator != planeCoefficients.end(); ++planeIterator, ++i) {
        // generate name for plane
        std::stringstream planeName;
        planeName << "plane" << i;
        viewer->addPlane(*planeIterator, planeName.str());
    }

}

void FurnitureSuite::showClusters() {
    // vector to store min and max points of clusters
    std::vector<pcl::PointXYZRGB> mins;
    std::vector<pcl::PointXYZRGB> maxs;
    
    // initialize iterator and counter
    std::vector<pcl::PointIndices>::const_iterator clusterIterator;
    int i;
    for (clusterIterator = clusterIndices.begin(), i = 0; clusterIterator != clusterIndices.end(); ++clusterIterator, ++i) {
        // initialize temporary cloud to store points of cluster
        PointCloud::Ptr tmpCloud(new PointCloud);
        for (std::vector<int>::const_iterator pointIterator = clusterIterator->indices.begin(); pointIterator != clusterIterator->indices.end(); ++pointIterator) {
            tmpCloud->points.push_back(cloud->points[*pointIterator]);
        }

        // find extremities of cluster
        pcl::PointXYZRGB minPt, maxPt;
        pcl::getMinMax3D(*tmpCloud, minPt, maxPt);
        mins.push_back(minPt);
        maxs.push_back(maxPt);

        // generate name for bounding box
        std::stringstream cubeName;
        cubeName << "cube" << i;
        
        // add bounding box to viewer
        viewer->addCube(minPt.x, maxPt.x, minPt.y, maxPt.y, minPt.z, maxPt.z, 1.0f, 1.0f, 1.0f, cubeName.str());

    }
}

void FurnitureSuite::showViewer() {
    while (!viewer->wasStopped()) {
        viewer->spinOnce(100);
        boost::this_thread::sleep (boost::posix_time::microseconds (100000));
    }
}
