#include "cuboid_approx/primitives/Contact.h"

#include "cuboid_approx/util/AMUtil2.h"
#include "primitives/Cuboid.h"
#include "AnnotationController.h"

namespace am
{
    // STATIC FIELDs
    const std::string Contact::pattern              = "proxy_";
    const int         Contact::number_width         = 2;
    const double      Contact::CONTACT_LINE_WIDTH   = 12.;

    // OPERATORs
    bool
    Contact::operator==( Contact const& other ) const
    {
        if ( (other.ids.size()!=2) || (ids.size()!=2) )
        {
            std::cerr << "[" << __func__ << "]: " << "(other.ids.size()!=2) || (ids.size()!=2)...exiting\n";
            return false;
        }

        return ( (other.type == type)
                 && (
                         ( (other.ids[0] == ids[0]) && (other.ids[1] == ids[1]) )
                      || ( (other.ids[0] == ids[1]) && (other.ids[1] == ids[0]) )
                    )
               );
    } // ... operator==

    bool Contact::operator< ( Contact const& other ) const // sort
    {
        if ( ids[0] != other.ids[0] )
            return ids[0] < other.ids[0];
        else
            return ids[1] < other.ids[1];
    } // ... operator<

    // SET, GET
    Eigen::Vector4f
    Contact::getColour() const
    {
        switch ( type )
        {
            case TYPE::TOUCHING:
                //return AnnotationController::CONTACT_TOUCHING_COLOUR;
                return ca::CONTACT_TOUCHING_COLOUR;
                break;
            case TYPE::FIXED:
                //return AnnotationController::CONTACT_FIXED_COLOUR;
                return ca::CONTACT_FIXED_COLOUR;
                break;
            case TYPE::UNDEFINED:
                return Eigen::Vector4f::Zero();
                break;
            default:
                std::cerr << __func__ << "  dunno..." << std::endl;
                return Eigen::Vector4f::Zero();
                break;
        }
    } // ... getColour

    std::string
    Contact::toString() const
    {
        char format[512];
        sprintf( format, "%%s%%0%dd %%s%%0%dd %%s", number_width, number_width );
        char line[512];
        sprintf( line, format
                 , pattern.c_str(), ids[0], pattern.c_str(), ids[1], ((type==TOUCHING)?"touching":"fixed") );

        return std::string( line );
    } // ... toString

    // STATICs
    int
    Contact::read( std::vector<am::Contact>      & contacts
                   , std::string            const& fileName
                   , int                         * max_proxy_id_arg
                   , bool                   const  silent           )
    {
        int max_proxy_id = 0;
        if ( fileName.empty() )
        {
            std::cerr << "[" << __func__ << "]: " << "fileName " << fileName << " cannot be opened...\n";
            return EXIT_FAILURE;
        }

        std::string line;
        std::ifstream myfile( fileName );
        if ( !myfile.is_open() )
        {
            std::cerr << "[" << __func__ << "]: " << "fileName " << fileName << " cannot be opened...\n";
            return EXIT_FAILURE;
        }

        while ( getline (myfile,line) )
        {
            if ( line.empty() )
                continue;

            am::Contact contact;

            if ( !silent )
            {
                std::cout << "line: " << line << "\t";
                std::cout << "matches: ";
            }

            size_t pos;
            while ( (pos = line.find(pattern)) != std::string::npos )
            {
                std::string number = line.substr( pos+pattern.size(), number_width );
                int id = atoi( number.c_str() );

                if ( !silent )  std::cout << "[" << id << "], ";

                contact.ids.push_back( id );
                line = line.substr( pos + pattern.size() );

                if ( id > max_proxy_id )    max_proxy_id = id;
            }

            if ( contact.ids.size() != 2 )
            {
                std::cerr << __func__ << ": parse error("<<line<<"), 2 ids required for a contact..." << std::endl;
                continue;
            }
            else // order
            {
                if ( contact.ids[1] < contact.ids[0] ) std::swap( contact.ids[0], contact.ids[1] );
            }

            // parse contact
            if ( line.find("touching") != std::string::npos )
            {
                contact.type = am::Contact::TOUCHING;
                if ( !silent )  std::cout << " -> TOUCHING";
            }
            else if ( line.find("fixed") != std::string::npos )
            {
                contact.type = am::Contact::FIXED;
                if ( !silent )  std::cout << " -> FIXED";
            }
            else if ( line.find("removed") != std::string::npos )
            {
                contact.type = am::Contact::REMOVED;
                if ( !silent )  std::cout << " -> REMOVED";
            }
            else
            {
                std::cerr << __func__ << ": parse error, not touching or fixed..." << std::endl;
                if ( !silent )  std::cout << std::endl;
                continue;
            }
            if ( !silent )  std::cout << std::endl;

            // store
            contacts.push_back( contact );
        } // ... while getline

        // finish
        myfile.close();

        // output
        if ( max_proxy_id_arg )
            *max_proxy_id_arg = max_proxy_id;

        // return
        return EXIT_SUCCESS;
    } // ... read

    int
    Contact::write( std::vector<Contact> const& contacts, std::string const& fileName )
    {
        // open
        std::ofstream f( fileName );
        if ( !f.is_open() )
        {
            std::cerr << "[" << __func__ << "]: " << "could not open " << fileName << std::endl;
            return EXIT_FAILURE;
        }

        // write
        for ( size_t i = 0; i != contacts.size(); ++i )
            f << contacts[i].toString() << std::endl;

        // cleanup
        f.close();

        return EXIT_SUCCESS;
    } // ... write

    int
    Contact::compareFiles( std::string   const& path
                           , std::string const& path_gt
                           , bool               verbose
                           , float            * p_precision
                           , float            * p_recall
                           , int              * max_allowed_box_id
                           , std::string      * map_path
                           , int              * contact_count
                           , int              * cuboid_count )
    {
        std::vector<Contact> contacts, contacts_gt;

        // contacts
        int err = Contact::read( contacts, path, NULL, true );
        if ( EXIT_SUCCESS != err ) { std::cerr << "[" << __func__ << "]: " << " Contact::read returned error " << err << std::endl; return err; }

        // gt
        err = Contact::read( contacts_gt, path_gt, NULL, true );
        if ( EXIT_SUCCESS != err ) { std::cerr << "[" << __func__ << "]: " << " Contact::read returned error " << err << std::endl; return err; }

        // map
        std::map<int,int> mapping;
        if ( map_path )
        {
            std::ifstream myfile( *map_path );
            if ( !myfile.is_open() )
            {
                std::cerr << "[" << __func__ << "]: " << "fileName " << *map_path << " cannot be opened...\n";
                return EXIT_FAILURE;
            }

            std::string line;
            while ( getline(myfile, line) )
            {
                if ( !line.empty() && (line[0] == '#') )
                    continue;

                std::istringstream iss( line );
                std::string word;

                std::vector<int> ids;
                while ( std::getline(iss, word, ',') )
                {
                    ids.push_back( atoi(word.c_str()) );
                } // ... while word

                if ( ids.size() != 2 )
                {
                    std::cerr << "[" << __func__ << "]: " << "skipping line " << line << "..." << std::endl;
                    continue;
                }

                if ( mapping.find(ids[0]) != mapping.end() )
                {
                    std::cerr << "[" << __func__ << "]: "
                              << "duplicate mapping entry: " << line
                              << " and " << ids[0] << "," << mapping.at(ids[0])
                              << ", mapping.find(ids[0]): " << mapping.find(ids[0])->first << ", " << mapping.find(ids[0])->second
                              << "..." << std::endl;
                    continue;
                }

                mapping[ids[0]] = ids[1];
            } // ... while getline
        } // ... if map_path

        // filter
        int filter_at = std::numeric_limits<int>::max();
        if ( max_allowed_box_id )
            filter_at = *max_allowed_box_id;
        int valid_contacts = 0, valid_gt_contacts = 0;
        int max_gt_box_id = 0;

        float precision = 0, recall = 0;
        for ( size_t i = 0; i != contacts.size(); ++i )
        {
            am::Contact &contact = contacts[i];
            if ( map_path )
            {
                if ( mapping.find(contact.ids[0]) != mapping.end() )
                    contact.ids[0] = mapping[ contact.ids[0] ];
                if ( mapping.find(contact.ids[1]) != mapping.end() )
                    contact.ids[1] = mapping[ contact.ids[1] ];
            }

            if ( (contact.ids[0] > filter_at) || (contact.ids[1] > filter_at) )
                continue;
            else
                ++valid_contacts;

            if ( std::find(contacts_gt.begin(), contacts_gt.end(),contact) != contacts_gt.end() )
                precision += 1.f;
            else
                if ( verbose )
                    std::cout << "FP: " << contacts[i].toString() << std::endl;
        }

        // mapping is done by now
        for ( size_t i = 0; i != contacts_gt.size(); ++i )
        {
            if ( (contacts_gt[i].ids[0] > filter_at) || (contacts_gt[i].ids[1] > filter_at) )
                continue;
            else
                ++valid_gt_contacts;

            max_gt_box_id = std::max( std::max(max_gt_box_id, contacts_gt[i].ids[0] ), contacts_gt[i].ids[1] );

            if ( std::find(contacts.begin(), contacts.end(), contacts_gt[i]) != contacts.end() )
                recall += 1.f;
            else
                if ( verbose )
                    std::cout << "FN: " << contacts_gt[i].toString() << std::endl;
        }

//        precision /= static_cast<float>( contacts   .size() );
//        recall    /= static_cast<float>( contacts_gt.size() );

        if ( valid_contacts )
            precision /= static_cast<float>( valid_contacts );
        else
            precision = 0;

        if ( valid_gt_contacts )
            recall    /= static_cast<float>( valid_gt_contacts );
        else
            recall = 1;

        // final output
        if ( verbose ) std::cout << "precision recall\n";

        std::cout << precision
                  << "\t" //( (verbose) ? "\t" : " ")
                  << recall;

        if ( p_precision    ) *p_precision      = precision;
        if ( p_recall       ) *p_recall         = recall;
        if ( contact_count  ) *contact_count    = valid_gt_contacts;
        if ( cuboid_count   ) *cuboid_count     = max_gt_box_id + 1;

        return err;
    } // ... compareFiles

    int
    Contact::filterByMaxCuboidID( am::Contact::Vec       &filtered
                                  , am::Contact::Vec      input // copy, so that it can be called in place
                                  , int                   max_cuboid_id ) // NOT size
    {
        filtered.clear();
        am::Contact::Vec::const_iterator end_it = input.end();
        for ( am::Contact::Vec::const_iterator it = input.begin(); it != end_it; ++it )
        {
            bool keep = true;
            for ( int id = 0; (id != it->ids.size()) && keep; ++id )
            {
                keep &= (it->ids[id] <= max_cuboid_id);
            }

            if ( keep )
                filtered.push_back( *it );
        }

        return EXIT_SUCCESS;
    }

#if 0
    int
    Contact::sort( std::vector<Contact> &contacts )
    {
        std::sort( contacts.begin(), contacts.end(), []( Contact const& a, Contact const& b){
            if ( a.ids[0] != b.ids[0] )
                return a.ids[0] < b.ids[0];
            else
                return a.ids[1] < b.ids[1];
        });

        return EXIT_SUCCESS;
    } // ... sort
#endif

    std::string
    drawSphere(
            Eigen::Matrix<float,3,1>                    const& center
            , float                                     const  radius
            , Eigen::Vector4f                           const& colour
            , int                                       const  cuboid_id
            , pcl::visualization::PCLVisualizer::Ptr           viewer
            , int                                       const  viewport
            , float                                     const  min_radius /*= _state._max_sphere_size*/
            , float                                     const  max_radius /*= .02f*/
            )
    {
        bool succ = false;

        std::string sphere1_name = am::util::sprintf( "contact_sphere%02d", cuboid_id );

        auto it = viewer->getShapeActorMap()->find( sphere1_name );
        if ( it == viewer->getShapeActorMap()->end() )
        {
            pcl::PointXYZ p; p.x = center(0); p.y = center(1); p.z = center(2);
#if CA_PCL_MODIFIED
            succ = viewer->addSphere( p
                                      , std::max( min_radius, std::min( max_radius, radius ) )
                                      , colour(0), colour(1), colour(2)
                                      , sphere1_name
                                      , viewport
                                      , 20 );
#else
            succ = viewer->addSphere( p
                                      , std::max( min_radius, std::min( max_radius, radius ) )
                                      , colour(0), colour(1), colour(2)
                                      , sphere1_name
                                      , viewport
                                      );
#endif

            viewer->setShapeRenderingProperties( pcl::visualization::PCL_VISUALIZER_SHADING
                                                 , pcl::visualization::PCL_VISUALIZER_SHADING_PHONG
                                                 , sphere1_name
                                                 , viewport );
            if ( succ )
                return sphere1_name;
        }

        return std::string("");
    } // ... drawSphere

    std::vector<std::string>
    Contact::draw( Cuboid                                   const& cubA
                   , Eigen::Matrix<float,4,1>               const& colourA
                   , Cuboid                                 const& cubB
                   , Eigen::Matrix<float,4,1>               const& colourB
                   , pcl::visualization::PCLVisualizer::Ptr        viewer
                   , int                                    const  viewport
                   , double                                 const  line_width
                   , float                                  const  max_radius
                   , float                                  const  min_radius ) const
    {
        std::vector<std::string> drawn_contacts;

        // LINE
        {
            std::stringstream ss;
            ss << "contact_" << std::setw(2) << std::setfill('0') << this->ids[0] << "_" << this->ids[1] << std::endl;
            std::string link_name( ss.str() );
            //std::string link_name( am::util::sprintf("contact%d",(contact.ids[0]+1)*100+(contact.ids[1]+1)) );

            auto link_colour = this->getColour();
            viewer->addLine(   am::util::pcl::asPointXYZ(cubA.Center())
                             , am::util::pcl::asPointXYZ(cubB.Center())
                             , link_colour(0), link_colour(1), link_colour(2)
                             , link_name
                             , viewport );
            viewer->setShapeRenderingProperties( pcl::visualization::PCL_VISUALIZER_LINE_WIDTH, line_width, link_name );
            viewer->setShapeRenderingProperties( pcl::visualization::PCL_VISUALIZER_OPACITY, link_colour(3), link_name );

            drawn_contacts.push_back( link_name );
        }

        // collect cuboids
        std::vector<Cuboid const*> cubs( {&cubA, &cubB} );

        // SPHEREs
        for ( int i = 0; i != ids.size(); ++i )
        {
            std::string sphere_name( "" );
            sphere_name = drawSphere( cubs[i]->Center()
                                      , static_cast<float>( cubs[i]->getSideLengths()[0]/2.f )
                                      , (i ? colourB : colourA) + (Eigen::Vector4f() << .2f, 0.f, 0.f,0.f).finished() // draw a bit lighter
                                      , this->ids[i]
                                      , viewer
                                      , viewport
                                      , min_radius
                                      , max_radius );

            if ( !sphere_name.empty() )
                drawn_contacts.push_back( sphere_name );
        }

        return drawn_contacts;
    } // ... draw

//    int
//    Contact::unDraw( pcl::visualization::PCLVisualizer::Ptr viewer )
//    {
//        for ( std::string &shape_name : _state._drawn_contacts )
//        {
//            std::cout << "removing " << shape_name << std::endl;
//            viewer->removeShape( shape_name, viewport );
//        }

//        if ( !_state._drawn_contacts.size() )
//            std::cerr << __func__ << ": nothing to remove" << std::endl;

//        _state._drawn_contacts.clear();
//    } // ... unDraw

} // ... ns am
