#include "Cuboid.h"

#if USE_WM5
    #include "Wm5Box3.h"
    #include "Wm5IntrBox3Box3.h"
#endif

#include <iostream>
#include <fstream>
#include <cstdio>

#include <pcl/common/transforms.h>          // transformPointCloud
#include "pcl/common/common.h"              // getMinMax3D
#include "pcl/filters/project_inliers.h"    // projectInliers
#include "pcl/common/centroid.h"            // compute3DCentroid
#include "pcl/filters/extract_indices.h"
#include "pcl/common/intersections.h"       // planeplaneintersection

#include "cuboid_approx/util/AMUtil2.h"

#include "DepthProcessor.h"

namespace am
{
    template <typename T> inline T
    safe_acos( T dot )
    {
        T val = acos( dot );
        if ( val != val ) val = T(0);

        return val;
    }

#if USE_WM5
    typedef Wm5::IntrBox3Box3f  INTR_BOX3BOX3_F;
    typedef Wm5::Box3<float>             BOX3_F;
#endif //USE_WM5

    Cuboid Cuboid::operator=( const Cuboid& other )
    {
        if ( this != &other )
        {
            other.cloneTo( *this );
        }

        return *this;
    }

    /// STATIC METHODS
    int
    Cuboid::calculateCorners( std::vector<pcl::PointXYZ> &corners,
                              MyPoint                    &min_pt,
                              MyPoint                    &max_pt,
                              Cuboid               const &cuboid )
    {
        // move the points to the that reference frame
        pcl::PointCloud<MyPoint> cPoints;
        {
            Eigen::Matrix4f p2w( Eigen::Matrix4f::Identity() );
            p2w.block<3,3>(0,0) = cuboid.R().transpose();
            p2w.block<3,1>(0,3) = -1.f * ( p2w.block<3,3>(0,0) * cuboid.Center() );
            pcl::transformPointCloud( *cuboid.getROICloud(), cPoints, p2w );
        }

        // get bounds
        {
            pcl::getMinMax3D<MyPoint>( cPoints, min_pt, max_pt );
        }

        // create corners in cuboid space
        std::vector<Eigen::Vector3f> corners_eigen =
        {
            // right handed: x -> right, y -> up, z -> towards viewpoint
            {min_pt.x, min_pt.y, max_pt.z}, // 0,0,1
            {max_pt.x, min_pt.y, max_pt.z}, // 1,0,1
            {max_pt.x, max_pt.y, max_pt.z}, // 1,1,1
            {min_pt.x, max_pt.y, max_pt.z}, // 0,1,1
            {min_pt.x, min_pt.y, min_pt.z}, // 0,0,0
            {max_pt.x, min_pt.y, min_pt.z}, // 1,0,0
            {max_pt.x, max_pt.y, min_pt.z}, // 1,1,0
            {min_pt.x, max_pt.y, min_pt.z}  // 0,1,0
        };

        // output to world space
        corners.resize( corners_eigen.size() );
        for ( int corner_id = 0; corner_id < corners_eigen.size(); ++corner_id )
            corners[corner_id] = am::util::pcl::asPointXYZ( cuboid.R() * corners_eigen[corner_id] + cuboid.Center() );
    }

    int
    Cuboid::draw( Cuboid                                 const &cuboid
                  , pcl::visualization::PCLVisualizer::Ptr     v
                  , std::string                                cloud_name
                  , Eigen::Vector3f                            colour
                  , double                                     line_width
                  , bool                                       colour_edges
                  , bool                                       draw_name   )
    {
        Eigen::Vector4f colour_with_opacity;
        colour_with_opacity.head<3>() = colour;
        colour_with_opacity(3)        = .15;
        Cuboid::draw4( cuboid, v, cloud_name, colour_with_opacity, line_width, colour_edges, draw_name );
    }

    int
    Cuboid::draw4( Cuboid                                 const &cuboid,
                   pcl::visualization::PCLVisualizer::Ptr       v,
                   std::string                                  cloud_name,
                   Eigen::Vector4f                              colour,
                   double                                       line_width
                   , bool                                       colour_edges
                   , bool                                       draw_name
                   , TYPE                                 const shape )
    {
        bool draw_frame = false; // cuboids don't draw frames by default

        if ( shape == Cuboid::TYPE::SPHERE )
        {
            float radius = cuboid.getExtent()[0];
            std::cout << "[" << __func__ << "]: " << "radius: " << radius << std::endl;
            std::string sphere_name = cloud_name + "_sphere";
#if CA_PCL_MODIFIED
            v->addSphere( am::util::pcl::asPointXYZ(cuboid.Center()), radius, colour(0), colour(1), colour(2), sphere_name, 0, 20 );
#else
            v->addSphere( am::util::pcl::asPointXYZ(cuboid.Center()), radius, colour(0), colour(1), colour(2), sphere_name, 0 );
#endif
            v->setShapeRenderingProperties( pcl::visualization::PCL_VISUALIZER_OPACITY, colour(3), sphere_name );

            // if sphere, draw frame, if ID visible
            if ( draw_name )
                draw_frame = true;
        }
        else if ( shape == Cuboid::TYPE::CYLINDER )
        {
            Eigen::Vector3f extent = cuboid.getExtent();
            float height = extent(0) * 2;
            float radius = extent(1);
            std::cout << "[" << __func__ << "]: " << "radius: " << radius << std::endl;
            std::string sphere_name = cloud_name + "_sphere";

            Eigen::Vector3f axis = cuboid.R().col(0) * height;
            Eigen::Vector3f bottom = cuboid.Center() - axis/2;

            pcl::ModelCoefficients cylinder_coeff;
            cylinder_coeff.values.resize(7);
            cylinder_coeff.values[0] = bottom(0);
            cylinder_coeff.values[1] = bottom(1);
            cylinder_coeff.values[2] = bottom(2);

            // first eigen axis is cylinder axis
            cylinder_coeff.values[3] = axis(0);
            cylinder_coeff.values[4] = axis(1);
            cylinder_coeff.values[5] = axis(2);

            cylinder_coeff.values[6] = radius; // second axis dim is radius, should be the same as 3rd

            //v->addCylinder( am::util::pcl::asPointXYZ( colour(0), colour(1), colour(2), sphere_name, 0, 20 );
#if CA_PCL_MODIFIED
            v->addCylinder( cylinder_coeff, colour(0), colour(1), colour(2), sphere_name, 0 );
#else
            v->addCylinder( cylinder_coeff, sphere_name, 0 );
#endif
            v->setShapeRenderingProperties( pcl::visualization::PCL_VISUALIZER_OPACITY, colour(3), sphere_name );
            v->setShapeRenderingProperties( pcl::visualization::PCL_VISUALIZER_REPRESENTATION, pcl::visualization::PCL_VISUALIZER_REPRESENTATION_SURFACE, sphere_name );

            // if sphere, draw frame, if ID visible
            if ( draw_name )
                draw_frame = true;
        }
        else
        {
            /// drawCuboid
            // edge lines
            {
                const std::vector<std::pair<int,int>> cuboid_lines =
                {
                    {0,1}, {1,2}, {2,3}, {3,0},
                    {4,5}, {5,6}, {6,7}, {7,4},
                    {0,4}, {1,5}, {2,6}, {3,7}
                };

                for ( int line_id = 0;
                      line_id < cuboid_lines.size();
                      ++line_id )
                {
                    std::string line_name = cloud_name + Cuboid::cuboidLineName(-1, line_id);

                    Eigen::Vector4f line_colour = colour_edges ? colour : (Eigen::Vector4f() << .3f, .3f, .3f, 1.f).finished();
                    v->addLine( cuboid.Corners()[ cuboid_lines[line_id].first ]
                            , cuboid.Corners()[ cuboid_lines[line_id].second]
                            //, colour(0), colour(1), colour(2)
                            , line_colour(0), line_colour(1), line_colour(2) //.9, .9, 0.,
                            //.3, .3, .3,
                            , line_name                                        );
                    v->setShapeRenderingProperties( pcl::visualization::PCL_VISUALIZER_LINE_WIDTH, line_width, line_name );
                }
            } // ... edge lines

            // faces
            {
                std::vector<std::vector<int> > corner_ids =
                { //{3,0,7,4},
                  //{7,3,0},
                  {0,7,3},
                  {0,7,4},
                  {0,1,4},
                  {1,4,5},
                  {0,3,1},
                  {3,1,2},
                  {1,2,5},
                  {2,5,6},
                  {2,3,6},
                  {3,6,7},
                  {4,5,7},
                  {5,7,6}
                };

                for ( int face_id = 0;
                      face_id != corner_ids.size();
                      ++face_id )
                {
                    std::string face_name = cloud_name + am::util::sprintf("face%d",face_id);

                    pcl::PointCloud<pcl::PointXYZ> polygon_cloud;
                    for ( int cid = 0; cid != corner_ids[face_id].size(); ++cid )
                    {
                        pcl::PointXYZ pnt = cuboid.Corners()[ corner_ids[face_id][cid] ];
                        polygon_cloud.push_back( pnt );
                    }
                    v->addPolygon<pcl::PointXYZ>( polygon_cloud.makeShared(), colour(0), colour(1), colour(2), face_name );

                    v->setShapeRenderingProperties( pcl::visualization::PCL_VISUALIZER_OPACITY, colour(3), face_name );
                    v->setShapeRenderingProperties( pcl::visualization::PCL_VISUALIZER_REPRESENTATION, pcl::visualization::PCL_VISUALIZER_REPRESENTATION_SURFACE, face_name );
                } // ... for faces
            } // ... draw faces
        } // ... if cuboid shape

        // draw cuboid name
        if ( draw_name )
        {
            v->addText3D( cloud_name.substr(cloud_name.size()-2),
                          am::util::pcl::asPointXYZ(cuboid._center),
                          .05,
                          1., 0., 0.,
                          cloud_name+"_cloud_name" );
        }

        // frame
        if ( draw_frame )
            for ( int col = 0; col < 3; ++col )
            {
                char tit[255];
                sprintf( tit, "%s_arrow%d", cloud_name.c_str(), col );
                v->addArrow( am::util::pcl::asPointXYZ( cuboid.Center() + cuboid.R().col(col) )
                             , am::util::pcl::asPointXYZ( cuboid.Center() )
                             , (col == 0) * 1.f, (col == 1) * 1.f, (col == 2) * 1.f
                             , false
                             , tit
                             );
            }

#if 0

        Eigen::Vector3f local_offset = cuboid.R() * (Eigen::Vector3f() << .1f, 0.f, 0.f).finished();
        for ( int corner_id = 0; corner_id < cuboid.Corners().size(); ++corner_id )
        {
            Eigen::Vector3f corner = cuboid.Corners()[corner_id].getVector3fMap();

            char tit[255];
            sprintf( tit, "%s_dbg_arrow%d", cloud_name.c_str(), corner_id );
            v->addArrow( am::util::pcl::asPointXYZ( cuboid.Center() + local_offset )
                         , am::util::pcl::asPointXYZ( cuboid.Center() )
                         , 1.f, 1.f, 0.f
                         , false
                         , tit
                         );
            sprintf( tit, "%s_dbg2_arrow%d", cloud_name.c_str(), corner_id );
            v->addArrow( am::util::pcl::asPointXYZ( corner )
                         , am::util::pcl::asPointXYZ( cuboid.Center() )
                         , 0.f, 1.f, 1.f
                         , false
                         , tit
                         );
        }

        for ( int i = 0; i < cuboid._corners.size(); ++i )
        {
            char tit[255];
            sprintf(tit,"C%d_%s",i,cloud_name.c_str());
            v->addText3D( tit, cuboid._corners[i], .05 );
        }
#endif

#if 0
        // draw plane normals
        {
            for ( int pid = 0; pid < cuboid.Planes().size(); ++pid )
            {
                char plane_prefix[255];
                sprintf( plane_prefix, "%s_plane%d", cloud_name.c_str(), pid );
                cuboid.Planes()[pid].draw( v, plane_prefix, .5, .0, .0, true, true );
            }
        }
#endif
    }

    int
    Cuboid::unDraw( pcl::visualization::PCLVisualizer::Ptr v,
                    std::string cloud_name                   )
    {

        v->removeShape( cloud_name + "_sphere" );
        for ( int lid = 0;
              lid < 12;
              ++lid )
        {
            Cuboid::cuboidLineName( -1, lid );
            //std::cout <<  "removing " << cloud_name + Cuboid::cuboidLineName( -1, lid ) << std::endl;
            v->removeShape( cloud_name + Cuboid::cuboidLineName( -1, lid ) );
        }

        for ( int col = 0; col < 3; ++col )
        {
            char tit[255];
            sprintf( tit, "%s_arrow%d", cloud_name.c_str(), col );
            v->removeShape( tit );
        }

        for ( int i = 0; i < 8; ++i )
        {
            char tit[255];
            sprintf(tit,"C%d%s",i,cloud_name.c_str());
            v->removeShape( tit );
        }

        for ( int face_id = 0; face_id != 12; ++face_id )
        {
            std::string name = cloud_name + am::util::sprintf("face%d",face_id);
            v->removeShape( name );
        }
        //v->removeShape( cloud_name+"cloud_name" );
        v->removeText3D( cloud_name + "_cloud_name" );
    }

    int
    Cuboid::cloneTo( Cuboid &cuboid ) const
    {
        int err;
        err = Primitive::cloneTo( cuboid );

        cuboid._center  = _center;
        cuboid._R       = _R;
        cuboid._corners = _corners;
        cuboid._planes  = _planes;

        return err;
    }

    int
    Cuboid::writeToFile( Cuboid const& cuboid, std::string path, bool append )
    {
        std::ofstream myfile;
        myfile.open ( path, std::ios_base::binary | std::ios_base::out | (append ? std::ios_base::app : std::ios_base::trunc) );
        if ( !myfile.is_open() )
        {
            std::cerr << "could not open file " << path << std::endl;
            return EXIT_FAILURE;
        }

        if ( !append )
        {
            int cuboid_count = 1;
            myfile.write( (char*)&cuboid_count, sizeof(int) );
        }

        for ( pcl::PointXYZ const& corner : cuboid.Corners() )
        {
            //std::cout << "writing " << corner << std::endl;
            myfile.write( (char*)(&corner.x), sizeof(float) );
            myfile.write( (char*)(&corner.y), sizeof(float) );
            myfile.write( (char*)(&corner.z), sizeof(float) );
        }

        myfile.close();
        return 0;
    }

    int
    Cuboid::writeToTextFile( Cuboid const& cuboid, std::string path, bool append )
    {
        std::ofstream myfile;
        myfile.open ( path, std::ios_base::out | (append ? std::ios_base::app : std::ios_base::trunc) );
        if ( !myfile.is_open() )
        {
            std::cerr << "could not open file " << path << std::endl;
            return EXIT_FAILURE;
        }

        if ( !append )
        {
            int cuboid_count = 1;
            //myfile.write( (char*)&cuboid_count, sizeof(int) );
            myfile << cuboid_count << std::endl;
        }

        for ( pcl::PointXYZ const& corner : cuboid.Corners() )
        {
            //std::cout << "writing " << corner << std::endl;
//            myfile.write( (char*)(&corner.x), sizeof(float) );
//            myfile.write( (char*)(&corner.y), sizeof(float) );
//            myfile.write( (char*)(&corner.z), sizeof(float) );
            myfile << corner.x << "," << corner.y << "," << corner.z << ",";
        }

        myfile.close();
        return 0;
    }

    std::string
    Cuboid::cuboidName( int cube_id )
    {
        char cube_name[255];
        sprintf( cube_name, "cube%02d", cube_id );
        return std::string( cube_name );
    }

    std::string
    Cuboid::cuboidLineName( int cube_id, int line_id )
    {
        char line_name[255];
        sprintf( line_name, "cuboid_line%02d", line_id );

        return (cube_id > -1) ? (cuboidName(cube_id) + line_name) : line_name ;
    }

    int
    Cuboid::verify( am::Cuboid const& cuboid )
    {
        if ( cuboid.Corners().size() != 8 )
        {
            std::cerr << "Cuboid::Verify(): cuboid.Corners().size() != 8" << std::endl;
            return EXIT_FAILURE;
        }
        auto corners = cuboid.Corners();

        std::vector< std::vector<int> > parallels =
        {
            { 0,1,  6,7 },
            { 0,1,  4,5 },
            { 0,1,  2,3 },
            { 1,2,  3,0 },
            { 1,2,  5,6 },
            { 1,2,  7,4 },
            { 0,4,  1,5 },
            { 0,4,  2,6 },
            { 0,4,  3,7 }
        };

        std::vector< std::vector<int> > perpendicular =
        {
            { 0,1,  0,3 },
            { 0,1,  0,4 },
            { 0,3,  0,4 }
        };

        for ( int i = 0; i < parallels.size(); ++i )
        {
            Eigen::Vector3f line0 = corners[ parallels[i][1] ].getVector3fMap()
                                  - corners[ parallels[i][0] ].getVector3fMap();
            Eigen::Vector3f line1 = corners[ parallels[i][3] ].getVector3fMap()
                                  - corners[ parallels[i][2] ].getVector3fMap();
            if ( fabs(line0.dot(line1)) < .9f )
                printf( "p%dp%d . p%dp%d not parallel!!! (%f)\n", parallels[i][0],parallels[i][1],parallels[i][2],parallels[i][3], line0.dot(line1) );
            else
                printf( "p%dp%d . p%dp%d parallel\n", parallels[i][0],parallels[i][1],parallels[i][2],parallels[i][3] );
        }

        for ( int i = 0; i < perpendicular.size(); ++i )
        {
            Eigen::Vector3f line0 = corners[ perpendicular[i][1] ].getVector3fMap()
                                  - corners[ perpendicular[i][0] ].getVector3fMap();
            Eigen::Vector3f line1 = corners[ perpendicular[i][3] ].getVector3fMap()
                                  - corners[ perpendicular[i][2] ].getVector3fMap();
            if ( fabs(line0.dot(line1)) > .1f )
                printf( "p%dp%d . p%dp%d not perpendicular!!! (%f)\n", perpendicular[i][0],perpendicular[i][1],perpendicular[i][2],perpendicular[i][3], line0.dot(line1) );
            else
                printf( "p%dp%d . p%dp%d perpendicular\n", perpendicular[i][0],perpendicular[i][1],perpendicular[i][2],perpendicular[i][3] );
        }

        //am::CloudViewer viewer( "cuboidverify" );
        //cuboid.draw( viewer.ViewerPtr(), "cuboid" );
        //viewer.ViewerPtr()->spin();

        return EXIT_SUCCESS;
    }

    /// CONSTR
    int
    Cuboid::from_cloud( Cuboid                       &cuboid,
                        PCloud::ConstPtr              cloud_ptr,
                        pcl::PointIndices::Ptr const &indices_ptr,
                        double                        distanceThreshold,
                        bool                          debug )
    {
        // cache var
        am::Planes &planes      = cuboid.PlanesRef();
        cuboid.CloudPtrRef()    = cloud_ptr;
        cuboid.IndicesPtrRef()  = indices_ptr;

        am::Params params;
        params.setDouble("min_remainder_percentage",.1);

        if ( debug )
            am::CloudViewer::showCloud( cloud_ptr, "maskedindices", indices_ptr );

        int err = am::Planes::from_cloud( planes,
                                cloud_ptr,
                                indices_ptr,
                                params,
                                distanceThreshold,
                                debug );

        if ( planes.size() == 1 )
        {
            std::cout << __func__ << ": planes.size() == 1, using Cuboid::form1Plane..." << std::endl;
            if ( planes[0].Normal().dot( (Eigen::Vector3f()<<0.f,0.f,1.f).finished()) > 0.f )
                 planes[0].flipNormal();
            am::Cuboid::from1Plane( cuboid, planes[0], distanceThreshold );
            return 2;
        }
        else

        if ( (err != EXIT_SUCCESS) || (planes.size() != 3) )
        {
            std::cerr << "planes.size: " << planes.size() << std::endl;
            return std::max( err, EXIT_FAILURE );
        }

        std::vector< am::Histogram > histograms;
        {
            histograms.resize( planes.size(), am::Histogram(50) );
            for ( int plane_id = 0; plane_id < planes.size(); ++plane_id )
            {
                std::cout << "projectPointsToNormal plane_id: " << plane_id << " indices_ptr->size: " << planes[plane_id].IndicesPtr()->indices.size() << std::endl;

                am::Plane::projectPointsToNormal( /*     out_hist: */ histograms[plane_id]
                                                  , /*   in_plane: */ planes[plane_id]
                                                  , /*   in_cloud: */ planes[plane_id].CloudPtr()
                                                  , /* in_indices: */ indices_ptr
                                                  , /*  bin_count: */ 50
                                                  , /* distance_threshold: */ distanceThreshold
                                                  , /* bool debug? */ debug );
            }
        }

        std::vector<double> offsets;
        {
            am::Plane::processHistogram( offsets, histograms, .2 );
        }

        // set frames from neighbouring plane normals
        for ( int plane_id = 0; plane_id < planes.size(); ++plane_id )
        {
            // normals
            Eigen::Matrix4f frame;
            for ( int plane_id2 = 0; plane_id2 < planes.size(); ++plane_id2 )
            {
                frame.block<3,1>(0,plane_id2) = planes[ (plane_id+plane_id2+1)%planes.size() ].Normal();
            }
            // centroid
            frame.block<3,1>(0,3) = planes[plane_id].getCentroid();

            // store
            planes[plane_id].Frame() = frame;

            // print histograms
            {
                std::cout << "planes{" << plane_id+1 << "} = [";
                for ( auto const& val : histograms[plane_id].histogram )
                {
                    std::cout << val << ", ";
                }
                std::cout << "];" << std::endl;
            }
        }

        // dbg
        {
            am::util::printVector( offsets, "offsets" );
        }

        // move planes based on histograms
        for ( int plane_id = 0; plane_id < 3; ++plane_id )
        {
            // copy
            planes.add( planes[plane_id].clone() );
            planes[3+plane_id].offsetAlongNormal( offsets[3+plane_id] );
            planes[3+plane_id].flipNormal();

            // adjust
            planes[plane_id].offsetAlongNormal( offsets[plane_id] );

            am::Plane::calculateCorners( planes[plane_id],
                                         planes[plane_id].Frame() );
            am::Plane::calculateCorners( planes[3+plane_id],
                                         planes[3+plane_id].Frame() );
        }

        // get lines
        std::vector<Eigen::VectorXf> lines(4);
        std::vector<int>             pids = { 0, 1, 3, 4 }; // 4 plane ids forming a column
        for ( int pid = 0; pid < pids.size(); ++pid )
        {
            std::cout << "planewplane: " << pid
                      << "\n\tplanes[pids[ "<<pid<<"               ]].Coeffs().transpose(): " << planes[pids[ pid               ]].Coeffs().transpose()
                      << ",\n\tplanes[pids["<<(pid+1)%pids.size()<<"]].Coeffs().transpose(): " << planes[pids[(pid+1)%pids.size()]].Coeffs().transpose()
                      << ",\n\tlines[" << pid << "]: " << lines[pid].transpose()
                      << std::endl;
            pcl::planeWithPlaneIntersection(
                        planes[pids[ pid               ]].Coeffs(),
                        planes[pids[(pid+1)%pids.size()]].Coeffs(),
                        lines[pid],
                        .05 );
            std::cout << "[" << __func__ << "]: " << "planeWithPlane ok" << std::endl; fflush(stdout);
        }

        // get corners
        cuboid.Corners().resize(8);
        for ( int lid = 0; lid < lines.size(); ++lid )
        {
            cuboid.Corners()[  lid] = am::util::pcl::asPointXYZ( am::geom::lineToPlaneIntersection(lines[lid], planes[2].Coeffs()) );
            cuboid.Corners()[4+lid] = am::util::pcl::asPointXYZ( am::geom::lineToPlaneIntersection(lines[lid], planes[5].Coeffs()) );
        }

        // get centroid and R
        calculateFrameFromPlanes( cuboid );

        return EXIT_SUCCESS;
    }

    int
    Cuboid::fromCorners( am::Cuboid                 &cuboid,
                         std::vector<float>          corner_coords,
                         PCloud::ConstPtr            cloud_ptr,
                         pcl::PointIndices::ConstPtr indices_ptr
                         , bool debug )
    {
        if ( corner_coords.size() < 24 )
        {
            std::cerr << "Cuboid::fromCorners needs 24 corner coordinates, " << corner_coords.size() << " is not enough..." << std::endl;
            return EXIT_FAILURE;
        }
        cuboid.CloudPtrRef() = cloud_ptr;

        // store corners
        cuboid.Corners().resize( 8 );
        int corner_id = 0;
        for ( int coord_id = 0; coord_id < 24; coord_id+=3, ++corner_id )
        {
            cuboid.Corners()[ corner_id ] = pcl::PointXYZ( corner_coords[coord_id  ],
                                                           corner_coords[coord_id+1],
                                                           corner_coords[coord_id+2] );
        }

        // create planes
        cuboid.PlanesRef().clear();
        std::vector<std::vector<int>> corner_ids =
        { {3,0,7,4},
          {0,1,4,5},
          {0,3,1,2},
          {1,2,5,6},
          {2,3,6,7},
          {4,5,7,6},
        };

        am::Plane plane;
        for ( int i = 0; i < corner_ids.size(); ++i )
        {
            Plane::fromCorners( plane, cuboid.Corners(), corner_ids[i], cuboid.CloudPtr(), indices_ptr );
            cuboid.PlanesRef().add( plane );
        }

        // flipnormals towards inside
        {
            // get centroid and R
            int err = calculateFrameFromPlanes( cuboid, debug );

            if ( debug && (err != EXIT_SUCCESS) )
            {
                // debug
                {
                    pcl::visualization::PCLVisualizer::Ptr v( new pcl::visualization::PCLVisualizer() );
                    v->setBackgroundColor( .5, .6, .6 );
                    for ( int i = 0; i < corner_ids.size(); ++i )
                    {
                        char tmp[64];
                        sprintf( tmp, "plane%d", i );
                        cuboid.Planes().at( i ).draw( v, tmp, .2, .0, .0, true, true );
                        Eigen::Vector3f centroid(0,0,0);
                        for ( int j = 0; j != 4; ++j )
                            centroid += cuboid.Planes().at( i ).Corners().at(j).getVector3fMap();
                        centroid /= 4.f;
                        v->addText3D( tmp, pcl::PointXYZ( centroid(0), centroid(1), centroid(2)), .01, .2, .0, .0, std::string(tmp)+"text", 0 );
                    }
                    std::cout << "[" << __func__ << "]: " << "corners:\n";
                    for ( int i = 0; i < cuboid.Corners().size(); ++i )
                    {
                        char tmp[64];
                        sprintf( tmp, "corner_arrow%d", i );
                        v->addArrow( cuboid.Corners()[i]
                                     , pcl::PointXYZ(0,0,0)
                                     , .8, .0, .0
                                     , false
                                     , tmp
                                     , 0 );
                        v->addText3D( tmp, cuboid.Corners()[i], .01, .8, .0, .0, std::string(tmp)+"cid", 0 );
                        std::cout << "\t" << cuboid.Corners()[i].getVector3fMap().transpose() << std::endl;
                    }
                    std::cout << std::endl;

                    CloudViewer::addCloud( v, cloud_ptr, "cloud", indices_ptr );
                    v->spin();
                }

                // debug
                {
                    std::cout << "[" << __func__ << "]: sidelengths:\n";
                    for ( int cid = 0; cid != corner_ids.size(); ++cid )
                    {
                        std::cout << (cuboid.Corners()[ corner_ids[cid][1] ].getVector3fMap() - cuboid.Corners()[ corner_ids[cid][0] ].getVector3fMap()).norm() << ", ";
                        std::cout << (cuboid.Corners()[ corner_ids[cid][2] ].getVector3fMap() - cuboid.Corners()[ corner_ids[cid][1] ].getVector3fMap()).norm() << ", ";
                        std::cout << (cuboid.Corners()[ corner_ids[cid][3] ].getVector3fMap() - cuboid.Corners()[ corner_ids[cid][2] ].getVector3fMap()).norm() << ", ";
                        std::cout << (cuboid.Corners()[ corner_ids[cid][0] ].getVector3fMap() - cuboid.Corners()[ corner_ids[cid][3] ].getVector3fMap()).norm() << std::endl;
                    }
                }
            } // ... debug

            Eigen::Vector3f centroid = cuboid._center;
            for ( int plane_id = 0; plane_id < cuboid._planes.size(); ++plane_id )
            {
                // check, if flip needed
                if ( cuboid._planes[plane_id].Normal().dot(centroid) + cuboid._planes[plane_id].D() < 0.f )
                    cuboid._planes[plane_id].flipNormal();
            }
        }

        if ( cloud_ptr != nullptr )
        {
            // calc cuboid inliers
            cuboid.getInlierIndices( cuboid.IndicesPtrRef(),
                                     cuboid.CloudPtr(),
                                     indices_ptr,
                                     0.f,
                                     false );
        }
    }

    // calls fromCorners
    int
    Cuboid::from1Plane( am::Cuboid &cuboid, am::Plane const& plane_arg, float assumed_depth )
    {
        am::Plane plane( plane_arg );
        if ( plane.Corners().size() != 4 )
        {
            am::Plane::calculateCorners( plane, plane.Frame() );
        }

        std::vector<float> corner_coords(24);
        am::Plane plane2( plane );
        for ( int i = 0; i < plane2.Corners().size(); ++i )
        {
            plane2.Corners()[ i ] = am::util::pcl::asPointXYZ( plane2.Corners()[ i ].getVector3fMap() - plane.Normal() * assumed_depth );
            corner_coords[ i   *3  ] = plane .Corners()[i].x;
            corner_coords[ i   *3+1] = plane .Corners()[i].y;
            corner_coords[ i   *3+2] = plane .Corners()[i].z;
            corner_coords[(i+4)*3  ] = plane2.Corners()[i].x;
            corner_coords[(i+4)*3+1] = plane2.Corners()[i].y;
            corner_coords[(i+4)*3+2] = plane2.Corners()[i].z;
        }
        am::Cuboid::fromCorners( cuboid, corner_coords, plane.CloudPtr(), plane.IndicesPtr() );
        if ( !cuboid.IndicesPtr() ) cuboid.IndicesPtrRef().reset( new pcl::PointIndices() );
        if ( !plane_arg.IndicesPtr() ) { std::cerr << "[" << __func__ << "]: " << "plane_arg.indices_empty..." << std::endl; }
        else                             cuboid.IndicesPtrRef()->indices = plane_arg.IndicesPtr()->indices;

        return EXIT_SUCCESS;
    }

    int
    Cuboid::fromFrame( Cuboid                             & cuboid
                       , Eigen::Matrix4f             const& frame
                       , PCloud::ConstPtr                   cloud
                       , pcl::PointIndices::ConstPtr        indices )
    {
        // calc bounding box
        MyPoint min_pt, max_pt;
        am::DepthProcessor::calculateExtremaInFrame( min_pt, max_pt, frame, cloud, indices );
        std::vector<float> extrema = { min_pt.x, min_pt.y, min_pt.z,
                                       max_pt.x, max_pt.y, max_pt.z };

        std::vector<float> corner_coords( 24 );
        for ( int x = 0; x < 2; ++x)
            for ( int y = 0; y < 2; ++y )
                for ( int z2 = 0; z2 < 2; ++z2 )
                {
                    int corner_id = ( x*4 + y*2 + z2 ) * 3;

                    int z = ( y + z2 ) % 2; // reorder along one dim
                    pcl::PointXYZ local_pnt( extrema[ x*3 ], extrema[ y*3+1 ], extrema[ z*3+2 ] );
                    pcl::PointXYZ world_pnt = toWorld( local_pnt, frame );

                    corner_coords[ corner_id   ] = world_pnt.x;
                    corner_coords[ corner_id+1 ] = world_pnt.y;
                    corner_coords[ corner_id+2 ] = world_pnt.z;
                }

        return am::Cuboid::fromCorners( cuboid, corner_coords, cloud, indices );
    }

    /// METHODS
    int
    Cuboid::calculateCentroidFromCorners( am::Cuboid &cuboid )
    {
        // check
        if ( cuboid._corners.size() != 8 )
        {
            std::cerr << __PRETTY_FUNCTION__ << "cuboid._corners.size() != 8" << std::endl;
            return EXIT_FAILURE;
        }

        // calculate centroid
        Eigen::Vector3f centroid( 0.f, 0.f, 0.f );
        for ( int corner_id = 0; corner_id < cuboid._corners.size(); ++corner_id )
        {
            centroid += cuboid._corners[corner_id].getVector3fMap();
        }
        cuboid._center = centroid / cuboid._corners.size();

        return EXIT_SUCCESS;
    }

    int
    Cuboid::calculateFrameFromPlanes( am::Cuboid &cuboid, bool debug )
    {
        // centroid
        calculateCentroidFromCorners( cuboid );

        // frame (_R)
        std::vector<Eigen::Vector3f> normals;
        for ( int plane_id = 0; (plane_id < cuboid._planes.size()) && (normals.size() < 3); ++plane_id )
        {
            // assemble frame
            auto current_normal = cuboid._planes[plane_id].Normal();

            bool perpendicular = true;
            for ( int normal_id = 0; (normal_id < normals.size()) && perpendicular; ++normal_id )
            {
                perpendicular &= ( fabs(normals[normal_id].dot(current_normal)) < .3f );
                if ( !perpendicular )
                {
                    std::cerr << "fabs(normals[" << normal_id << "].dot( plane" << plane_id << ".normal)): "
                              << fabs(normals[normal_id].dot(current_normal))
                              << " <? " << .3f << std::endl; fflush(stderr);

                    if ( debug )
                    {
                        pcl::visualization::PCLVisualizer::Ptr vptr( new pcl::visualization::PCLVisualizer() );
                        for ( int plane_id = 0; plane_id != cuboid._planes.size(); ++plane_id )
                        {
                            char tmp[64];
                            sprintf( tmp, "plane_%d", plane_id );
                            cuboid._planes.at( plane_id ).draw( vptr, tmp, .2, .0, .0, true, true );
                            for ( int corner_id = 0; corner_id != cuboid._planes.at( plane_id ).Corners().size(); ++corner_id )
                            {
                                std::cout << cuboid._planes.at( plane_id ).Corners()[corner_id].getVector3fMap().transpose() << std::endl;
                            }
                            std::cout << "[" << __func__ << "]: " << "frame:\n" << cuboid._planes.at( plane_id ).Frame() << std::endl;
                        }
                        vptr->spin();
                    }
                }
            }

            if ( perpendicular )
                normals.push_back( current_normal );
        }

        // save
        cuboid._R = Eigen::Matrix3f::Zero();

        if ( normals.size() == 3 )
        {
            for ( int normal_id = 0; normal_id < normals.size(); ++normal_id )
                cuboid._R.col( normal_id ) = normals[ normal_id ].normalized();
        }
        else
        {
            if ( cuboid._planes.size() < 3 )
            {
                std::cerr << "[" << __func__ << "]: " << "can't even fallback...cuboid_planes.size() < 3" << std::endl;
                return EXIT_FAILURE;
            }

            if ( debug )
            {
                std::cout << "safe_acos( 1.f ): " << safe_acos(1.f) << std::endl;
                std::cout << "safe_acos( -1.f ): " << safe_acos(-1.f) << std::endl;
                std::cout << "safe_acos( 0.f ): " << safe_acos(0.f) << std::endl;
            }

            // extract most perpendicular triplet
            float best_cos = 0.f;
            int   ijk[3]   = {0, 1, 2};
            const int plane_count = cuboid._planes.size();
            for (           int i = 0  ; i != plane_count - 2; ++i )
                for (       int j = i+1; j != plane_count - 1; ++j )
                    for (   int k = j+1; k != plane_count    ; ++k )
            {
                float tmp = safe_acos(cuboid._planes.at(i).Normal().dot(cuboid._planes.at(j).Normal())) +
                            safe_acos(cuboid._planes.at(i).Normal().dot(cuboid._planes.at(k).Normal())) +
                            safe_acos(cuboid._planes.at(k).Normal().dot(cuboid._planes.at(j).Normal()));
                if ( tmp < best_cos )
                {
                    best_cos = tmp;
                    ijk[0] = i;
                    ijk[1] = j;
                    ijk[2] = k;
                }
            }

            for ( int i = 0; i != 3; ++i )
                cuboid._R.col( i ) = cuboid._planes.at( ijk[i] ).Normal().normalized();

            std::cerr << "[" << __func__ << "]: "
                      << "not enough perpendicular normals..."
                      << " planes.size(): " << cuboid._planes.size()
                      << ", normals.size(): " << normals.size()
                      << ", best effort R:\n"
                      << cuboid._R << std::endl;

            return EXIT_FAILURE;
        }

        return EXIT_SUCCESS;
    }

    int
    Cuboid::draw( pcl::visualization::PCLVisualizer::Ptr visualizer,
                  std::string                            cloud_name,
                  //int                                    cuboid_id,
                  Eigen::Vector3f                        colour,
                  double                                 line_width,
                  bool                                   colour_edges,
                  bool                                   draw_name ) const
    {
        return Cuboid::draw( *this, visualizer, cloud_name, colour, line_width, colour_edges, draw_name );
    }

    int
    Cuboid::draw4( pcl::visualization::PCLVisualizer::Ptr visualizer
                   , std::string                          cloud_name
                   , Eigen::Vector4f                      colour
                   , double                               line_width
                   , bool                                 colour_edges
                   , bool                                 draw_name
                   , Cuboid::TYPE                   const shape         ) const
    {
        return Cuboid::draw4( *this, visualizer, cloud_name, colour, line_width, colour_edges, draw_name, shape );
    }

    int
    Cuboid::drawPlanes( pcl::visualization::PCLVisualizer::Ptr visualizer,
                        std::string                            cloud_name ) const
    {
        int err = EXIT_SUCCESS;
        std::cout << "drawPlanes: " << _planes.size() << " planes inside" << std::endl;

        char plane_name[255];
        for ( int plane_id = 0; plane_id < _planes.size(); ++plane_id )
        {
            sprintf( plane_name, "%s_plane%d", cloud_name.c_str(), plane_id );
            std::cout << "adding plane " << plane_name << std::endl;
            err = err + _planes[plane_id].draw( visualizer, plane_id );
        }

        return err;
    }

    int
    Cuboid::reassignPlanes( am::Planes &planes,
                            PCloud::ConstPtr cloud_ptr,
                            pcl::PointIndices::ConstPtr indices_ptr,
                            double distanceThreshold )
    {
        std::cout << "calculateplanes" << std::endl;
        pcl::SampleConsensusModelPlane<MyPoint> sac_model( cloud_ptr );
        if ( indices_ptr )
            sac_model.setIndices( indices_ptr->indices );

        std::vector< std::vector<double> > distances(planes.size());
        std::vector<int> closest_plane_ids( indices_ptr->indices.size() );
        std::fill( closest_plane_ids.begin(), closest_plane_ids.end(), -1 );
        for ( int plane_id = 0; plane_id < planes.size(); ++plane_id )
        {
            sac_model.getDistancesToModel( planes[plane_id].Coeffs(), distances[plane_id] );
            std::cout << "got " << distances[plane_id].size() << " distances for plane " << plane_id << std::endl;
            for ( int point_id = 0; point_id < distances[plane_id].size(); ++point_id )
            {
                if ( distances[plane_id][point_id] < distanceThreshold )
                    if ( closest_plane_ids[point_id] == -1 )
                        closest_plane_ids[point_id] = plane_id;
                    else if ( distances[plane_id][point_id] < distances[ closest_plane_ids[point_id] ][ point_id ] )
                        closest_plane_ids[point_id] = plane_id;
            }

            planes[plane_id].IndicesPtrRef()->indices.clear();
        }

        // reassign points

        for ( int point_id = 0; point_id < indices_ptr->indices.size(); ++point_id )
        {
            if ( closest_plane_ids[point_id] > -1 )
                planes[ closest_plane_ids[point_id] ].IndicesPtrRef()->indices.push_back( point_id );
        }

        //for ( int plane_id = 0; plane_id < planes.size(); ++plane_id )
        //{
        //    Plane::calculateCorners( planes[plane_id] );
        //}
    }

    int
    Cuboid::clampPointsToOrientedPlanes( Cuboid            & out_cuboid
                                         , Cuboid     const& in_cuboid
                                         , am::Planes const& planes
                                         , double     const  distanceThreshold )
    {
        std::vector<Eigen::Vector3f> normals(planes.size());
        for ( int plane_id = 0; plane_id < planes.size(); ++plane_id )
        {
            //normals[plane_id] = (planes[plane_id].Normal() + planes[plane_id].getCentroid()).normalized();
            normals[plane_id] = planes[plane_id].Normal();
        }

        //out_cuboid.CloudPtrRef() = in_cuboid.CloudPtr();
        in_cuboid.cloneTo( out_cuboid );

        out_cuboid.IndicesPtrRef() = pcl::PointIndices::Ptr( new pcl::PointIndices() );
        out_cuboid.IndicesPtr()->indices.reserve( in_cuboid.IndicesPtr()->indices.size() );

        for ( int point_id = 0;
                  point_id < in_cuboid.IndicesPtr()->indices.size();
                ++point_id )
        {
            Eigen::Vector3f point = in_cuboid.CloudPtr()->at( in_cuboid.IndicesPtr()->indices[point_id] ).getVector3fMap();
            bool keep = true;
            for ( int plane_id = 0; plane_id < planes.size() && keep; ++plane_id )
                keep &= (normals[plane_id].dot(point) + planes[plane_id].D() > distanceThreshold );
            if ( keep )
                out_cuboid.IndicesPtr()->indices.push_back( in_cuboid.IndicesPtr()->indices[point_id] );
        }
    }



    int
    Cuboid::getInlierIndices( pcl::PointIndices::Ptr               &indices_ptr
                              , PCloud::ConstPtr                    cloud_ptr
                              , pcl::PointIndices::ConstPtr         in_indices_ptr
                              , float                               outside_distance_threshold
                              , bool                          const debug                      ) const
    {
        std::vector<Eigen::Vector3f> normals( _planes.size() );
        for ( int plane_id = 0; plane_id < _planes.size(); ++plane_id )
        {
            //normals[plane_id] = (planes[plane_id].Normal() + planes[plane_id].getCentroid()).normalized();
            normals[plane_id] = _planes[plane_id].Normal();
        }
        const int out_size = in_indices_ptr ? in_indices_ptr->indices.size() : cloud_ptr->size();

        if ( !indices_ptr )     indices_ptr = pcl::PointIndices::Ptr( new pcl::PointIndices() );
        indices_ptr->indices.clear();
        indices_ptr->indices.reserve( out_size );

        if ( outside_distance_threshold == 0.f )    outside_distance_threshold = -FLT_EPSILON;
        if ( outside_distance_threshold > 0.f )
        {
            std::cerr << "warning: Cuboid::getInlierIndices() changing outside_distance_threshold "
                         << outside_distance_threshold << " to -" << outside_distance_threshold << std::endl;
            outside_distance_threshold  = -outside_distance_threshold;
        }

        for ( int point_id = 0;
                  point_id < out_size;
                ++point_id )
        {
            Eigen::Vector3f point = cloud_ptr->at( in_indices_ptr ? in_indices_ptr->indices[point_id] : point_id ).getVector3fMap();
            bool keep = true;
            for ( int plane_id = 0; plane_id < _planes.size() && keep; ++plane_id )
            {
                if ( debug && plane_id > 0)
                std::cout << "plane_id: " << plane_id
                          << "\n\tnormals[plane_id]: " << normals[plane_id].transpose()
                          << "\n\t_planes[plane_id].D(): " << _planes[plane_id].D()
                          << "\n\tnormals[plane_id].dot(point) + _planes[plane_id].D(): " << normals[plane_id].dot(point) + _planes[plane_id].D()
                          << std::endl;
                keep &= (normals[plane_id].dot(point) + _planes[plane_id].D() > outside_distance_threshold );
            }
            if ( keep )
            {
                indices_ptr->indices.push_back( point_id );
            }
        }
    }
    int
    Cuboid::writeLabelsToFile( std::string dat_path,
                               int width ) const
    {
        FILE *fp = fopen( dat_path.c_str(), "ab" );
        if ( !fp )
        {
            printf("Cuboid::writeLabelsTofile failed when opening %s\n", dat_path.c_str() );
            return false;
        }

        // write point count
        int count = _indices_ptr->indices.size();
        fwrite( &count, sizeof(int), 1, fp );

        // generate
        int *coords = new int[count * 2];
        for ( unsigned pid = 0; pid < count; ++pid )
        {
            auto pnt = _cloud_ptr->at( _indices_ptr->indices[pid] );
            coords[pid*2  ] = pnt.label / width; // y
            coords[pid*2+1] = pnt.label % width; // x
        }

        // save
        fwrite( coords, sizeof(int), count*2, fp );

        // close
        fclose( fp );

        // cleanup
        if ( coords )
            delete [] coords;

        return EXIT_SUCCESS;
    }


    std::vector<double>
    Cuboid::getSideLengths() const
    {
        std::vector<double> lengths, out_lengths;

        int plane_id = 0;
        for ( auto const& plane : _planes.Primitives() )
        {
            if ( plane.Corners().size() != 4 )
            {
                std::cerr << __func__ << ": plane["<<plane_id<<"].Corners().size() != 4: " << plane.Corners().size() << std::endl;
            }

            std::vector<double> plane_lengths;
            for ( int corner_id = 1; corner_id < plane.Corners().size(); ++corner_id )
            {
                plane_lengths.push_back( (plane.Corners()[corner_id].getVector3fMap()-plane.Corners()[0].getVector3fMap()).norm() );
                //std::cout << "plane side length: " << plane_lengths.back() << std::endl;
            }

            // the longest length is the diagonal, skip that one
            std::sort( plane_lengths.begin(), plane_lengths.end() );
            lengths.insert( lengths.end(), plane_lengths.begin(), plane_lengths.begin()+2 );
            ++plane_id;
        }

        auto sorted_indices = am::util::algo::sort_indices( lengths );
        //am::util::printVector( sorted_indices, "sorted_indices" );
        // { 8, 2, 4, 10,}  { 0, 5, 11, 6,}  { 1, 3, 7, 9 }

        std::sort( lengths.begin(), lengths.end() );

        out_lengths.resize(3,0);
        for ( int i = 0; i < 3; ++i )
        {
            for ( int j = i*4; j < (i+1)*4; ++j )
                out_lengths[i] += lengths[j];
            out_lengths[i] /= 4.;
        }

        return out_lengths;
    }

    Eigen::Vector3f
    Cuboid::getExtent() const
    {
        Eigen::Vector3f extent ( Eigen::Vector3f::Zero() );

        // for all three frame vectors
        for ( int f = 0; f != 3; ++ f )
        {
            Eigen::Vector3f vec = _R.col( f ).normalized();
            // get greatest extent in this direction
            for ( int c = 0; c != _corners.size(); ++c )
            {
                extent(f) = std::max( extent(f), vec.dot(_corners[c].getVector3fMap()-_center) );
            }
        }

        return extent;
    }

    int
    Cuboid::rotateAroundFrameAxis( Eigen::Vector3f angles )
    {
        for ( int i = 0; i != 3; ++i )
        {
            if ( angles(i) == 0.f ) continue;

            Eigen::AngleAxisf rot( angles(i), _R.col(i).normalized() );

            for ( int corner_id = 0; corner_id < _corners.size(); ++corner_id )
            {
                _corners[ corner_id ] = am::util::pcl::asPointXYZ( rot * (_corners[corner_id].getVector3fMap() - _center) + _center );
            }

            _R.col( (i+1)%3 ) = rot * _R.col( (i+1)%3 );
            _R.col( (i+2)%3 ) = rot * _R.col( (i+2)%3 );
        }

        return EXIT_SUCCESS;
    }

    int
    Cuboid::moveAlongFrame( Eigen::Vector3f offset )
    {
        Eigen::Vector3f local_offset = _R * offset;

        for ( int corner_id = 0; corner_id < _corners.size(); ++corner_id )
        {
            _corners[ corner_id ] = am::util::pcl::asPointXYZ( _corners[corner_id].getVector3fMap() + local_offset );
        }
        _center += local_offset;

        return EXIT_SUCCESS;
    }

    int
    Cuboid::resizeAlongFrame( Eigen::Vector3f offset )
    {
//        std::cout << "offset: " << offset << std::endl;
        float sign = ( offset.dot(Eigen::Vector3f::Ones()) > 0.f ) ? 1.f : -1.f;

        Eigen::Vector3f local_offset = _R * offset;

        for ( int corner_id = 0; corner_id < _corners.size(); ++corner_id )
        {
            Eigen::Vector3f corner = _corners[corner_id].getVector3fMap();
            if ( (corner-_center).normalized().dot((local_offset).normalized()) > 0.f )
                _corners[ corner_id ] = am::util::pcl::asPointXYZ( _corners[corner_id].getVector3fMap() + local_offset * sign );
            else
                _corners[ corner_id ] = am::util::pcl::asPointXYZ( _corners[corner_id].getVector3fMap() - local_offset * sign );
        }

        return EXIT_SUCCESS;
    }

    int
    Cuboid::getMinMaxCorner( pcl::PointXYZ &mn, pcl::PointXYZ &mx )
    {
        mn = { FLT_MAX, FLT_MAX, FLT_MAX };
        mx = { -FLT_MAX, -FLT_MAX, -FLT_MAX };

        for ( pcl::PointXYZ const& corner : _corners )
        {
            if ( corner.data[0] != corner.x ) std::cerr << "data not ok..." << "corner.data[0]: " << corner.data[0] << " != corner.x: " << corner.x << std::endl;
            if ( corner.data[1] != corner.y ) std::cerr << "data not ok..." << "corner.data[1]: " << corner.data[1] << " != corner.y: " << corner.y << std::endl;
            if ( corner.data[2] != corner.z ) std::cerr << "data not ok..." << "corner.data[2]: " << corner.data[2] << " != corner.z: " << corner.z << std::endl;

            for ( int d = 0; d != 3; ++d )
            {
                std::cout << "checking " << d << " corner.data[d]: " << corner.data[d] << " vs " << mn.data[d] << std::endl;
                if ( corner.data[d] < mn.data[d] )  mn.data[d] = corner.data[d];
                if ( corner.data[d] > mx.data[d] )  mx.data[d] = corner.data[d];
            }
        }
        return EXIT_SUCCESS;
    }

    Eigen::Matrix<Cuboid::Scalar,3,1>
    Cuboid::getOutwardsNormalOfLargestPlane() const
    {
        Scalar  max_area        = Scalar(0);
        int     max_area_plane_id    = 0;

        const int planes_size = Planes().size();
        for ( int plane_id = 1; plane_id != planes_size; ++plane_id )
        {
            Scalar area = Planes().at( plane_id ).getArea();
            if ( area > max_area )
            {
                max_area             = area;
                max_area_plane_id    = plane_id;
            }
        }

        // cache selection
        Plane const& plane                       = Planes().at( max_area_plane_id );
        Eigen::Matrix<Scalar,3,1> plane_centroid = plane.getCentroidFromCorners();

        if ( (plane_centroid - Center()).dot( plane.Normal() ) < 0.f )
            return plane.Normal() * Scalar(-1);
        else
            return plane.Normal();
    }

#if USE_WM5
    int
    Cuboid::toBox3( std::shared_ptr<Wm5::Box3<float>>        &box
                    , am::Cuboid                       const &cub
                    , float                            const  offset )
    {

        Eigen::Matrix3f frame  = cub.R();                   // three orthogonal normals from three planes of the cuboid
        Eigen::Vector3f center = cub.Center();              // cuboid centroid
        auto n0 = frame.block<3,1>( 0, 0 ).normalized();    // make sure they are normalized
        auto n1 = frame.block<3,1>( 0, 1 ).normalized();
        auto n2 = frame.block<3,1>( 0, 2 ).normalized();

        // create Wm5::Box
        box = std::make_shared<Wm5::Box3<float>>( Wm5::Vector3<float>( center(0), center(1), center(2) ),               // center
                                                  Wm5::Vector3<float>(     n0(0),     n0(1),     n0(2) ),               // normalized frame
                                                  Wm5::Vector3<float>(     n1(0),     n1(1),     n1(2) ),
                                                  Wm5::Vector3<float>(     n2(0),     n2(1),     n2(2) ),
                                                  fabs((cub.Corners()[0].getVector3fMap()-center).dot( n0 ))+offset,    // extent = vector from centroid to first corner projected on one of the normals + size increment
                                                  fabs((cub.Corners()[0].getVector3fMap()-center).dot( n1 ))+offset,
                                                  fabs((cub.Corners()[0].getVector3fMap()-center).dot( n2 ))+offset  );

        return EXIT_SUCCESS;
    }

    bool
    Cuboid::intersect( am::Cuboid const& a, am::Cuboid const& b, float offset )
    {
        std::shared_ptr<BOX3_F> cell_box, proxy_box;
        Cuboid::toBox3( cell_box , a, offset );
        Cuboid::toBox3( proxy_box, b, offset );

        INTR_BOX3BOX3_F intr_box_box( *cell_box.get(), *proxy_box.get() );

        return intr_box_box.Test();

        if ( intr_box_box.Test() )   //intersect
        {
            std::cout << "intr_box_box.GetIntersectionType(): " << intr_box_box.GetIntersectionType() << std::endl;
        }
    }
#endif

    /// SET, GET
    // _center
    Eigen::Vector3f const&
    Cuboid::Center() const
    {
        return _center;
    }

    Eigen::Vector3f&
    Cuboid::CenterRef()
    {
        return _center;
    }

    // _R
    Eigen::Matrix3f const&
    Cuboid::R() const
    {
        return _R;
    }

    Eigen::Matrix3f &
    Cuboid::RRef()
    {
        return _R;
    }

    // _planes
    am::Planes const&
    Cuboid::Planes() const
    {
        return _planes;
    }

    am::Planes &
    Cuboid::PlanesRef()
    {
        return _planes;
    }

    // _corners
    std::vector<::pcl::PointXYZ> const&
    Cuboid::Corners() const
    {
        return _corners;
    }

    std::vector<::pcl::PointXYZ> &
    Cuboid::Corners()
    {
        return _corners;
    }

#if 0
    bool
    Cuboid::intersection( std::shared_ptr<am::Cuboid> const& a,
                          std::shared_ptr<am::Cuboid> const& b )
    {
        pcl::PointXYZ minA, maxA, minB, maxB;
        a->getMinMaxCorner( minA, maxA );
        b->getMinMaxCorner( minB, maxB );
        std::cout << "minA: " << minA << std::endl;
        std::cout << "maxA: " << maxA << std::endl;
        std::cout << "minB: " << minB << std::endl;
        std::cout << "maxB: " << maxB << std::endl;

        for ( size_t i = 0;  i != 3; ++i )
        {
            float a = std::max( minA.data[i], minB.data[i] );
            float b = std::min( maxA.data[i], maxB.data[i] );

            if ( a > b )    return false;
        }

        return true;
    }


    int
    Cuboid::from_vertex_list( Cuboid                 &cuboid,
                              Eigen::Matrix3Xf const &vertex_list )
    {
        //std::cout << "DepthProcessor::vertices2cuboid() starting...\n";

        // center
        cuboid.CenterRef() = vertex_list.rowwise().mean();
        Eigen::MatrixXf vertices = vertex_list.colwise() - cuboid.Center();

        // covariance
        Eigen::MatrixXf cov = vertices * vertices.adjoint();

        // eigen solve
        Eigen::SelfAdjointEigenSolver<Eigen::MatrixXf> eigen_solver( cov );
        // debug
        {
            //std::cout << "eigen_values:\n\t" <<  eigen_solver.eigenvalues().transpose() << std::endl;
            //std::cout << "eigen_vectors:\n " << eigen_solver.eigenvectors().rightCols(3) << std::endl;
        }

        Eigen::Matrix3f    eigen_vectors = eigen_solver.eigenvectors().rightCols(3);
        std::vector<float> eigen_values  = { eigen_solver.eigenvalues()(0),
                                             eigen_solver.eigenvalues()(1),
                                             eigen_solver.eigenvalues()(2) };

        std::vector<int> eigen_vector_ids = { 0, 1, 2 };
        am::util::algo::sort3_desc( eigen_values[0], eigen_values[1], eigen_values[2], &eigen_vector_ids );

        cuboid.RRef().col(0) = eigen_vectors.col( eigen_vector_ids[0] );
        cuboid.RRef().col(1) = eigen_vectors.col( eigen_vector_ids[1] );
        cuboid.RRef().col(2) = eigen_vectors.col( eigen_vector_ids[2] );

        cuboid.RRef().col(2) = cuboid.RRef().col(0).cross( cuboid.RRef().col(1) );

        return EXIT_SUCCESS;
    }

    int
    Cuboid::from_cloud_old( Cuboid                       &cuboid,
                        PCloud::ConstPtr             cloud_ptr,
                        pcl::PointIndices::Ptr const &indices_ptr )
    {

        PCloud::Ptr masked_cloud_ptr( new PCloud );
        pcl::ExtractIndices<MyPoint> extraction;
        {
            extraction.setInputCloud( cloud_ptr );
            extraction.setIndices   ( indices_ptr );
            extraction.setNegative  ( false );
            extraction.filter       ( *masked_cloud_ptr );
        }

        Eigen::Matrix3Xf vertices;
        DepthProcessor::cloud_2_vertex_list( vertices, masked_cloud_ptr );

        //Cuboid cuboid;
        Cuboid::from_vertex_list( cuboid, vertices );
        cuboid.CloudPtrRef()    = PCloud::ConstPtr( cloud_ptr );
        cuboid.IndicesPtrRef()  = indices_ptr;

        return EXIT_SUCCESS;
    }
#endif

} // ns am

// PreCompile
template int
am::Cuboid::fromCloudWithPCA( am::Cuboid &cuboid_out
                              , boost::shared_ptr<pcl::PointCloud<MyPoint> > cloud
                              , boost::shared_ptr<pcl::PointIndices > indices_arg
                              , int   const K
                              , float const radius
                              , int   const filter_nmeans
                              , float const filter_stddev
                              , float const density_cap_median_multiplier
                              , bool  const debug                        );
