#include "Cuboids.h"

#include "cuboid_approx/primitives/Contact.h"
#include "pcl/sample_consensus/sac_model_plane.h"
#include "CuboidUtil.h" // cuboidstowallspath
#include "cuboid_approx/io/walls_io.h" // writeWalls

namespace am
{
    int
    Cuboids::readFromFile( Cuboids                             &cuboids
                           , std::string                     const &path
                           , std::string                     const *inliers_path
                           , PCloud::ConstPtr                       cloud_ptr
                           , pcl::PointIndices::ConstPtr            indices_ptr
                           , std::vector<am::Contact>              *contacts
                           , std::vector<am::Primitive::Ptr>       *statics
                           , double                          const *plane_inlier_threshold
                           , bool verbose
                           )
    {
        // read cuboid corners
        int cuboid_count = 0;
        {
            //float *corners      = NULL;
            const std::vector< std::string                     > data_labels  = { "boxes", "floors", "walls", "ceilings" }; // boxes has to be first for statics to work
            const std::map   < std::string, int                > data_lengths = { {"boxes", 24}, {"floors", 4}, {"walls", 4}, {"ceilings", 4} };
            /* */ std::map   < std::string, std::vector<float> > data;
            /* */ int                                            read_count   = 0;

            Cuboids static_cuboids; // holds statics->cuboids until insertion

            // (0) open file
            FILE *fp = fopen( path.c_str(), "rb" );
            if ( fp == NULL )
            {
                printf( "Cuboids::readFromFile failed, when opening %s\n", path.c_str() );
                return EXIT_FAILURE;
            }

            // (1) read cuboids
            // read number of cuboids
            read_count = fread( &cuboid_count, sizeof(int), 1, fp );
            if ( (cuboid_count < 1) || (read_count != 1) )
            {
                std::cerr << "Cuboids::readFromfile failed, count < 1 || read_count != 1..." << cuboid_count << ", " << read_count << std::endl;
                if ( fp )  fclose( fp );

                return EXIT_FAILURE;
            }
            // allocate
            //corners = new float[ cuboid_count * 8 * 3 ];
            data["boxes"].resize( cuboid_count * data_lengths.at("boxes"), 0.f );

            // Youyi's cuboid format (.box containing [c s x y z] doubles)
            typedef double Scalar;
            if ( path.find(".box") != std::string::npos )
            {
                Scalar *box_data = new Scalar[ cuboid_count * 15 ]; // center (3), frame_v0 (3), frame_v1(3), frame_v2(3), extent(3)
                read_count = fread( box_data, sizeof(Scalar), cuboid_count * 15, fp );
                if ( read_count != cuboid_count * 15 )
                {
                    std::cerr << "[" << __func__ << "]: " << "wanted to read " << cuboid_count * 15 << " doubles, but found only " << read_count << std::endl;
                    return EXIT_FAILURE;
                }

                int data_offset = 0, corners_offset = 0;
                for ( size_t cuboid_id = 0; cuboid_id != cuboid_count; ++cuboid_id )
                {
                    Eigen::Vector3d center;                                      for ( int d = 0; d != 3; ++d ) center( d      ) = box_data[data_offset++];
                    Eigen::Vector3d extent;                                      for ( int d = 0; d != 3; ++d ) extent( d      ) = box_data[data_offset++];
                    Eigen::Matrix3d frame;  for ( int vid = 0; vid != 3; ++vid ) for ( int d = 0; d != 3; ++d ) frame ( d, vid ) = box_data[data_offset++]; // vectors in columns

                    for ( int x = 0; x != 2; ++x )
                        for ( int y = 0; y != 2; ++y )
                            for ( int z2 = 0; z2 != 2; ++z2 )
                            {
                                int z = (y + z2) % 2;
                                Eigen::Vector3d corner = center
                                                         + frame.col(0) * extent(0) * (x ? 1.f : -1.f)
                                                         + frame.col(1) * extent(1) * (y ? 1.f : -1.f)
                                                         + frame.col(2) * extent(2) * (z ? 1.f : -1.f);

                                data["boxes"][ corners_offset++ ] = corner(0);
                                data["boxes"][ corners_offset++ ] = corner(1);
                                data["boxes"][ corners_offset++ ] = corner(2);
                            }
                }

                if ( box_data ) { delete[] box_data; box_data = NULL; }
            }
            else if ( path.find(".gt") != std::string::npos )
            {
                std::map< std::string, int*                       > primitives_ids; // primitives_ids[ prim_type ][ id ] == matlab_id
                std::map< int        , std::pair<std::string,int> > ids_primitives; // ids_primitives[ matlab_id ]       == { prim_type, C++_id }

                // (2) read floors, walls, ceilings
                for ( int ent_id = 0; ent_id != data_labels.size(); ++ent_id )
                {
                    std::string const& label = data_labels[ ent_id ];
                    if ( verbose ) std::cout << "label is \"" << label << "\"" << std::endl;
                    int plane_count = 0,
                        full_n      = 0;

                    // (1) read data
                    if ( ent_id ) // not boxes
                    {
                        // read N
                        read_count = fread( &plane_count, sizeof(int), 1, fp );
                        if ( read_count != 1 )
                        {
                            std::cerr << "[" << __func__ << "]: " << " read " << label << "_count could not read " << 1 << " " << label << "_count\n";
                            return EXIT_FAILURE;
                        }

                        // planes: | nx, ny, nz, d | = 4
                        full_n = plane_count * data_lengths.at( label );

                        // allocate
                        //data[ label ] = new float[ full_n ];
                        data[ label ].resize( full_n );
                    }
                    else // boxes -> data already read, only reading ids
                    {
                        plane_count = cuboid_count;
                        full_n      = plane_count * data_lengths.at( "boxes" );
                    }

                    // read
                    read_count = fread( data[label].data(), sizeof(float), full_n, fp );
                    if ( read_count != full_n )
                    {
                        std::cerr << "[" << __func__ << "]: " << " read " << label << " could not read " << full_n << " " << label << " coordinates\n";
                        return EXIT_FAILURE;
                    }

                    // (2) read ids
                    // allocate ids
                    primitives_ids[ label ] = new int[ plane_count ];
                    // read ids
                    read_count = fread( primitives_ids[label], sizeof(int), plane_count, fp );
                    if ( read_count != plane_count )
                    {
                        std::cerr << "[" << __func__ << "]: " << " read " << label << " could not read " << plane_count << " ids\n";
                        return EXIT_FAILURE;
                    }
                    else
                        std::cout<<"primitives_ids["<<label<<"]:";for(size_t vi=0;vi!=plane_count;++vi)std::cout<<primitives_ids[label][vi]<<" ";std::cout << "\n";

                    // create reverse map
                    for ( int id = 0; id != plane_count; ++id )
                    {
                        if ( verbose )
                            std::cout << "adding ids_primitives[" << primitives_ids[label][id] << "] = \"" << label << "\"," << id << std::endl;
                        ids_primitives[ /* matlab_id: */ primitives_ids[label][id] ] = std::pair<std::string,int>( label, /* C++_id: */ id );
                    }
                }

                std::vector<int> counts = { int(data["boxes"   ].size()) / data_lengths.at( "boxes"    ),
                                            int(data["floors"  ].size()) / data_lengths.at( "floors"   ),
                                            int(data["walls"   ].size()) / data_lengths.at( "walls"    ),
                                            int(data["ceilings"].size()) / data_lengths.at( "ceilings" ) };

                // (3) read relationships
                std::vector<int> parentIDs;
                {
                    int parent_count = 0;
                    // read count
                    read_count = fread( &parent_count, sizeof(int), 1, fp );
                    if ( read_count != 1 )
                    {
                        std::cerr << "[" << __func__ << "]: " << " read " << "parent" << "_count could not read " << 1 << " " << "parent" << "_count\n";
                        return EXIT_FAILURE;
                    }

                    // allocate ids
                    //parentIDs = new int[ parent_count ];
                    parentIDs.resize( parent_count );
                    std::cout << "parent_count " << parent_count << " vs. "
                              << (   data["boxes"   ].size() / data_lengths.at( "boxes"  )
                                   + data["floors"  ].size() / data_lengths.at("floors"  )
                                   + data["walls"   ].size() / data_lengths.at("walls"   )
                                   + data["ceilings"].size() / data_lengths.at("ceilings") ) << " cuboids.size() + statics->size()\n";

                    // read ids
                    read_count = fread( parentIDs.data(), sizeof(int), parent_count, fp );
                    if ( read_count != parent_count )
                    {
                        std::cerr << "[" << __func__ << "]: " << " read " << "parent_ids" << " could not read " << parent_count << " ids\n";
                        return EXIT_FAILURE;
                    }
                    else
                        std::cout<<"parentIDs:";for(size_t vi=0;vi!=parentIDs.size();++vi)std::cout<<parentIDs[vi]<<" ";std::cout << "\n";
                }

                //if ( statics )
                {
                    for ( int label_id = 1; label_id != data_labels.size(); ++label_id ) // start from non-boxes
                    {
                        std::string                  const& label = data_labels[ label_id ];
                        pcl::ModelCoefficients::Ptr         coeffs( new pcl::ModelCoefficients() );
                        coeffs->values.resize( data_lengths.at(label) );

                        // every primitive
                        for ( int prim_id = 0; prim_id != data[label].size(); )
                        {
                            for ( size_t dim = 0; dim != coeffs->values.size(); ++dim, ++prim_id )
                            {
                                coeffs->values[ dim ] = data[label][prim_id];
                            }

                            Primitive::Ptr primitive( new Plane(coeffs,cloud_ptr) );
                            Plane::Ptr     plane = std::static_pointer_cast<Plane>( primitive );
#if 0
                            if ( plane_inlier_threshold )
                                plane->getInlierIndices( plane->IndicesPtrRef()
                                                         , plane->CloudPtr()
                                                         , indices_ptr
                                                         , *plane_inlier_threshold );
                            else
                                plane->getInlierIndices( plane->IndicesPtrRef()
                                                         , plane->CloudPtr()
                                                         , indices_ptr
                                                         );
                            std::cout << "inlier count: " << plane->IndicesPtr()->indices.size() << std::endl;
#else
                            plane->IndicesPtrRef().reset( new pcl::PointIndices() );
#endif

//                            if ( plane->CloudPtr() )
//                            {
//                                auto v = am::CloudViewer::showCloud( plane->CloudPtr(), "planecloud", nullptr, 2., false );
//                                v->VP()->addPlane( *coeffs, "plane" );
//                                v->VP()->spin();
//                            }

                            if ( statics )
                                statics->push_back( primitive );

                            Cuboid cuboid;
#                           warning "TODO: disabled option - enable after PCA fixed"
                            if ( false && plane->IndicesPtr() && plane->IndicesPtr()->indices.size() )
                            {
                                plane->Corners().clear(); // is inited to dummies, and the next function checks, if calc is necessary.
                                // TODO: calculate frame used by "from1Plane"
                                Cuboid::from1Plane( cuboid, *plane, .05f );
//                                auto it = static_cuboids.back().Corners().begin();
//                                auto end_it = static_cuboids.back().Corners().end();
//                                for ( ; it != end_it; ++it )
//                                    std::cout << "corner: " << it->x << ", " << it->y << ", " << it->z << std::endl;
                            }
                            else
                            {
                                PCloud::ConstPtr the_cloud( nullptr );
                                pcl::PointIndices::ConstPtr the_indices( NULL );
                                if ( cloud_ptr )
                                {
                                    the_cloud   = cloud_ptr;
                                    the_indices = indices_ptr;
                                }
                                else
                                {
                                    PCloud::Ptr tmp_cloud( new PCloud() );

                                    // use cuboid corners as "cloud"
//                                    for ( int cuboid_id = 0; cuboid_id != cuboids.size(); ++cuboid_id )
//                                    {
//                                        Cuboid const& tmp_cuboid = cuboids.at( cuboid_id );
//                                        for ( int corner_id = 0; corner_id != cuboid.Corners().size(); ++corner_id )
//                                        {
                                    for ( int corner_id = 0; corner_id < data["boxes"].size(); corner_id += 3 )
                                    {
                                            MyPoint pnt;
                                            pnt.x = data["boxes"][ corner_id ];
                                            pnt.y = data["boxes"][ corner_id + 1 ];
                                            pnt.z = data["boxes"][ corner_id + 2 ];
                                            tmp_cloud->push_back( pnt );
                                    }
//                                        }
//                                    }
                                    the_cloud = boost::const_pointer_cast<const PCloud>( tmp_cloud );
                                    //std::cout << "ok, cloud->size(): " << the_cloud->size() << std::endl; fflush(stdout);
                                }

                                // project whole cloud to frame, and create a too big plane
                                PCloud::Ptr     projected_cloud( new PCloud() );
                                DepthProcessor::PCA_withPlane( plane->Frame(), plane->Coeffs(), the_cloud, the_indices, false, &projected_cloud );
                                //std::cout << " frame: " << plane->Frame() << std::endl;
                                MyPoint         min_pt, max_pt;
                                DepthProcessor::calculateExtremaInFrame( min_pt, max_pt, plane->Frame(), projected_cloud, NULL );

                                std::vector<Eigen::Vector4f> local_extrema( 4 );
                                const float z = (min_pt.z + max_pt.z)/2.f;
                                local_extrema[0] << min_pt.x, min_pt.y, z, 1;
                                local_extrema[1] << min_pt.x, max_pt.y, z, 1;
                                local_extrema[2] << max_pt.x, max_pt.y, z, 1;
                                local_extrema[3] << max_pt.x, min_pt.y, z, 1;
                                for ( int i = 0; i != local_extrema.size(); ++i )
                                {
                                    pcl::PointXYZ pnt;
                                    Eigen::Vector4f pnt_eigen = plane->Frame() * local_extrema[i];
                                    pnt.x = pnt_eigen(0); pnt.y = pnt_eigen(1); pnt.z = pnt_eigen(2);
                                    plane->Corners()[i] = pnt;
                                    //std::cout << "corner: " << pnt.getVector3fMap().transpose() << std::endl;
                                } // ... for local_extrema

                                bool debug2 = false;
                                if ( debug2 )
                                {
                                    pcl::visualization::PCLVisualizer::Ptr vptr( new pcl::visualization::PCLVisualizer("plane") );
                                    vptr->setBackgroundColor( .5, .6, .6 );
                                    CloudViewer::addPointCloud( vptr, projected_cloud, "proj_cloud" );
                                    plane->draw( vptr, "plane" );
                                    vptr->spin();
                                }
                            } // ... else plane->indices

                            Cuboid::from1Plane( cuboid, *plane, .05f );
                            static_cuboids.add( cuboid );

                        } // ... every primitive
                    }
                }

                // (4) relationships -> contacts
                if ( parentIDs.size() && contacts )
                {
                    contacts->reserve( parentIDs.size() );

                    int accum_c_id = 0;
                    for ( int label_id = 0; label_id != data_labels.size(); ++label_id )
                    {
                        std::string const& label = data_labels[ label_id ];
                        int         const  count = data[ label ].size() / data_lengths.at( label );

                        for ( int c_id = 0; c_id != count; ++c_id, ++accum_c_id )
                        {
                            if ( parentIDs.size() > c_id )
                            {
                                int matlab_id        = primitives_ids[ label ][ c_id ];
                                int parent_matlab_id = parentIDs[ matlab_id - 1 ];

                                std::cout << label.substr(0 ,label.size()-2) << c_id
                                          << " has matlab_id: " << matlab_id
                                          << ", parent_matlab_id: " << parent_matlab_id;
                                if ( parent_matlab_id )
                                    std::cout << ", parent == data[\"" << ids_primitives[ parent_matlab_id ].first << "\"][" << ids_primitives[ parent_matlab_id ].second << "]";
                                std::cout << std::endl;

                                // if has parent
                                if ( parent_matlab_id )
                                {
                                    int offset = 0;
                                    int ent_index = std::distance( data_labels.begin(), std::find( data_labels.begin(), data_labels.end(), ids_primitives[parent_matlab_id].first) );
                                    std::cout << "[reindex]: " << "type: " << ids_primitives[parent_matlab_id].first << " -> ent_id: " << ent_index << std::endl;
                                    for ( int ent_id = 0; ent_id != ent_index; ++ent_id )
                                    {
                                        std::string label = data_labels.at( ent_id );
                                        std::cout << "increasing offset " << offset;
                                        offset += data.at( label ).size() / data_lengths.at( label );
                                        std::cout << " by " << data[ label ].size() / data_lengths.at( label )
                                                  << " for " << data_labels.at( ent_id ) << std::endl;
                                    }
                                    int parent_id = ids_primitives[parent_matlab_id].second + offset;

                                    Contact::TYPE contact_type = Contact::TOUCHING;
                                    if (    (ids_primitives[ parent_matlab_id ].first.find("walls"   )  != std::string::npos)
                                         || (ids_primitives[ parent_matlab_id ].first.find("ceilings")  != std::string::npos)
                                         || (label                                   .find("walls")     != std::string::npos)
                                         || (label                                   .find("ceilings")  != std::string::npos) )
                                        contact_type = Contact::FIXED;

                                    contacts->push_back( Contact( {  std::min(accum_c_id,parent_id)
                                                                   , std::max(accum_c_id,parent_id)}
                                                                   , contact_type                   )
                                                         );
                                    std::cout << "added contact " << accum_c_id << " -> " << parent_id << std::endl;
                                }
                            }
                            else
                                std::cerr << "skipping cub_id " << c_id << std::endl;
                        }
                    }
                    // sort contacts
                    std::sort( contacts->begin(), contacts->end() );
                }

                // walls
                {
                    std::string walls_path = am::util::cuboidPathToWallsPath( path );
                    if ( !boost::filesystem::exists(walls_path) )
                    {
                        am::Cuboid::TypeMap walls;
                        int cuboid_id = counts[0]; // start at first non-box index
                        // for every non-box label
                        for ( int ent_id = 1; ent_id != 4; ++ent_id )
                            // add every cuboid
                            for ( int i = 0; i != counts[ent_id]; ++i, ++cuboid_id )
                                walls[ cuboid_id ] = ((ent_id == 1) ? am::Cuboid::FLOOR : am::Cuboid::WALL);

                        // write to disk
                        ca::io::writeWalls( walls, walls_path, /* cuboid_count: */ std::accumulate(counts.begin(),counts.end(),0) );
                    }
                    else
                    {
                        std::cout << "[" << __func__ << "]: " << "not saving walls, since they already exist" << std::endl;
                    }

                } // ... walls

                // (5) cleanup
                for ( auto it = primitives_ids.begin(); it != primitives_ids.end(); ++it )
                {
                    if ( it->second )
                    {
                        delete [] it->second;
                        it->second = NULL;
                    }
                }
            }
            else if ( path.find(".dat") != std::string::npos ) // Tianjia's format (.dat containing corner floats)
            {
                read_count = fread( data["boxes"].data(), sizeof(float), cuboid_count * 8 * 3, fp );
                if ( read_count != cuboid_count * 8 * 3 )
                {
                    std::cerr << "[" << __func__ << "]: " << " read corners could not read " << cuboid_count * 8 * 3 << " corners\n";
                    return EXIT_FAILURE;
                }
            }
            else
            {
                std::cerr << "[" << __func__ << "]: " << "could not recognize format of extension " << path << std::endl;
                return EXIT_FAILURE;
            }

            // convert corners to cuboids
            am::Cuboid cuboid;
            for ( int cuboid_id = 0; cuboid_id < cuboid_count; ++cuboid_id )
            {
                //std::cout << "[" << __func__ << "]: " << "cuboid" << cuboid_id << std::endl;
                std::vector<float> corner_coords( 24 );
                corner_coords.assign( data["boxes"].begin() +  cuboid_id   * 24,
                                      data["boxes"].begin() + (cuboid_id+1)* 24 );

//                for ( int i = 0; i != corner_coords.size(); ++i )
//                {
//                    if ( (i > 0) && !(i % 3) )
//                        std::cout << std::endl;
//                    std::cout << corner_coords[i] << ", (" << data["boxes"][cuboid_id*24+i] << ")" << ", ";
//                }
//                std::cout << std::endl;

                am::Cuboid::fromCorners( cuboid,
                                         corner_coords,
                                         cloud_ptr,
                                         indices_ptr ? indices_ptr : nullptr );
                cuboids.add( cuboid );
            }
            // add statics
            for ( size_t cuboid_id = 0; cuboid_id != static_cuboids.size(); ++cuboid_id )
            {
                cuboids.add( static_cuboids.at(cuboid_id) );
            }

            // cleanup
            if ( fp      )  fclose( fp );
            //if ( corners )  delete [] corners;
        } // ...read cuboids

        // read inliers
        if ( inliers_path && !inliers_path->empty() )
        {
            // open file
            FILE *fp = fopen( inliers_path->c_str(), "rb" );
            if ( !fp )
            {
                printf( "Cuboids::readFromFile failed, when opening %s\n", inliers_path->c_str() );
                return EXIT_FAILURE;
            }

            // read cuboids count
            int inliers_cuboid_count;
            int read_count = fread( &inliers_cuboid_count, sizeof(int), 1, fp );
            if ( read_count != 1 ) std::cerr << "[" << __func__ << "]: " << "read " << read_count << " instead of " << 1 << std::endl;

            if (    (inliers_cuboid_count <  1           )
                 || (inliers_cuboid_count != cuboid_count) )
            {
                std::cerr << "Cuboids::readFromfile failed reading inliers, count < 1 || count != cuboid_count..." << cuboid_count << std::endl;
                if ( fp ) fclose( fp );
                return EXIT_FAILURE;
            }
            // read dimensions
            int height, width;
            read_count = fread( &height, sizeof(int), 1, fp );
            if ( read_count != 1 ) std::cerr << "[" << __func__ << "]: " << "read " << read_count << " instead of " << 1 << std::endl;
            read_count = fread( &width , sizeof(int), 1, fp );
            if ( read_count != 1 ) std::cerr << "[" << __func__ << "]: " << "read " << read_count << " instead of " << 1 << std::endl;
            if ( (height < 1) || (width < 1) )
            {
                std::cerr << "cuboids::readfromfile failed reading inliers, since width < 1 or height < 1..." << width << ", " << height << std::endl;
                if ( fp ) fclose( fp );
                return EXIT_FAILURE;
            }

            // for every cuboid
            for ( int cid = 0; cid < inliers_cuboid_count; ++cid )
            {
                // inliers_count
                int inliers_count = 0;
                read_count = fread( &inliers_count, sizeof(int), 1, fp );
                if ( read_count != 1 ) std::cerr << "[" << __func__ << "]: " << "read " << read_count << " instead of " << 1 << std::endl;

                if ( inliers_count < 1 )
                {
                    std::cerr << "no inilers for cuboid " << cid << ", skipping..." << std::endl;
                    continue;
                }

                // read inliers
                int *data = new int[ inliers_count * 2 ];
                read_count = fread( data, sizeof(int), inliers_count * 2, fp );
                if ( read_count != inliers_count * 2 ) std::cerr << "[" << __func__ << "]: " << "read " << read_count << " instead of " << inliers_count * 2 << std::endl;

                // debug
                cv::Mat img = cv::Mat::zeros( height, width, CV_8UC3 );

                // for every inlier point
                am::Cuboid& cuboid = cuboids.at( cid );
                cuboid.IndicesPtrRef()->indices.resize( inliers_count );
                int prev_id = 0;
                for ( int pid = 0; pid < inliers_count; ++pid )
                {
                    int y = data[pid*2  ];
                    int x = data[pid*2+1];
                    const int label = y * width + x;

                    bool found = false;
                    // look forwards from previous
                    auto it = std::find_if( cloud_ptr->begin()+prev_id,
                                            cloud_ptr->end(),
                                            [&label] (MyPoint pnt) { return pnt.label == label; } );
                    // if found
                    if ( it != cloud_ptr->end() )
                    {
                        found = true;
                    }
                    else
                    {
                        // otherwise, look from beginning
                        std::cerr << "restarting search" << std::endl;
                        it = std::find_if( cloud_ptr->begin(),
                                           cloud_ptr->begin()+prev_id+1,
                                           [&label] (MyPoint pnt) { return pnt.label == label; } );
                        // if found
                        found |= ( it != cloud_ptr->begin()+prev_id+1 );
                    }

                    if ( !found )
                    {
                        std::cerr << "point " << pid
                                  << " with label " << label << "(" << y << "," << x << ") not found in pointcloud!"
                                  << std::endl;
                        continue;
                    }
                    cuboid.IndicesPtrRef()->indices[pid] = std::distance( cloud_ptr->begin(), it );

                    prev_id = cuboid.IndicesPtrRef()->indices[pid];

                    // debug
                    img.at<uchar>( y, x*3) = it->b;
                    img.at<uchar>( y, x*3+1) = it->g;
                    img.at<uchar>( y, x*3+2) = it->r;
                } // for every inlier point

                if ( data ) delete [] data;
            } // for every cuboid

            if ( fp )   fclose( fp );
        } // ... read inliers

        return EXIT_SUCCESS;
    }

    int
    Cuboids::writeToFile( Cuboids const& cuboids, std::string path, int height, int width )
    {
        // corners
        std::ofstream myfile;
        myfile.open ( path, std::ios_base::binary | std::ios_base::out | std::ios_base::trunc );
        if ( !myfile.is_open() )
        {
            std::cerr << "could not open file " << path << std::endl;
            return EXIT_FAILURE;
        }
        std::cout << "writing cuboids to " << path << std::endl;

        int cuboids_count = cuboids.size();
        myfile.write( (char*) &cuboids_count, sizeof(int) );
        myfile.close();

        for ( am::Cuboid const& cuboid : cuboids.Primitives() )
        {
            am::Cuboid::writeToFile( cuboid, path,
                                     /* append: */ true );
        }

        return writeLabels( cuboids, path, height, width );

        return EXIT_SUCCESS;
    }

    int
    Cuboids::writeToTextFile( Cuboids const& cuboids, std::string path, int height, int width )
    {
        // corners
        std::ofstream myfile;
        myfile.open ( path, std::ios_base::out | std::ios_base::trunc );
        if ( !myfile.is_open() )
        {
            std::cerr << "could not open file " << path << std::endl;
            return EXIT_FAILURE;
        }
        std::cout << "writing cuboids to " << path << std::endl;

        int cuboids_count = cuboids.size();
        //myfile.write( (char*) &cuboids_count, sizeof(int) );
        myfile << cuboids_count << std::endl;
        myfile.close();

        for ( am::Cuboid const& cuboid : cuboids.Primitives() )
        {
            am::Cuboid::writeToTextFile( cuboid, path,
                                         /* append: */ true );
        }

        return writeLabels( cuboids, path, height, width );
    }

    int
    Cuboids::writeLabels( Cuboids const& cuboids, std::string path, int height, int width )
    {
        const int cuboids_count = cuboids.size();
        if (    (height > 0)
             && (width  > 0)
             && (cuboids[0].CloudPtr()  ) && (cuboids[0].CloudPtr()  ->size()        )
             && (cuboids[0].IndicesPtr()) && (cuboids[0].IndicesPtr()->indices.size()) )
        {

            // labels
            std::string labels_path = Cuboids::getLabelsPath( path );
            std::cout << "labels_path: " << labels_path << std::endl;
            FILE *fp = fopen( labels_path.c_str(), "wb" );
            if ( !fp )
            {
                printf("Cuboids::writeLabelsTofile failed when opening %s\n", labels_path.c_str() );
                return EXIT_FAILURE;
            }

            // write cuboid count and dimensions
            fwrite( &cuboids_count, sizeof(int), 1, fp );
            fwrite( &height       , sizeof(int), 1, fp );
            fwrite( &width        , sizeof(int), 1, fp );
            fclose( fp );

            // write labels
            for ( am::Cuboid const& cuboid : cuboids.Primitives() )
            {
                if ( cuboid.CloudPtr() )
                    cuboid.writeLabelsToFile( labels_path, width );
            }
        }
        else
            std::cout << "[" << __func__ << "]: " << "Not writing labels to cuboids" << std::endl;

        return EXIT_SUCCESS;
    }

    /*int
    Cuboids::loadFromFile( )*/

    std::string
    Cuboids::getLabelsPath( std::string path )
    {
        return path.substr( 0, path.size()-4 ) + "_points.dat";
    }

    int
    Cuboids::testReadWrite()
    {
        Cuboids cuboids;
        am::Cuboid cuboid;
        //cuboid.Corners().resize(8);
        cuboid.Corners() = {{ 0.240607,-0.211485,2.41565},
                           { -0.00153208,-0.340846,2.7239},
                           { -0.0486627,0.498744,3.03922},
                           { 0.193476,0.628105,2.73098},
                           { -0.643829,-0.0289862,1.79753},
                           { -0.885968,-0.158347,2.10578},
                           { -0.933098,0.681243,2.4211},
                           { -0.69096,0.810604,2.11285} };
        cuboids.add( cuboid );
        cuboid.Corners() = {{ 0.240607,-0.211485,2.41565},
                           { -1.00153208,-0.340846,2.7239},
                           { -0.0486627,0.498744,3.03922},
                           { 0.193476,3.628105,2.73098},
                           { -0.643829,-0.0289462,1.79753},
                           { -0.885968,-0.158347,2.10578},
                           { -0.932098,0.681543,2.4211},
                           { -0.69196,0.810604,0.11285} };
        cuboids.add( cuboid );

        Cuboids::writeToFile( cuboids, "cuboid_test.dat", 480, 640 );
        Cuboids read_cuboids;
        Cuboids::readFromFile( read_cuboids, "cuboid_test.dat" );

        float diff = 0;
        for ( int cuboid_id = 0; cuboid_id < cuboids.size(); ++cuboid_id )
        {
            for ( int corner_id = 0; corner_id < 8; ++corner_id )
            {
                diff += fabs(read_cuboids.at(cuboid_id).Corners()[corner_id].x - cuboids.at(cuboid_id).Corners()[corner_id].x);
                diff += fabs(read_cuboids.at(cuboid_id).Corners()[corner_id].y - cuboids.at(cuboid_id).Corners()[corner_id].y);
                diff += fabs(read_cuboids.at(cuboid_id).Corners()[corner_id].z - cuboids.at(cuboid_id).Corners()[corner_id].z);
            }
        }
        if ( diff > FLT_EPSILON )
        {
            std::cerr << "Cuboids::testReadWrite failed!" << std::endl;
            for ( int cuboid_id = 0; cuboid_id < cuboids.size(); ++cuboid_id )
            {
                for ( int corner_id = 0; corner_id < 8; ++corner_id )
                {
                    std::cerr << read_cuboids.at(cuboid_id).Corners()[corner_id].x << " !=? " << cuboids.at(cuboid_id).Corners()[corner_id].x << std::endl;
                    std::cerr << read_cuboids.at(cuboid_id).Corners()[corner_id].y << " !=? " << cuboids.at(cuboid_id).Corners()[corner_id].y << std::endl;
                    std::cerr << read_cuboids.at(cuboid_id).Corners()[corner_id].z << " !=? " << cuboids.at(cuboid_id).Corners()[corner_id].z << std::endl;
                }
            }
        }
        else
        {
            std::cout << "Cuboids::testReadWrite test PASSED...\n" << std::endl;
        }

        return diff >= FLT_EPSILON;
    }

} // ns am
