#include "Plane.h"

#include "CloudViewer.h"

#include "cuboid_approx/util/AMUtil2.h"

#include <pcl/common/transforms.h>          // transformPointCloud
#include "pcl/common/common.h"              // getMinMax3D
#include "pcl/filters/project_inliers.h"    // projectInliers
#include "pcl/common/centroid.h"            // compute3DCentroid
#include "pcl/segmentation/sac_segmentation.h" // sacsegmentation
#include "pcl/filters/filter_indices.h"
#include "pcl/correspondence.h"
#include "Segmentation.h"

#include "DepthProcessor.h"
#include "Cuboid.h"

namespace am
{
    Plane::Plane()
        : Primitive()
        , _coeffs_ptr( new pcl::ModelCoefficients() )
        , _frame( Eigen::Matrix4f::Identity() )
    {
        _coeffs_ptr->values.resize(4,0.f);
        _corners = { pcl::PointXYZ(0.f,0.f,0.f)
                     , pcl::PointXYZ(0.f,1.f,0.f)
                     , pcl::PointXYZ(1.f,1.f,0.f)
                     , pcl::PointXYZ(1.f,0.f,0.f) };
    }

    Plane::Plane( pcl::ModelCoefficients::Ptr coeffs, PCloud::ConstPtr cloud_ptr )
        : Primitive( cloud_ptr )
        , _coeffs_ptr( coeffs )
        , _frame( Eigen::Matrix4f::Identity() )
    {
        _corners = { pcl::PointXYZ(0.f,0.f,0.f)
                     , pcl::PointXYZ(0.f,1.f,0.f)
                     , pcl::PointXYZ(1.f,1.f,0.f)
                     , pcl::PointXYZ(1.f,0.f,0.f) };
    }

    Plane& Plane::operator=( Plane const& other )
    {
        if ( this != &other )
        {
            this->_coeffs_ptr    = pcl::ModelCoefficients::Ptr( new pcl::ModelCoefficients() );
            *(this->_coeffs_ptr) = *(other._coeffs_ptr);
            _corners             = other._corners;
            _frame               = other._frame;
        }
        return *this;
    }

    Plane::Plane( Plane const& theOther )
        : Primitive( theOther )
    {
        *this = theOther;
    }

    // outputs a plane containing the original cloud, and the indices changed to plane inliers
    int
    Plane::fromCloud( Plane                        &plane,
                      PCloud::ConstPtr             cloud_ptr,
                      pcl::PointIndices::Ptr const indices_ptr,
                      double                       distanceThreshold )
    {
        int err = EXIT_SUCCESS;

        // copy big cloud reference
        plane.CloudPtrRef()   = cloud_ptr;

        err = Segmentation::extract_plane<MyPoint>( /*        out_coeffs: */ plane.CoeffsPtrRef(),
                                                    /*          in_cloud: */ cloud_ptr,
                                                    /*        in_indices: */ indices_ptr,
                                                    /* distanceThreshold: */ distanceThreshold );
        if ( err != EXIT_SUCCESS )
        {
            std::cerr << "Segmentation::extract_plane() returned error with " << (int)err << "\n";
            return err;
        }

        plane.getInlierIndices( /*     out indices: */ plane.IndicesPtrRef(),
                                /*        in cloud: */ cloud_ptr,
                                /*      in indices: */ indices_ptr,
                                /* distance Thresh: */ distanceThreshold );

        // set frame
        {
            am::PCloud::Ptr _2d_cloud_ptr;
            am::DepthProcessor::projectPlaneInliers( _2d_cloud_ptr,
                                                     plane.CloudPtr(),
                                                     plane.IndicesPtr(),
                                                     plane.CoeffsPtr() );
            am::DepthProcessor::PCA( plane.Frame(),
                                     _2d_cloud_ptr,
                                     nullptr );
            std::cerr << "plane.Frame().block<3,1>(0,2).dot( plane.Normal() ): " << plane.Frame().block<3,1>(0,2).dot( plane.Normal() ) << std::endl;
            if ( plane.Frame().block<3,1>(0,2).dot( plane.Normal() ) < 0.f )
            {
                std::cerr << " flipping frame " << std::endl;
                plane.Frame().block<3,1>(0,0) *= -1.f;
                plane.Frame().block<3,1>(0,2) *= -1.f;
            }
        }

        return err;
    }

    int
    Plane::getInlierIndices( pcl::PointIndices::Ptr       &inliers,
                             PCloud::ConstPtr              cloud_ptr,
                             pcl::PointIndices::ConstPtr   indices_ptr,
                             double                        threshold   ) const
    {
        if ( !cloud_ptr )
            return EXIT_FAILURE;

        if ( !inliers ) inliers = pcl::PointIndices::Ptr( new pcl::PointIndices );

        pcl::SampleConsensusModelPlane<MyPoint> sacplane( cloud_ptr );
        if ( indices_ptr )
            sacplane.setIndices( indices_ptr->indices );
        sacplane.selectWithinDistance( this->Coeffs(), threshold, inliers->indices );

        return EXIT_SUCCESS;
    }

    int
    Plane::getOutlierIndices( pcl::PointIndices::Ptr &outlier_indices, PCloud::ConstPtr cloud_ptr, double threshold ) const
    {
        if ( !outlier_indices ) outlier_indices = pcl::PointIndices::Ptr( new pcl::PointIndices );

        pcl::SampleConsensusModelPlane<MyPoint> sacplane( cloud_ptr );
        std::vector<double> distances;
        sacplane.getDistancesToModel( this->Coeffs(), distances );

        for ( int point_id = 0; point_id < distances.size(); ++point_id )
        {
            if ( distances[point_id] > threshold )
            {
                outlier_indices->indices.push_back( point_id );
            }
        }

        return EXIT_SUCCESS;
    }

    int
    Plane::getInlierPoints( PCloud::Ptr &inlier_points, PCloud::ConstPtr cloud_ptr, double threshold ) const
    {
        int err = EXIT_SUCCESS;

        // allocate, if necessary
        if ( !inlier_points ) inlier_points = PCloud::Ptr( new PCloud );

        // fetch indices
        pcl::PointIndices::Ptr inlier_indices;
        err = getInlierIndices( inlier_indices,
                                cloud_ptr,
                                /* indices: */ nullptr,
                                threshold );
        if ( err != EXIT_SUCCESS )
        {
            std::cerr << "Plane::getInlierPoints(): getInlierIndices returned error " << (int)err << std::endl;
            return err;
        }

        // extract points
        pcl::ExtractIndices<MyPoint> extraction;
        extraction.setInputCloud( cloud_ptr );
        extraction.setIndices   ( inlier_indices );
        extraction.setNegative  ( false );
        extraction.filter       ( *inlier_points );

        return err;
    }

    int
    Plane::draw( pcl::visualization::PCLVisualizer::Ptr v
                 , int plane_id
                 , double r
                 , double g
                 , double b
                 , bool drawPlane
                 , bool drawNormal
                 , bool drawFrame
                 , int viewport_id ) const
    {
        Plane::draw( *this, v, plane_id, r, g, b, drawPlane, drawNormal, drawFrame, viewport_id );
    }

    int
    Plane::draw( pcl::visualization::PCLVisualizer::Ptr v
                 , std::string plane_name
                 , double r
                 , double g
                 , double b
                 , bool drawPlane
                 , bool drawNormal
                 , bool drawFrame
                 , int viewport_id ) const
    {
        Plane::draw( *this, v, plane_name, r, g, b, drawPlane, drawNormal, drawFrame, viewport_id );
    }

    int
    Plane::draw( am::Plane const& plane
                 , pcl::visualization::PCLVisualizer::Ptr v
                 , int plane_id
                 , double r
                 , double g
                 , double b
                 , bool drawPlane
                 , bool drawNormal
                 , bool drawFrame
                 , int viewport_id
                 )
    {
        char plane_name[255];
        sprintf( plane_name, "plane%d", plane_id );
        return draw( plane, v, r, g, b, drawPlane, drawNormal, drawFrame, viewport_id );
    }

    int
    Plane::draw( am::Plane const& plane
                 , pcl::visualization::PCLVisualizer::Ptr v
                 , std::string plane_name
                 , double r
                 , double g
                 , double b
                 , bool drawPlane
                 , bool drawNormal
                 , bool drawFrame
                 , int viewport_id
                 )
    {
        if ( plane.Corners().size() < 4 )
        {
            std::cerr << "Plane::Draw(): corners size < 4, cannot draw!" << std::endl;
            return EXIT_FAILURE;
        }

        // polygon containing the for corners spanning the plane's face
        pcl::PointCloud<pcl::PointXYZ>::Ptr plane_polygon_cloud_ptr( new pcl::PointCloud<pcl::PointXYZ> );

        std::vector<pcl::PointXYZ> corners = plane.Corners();
        char title[255];
        for ( int corner_id = 0;
              corner_id < corners.size();
              ++corner_id )
        {
            sprintf( title, "%s_line%d", plane_name.c_str(), corner_id );

            v->addLine( corners[corner_id],
                        corners[(corner_id+1) % corners.size()],
                    r, g, b, title, viewport_id );
            v->setShapeRenderingProperties( pcl::visualization::PCL_VISUALIZER_LINE_WIDTH, 2., title, viewport_id );

            plane_polygon_cloud_ptr->push_back( corners[corner_id] );
        }

        // draw plane polygon
        sprintf( title, "%s_polygon", plane_name.c_str() );
        if ( drawPlane )  v->addPolygon<pcl::PointXYZ>( plane_polygon_cloud_ptr, r,g,b, title, viewport_id );

        // show normal
        if ( drawNormal )
        {
            v->addArrow( am::util::pcl::asPointXYZ( plane.getCentroid() + plane.Normal() * .1f ), // end point
                         am::util::pcl::asPointXYZ( plane.getCentroid()                        ), // start point
                         r*3., g*3., b*3.,
                         false,                                                                   // show length
                         std::string(title) + "_normal"                                                  // cloud id
                         , viewport_id );
        }

        // draw coorindate system
        if ( drawFrame )
        {
            for ( int axis_id = 0; axis_id < 3; ++axis_id )
            {
                char arrow_name[1024];
                sprintf( arrow_name, "_arrow%d", axis_id );

                v->addArrow( am::util::pcl::asPointXYZ( plane.Frame().block<3,1>(0,3) + plane.Frame().block<3,1>(0,axis_id).normalized() *.2f ), // end point
                             am::util::pcl::asPointXYZ( plane.Frame().block<3,1>(0,3)                             ), // start point
                             (2-axis_id  )%3 *.45 + .1,                                               // r
                             (2-axis_id+1)%3 *.45 + .1,                                               // g
                             (2-axis_id+2)%3 *.45 + .1,                                               // b
                             false,                                                                   // show length
                             std::string(title) + arrow_name                                          // cloud id
                             , viewport_id );
                //v->setShapeRenderingProperties( pcl::visualization::PCL_VISUALIZER_, 3., std::string(title) + arrow_name );
            }
        }

    }

    std::vector<Eigen::Vector3f>
    Plane::getSides() const
    {
        std::vector<Eigen::Vector3f> sides;
        const int corners_count = Corners().size();
        for ( int side_id = 0; side_id != corners_count; ++side_id )
        {
            sides.push_back(   Corners().at( (side_id + 1 == corners_count) ? 0 : side_id + 1 ).getVector3fMap()
                             - Corners().at( side_id                                          ).getVector3fMap() );
        }
        return sides;
    }

    Eigen::Vector3f
    Plane::getNormalFromCorners() const
    {
        if ( _corners.size() != 4 )
        {
            std::cerr << "[" << __func__ << "]: " << "need 4 corners for this...returning 0 vector\n" << std::endl;
            return Eigen::Vector3f::Zero();
        }
//        Eigen::Vector3f v1 = (corners[corner_ids[1]].getVector3fMap() - corners[corner_ids[0]].getVector3fMap()).normalized();
//        Eigen::Vector3f v2 = (corners[corner_ids[2]].getVector3fMap() - corners[corner_ids[0]].getVector3fMap()).normalized();
//        Eigen::Vector3f normal = v1.cross( v2 ).normalized();

        Eigen::Vector3f normal( Eigen::Vector3f::Zero() );
        std::vector<Eigen::Vector3f> sides = getSides();
        for ( int i = 0; i != sides.size(); ++i )
        {
            normal += sides[ (i+1 == sides.size()) ? 0 : i+1 ].cross( -1.f * sides[i] );
        }

        return normal.normalized();
    }

    int
    Plane::fromCorners( am::Plane                        &plane,
                        std::vector<pcl::PointXYZ> const &corners,
                        std::vector<int>           const &corner_ids,
                        PCloud::ConstPtr                 cloud_ptr,
                        pcl::PointIndices::ConstPtr      indices_ptr,
                        double                           distanceThreshold
                        )
    {
        if ( corners.size() < 4 || corner_ids.size() < 3 )
        {
            std::cerr << "Plane::fromCorners: corners.size() < 4 or corner_ids.size() < 3 ..." << std::endl;
            return EXIT_FAILURE;
        }

        // corners
        plane.Corners().resize(corner_ids.size());
        plane._corners[0] = corners[corner_ids[0]];
        plane._corners[1] = corners[corner_ids[1]];
        plane._corners[2] = corners[corner_ids[3]];
        plane._corners[3] = corners[corner_ids[2]];

//        Eigen::Vector3f v1 = (corners[corner_ids[1]].getVector3fMap() - corners[corner_ids[0]].getVector3fMap()).normalized();
//        Eigen::Vector3f v2 = (corners[corner_ids[2]].getVector3fMap() - corners[corner_ids[0]].getVector3fMap()).normalized();
//        Eigen::Vector3f normal = v1.cross( v2 ).normalized();
        Eigen::Vector3f normal   = plane.getNormalFromCorners();
        Eigen::Vector3f centroid = plane.getCentroidFromCorners();
//        Eigen::Vector3f centroid( {0.f,0.f,0.f} );
//        for ( int i = 0; i < 4; ++i )
//            centroid += corners[corner_ids[i]].getVector3fMap();
//        centroid /= 4.f;

        float d = -1.f * centroid.dot( normal );

        // construct
        plane.setNormal( normal );
        plane.CoeffsPtrRef()->values[3] = d;

        if ( cloud_ptr )
        {
            plane.getInlierIndices( plane.IndicesPtrRef(),
                                    cloud_ptr,
                                    indices_ptr,
                                    distanceThreshold );
        }
        plane.CloudPtrRef() = cloud_ptr;

        // frame
        Eigen::Vector3f v1 = (corners[corner_ids[1]].getVector3fMap() - corners[corner_ids[0]].getVector3fMap()).normalized();
        Eigen::Vector3f v2 = (corners[corner_ids[2]].getVector3fMap() - corners[corner_ids[0]].getVector3fMap()).normalized();

        plane._frame.block<3,1>(0,0) = v1.normalized();
        plane._frame.block<3,1>(0,1) = v2.normalized();
        plane._frame.block<3,1>(0,2) = normal;
        plane._frame.block<3,1>(0,3) = centroid;

#if 0
        {
            pcl::visualization::PCLVisualizer::Ptr vptr( new pcl::visualization::PCLVisualizer("Plane::fromCorners") );
            vptr->setBackgroundColor( .5, .6, .6 );
            vptr->addArrow( am::util::pcl::asPointXYZ( plane._frame.block<3,1>(0,3) + plane._frame.block<3,1>(0,0) )
                            , am::util::pcl::asPointXYZ( plane._frame.block<3,1>(0,3))
                            , 1., 0., 0.
                            , false
                            , "frame0"
                            , 0 );
            vptr->addArrow( am::util::pcl::asPointXYZ( plane._frame.block<3,1>(0,3) + plane._frame.block<3,1>(0,1) )
                            , am::util::pcl::asPointXYZ( plane._frame.block<3,1>(0,3))
                            , 0., 1., 0.
                            , false
                            , "frame1"
                            , 0 );
            vptr->addArrow( am::util::pcl::asPointXYZ( plane._frame.block<3,1>(0,3) + plane._frame.block<3,1>(0,2) )
                            , am::util::pcl::asPointXYZ( plane._frame.block<3,1>(0,3))
                            , 0., 0., 1.
                            , false
                            , "frame2"
                            , 0 );

            vptr->addArrow( am::util::pcl::asPointXYZ( (plane.Normal() * 2.f + plane._frame.block<3,1>(0,3)) )
                            , am::util::pcl::asPointXYZ( plane._frame.block<3,1>(0,3) )
                            , 1., 1., 0.
                            , false
                            , "coeffnormal"
                            , 0 );
            plane.draw( vptr, "plane" );
            vptr->spin();
        }
#endif

    }

    int
    Plane::countFinitePoints() const
    {
        int cnt = 0;
        for ( int pid = 0; pid < _indices_ptr->indices.size(); ++pid )
        {
            MyPoint pnt = _cloud_ptr->at( _indices_ptr->indices[pid] );

            if ( pcl_isfinite(pnt.x) && pcl_isfinite(pnt.y) && pcl_isfinite(pnt.z) )
            {
                ++cnt;
            }
        }

        return cnt;
    }

    int
    Plane::writeToFile( Plane const& plane, std::string path, bool append )
    {
        std::ofstream myfile;
        myfile.open ( path, std::ios_base::binary | std::ios_base::out | (append ? std::ios_base::app : std::ios_base::trunc) );
        if ( !myfile.is_open() )
        {
            std::cerr << "could not open file " << path << std::endl;
            return EXIT_FAILURE;
        }

        // write "plane count = 1" at top if not appending
        if ( !append )
        {
            int plane_count = 1;
            myfile.write( (char*)&plane_count, sizeof(int) );
        }

        std::cout << "writing plane: " << plane.Coeffs().transpose() << std::endl;
        for ( int coeff_id = 0; coeff_id < 4; ++coeff_id )
            myfile.write( (char*)&(plane.CoeffsPtr()->values[coeff_id]), sizeof(float) );

        myfile.close();
        return EXIT_SUCCESS;
    }

    Eigen::Vector3f
    Plane::getCentroidFromCorners() const
    {
        if ( _corners.size() < 4 )
        {
            std::cerr << "[" << __func__ << "]: " << "no cloud_ptr, no corners, returning {0,0,0}\n";
            return (Eigen::Vector3f() << 0.f, 0.f, 0.f).finished();
        }
        else
        {
            Eigen::Vector3f centroid( Eigen::Vector3f::Zero() );
            for ( int corner_id = 0; corner_id != _corners.size(); ++corner_id )
            {
                centroid += _corners[ corner_id ].getVector3fMap();
            }
            centroid /= float( _corners.size() );
            return centroid;
        }
    }

    Eigen::Vector3f
    Plane::getCentroid( pcl::PointIndices::ConstPtr indices_ptr ) const
    {
        Eigen::Vector4d centroid;
        if ( !_cloud_ptr )
            return getCentroidFromCorners();

        pcl::PointIndices::ConstPtr indices = indices_ptr ? indices_ptr
                                                          : (_indices_ptr ? _indices_ptr
                                                                          : nullptr);
        if ( indices )
        {
            if ( !indices->indices.size() )
            {
                std::cerr << __PRETTY_FUNCTION__ << " WARNING! no indices to work from (indices_ptr argument)...." << std::endl;
                pcl::compute3DCentroid( *_cloud_ptr,
                                        centroid );
            }
            else
            {
                std::cout << "[" << __func__ << "]: " << "indices.size(): " << indices->indices.size()
                          << ", maxid: " << *std::max_element( indices->indices.begin(), indices->indices.end() )
                          << ", cloud.size(): " << _cloud_ptr->size()
                          << std::endl; fflush(stdout);
                pcl::compute3DCentroid( *_cloud_ptr,
                                        *indices,
                                        centroid );
                std::cout << "[" << __func__ << "]: " << "compute centroid ok" << std::endl; fflush(stdout);
            }

        }
        //        else if ( _indices_ptr )
        //        {
        //            if ( !_indices_ptr->indices.size() )
        //            {
        //                std::cerr << "[" << __func__ << "]: " << "WARNING! no indices to work from (_indices_ptr)...." << std::endl;
        //                pcl::compute3DCentroid( *_cloud_ptr,
        //                                        centroid );
        //            }
        //            else
        //                pcl::compute3DCentroid( *_cloud_ptr,
        //                                        *_indices_ptr,
        //                                        centroid );

        //        }
        else
        {
            pcl::compute3DCentroid( *_cloud_ptr,
                                    centroid );
        }


        //std::cout << "centroid: " << centroid.transpose();
        pcl::PointXYZ p0;
        pcl::projectPoint( am::util::pcl::v4asPointXYZ(centroid.cast<float>(),true), this->Coeffs(), p0 );
        //std::cout << " p0: " << p0 << std::endl;
        return {p0.x, p0.y, p0.z};
}

    int
    Plane::flipNormal()
    {
        for ( int coeff_id = 0; coeff_id < 4; ++coeff_id )
            _coeffs_ptr->values[coeff_id] *= -1.f;
        _frame.block<3,1>(0,2) *= -1.f;
        _frame.block<3,1>(0,0) *= -1.f;
    }

    int
    Plane::offsetAlongNormal( double dist )
    {
        std::cout << "offsetting by " << dist << std::endl;

        Eigen::Vector3f centroid = getCentroid();
        Eigen::Vector3f normal   = Normal().normalized();
        const auto offset = dist * normal;
        centroid += offset;
        _coeffs_ptr->values[3]   = -normal.dot( centroid );

        _frame.block<3,1>(0,3)   = centroid;

        //for ( int corner_id = 0; corner_id != _corners.size(); ++corner_id )
        //    _corners[corner_id] += am::utioffset;
    }

    int
    Plane::checkNormalFlip( Plane                       &plane,
                            PCloud::ConstPtr            cloud_ptr,
                            pcl::PointIndices::ConstPtr indices_ptr,
                            double                      distanceThreshold )
    {
        //Eigen::Vector3f p0 = plane.getCentroid();
        Eigen::Vector3f world_normal = plane.Normal();

        float positive_distances = 0.f,
                negative_distances = 0.f;

        // for every point in cloud (hopefully cuboid cloud)
        for ( int index_id = 0;
              index_id < indices_ptr->indices.size();
              ++index_id )
        {
            // get point
            Eigen::Vector3f pt = cloud_ptr->at( indices_ptr->indices[ index_id ] ).getVector3fMap();
            // get p2p distance
            float pnt2plane_distance = world_normal.dot(pt) + plane.D();
            // if outlier
            //if ( fabs(pnt2plane_distance) > distanceThreshold )
            //{
            // accumulate on correct side
            if ( pnt2plane_distance >= 0.f ) // if "in front of"
                positive_distances += pnt2plane_distance;
            else
                negative_distances -= pnt2plane_distance;
            //}
        }
        std::cout << "pos_distances: " << positive_distances << " negative_distances: " << negative_distances << " -> ";

        // flip, if "behind" has more weight
        if ( positive_distances < negative_distances )
        {
            std::cout << "FLIPPING" << std::endl;
            plane.flipNormal();
        }
        else
            std::cout << "NOT FLIPPING" << std::endl;
    }

    int
    Plane::calculateCorners( am::Plane             &plane,
                             Eigen::Matrix4f const &frame )
    {
        PCloud::Ptr inliers_cloud_ptr;
        DepthProcessor::projectPlaneInliers( inliers_cloud_ptr,
                                             plane.CloudPtr(),
                                             plane.IndicesPtr(),
                                             plane.CoeffsPtr() );
        //Eigen::Matrix4f frame;
        if ( frame.block<3,1>(0,2).dot( plane.Normal() ) < .9f )
        {
            std::cerr << "Plane::CalculateCorners(): third eigen vector not parallel to normal!" << std::endl;
            std::cout << "calculateCorners: ";
            std::cout << "frame.block<3,1>(0,2): " << frame.block<3,1>(0,2).transpose() << std::endl;
            std::cout << "plane.Normal().transpose(): " << plane.Normal().transpose() << std::endl;
            std::cout << "dotproduct: " << frame.block<3,1>(0,2).dot( plane.Normal() ) << std::endl;
        }

        std::cout << "frame: " << frame << std::endl; fflush(stdout);
        MyPoint min_pt, max_pt;
        DepthProcessor::calculateExtremaInFrame( min_pt, max_pt, frame, inliers_cloud_ptr, nullptr );

        auto min_pt_world = toWorld( min_pt, frame );
        auto c1_world     = toWorld( pcl::PointXYZ( min_pt.x, max_pt.y, min_pt.z ), frame );
        auto max_pt_world = toWorld( max_pt, frame );
        auto c3_world     = toWorld( pcl::PointXYZ( max_pt.x, min_pt.y, max_pt.z ), frame );

        plane.Corners().resize(4);
        plane.Corners().at(0) = min_pt_world;
        plane.Corners().at(1) = c1_world;
        plane.Corners().at(2) = max_pt_world;
        plane.Corners().at(3) = c3_world;
        plane.Frame() = frame;
    }

    int
    Plane::makePerpendicular( am::Plane                   &plane,
                              Eigen::Vector3f             normal,
                              PCloud::ConstPtr            cloud_ptr,
                              pcl::PointIndices::ConstPtr indices_ptr,
                              double                      distanceThreshold
                              )
    {

        Eigen::Vector3f centroid = plane.getCentroid();

        Eigen::Vector3f n2 = plane.Normal();
        plane.setNormal( (n2 - normal * normal.dot(n2)).normalized() );

        if ( fabs(plane.Normal().dot(normal)) > 0.001 )
            std::cerr << "WTFFFFFF plane.Normal().dot( n1 ): " << plane.Normal().dot( normal ) << std::endl;

        std::cout << "[" << __func__ << "]: " << "qua" << std::endl; fflush(stdout);
        plane.CoeffsPtrRef()->values[3] = -plane.Normal().dot( centroid );
        std::cout << "[" << __func__ << "]: " << "ok" << std::endl; fflush(stdout);

        Plane::checkNormalFlip( plane,
                                cloud_ptr,
                                indices_ptr,
                                distanceThreshold );
        std::cout << "[" << __func__ << "]: " << "flip ok" << std::endl; fflush(stdout);

        plane.getInlierIndices( plane.IndicesPtrRef(),
                                plane.CloudPtr(),
                                indices_ptr,
                                distanceThreshold );
        std::cout << "[" << __func__ << "]: " << "inliers ok" << std::endl; fflush(stdout);
    }

    int
    Plane::projectPointsToNormal( Histogram                  &histogram,
                                  am::Plane            const &plane,
                                  PCloud::ConstPtr            cloud_ptr,
                                  pcl::PointIndices::ConstPtr indices_ptr,
                                  int                         N_bins
                                  , double                      distanceThreshold
                                  , bool                        debug )
    {
        //debug = true;
        if ( !indices_ptr->indices.size() )
        {
            std::cerr << "projectPointsToNormal(): indices_ptr.size() == 0" << std::endl;
            return EXIT_FAILURE;
        }

        // debug
        std::cout << "N_bins: " << N_bins << std::endl;
        boost::shared_ptr<am::CloudViewer> vp(nullptr);
        if ( debug  )
            vp = am::CloudViewer::showCloud( cloud_ptr, "plane", plane.IndicesPtr(), 5., false );

        // project points to line
        PCloud::Ptr projected_points_cloud_ptr( new PCloud() );
        Eigen::VectorXf line_coeffs( 6 );
        {
            // SAC
            pcl::SampleConsensusModelLine<MyPoint> sac_model_line( cloud_ptr );
            // create line
            if ( plane._indices_ptr->indices.size() )
                line_coeffs.head<3>()     = plane.getCentroid(); // P0
            else
            {
                pcl::PointIndices::Ptr tmp_indices_ptr;
                plane.getInlierIndices( tmp_indices_ptr, cloud_ptr, indices_ptr, distanceThreshold );
                if ( tmp_indices_ptr->indices.size() )
                    line_coeffs.head<3>()     = plane.getCentroid( tmp_indices_ptr ); // P0
                else
                {
                    std::cerr << __PRETTY_FUNCTION__ << " no indices, heuristic fix..." << std::endl;
                    line_coeffs.head<3>()     = plane.Coeffs()(3) * plane.Normal();
                }
            }
            line_coeffs.segment<3>(3) = plane.Normal();      // V

            // debug
            //        {
            //            std::cout << "line_coeffs: " << line_coeffs << std::endl;
            //            std::cout << "plane.coeffs: " << plane.Coeffs() << std::endl;
            //            std::cout << "plane.normal:  " << plane.Normal().transpose() << std::endl;
            //            auto vptr = am::CloudViewer::showCloud( cloud_ptr, "showCloud", nullptr, 2., false );
            //            vptr->VP()->addArrow( am::util::pcl::asPointXYZ( line_coeffs.head<3>() + line_coeffs.segment<3>(3))
            //                                  , am::util::pcl::asPointXYZ( line_coeffs.head<3>() )
            //                                  , 1., 0., 0.
            //                                  , false
            //                                  , "line" );
            //            vptr->VP()->spin();

            //        }

            // projection
            sac_model_line.projectPoints( /* in indices: */ indices_ptr->indices,
                                          /*    in line: */ line_coeffs,
                                          /* out points: */ *projected_points_cloud_ptr,
                                          false );
        }

        if ( debug )
        {
            am::CloudViewer::addPointCloud( vp->VP(), projected_points_cloud_ptr, "line" );
            vp->ViewerPtr()->spin();
        }

        // calculate projected points' distances from P0 along the normal
        std::vector<double> distances( projected_points_cloud_ptr->size() );
        double min_val = DBL_MAX,  // farthest point distance along Normal "behind" P0
                max_val = -DBL_MAX; // farthest point distance along Normal from P0
        // all projected points
        for ( int point_id = 0;
              point_id < projected_points_cloud_ptr->size();
              ++point_id )
        {
            // get vector P1-P0
            Eigen::Vector3f diff = (projected_points_cloud_ptr->at(point_id).getVector3fMap() - line_coeffs.head<3>());
            // length
            distances[point_id] = diff.norm();
            // direction
            if ( diff.dot( line_coeffs.head<3>() + line_coeffs.segment<3>(3)) < 0.f ) // if dot product doesn't align with local normal
                distances[point_id] *= -1.f;                                          // distance is behind (negative)
            // min, max
            if ( distances[point_id] > max_val )    max_val = distances[point_id];
            if ( distances[point_id] < min_val )    min_val = distances[point_id];
        }

        // debug
        {
            if ( max_val < -1000 || min_val > 1000 )
            {
                std::cerr << "max_val: " << max_val << " , min_val: " << min_val << std::endl;
                for ( int i = 0; i < distances.size(); ++i )
                    std::cerr << distances[i] << ", ";
                std::cerr << std::endl;
            }

        }

        // recurr from here

        // histogram bin width
        bool needRestart = true;
        int beg_pos = 0;
        int end_pos = N_bins-1;
        double step;
        //planes{1} = [4981, 1188, 525, 491, 267, 41, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 4, ];
        //planes{2} = [325, 1028, 469, 416, 421, 389, 346, 390, 393, 335, 361, 348, 347, 336, 335, 290, 341, 331, 223, 66, 1, 2, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 2, 2, ];
        //planes{3} = [25, 184, 412, 309, 244, 204, 202, 188, 197, 206, 204, 190, 196, 192, 180, 191, 193, 199, 182, 195, 175, 194, 177, 161, 169, 170, 178, 169, 179, 172, 164, 156, 168, 170, 181, 176, 154, 159, 131, 113, 74, 9, 0, 0, 0, 3, 1, 0, 0, 1, ];

        //planes{2} = [325, 1028, 469, 416, 421, 389, 346, 390, 393, 335, 361, 348, 347, 336, 335, 290, 341, 331, 223, 66, 1, 2, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 2, 2, ];
        //planes{3} = [25, 184, 412, 309, 244, 204, 202, 188, 197, 206, 204, 190, 196, 192, 180, 191, 193, 199, 182, 195, 175, 194, 177, 161, 169, 170, 178, 169, 179, 172, 164, 156, 168, 170, 181, 176, 154, 159, 131, 113, 74, 9, 0, 0, 0, 3, 1, 0, 0, 1, ];

        while ( needRestart )
        {
            needRestart = false;

            step = (max_val - min_val) / N_bins;
            // prepare histogram
            histogram = Histogram( N_bins );

            // bin points
            for ( int point_id = 0;
                  point_id < projected_points_cloud_ptr->size();
                  ++point_id )
            {
                // calculate bin index
                int index = (int) floor((distances[point_id] - min_val) / step);
                // check for exactly at end
                //if ( index == histogram.histogram.size() )  --index;
                // check, if histogram has been clamped, and this doesn't fit
                if ( index < 0 || index >= histogram.histogram.size() )
                    continue;

                // increment bin
                ++histogram.histogram[ index ];

                // track min, max distance in each bin
                if ( histogram.mins[index] > distances[point_id] )  histogram.mins[index] = distances[point_id];
                if ( histogram.maxs[index] < distances[point_id] )  histogram.maxs[index] = distances[point_id];
            }

            // check histogram, recurr if necessary
            if ( (beg_pos == 0) && (end_pos == histogram.histogram.size()-1) )
            {
                // calculate minimum bin size limit
                int limit;
                {
                    // get hist max
                    int h_max_val, h_max_pos;
                    histogram.getMax( &h_max_val, &h_max_pos );

                    limit = std::max(1,std::min( (int)ceil(h_max_val * .02), 5 ));
                    std::cout << "limit is: " << limit << std::endl;
                }

                // check from beginning for first >limit
                while ( (beg_pos < histogram.histogram.size()) && (histogram.histogram[beg_pos] < limit) )
                    ++beg_pos;
                // check from end for first <limit
                while ( (end_pos > 0) && (histogram.histogram[end_pos] < limit) )
                    --end_pos;

                //TODO: find zeros, and restart: 6, 2, 0, 0, 0, 0, 0, 1, 9, 22, 567, 5683, 2701, 216, 154, 116, 96, 118, 97, 88, 98, 88, 96, 97, 89, 87, 82, 102, 77, 65, 71, 81, 79, 60, 80, 80, 65, 81, 80, 77, 84, 82, 82, 97, 85, 76, 54, 41, 14, 4

                // if there is something to clamp, restart
                if ( (beg_pos > 0) || (end_pos < histogram.histogram.size() - 1) )
                {
                    if ( beg_pos >= end_pos )
                    {
                        std::cerr << "not rescaling, since beg>= end" << std::endl;
                        break;
                    }

                    std::cerr << "rescaling histogram, since (beg,end) = " << beg_pos << "," << end_pos
                              << "min_val: " << min_val << ", max_val: " << max_val << std::endl;

                    min_val = histogram.mins[beg_pos];
                    max_val = histogram.maxs[end_pos];
                    std::cerr << "new min_val: " << min_val << ", max_val: " << max_val << std::endl;

                    am::util::printVector( histogram.maxs, "histogram.maxs" );
                    am::util::printVector( histogram.histogram, "histogram.histogram" );

                    needRestart = true;
                }
            }
            else
            {
                std::cout << "after correction: " << std::endl;
                am::util::printVector( histogram.maxs, "histogram.maxs" );
                am::util::printVector( histogram.histogram, "histogram.histogram" );
            }
        }
    }

    // histogram has peak, if the integral is < 15% of the whole area around curve
    std::pair<bool,std::vector<int>::iterator>
    hasPeak( std::vector<int> &histogram, float thresh = .15f )
    {
        if ( histogram.size() < 2 )
            return std::make_pair<bool,std::vector<int>::iterator>( false, histogram.begin() );

        auto max_pos_it = std::max_element( histogram.begin(), histogram.end() );
        float max_val = *max_pos_it;

        int sum = 0;
        for ( auto const& val : histogram )
        {
            sum += val;
        }
        float avg = (float)sum / max_val / (float)histogram.size();

        return std::pair<bool,std::vector<int>::iterator>( avg < thresh, max_pos_it );
    }

    // keeps removing maximum elements until no peak anymore
    // then returns median
    int
    getMedianWithoutPeak( std::vector<int> histogram )
    {
        am::util::printVector( histogram, "getMedianWithoutPeak" );
        for ( auto has_peak_with_pos = hasPeak(histogram);
              has_peak_with_pos.first;
              has_peak_with_pos = hasPeak(histogram) )
        {
            histogram.erase( has_peak_with_pos.second );
        }
        return am::util::algo::median<int>( histogram );
    }

    int
    Plane::processHistogram( std::vector<double>              &offsets,
                             std::vector<am::Histogram> const &histograms,
                             double                            threshold  )
    {
        // error check
        if ( histograms.size() < 3 )
        {
            std::cerr << "Histograms size < 3...exiting\n";
            return EXIT_FAILURE;
        }

        // prep output
        offsets.resize( 6 );

        // get histogram maximums
        std::vector<std::pair<int,int> > max_vals;
        for ( int hid = 0; hid < histograms.size(); ++hid )
        {
            int val, pos;
            histograms[hid].getMax( &val, &pos );
            max_vals.push_back( std::make_pair( val, pos ) );
        }

        // TODO: detect 2planes/vs/3planes

        // calc dimensions
        double w = sqrtf( max_vals[0].first * max_vals[1].first / max_vals[2].first ); // w = (m0 * m1 / m2);
        double h = max_vals[0].first / w;                                              // h = m0 / w;
        double l = max_vals[1].first / w;                                              // l = m1 / w;

        // prepare cutoff limits
        std::vector<double> limits(3);
    #if 1
        for ( int hid = 0; hid < histograms.size(); ++hid )
            limits[hid] = std::max(1.,getMedianWithoutPeak(histograms[hid].histogram)*threshold);
    #else
        limits = { (w+h)*threshold, (w+l)*threshold, (h+l)*threshold };
    #endif

        // look both ways starting from max to find crossing of limits
        for ( int hid = 0; hid < histograms.size(); ++hid )
        {
            // cache reference
            auto &histogram = histograms[hid].histogram;

            // FORWARD
            // find first element smaller than limit
            auto it = std::find_if( histogram.begin() + max_vals[hid].second,                   // position of maximum
                                    histogram.end(),                                            // look until reaching end
                                    [&limits,hid] (int val) { return val < limits[hid]; } );    // smaller than limit
            // calculate position (at least end-1)
            int pos = std::min( (int)std::distance( histogram.begin(), it ), (int)histogram.size()-1 );
            // roll back to first non-zero
            while ( (histogram[pos] == 0) && (pos > 0) )
                --pos;

            // store value
            offsets[histograms.size()+hid] = histograms[hid].maxs[pos];
            std::cout << "selecting histogram for plane" << histograms.size()+hid
                      << " at " << pos
                      << " with bin value " << histogram[pos]
                         << " prev bin value " << ( pos > 0 ? histogram[pos-1] : -1 )
                         << " and clamp value " << offsets[histograms.size()+hid]
                         << " since limit is " << limits[hid] << std::endl;

            // BACKWARD
            // find first element smaller than limit
            auto rit = std::find_if( histogram.rend() - max_vals[hid].second,                   // position of maximum
                                     histogram.rend(),                                          // look backwards, until reaching beginning
                                     [&limits,hid] (int val) { return val < limits[hid]; } );   // smaller than limit
            // calculate found position (at least 0)
            pos = std::max( (int)std::distance( rit, histogram.rend() )-1, 0 );
            // roll forward to first non-zero
            while ( (histogram[pos] == 0) && (pos < histogram.size()-1) )
                ++pos;

            // store value
            offsets[hid] = histograms[hid].mins[pos];
            std::cout << "selecting histogram for plane" << hid
                      << " at " << pos
                      << " with bin value " << histogram[pos]
                         << " prev bin value " << (pos < histogram.size()-1 ? histogram[pos+1] : -1 )
                         << " and clamp value " << offsets[hid]
                            << " since limit is " << limits[hid] << std::endl;

            if ( offsets[hid] > 0.f )
            {
                std::vector<double> mins = histograms[hid].mins;
                am::util::printVector( mins, "histograms[hid].mins" );
                std::vector<double> maxs = histograms[hid].maxs;
                am::util::printVector( maxs, "histograms[hid].maxs" );
                std::vector<int> bins = histograms[hid].histogram;
                am::util::printVector( bins, "histograms[hid].histogram" );
            }
            if ( offsets[hid+3] < 0.f )
            {
                std::vector<double> mins = histograms[3+hid].mins;
                am::util::printVector( mins, "histograms[3+hid].mins" );
                std::vector<double> maxs = histograms[3+hid].maxs;
                am::util::printVector( maxs, "histograms[3+hid].maxs" );
                std::vector<int> bins = histograms[3+hid].histogram;
                am::util::printVector( bins, "histograms[3+hid].histogram" );
            }
        }

        return EXIT_SUCCESS;
    }

    float
    Plane::getArea() const
    {
        return  (   (_corners[1].getVector3fMap() - _corners[0].getVector3fMap()).norm()
                  + (_corners[3].getVector3fMap() - _corners[2].getVector3fMap()).norm() / 2.f )
                *
                (   (_corners[2].getVector3fMap() - _corners[1].getVector3fMap()).norm()
                  + (_corners[0].getVector3fMap() - _corners[3].getVector3fMap()).norm() / 2.f );
    }

    // SET, GET

    pcl::ModelCoefficients::ConstPtr
    Plane::CoeffsPtr() const
    {
        return _coeffs_ptr;
    }

    pcl::ModelCoefficients::Ptr&
    Plane::CoeffsPtrRef()
    {
        return _coeffs_ptr;
    }

    Eigen::Vector4f
    Plane::Coeffs() const
    {
        return { _coeffs_ptr->values[0], _coeffs_ptr->values[1], _coeffs_ptr->values[2], _coeffs_ptr->values[3] };
    }

    Eigen::Vector3f
    Plane::Normal() const
    {
        return { _coeffs_ptr->values[0], _coeffs_ptr->values[1], _coeffs_ptr->values[2] };
    }

    int
    Plane::setNormal( Eigen::Vector3f const& normal )
    {
        _coeffs_ptr->values[0] = normal(0);
        _coeffs_ptr->values[1] = normal(1);
        _coeffs_ptr->values[2] = normal(2);

        return EXIT_SUCCESS;
    }

    float
    Plane::D() const
    {
        return _coeffs_ptr->values[3];
    }

    std::vector<::pcl::PointXYZ> const&
    Plane::Corners() const
    {
        return _corners;
    }

    std::vector<::pcl::PointXYZ> &
    Plane::Corners()
    {
        return _corners;
    }

    Eigen::Matrix4f const&
    Plane::Frame() const
    {
        return _frame;
    }

    Eigen::Matrix4f&
    Plane::Frame()
    {
        return _frame;
    }

} // ns am
