#include "Planes.h"

#include "pcl/PointIndices.h"
#include "pcl/common/centroid.h"
#include "CloudViewer.h"
#include "Params.h"

#include "Cuboid.h"
#include "DepthProcessor.h"
#include "Segmentation.h"

namespace am
{

    inline int
    showCorner( pcl::PointXYZ pnt,
                std::string name,
                pcl::visualization::PCLVisualizer::Ptr viewer_ptr )
    {
        viewer_ptr->addText3D( name,
                               pnt,
                               .1,
                               1.,0.,1.,
                               name + "_text" );

        viewer_ptr->addArrow( pnt,
                              pcl::PointXYZ(0.,0.,0.),
                              1., 0., 0.,
                              false,
                              name + "_arrow" );

        return EXIT_SUCCESS;
    }

    int
    Planes::from_cloud( am::Planes                   &planes
                        //, am::Cuboid           const &cuboid,
                        , PCloud::ConstPtr            cloud_ptr
                        , pcl::PointIndices::ConstPtr indices_ptr
                        , am::Params            const &params
                        , double                      distanceThreshold
                        , bool                        debug )
    {
        std::cout << "from_cloud with threshold: " << distanceThreshold << std::endl;
        //int    max_planes     = params.getInt   ( "max_planes"              , 6   );
        //double min_percentage = params.getDouble( "min_remainder_percentage", .1 );
        if ( planes.size() ) std::cerr << "warning...Planes::from_cuboid \"planes\" output is not empty initially" << std::endl;

        // init remaining cloud
        pcl::PointIndices::Ptr remainder_indices_ptr( new pcl::PointIndices );
        {
            remainder_indices_ptr->indices = indices_ptr->indices;
        }

        while ( (planes.size() < 2) && (remainder_indices_ptr->indices.size() > 0) )
        {
            am::Plane plane;
            Plane::fromCloud( plane
                              , cloud_ptr //cuboid.CloudPtr()
                              , remainder_indices_ptr
                              , distanceThreshold );

            // if first plane
            if ( !planes.size() )
            {
                Plane::checkNormalFlip( plane,
                                        cloud_ptr,
                                        indices_ptr,
                                        distanceThreshold );
            }
            else
            {
                am::Plane::makePerpendicular( plane,
                                              planes[0].Normal(),
                                              cloud_ptr,
                                              indices_ptr,
                                              distanceThreshold );
            }

            // select new plane outliers
            pcl::PointIndices::Ptr outlier_indices_ptr;
            plane.getOutlierIndices( /* out_indices: */ outlier_indices_ptr,
                                     /* input_cloud: */ plane.CloudPtr(),
                                     /*   threshold: */ distanceThreshold );

            // select outliers from all planes
            am::util::algo::vectorIntersection( /*  out: */ remainder_indices_ptr->indices,
                                                /* set1: */ remainder_indices_ptr->indices,
                                                /* set2: */ outlier_indices_ptr->indices  );
            //Plane::calculateCorners( plane );
            std::cout << "remainder_indices.count: " << remainder_indices_ptr->indices.size() << std::endl;

            if ( !plane.CoeffsPtr() || plane.CoeffsPtr()->values.size() < 4 )
            {
                std::cerr << __PRETTY_FUNCTION__ << ":\t plane no coeffs!, exiting..." << std::endl;
                return EXIT_FAILURE;
            }

            planes.add( plane );
        }
        if ( planes.size() < 2 )
        {
            std::cerr << __PRETTY_FUNCTION__ << ": planes.size() < 2, probably not enough points, exiting..." << std::endl;
            return EXIT_FAILURE;
        }

        /// 3rd planeq
        {
            am::Plane &plane0 = planes[0];
            am::Plane &plane1 = planes[1];
            std::cout << "[" << __func__ << "]: " << "here" << std::endl; fflush(stdout);

            am::Plane plane2;
            plane2.CloudPtrRef() = cloud_ptr;
            // Normal
            plane2.setNormal(
                        plane0.Normal().cross(plane1.Normal())
                        .normalized() );

            auto axis = plane0.Normal();
            int err = Segmentation::extract_plane( plane2.CoeffsPtrRef(),
                                                   cloud_ptr,
                                                   //remainder_indices_ptr->size() ? remainder_indices_ptr : indices_ptr,
                                                   remainder_indices_ptr,
                                                   distanceThreshold,
                                                   pcl::SACMODEL_NORMAL_PARALLEL_PLANE, // I want a plane parallel to this normal
                                                   &axis );

            // TODO: use normal and centroids
            if ( err != EXIT_SUCCESS )
            {
                std::cerr << __PRETTY_FUNCTION__ << ": placing third plane in between the two other by heuristic..." << std::endl;
                Eigen::Vector3f normal_plane2 = plane0.Normal().cross( plane1.Normal() );
                auto &values = plane2.CoeffsPtrRef()->values;
                values.resize( 4 );
                values[0] = normal_plane2(0);
                values[1] = normal_plane2(1);
                values[2] = normal_plane2(2);
                values[3] = -normal_plane2.dot( (planes.at(1).getCentroid() + planes.at(0).getCentroid())/2.f );

                std::cout << "heuristic 3rd plane: " << plane2.Coeffs().transpose() << std::endl;
            }

            if ( plane2.CoeffsPtr()->values.size() < 4 )
            {
                std::cerr << __PRETTY_FUNCTION__ << ":\t 3rd plane no coeffs!, exiting..." << std::endl;
                return EXIT_FAILURE;
            }

            // finish
            plane2.getInlierIndices( plane2.IndicesPtrRef(),
                                     cloud_ptr,
                                     indices_ptr, // changed this from nullptr!
                                     distanceThreshold );


            am::Plane::makePerpendicular( plane2, plane0.Normal(), cloud_ptr, indices_ptr, distanceThreshold );
            am::Plane::makePerpendicular( plane2, plane1.Normal(), cloud_ptr, indices_ptr, distanceThreshold );

            planes.add( plane2 );
        }

        std::cout << "END Planes::from_cloud" << std::endl;
        return EXIT_SUCCESS;
    }

    int
    Planes::from_cuboid2( am::Planes       &planes,
                          am::Cuboid const &cuboid,
                          am::Params const &params,
                          double            distanceThreshold )
    {
        int    max_planes     = params.getInt   ( "max_planes"              , 6   );
        double min_percentage = params.getDouble( "min_remainder_percentage", .25 );

        // init remaining cloud
        pcl::PointIndices::Ptr remainder_indices_ptr( new pcl::PointIndices );
        {
            remainder_indices_ptr->indices = cuboid.IndicesPtr()->indices;
        }

        planes.PrimitivesRef().resize(3);
        Segmentation::extract_plane<MyPoint>(
                    /* out_coeffs: */ planes.at(0).CoeffsPtrRef(),
                    /*   in_cloud: */ cuboid.CloudPtr(),
                    /* in_indices: */ remainder_indices_ptr );
        planes.at(0).getInlierIndices( /*     out indices: */ planes.at(0).IndicesPtrRef(),
                                       /*        in cloud: */ cuboid.CloudPtr(),
                                       /*      in indices: */ remainder_indices_ptr,
                                       /* distance Thresh: */ distanceThreshold );
        std::cout << "inlier plane0: " << planes.at(0).IndicesPtr()->indices.size() << std::endl;

        Eigen::Vector3f normal0 = planes.at(0).Normal();
        Segmentation::extract_plane<MyPoint>(
                    /* out_coeffs: */ planes.at(1).CoeffsPtrRef(),
                    /*   in_cloud: */ cuboid.CloudPtr(),
                    /* in_indices: */ remainder_indices_ptr,
                    distanceThreshold,
                    pcl::SACMODEL_PERPENDICULAR_PLANE,
                    &normal0
                    );
        planes.at(1).getInlierIndices( /*     out indices: */ planes.at(1).IndicesPtrRef(),
                                       /*        in cloud: */ cuboid.CloudPtr(),
                                       /*      in indices: */ remainder_indices_ptr,
                                       /* distance Thresh: */ distanceThreshold );
        std::cout << "inlier plane1: " << planes.at(1).IndicesPtr()->indices.size() << std::endl;

        Eigen::Vector3f up = planes.at(0).Normal().cross( planes.at(1).Normal() );
        Segmentation::extract_plane<MyPoint>(
                    /* out_coeffs: */ planes.at(2).CoeffsPtrRef(),
                    /*   in_cloud: */ cuboid.CloudPtr(),
                    /* in_indices: */ remainder_indices_ptr,
                    distanceThreshold,
                    pcl::SACMODEL_NORMAL_PARALLEL_PLANE,
                    &up
                    );
        planes.at(2).getInlierIndices( /*     out indices: */ planes.at(2).IndicesPtrRef(),
                                       /*        in cloud: */ cuboid.CloudPtr(),
                                       /*      in indices: */ remainder_indices_ptr,
                                       /* distance Thresh: */ distanceThreshold );
        std::cout << "inlier plane2: " << planes.at(2).IndicesPtr()->indices.size() << std::endl;

#if 0
        // debug
        {
            am::CloudViewer viewer0("cuboid_from_cloud input indices vs. masked_indices", 4);

            PCloud::Ptr tmp_cloud_ptr( new PCloud() );
            am::DepthProcessor::extractCloud<MyPoint>( tmp_cloud_ptr,
                                                       cuboid.CloudPtr(),
                                                       remainder_indices_ptr );
            viewer0.getVPtr()->addPointCloud( tmp_cloud_ptr, "input", viewer0.getVid(0) );
            viewer0.getVPtr()->setPointCloudRenderingProperties(
                        pcl::visualization::PCL_VISUALIZER_POINT_SIZE, 5, "input" );


            PCloud::Ptr tmp2_cloud_ptr( new PCloud() );
            am::DepthProcessor::extractCloud<MyPoint>( tmp2_cloud_ptr,
                                                       planes.at(0).CloudPtr(),
                                                       planes.at(0).IndicesPtr() );
            viewer0.getVPtr()->addPointCloud( tmp2_cloud_ptr, "plane0", viewer0.getVid(1) );
            viewer0.getVPtr()->setPointCloudRenderingProperties(
                        pcl::visualization::PCL_VISUALIZER_POINT_SIZE, 5, "plane0" );

            PCloud::Ptr tmp3_cloud_ptr( new PCloud() );
            am::DepthProcessor::extractCloud<MyPoint>( tmp3_cloud_ptr,
                                                       planes.at(1).CloudPtr(),
                                                       planes.at(1).IndicesPtr() );
            viewer0.getVPtr()->addPointCloud( tmp3_cloud_ptr, "plane1", viewer0.getVid(2) );
            viewer0.getVPtr()->setPointCloudRenderingProperties(
                        pcl::visualization::PCL_VISUALIZER_POINT_SIZE, 5, "plane1" );

            PCloud::Ptr tmp4_cloud_ptr( new PCloud() );
            am::DepthProcessor::extractCloud<MyPoint>( tmp4_cloud_ptr,
                                                       planes.at(2).CloudPtr(),
                                                       planes.at(2).IndicesPtr() );
            viewer0.getVPtr()->addPointCloud( tmp4_cloud_ptr, "plane2", viewer0.getVid(3) );
            viewer0.getVPtr()->setPointCloudRenderingProperties(
                        pcl::visualization::PCL_VISUALIZER_POINT_SIZE, 5, "plane2" );

            viewer0.getVPtr()->spin();
        }
#endif
    }

    int
    Planes::from_cuboid( am::Planes       &planes,
                         am::Cuboid const &cuboid,
                         am::Params const &params,
                         double            distanceThreshold,
                         bool debug )
    {
        int    max_planes     = params.getInt   ( "max_planes"              , 6   );
        double min_percentage = params.getDouble( "min_remainder_percentage", .25 );

        // init remaining cloud
        pcl::PointIndices::Ptr remainder_indices_ptr( new pcl::PointIndices );
        remainder_indices_ptr->indices = cuboid.IndicesPtr()->indices;

        // get planes
        while ( planes.Primitives().size() < max_planes )
        {
            // enqueue
            Plane plane;

            // create plane from remainder points
            am::Plane::fromCloud( /*           out: */ plane,
                                  /*   input_cloud: */ cuboid.CloudPtr(),
                                  /* input_indices: */ remainder_indices_ptr,
                                  distanceThreshold );
            // accept plane?
            {
                std::cout << "plane" << planes.size() << " normal: " << plane.Normal().transpose() << ", d: " << plane.D() << std::endl;
            }

            planes.add( plane );
            std::cout << "plane has " << plane.countFinitePoints() << " finite points" << std::endl;
            std::cout << "planes.back() has " << planes.Primitives().back().countFinitePoints() << " finite points" << std::endl;

            // debug
            if ( debug )
            {
                am::CloudViewer viewer0("cuboid_from_cloud input indices vs. masked_indices", 2);

                PCloud::Ptr tmp_cloud_ptr( new PCloud() );
                am::DepthProcessor::extractCloud<MyPoint>( tmp_cloud_ptr,
                                                           cuboid.CloudPtr(),
                                                           remainder_indices_ptr );
                std::cout << "indices: " << remainder_indices_ptr->indices.size() << std::endl;
                viewer0.addCloud( tmp_cloud_ptr, "input", nullptr, viewer0.getVid(0), 10 );

                PCloud::Ptr tmp2_cloud_ptr( new PCloud() );
                am::DepthProcessor::extractCloud<MyPoint>( tmp2_cloud_ptr,
                                                           plane.CloudPtr(),
                                                           plane.IndicesPtr() );
                viewer0.addPointCloud( tmp2_cloud_ptr, "cuboid", nullptr, viewer0.getVid(1) );
                std::cout << "plane: " << plane.IndicesPtr()->indices.size() << std::endl;
                viewer0.getVPtr()->spin();
            }

            // select new plane outliers
            pcl::PointIndices::Ptr outlier_indices_ptr;
            plane.getOutlierIndices( /* out_indices: */ outlier_indices_ptr,
                                     /* input_cloud: */ plane.CloudPtr(),
                                     /*   threshold: */ distanceThreshold );

            // select outliers from all planes
            am::util::algo::vectorIntersection( /*  out: */ remainder_indices_ptr->indices,
                                                /* set1: */ remainder_indices_ptr->indices,
                                                /* set2: */ outlier_indices_ptr->indices  );

            // statistical clean
            //            pcl::PointIndices::Ptr removed_indices( new pcl::PointIndices );
            //            pcl::PointIndices::Ptr cleaned_indices( new pcl::PointIndices );
            //            DepthProcessor::statistical_clean<MyPoint>( /*     out clean cloud: */ cleaned_indices,
            //                                                        /*       in pointcloud: */ plane.CloudPtr(),
            //                                                        /*     roi of in cloud: */ &remainder_indices_ptr,
            //                                                        /* out removed indices: */ &removed_indices );
#if 0
            {
                am::CloudViewer tmpviewer("compare",2);

                PCloud::Ptr cleaned_points( new PCloud );
                DepthProcessor::extractCloud( cleaned_points, plane.CloudPtr(), cleaned_indices );
                tmpviewer.ViewerPtr()->addPointCloud( cleaned_points, "cleaned_points", tmpviewer.getVid(0) );
                std::cout << "cleaned_points count: " << cleaned_points->size() << std::endl;

                PCloud::Ptr orig_points( new PCloud );
                DepthProcessor::extractCloud( orig_points, plane.CloudPtr(), remainder_indices_ptr );
                tmpviewer.ViewerPtr()->addPointCloud( orig_points, "remainder", tmpviewer.getVid(1) );
                std::cout << "orig_points count: " << orig_points->size() << std::endl;

                tmpviewer.ViewerPtr()->spin();
            }


            // debug
            std::cout << "remainder_indices_ptr->indices.size(): "
                      << remainder_indices_ptr->indices.size()
                      << std::endl;
#endif

            // count rest, stop if too little
            if ( remainder_indices_ptr->indices.size() < cuboid.IndicesPtr()->indices.size() * min_percentage )
            {
                std::cout << "reached " << min_percentage*100. << "%, breaking" << std::endl;
                break;
            }
        }
#if 0
        // display
        {
            am::CloudViewer planeviewer("plane_viewer");
            planeviewer.ViewerPtr()->addPointCloud( cuboid.getROICloud(), "cuboid_cloud" );
            int plane_id = 0;
            for ( auto const& plane : planes.Primitives() )
            {
                std::cout << "plane has " << plane.countFinitePoints() << " finite points" << std::endl;

                char plane_name[255];
                sprintf( plane_name, "plane%02d", plane_id++ );
                plane.draw( planeviewer.ViewerPtr(), plane_name );

            }

            std::cout << "spinning" << std::endl;
            planeviewer.ViewerPtr()->spin();
            std::cout << "closing" << std::endl;
            planeviewer.ViewerPtr()->close();
            std::cout << "closed" << std::endl;
        }
#endif
    }



    int
    Planes::colourCloud( PCloud::Ptr cloud_ptr,
                         std::vector<cv::Point3f> const* arg_p_colors ) const
    {
        std::vector<cv::Point3f> const* p_colors = arg_p_colors
                                                   ? arg_p_colors
                                                   : new std::vector<cv::Point3f>( am::util::nColoursCv(6) );

        //        if ( arg_p_colors )
        //        {
        //            p_colors = arg_p_colors;
        //        }
        //        else
        //        {
        //            p_colors = new std::vector<cv::Point3f>();
        //            *p_colors = am::util::nColours( 6 );
        //        }

        // show planes
        for ( int plane_id = 0;
              plane_id < this->size();
              ++plane_id )
        {
            Plane const& plane = this->at( plane_id );

            // name
            char plane_name[255];
            {
                sprintf( plane_name, "plane%02d", plane_id );
            }

            // colour inliers
            {
                cv::Point3f c = p_colors->at(plane_id % p_colors->size());
                for ( int point_id = 0;
                      point_id < plane.IndicesPtr()->indices.size();
                      ++point_id )
                {
                    int idx = plane.IndicesPtr()->indices[ point_id ];
                    if ( idx < cloud_ptr->size() )
                    {
                        MyPoint &pnt = cloud_ptr->at( idx );
                        am::util::pcl::setPointColor( pnt, c.x, c.y, c.z ); // needs 255
                    }
                    else
                    {
                        std::cerr << __PRETTY_FUNCTION__ << " pnt_id > cloud_ptr->size(): " << idx << " >= " << cloud_ptr->size() << std::endl;
                    }
                }
            }
        }

        if ( !arg_p_colors )
        {
            delete p_colors;
        }
    }


    int
    Planes::readFromFile( Planes &planes, std::string const& path )
    {
        FILE *fp = fopen( path.c_str(), "rb" );
        if (!fp)
        {
            printf( "Planes::readFromFile failed, when opening %s\n", path.c_str() );
            return EXIT_FAILURE;
        }

        int count;
        int read_bytes = fread( &count, sizeof(int), 1, fp );
        if ( (read_bytes != 1) || (count < 1) )
        {
            std::cerr << "Planes::readFromfile failed, count < 1..." << std::endl;
            return EXIT_FAILURE;
        }

        float *coeffs = new float[ count * 4 ];
        read_bytes = fread( coeffs, sizeof(float), count * 4, fp );
        if ( read_bytes != count * 4 )
        {
            std::cerr << "[" << __func__ << "]: " << "not enough coeffs read: " << read_bytes << " is supposed to be " << count * 4 << std::endl;
        }

        int coeffs_index = 0;
        for ( int plane_id = 0; plane_id < count; ++plane_id )
        {
            am::Plane plane;
            plane.CoeffsPtrRef() = pcl::ModelCoefficients::Ptr( new pcl::ModelCoefficients() );
            plane.CoeffsPtrRef()->values.resize(4);
            for ( int i = 0; i < 4; ++i, ++coeffs_index )
                plane.CoeffsPtrRef()->values[i] = coeffs[ coeffs_index ];

            planes.add( plane );
        }

        if ( fp )
            fclose( fp );
        if ( coeffs )
            delete [] coeffs;

    }

    int
    Planes::writeToFile( Planes const& planes, std::string path )
    {
        // open file
        std::ofstream myfile;
        {
            myfile.open ( path, std::ios_base::binary | std::ios_base::out | std::ios_base::trunc );
            if ( !myfile.is_open() )
            {
                std::cerr << "could not open file " << path << std::endl;
                return EXIT_FAILURE;
            }
            std::cout << "writing planes to " << path << std::endl;
        }

        // write plane count
        int planes_count = planes.size();
        myfile.write( (char*) &planes_count, sizeof(int) );
        myfile.close();

        // write planes
        for ( am::Plane const& plane : planes.Primitives() )
        {
            am::Plane::writeToFile( plane, path, /* append: */ true );
        }

        // cleanup
        //myfile.close();
        return EXIT_SUCCESS;
    }

    int
    Planes::testReadWrite()
    {
        am::Planes planes;
        am::Plane plane;
        //cuboid.Corners().resize(8);
        plane.CoeffsPtrRef()->values = { 0.240607,-0.211485,2.41565, -0.00153208,-0.340846,2.7239};
        planes.add( plane );
        plane = am::Plane();
        plane.CoeffsPtrRef()->values = { 0.193476,0.628105,2.73098, -0.643829 };
        planes.add( plane );
        plane = am::Plane();
        plane.CoeffsPtrRef()->values = { -0.885968,-0.158347,2.10578, -123.456 };
        planes.add( plane );

        Planes::writeToFile( planes, "planes_test.dat" );
        Planes read_planes;
        Planes::readFromFile( read_planes, "planes_test.dat" );

        float diff = 0;
        for ( int plane_id = 0; plane_id < planes.size(); ++plane_id )
        {
            for ( int coeff_id = 0; coeff_id < 4; ++coeff_id )
            {
                diff += fabs( read_planes.at(plane_id).CoeffsPtrRef()->values[coeff_id] -
                                   planes.at(plane_id).CoeffsPtrRef()->values[coeff_id]   );
            }
        }
        if ( diff > FLT_EPSILON )
        {
            std::cerr << "Planes::testReadWrite failed!" << std::endl;
            for ( int plane_id = 0; plane_id < planes.size(); ++plane_id )
            {
                for ( int coeff_id = 0; coeff_id < 4; ++coeff_id )
                {
                    std::cerr << read_planes.at(plane_id).CoeffsPtrRef()->values[coeff_id] << " !=? " << planes.at(plane_id).CoeffsPtrRef()->values[coeff_id] << std::endl;
                }
            }
        }
        else
        {
            std::cout << "Planes::testReadWrite test PASSED...\n" << std::endl;
        }

        return diff >= FLT_EPSILON;
    }
} // ns am
