#include "CloudViewer.h"

#include "DepthProcessor.h"
#include "MyPointCloudGeometryHandlerCustom.h"

namespace am
{
    CloudViewer::CloudViewer(std::string title, int nviewports )
        : viewer_ptr_(new pcl::visualization::PCLVisualizer(title))
    {
        if ( nviewports > 2 )
        {
            std::cerr << "cloudviewer not implemented for sizes bigger than 2! " << std::endl;
        }

        _viewport_ids.resize( nviewports );

        int grid_width = 1;
        int grid_height = 1;
        while ( grid_width * grid_height < nviewports )
        {
            if ( grid_height < grid_width )
                grid_height++;
            else
                grid_width++;
        }
        double viewport_width  = 1. / grid_width;
        double viewport_height = 1. / grid_height;

        for ( int vid = 0; vid < _viewport_ids.size(); vid++ )
        {
            int vidx = vid % grid_width;
            int vidy = vid / grid_width;

            viewer_ptr_->createViewPort( viewport_width  *  vidx   ,
                                         1.f - viewport_height *  vidy   ,
                                         viewport_width  * (vidx+1),
                                         1.f - viewport_height * (vidy+1),
                                         _viewport_ids[vid] );

            //viewer_ptr_->setBackgroundColor( 0, 0, 0, getVid(vid) );
            viewer_ptr_->setBackgroundColor( .4, .4, .5, getVid(vid) );
            viewer_ptr_->setCameraClipDistances( 0.0001, 100.01, getVid(vid) );

            viewer_ptr_->addCoordinateSystem( 1.f, "coord_sys", getVid(vid) );

            //        viewer_ptr_->addCoordinateSystem(5.f, 0, 1, 0 );
            //        viewer_ptr_->addCoordinateSystem(3.f, 0, 0, 1 );
            //        viewer_ptr_->addCoordinateSystem(1.f, 0, 0, 0 );
            //viewer_ptr_->setPointCloudRenderingProperties (pcl::visualization::PCL_VISUALIZER_POINT_SIZE, 3 );

            //boost::function<void (const pcl::visualization::KeyboardEvent&, void*)> func_keyboard = boost::bind (&DepthViewer3D::keyboard_callback, this, _1, _2 );
            //viewer_ptr_->registerKeyboardCallback( depth_viewer3d_keyboard_callback, (void*)&viewer_ptr_ );
            //viewer_ptr_->registerMouseCallback   ( depth_viewer3d_mouse_callback   , (void*)this );
        }

        viewer_ptr_->setPosition( 0, 500 );
        viewer_ptr_->initCameraParameters();
        viewer_ptr_->setSize( 1280, 960 );
        viewer_ptr_->setShowFPS( true );


    }

    CloudViewer::~CloudViewer()
    {
        std::cout << "CloudViewer::~CloudViewer() closing window..." << std::endl;
        viewer_ptr_->close();
    }

    int
    CloudViewer::getVid( int vid )
    {
        return _viewport_ids[ vid ];
    }

    boost::shared_ptr<am::CloudViewer>
    CloudViewer::showCloud( PCloud::ConstPtr            cloud_ptr,
                            std::string                 title,
                            pcl::PointIndices::ConstPtr indices_ptr,
                            double                      point_size,
                            bool doSpin )
    {
        boost::shared_ptr<am::CloudViewer> viewer_ptr = boost::shared_ptr<am::CloudViewer>( new am::CloudViewer(title) );

        viewer_ptr->addPointCloud( cloud_ptr, title, indices_ptr, 0, point_size );
        viewer_ptr->getVPtr()->setBackgroundColor( .5, .5, .5 );

        if ( doSpin )
            viewer_ptr->getVPtr()->spin();

        return viewer_ptr;
    }

    boost::shared_ptr<am::CloudViewer>
    CloudViewer::showCloud( pcl::PointCloud<pcl::PointXYZRGB>::ConstPtr cloud_ptr,
                            std::string                 title,
                            pcl::PointIndices::ConstPtr indices_ptr,
                            double                      point_size,
                            bool doSpin )
    {
        boost::shared_ptr<am::CloudViewer> viewer_ptr = boost::shared_ptr<am::CloudViewer>( new am::CloudViewer(title) );

        if ( indices_ptr )
        {
            pcl::PointCloud<pcl::PointXYZRGB>::Ptr tmp_cloud_ptr( new pcl::PointCloud<pcl::PointXYZRGB>() );
            pcl::copyPointCloud( *cloud_ptr, indices_ptr->indices, *tmp_cloud_ptr );
            viewer_ptr->getVPtr()->addPointCloud( tmp_cloud_ptr, title, 0 );
        }
        else
        {
            viewer_ptr->getVPtr()->addPointCloud( cloud_ptr, title, 0 );
        }

        viewer_ptr->getVPtr()->setBackgroundColor( .5, .5, .5 );
        viewer_ptr->getVPtr()->setPointCloudRenderingProperties( pcl::visualization::PCL_VISUALIZER_POINT_SIZE, point_size, title, 0 );

        if ( doSpin )
            viewer_ptr->getVPtr()->spin();

        return viewer_ptr;
    }

    int
    CloudViewer::addPointCloud( PCloud::ConstPtr            cloud_ptr,
                                std::string                 title,
                                pcl::PointIndices::ConstPtr indices_ptr,
                                int                         viewport_id,
                                double                      point_size )
    {
        return addCloud( viewer_ptr_, cloud_ptr, title, indices_ptr, viewport_id, point_size );
    }

    int
    CloudViewer::addCloud( PCloud::ConstPtr            cloud_ptr,
                           std::string                 title,
                           pcl::PointIndices::ConstPtr indices_ptr,
                           int                         viewport_id,
                           double                      point_size )
    {
        return addCloud( viewer_ptr_,
                         cloud_ptr,
                         title,
                         indices_ptr,
                         viewport_id,
                         point_size );
    }

    int
    CloudViewer::addPointCloud( pcl::visualization::PCLVisualizer::Ptr viewer_ptr,
                                PCloud::ConstPtr            cloud_ptr,
                                std::string                 title,
                                pcl::PointIndices::ConstPtr indices_ptr,
                                int                         viewport_id,
                                double                      point_size )
    {
        return addCloud( viewer_ptr,
                         cloud_ptr,
                         title,
                         indices_ptr,
                         viewport_id,
                         point_size );
    }

    int
    CloudViewer::addCloud( pcl::visualization::PCLVisualizer::Ptr viewer_ptr,
                           PCloud::ConstPtr            cloud_ptr,
                           std::string                 title,
                           pcl::PointIndices::ConstPtr indices_ptr,
                           int                         viewport_id,
                           double                      point_size )
    {
        // roi cloud
        const int count = indices_ptr ? indices_ptr->indices.size() : cloud_ptr->size();
        pcl::PointCloud<pcl::PointXYZRGB>::Ptr tmp_cloud_ptr( new pcl::PointCloud<pcl::PointXYZRGB>() );
        tmp_cloud_ptr->reserve( count );

        // extract
        pcl::PointXYZRGB pnt;
        for ( int pid = 0; pid < count; ++pid )
        {
            // read
            auto tmp_pnt = cloud_ptr->at( indices_ptr ? indices_ptr->indices[pid] : pid );

            // copy
            pnt.x = tmp_pnt.x;
            pnt.y = tmp_pnt.y;
            pnt.z = tmp_pnt.z;
            pnt.rgb = tmp_pnt.rgb;

            // write
            tmp_cloud_ptr->push_back( pnt );
        }

        // action
        int ret = viewer_ptr->addPointCloud( tmp_cloud_ptr, title, viewport_id );

        // point size
        viewer_ptr->setPointCloudRenderingProperties( pcl::visualization::PCL_VISUALIZER_POINT_SIZE, point_size, title, viewport_id );

        // return
        return ret;
#if 0
        if ( indices_ptr )
        {
            pcl::visualization::PointCloudColorHandlerRGBField<MyPoint>  color_handler( tmp_cloud_ptr );
            //pcl::visualization::PointCloudGeometryHandlerCustom<MyPoint>::Ptr geom_handler
            //        (new pcl::visualization::PointCloudGeometryHandlerCustom<MyPoint>( tmp_cloud_ptr, "x","y","z" ) );
            am::MyPointCloudGeometryHandlerCustom<MyPoint> geom_handler ( tmp_cloud_ptr, "x","y","z" );
                   viewer_ptr->addPointCloud            ( tmp_cloud_ptr, *geom_handler , title, viewport_id );
            return viewer_ptr->updatePointCloud<MyPoint>( tmp_cloud_ptr, color_handler, title );
        }
        else
        {
            pcl::visualization::PointCloudColorHandlerRGBField<MyPoint>  color_handler( cloud_ptr );
            //pcl::visualization::PointCloudGeometryHandlerCustom<MyPoint> geom_handler ( cloud_ptr, "x","y","z" );
            am::PointCloudGeometryHandlerCustom<MyPoint> geom_handler ( cloud_ptr, "x","y","z" );
                   viewer_ptr->addPointCloud            ( cloud_ptr, geom_handler , title, viewport_id );
            return viewer_ptr->updatePointCloud<MyPoint>( cloud_ptr, color_handler, title );
        }
#endif
        //return viewer_ptr->addPointCloud( indices_ptr ? tmp_cloud_ptr : cloud_ptr, title, viewport_id );
    }

    int
    CloudViewer::updatePointCloud( pcl::visualization::PCLVisualizer::Ptr viewer_ptr,
                                       PCloud::ConstPtr            cloud_ptr,
                                       std::string                 title,
                                       pcl::PointIndices::ConstPtr indices_ptr,
                                       int                         viewport_id,
                                       double                      point_size )
    {
        // roi cloud
        const int count = indices_ptr ? indices_ptr->indices.size() : cloud_ptr->size();
        pcl::PointCloud<pcl::PointXYZRGB>::Ptr tmp_cloud_ptr( new pcl::PointCloud<pcl::PointXYZRGB>() );
        tmp_cloud_ptr->reserve( count );

        // extract
        pcl::PointXYZRGB pnt;
        for ( int pid = 0; pid < count; ++pid )
        {
            // read
            auto tmp_pnt = cloud_ptr->at( indices_ptr ? indices_ptr->indices[pid] : pid );

            // copy
            pnt.x = tmp_pnt.x;
            pnt.y = tmp_pnt.y;
            pnt.z = tmp_pnt.z;
            pnt.rgb = tmp_pnt.rgb;

            // write
            tmp_cloud_ptr->push_back( pnt );
        }

        // action
        int ret = viewer_ptr->updatePointCloud( tmp_cloud_ptr, title );

        // point size
        viewer_ptr->setPointCloudRenderingProperties( pcl::visualization::PCL_VISUALIZER_POINT_SIZE, point_size, title, viewport_id );

        // return
        return ret;
    }

    int
    CloudViewer::show2DFromLabels( PCloud::ConstPtr              cloud_ptr
                                   , int                         width
                                   , int                         height
                                   , std::string                 title
                                   , pcl::PointIndices::ConstPtr indices_ptr
                                   )
    {
        // roi cloud
        const int count = indices_ptr ? indices_ptr->indices.size() : cloud_ptr->size();
        cv::Mat img = cv::Mat::zeros( height, width, CV_8UC3 );

        // extract
        for ( int pid = 0; pid < count; ++pid )
        {
            // read
            auto tmp_pnt = cloud_ptr->at( indices_ptr ? indices_ptr->indices[pid] : pid );

            // copy
            int y = tmp_pnt.label / width;
            int x = tmp_pnt.label % width;

            // write
            img.at<uchar>( y, x*3   ) = tmp_pnt.r;
            img.at<uchar>( y, x*3+1 ) = tmp_pnt.g;
            img.at<uchar>( y, x*3+2 ) = tmp_pnt.b;
        }

        cv::imshow( title, img );
        int key = 0;
        while ( (key = cv::waitKey()) != 27 ) ;

        // return
        return EXIT_SUCCESS;
    }

} // ns am
