#include "cuboid_approx/proc3d/ContactsFiltering.h"

#include <sstream> // stringstream

namespace ca
{

    // (1) calculate below's top plane
    // (2) project above's centroid onto plane
    // (3) check, if projection inside selected plane boundaries
    ContactsFiltering::RELATIVE_POS
    ContactsFiltering::centroidAbove (   am::Cuboid         const& above
                                       , am::Cuboid         const& below
                                       , Eigen::Vector3f    const& up
                                       , float                   * angle )
    {
        typedef float Scalar;
        typedef Eigen::Matrix<Scalar,3,1> Vector3;

        int ret = NONE;

        Vector3 c_above = above.Center();
        Vector3 c_below = below.Center();

        Scalar cos_angle = up.dot( c_above - c_below );
        if ( angle )
        {
            *angle = acos( cos_angle );
            if ( *angle != *angle )
                *angle = Scalar(0);
        }

        // (0) is above at all?
        if ( cos_angle < Scalar(0) )
        {
            return static_cast<RELATIVE_POS>(ret);
        }

#if 0
        {
            std::stringstream ss;
            ss << the_dot;
            pcl::visualization::PCLVisualizer::Ptr vptr( new pcl::visualization::PCLVisualizer(ss.str()) );
            vptr->setBackgroundColor( .5, .6, .6 );
            above.draw( vptr, "above", {.0,.9,.0} );
            below.draw( vptr, "below", {.9,.0,.0} );
            auto tmp = (c_below + up );
            vptr->addArrow( pcl::PointXYZ( tmp(0), tmp(1), tmp(2))
                            , pcl::PointXYZ(c_below(0),c_below(1),c_below(2))
                            , 1., 0., 0.
                            , false
                            , "up"
                            , 0 );

            auto tmp2 = (c_above);
            vptr->addArrow( pcl::PointXYZ( tmp2(0), tmp2(1), tmp2(2))
                            , pcl::PointXYZ(c_below(0),c_below(1),c_below(2))
                            , 0., 1., 0.
                            , false
                            , "below_above"
                            , 0 );
            vptr->addSphere( pcl::PointXYZ(c_below(0),c_below(1),c_below(2))
                             , .1
                             , 1., 0., 0.,
                             "c_below_sphere"
                             , 0 );

            vptr->addSphere( pcl::PointXYZ(c_above(0),c_above(1),c_above(2))
                             , .1
                             , 0., 1., 0.,
                             "c_above_sphere"
                             , 0 );

            vptr->addArrow( pcl::PointXYZ( c_above(0), c_above(1), c_above(2) )
                            , pcl::PointXYZ(0,0,0)
                            , 0., 1., 0.
                            , false
                            , "c_above_arrow"
                            , 0 );

            vptr->spin();
        }
#endif

        // (1) calculate below's top plane
        // (1.1) project (plane_centroid - cuboid_centroid).normalized() onto up
        // (1.3) select max == smallest angle
        Scalar max_cos_theta       = Scalar(0);    // min_angle
        int    selected_plane_id   = 0;            // plane with min angle

        const int   planes_size         = below.Planes().size();
        for ( int plane_id = 1; plane_id != planes_size; ++plane_id )
        {
            // (1.1) project (plane_centroid - cuboid_centroid) onto up
            Vector3 vec         = (below.Planes().at( plane_id ).getCentroidFromCorners() - c_below).normalized();
            // (1.2) calculate angle
            Scalar  cos_theta   = vec.dot( up ); // |vec||up|cos_theta, |up| is the same for all, so can be omitted, |vec| == 1.

            // (1.3) select max == smallest angle
            if ( cos_theta > max_cos_theta )
            {
                max_cos_theta       = cos_theta;
                selected_plane_id   = plane_id;
            }
        }

        // cache plane
        am::Plane   const&  plane   = below.Planes().at( selected_plane_id );
        //Vector3     const   c_plane = plane.getCentroidFromCorners();
        //std::cout << "[" << __func__ << "]: " << "selected below.Planes().at( " << selected_plane_id << " ) with cos_theta " << max_cos_theta << std::endl;
        if ( up.dot( c_above - plane.getCentroidFromCorners() ) < Scalar(0) )
            return static_cast<RELATIVE_POS>(ret |= ABOVE);
        else
            ret |= ABOVE_TOP_PLANE;

#       warning "TODO: move to Plane.cpp"
        // (2) project above's centroid onto plane
        Vector3 c_above_on_plane;
        {
            // extract plane normal
            Vector3 plane_normal = Scalar(-1) * plane.Normal();
            if ( plane_normal.dot(up) < 0 )
            {
                std::cerr<< "[" << __func__ << "]: " << "WARNING, plane normal not pointing inwards, flipping..." << std::endl; fflush(stderr);
                plane_normal *= Scalar(-1);
            }

            // project
            c_above_on_plane = c_above - plane_normal * plane_normal.dot( c_above );
        }

#       warning "TODO: move to Plane.cpp"
        // (3) check, if projection inside selected plane boundaries
        const int corners_count = plane.Corners().size();
        if ( corners_count != 4 )
            std::cerr << "[" << __func__ << "]: " << "selected_plane.corners.size == " << corners_count << " != 4, continueing" << std::endl;
        bool inside = true;
        for ( int side_id = 0; side_id != corners_count && inside; ++side_id )
        {
            Vector3 toPoint = c_above_on_plane - plane.Corners().at(side_id).getVector3fMap();
            Vector3 side    = plane.Corners().at( (side_id + 1 == corners_count) ? 0 : side_id + 1 ).getVector3fMap()
                            - plane.Corners().at( side_id ).getVector3fMap();
            inside &= toPoint.dot( side ) > Scalar(0);
        }

        if ( inside )
            return static_cast<RELATIVE_POS>(ret |= SUPPORT );
        else
            return static_cast<RELATIVE_POS>(ret); // ABOVE_TOP_PLANE
    }

    int
    ContactsFiltering::filterContacts( am::Contact::Vec              & filtered_contacts
                                       , am::Contact::Vec       const& contacts
                                       , am::Cuboids            const& cuboids
                                       , std::map<int,am::Cuboid::TYPE> const& walls )
    {
#warning "TODO: keep one supporting edge out of many"
        typedef am::Cuboid::TypeMap::const_iterator const_iterator;
        typedef float Scalar;
        typedef Eigen::Matrix<Scalar,3,1> Vector3;

        std::cout << "[" << __func__ << "]: " << "filterContacts received " << cuboids.size() << " cuboids, and "
                  << contacts.size() << " contacts"
                  << ", and walls: " << walls.size()
                  << std::endl;

        bool keep = true;
        am::Contact::Vec deleted_contacts;

        //  box-box: is the top box's centroid above the bottom box? Y: keep contact
        // wall-box: is there anything below the box               ? Y: delete contact

        // cache end iterator
        const const_iterator walls_end = walls.end();

        // find floor:
        const_iterator  floor_it    = std::find_if( walls.begin(), walls.end(), [](std::pair<int,am::Cuboid::TYPE> const& entry)->bool { return entry.second == am::Cuboid::FLOOR; } );
        const Vector3   up          = cuboids.at( floor_it->first ).getOutwardsNormalOfLargestPlane();

        // stats
        std::vector<std::vector<int> > reasons( contacts.size() ); // contact_id did provide support to remove [ contact_id0, contact_id1, ...]

        // filter contacts
        for ( int contact_id = 0; contact_id != contacts.size(); ++contact_id )
        {
            keep = true;

            // always keep fixed
            if ( contacts[contact_id].type != am::Contact::FIXED )
            {

                const_iterator wall_it0 = walls.find( contacts[contact_id].ids[0] );
                const_iterator wall_it1 = walls.find( contacts[contact_id].ids[1] );
                const int box_in_contact_count = int(wall_it0 == walls_end) + int(wall_it1 == walls_end);
                std::cout << contacts[contact_id].toString() << ", boxes: " << box_in_contact_count << std::endl; fflush(stdout);

                // "box-box"
                if ( box_in_contact_count == 2 )
                {
                    int a = centroidAbove(
                                /* above: */ cuboids[contacts[contact_id].ids[0]]
                              , /* below: */ cuboids[contacts[contact_id].ids[1]]
                              , /*    up: */ up );
                    int b = centroidAbove(
                                /* above: */ cuboids[contacts[contact_id].ids[1]]
                              , /* below: */ cuboids[contacts[contact_id].ids[0]]
                              , /*    up: */ up );
                    keep = (a+b) & RELATIVE_POS::ABOVE;
                }
                // "wall-box" || "box-wall"
                else if ( box_in_contact_count == 1 )
                {
                    /// find a contact below that box

                    //const_iterator box_it = (wall_it1 != walls_end) ? wall_it2 : wall_it1;
                    // id box
                    const int box_in_contact = (wall_it0 == walls_end) ? 0 : 1;
                    const int box_id = contacts[contact_id].ids[box_in_contact];
                    std::cout << "wall" << contacts[contact_id].ids[!box_id] << " - box" << box_id << std::endl;

                    // find a box_id contact that's below box_id
                    for ( int i = 0; i != contacts.size() && keep; ++i )
                    {
                        // looking for another contact
                        if ( i == contact_id )
                            continue;

                        // get id of other box
                        int box2_id = -1;
                        if      ( contacts[i].ids[0] == box_id ) box2_id = contacts[i].ids[1];
                        else if ( contacts[i].ids[1] == box_id ) box2_id = contacts[i].ids[0];

                        // skip, if box_id not in contact
                        if ( box2_id < 0 )                       continue;

                        // filter, if box above
                        if ( centroidAbove( /*   above: */ cuboids.at(box_id)
                                            , /* below: */ cuboids.at(box2_id)
                                            , /*    up: */ up) )
                        {
                            keep = false;
                            reasons[ i ].push_back( contact_id );
                        }
                    }
                }
            } // if !FIXED

            if ( keep )
                filtered_contacts.push_back( contacts[contact_id] );
        }

        am::util::algo::vectorDifference( deleted_contacts, contacts, filtered_contacts );

        std::cout<<"[" << __func__ << "]: " << "kept_contacts:\n";
        for ( size_t vi = 0; vi != filtered_contacts.size(); ++vi )
            std::cout << filtered_contacts[vi].toString() << "\n";
        std::cout << "\n"; fflush(stdout);

        std::cout<<"[" << __func__ << "]: " << "deleted_contacts:\n";
        for ( size_t vi = 0; vi != deleted_contacts.size(); ++vi )
            std::cout << deleted_contacts[vi].toString() << "\n";
        std::cout << "\n"; fflush(stdout);

        std::cout<<"[" << __func__ << "]: " << "supporting contacts:\n";
        for ( size_t vi = 0; vi != reasons.size(); ++vi )
        {
            if ( !reasons[vi].size() ) continue;

            std::cout << "contact " << contacts[ vi ].toString() << " removed the following:\n";
            for ( size_t i = 0; i != reasons[vi].size(); ++i )
                std::cout << "\t" << contacts[ reasons[vi][i] ].toString() << "\n";
        }
        std::cout << "\n"; fflush(stdout);
//        am::Cuboid::TypeMap::const_iterator end_it = walls.end();
//        for ( am::Cuboid::TypeMap::const_iterator it = walls.begin(); it != end_it; ++it )
//        {
//            for ( int i = 0; )
//        }

        return EXIT_SUCCESS;
    } // ... filterContacts

} // ... namespace ca
