#include "DepthProcessor.h"

#include <pcl/common/common.h>
#include <pcl/common/transforms.h>
#include "pcl/filters/project_inliers.h"
#include <pcl/filters/voxel_grid.h>
#include "CloudViewer.h"

namespace am
{
    int
    DepthProcessor::cloud_2_vertex_list( Eigen::Matrix3Xf                            &vertex_list,
                                         PCloud::ConstPtr cloud        )
    {
        std::cout << "DepthProcessor::cloud2VertexList(): input cloud size is: " << cloud->size() << std::endl;
        vertex_list.resize( Eigen::NoChange, cloud->size() );

        int colId = 0;
        for ( auto it = cloud->begin(); it != cloud->end(); ++it, ++colId )
        {
            MyPoint pnt = *it;
            vertex_list( 0, colId ) = pnt.x;
            vertex_list( 1, colId ) = pnt.y;
            vertex_list( 2, colId ) = pnt.z;
        }
        std::cout << "DepthProcessor::cloud2VertexList(): " << colId << " points copied...\n";

        return EXIT_SUCCESS;
    }


    float kduToMM( const float depthInKDU,
                          const float dc1,
                          const float dc2)
    {
        return 1.0f / (dc2 * depthInKDU + dc1);
    }

    double
    disparity_to_depth( double disparity )
    {
        static const double CALIB_DC1 =  3.1635;
        static const double CALIB_DC2 = -0.0029;

        return kduToMM( disparity, CALIB_DC1, CALIB_DC2 );
    }


    bool
    DepthProcessor::isValidDepth( float d )
    {
        if ( d > 10.f )
            std::cerr << "wtf..." << d << std::endl;
        return d > .2f;
    }


    int
    DepthProcessor::projectPlaneInliers( PCloud::Ptr                       &inliers_cloud_ptr,
                                         PCloud::ConstPtr                   cloud_ptr,
                                         pcl::PointIndices::ConstPtr        indices_ptr,
                                         pcl::ModelCoefficients::ConstPtr   coeffs_ptr )

    {
        // project to plane
        if ( !inliers_cloud_ptr )
            inliers_cloud_ptr = PCloud::Ptr( new PCloud() );

        pcl::ProjectInliers<MyPoint> project_plane_inliers;
        project_plane_inliers.setModelType          ( pcl::SACMODEL_PLANE );
        project_plane_inliers.setInputCloud         ( cloud_ptr );
        if ( indices_ptr )
            project_plane_inliers.setIndices            ( indices_ptr );
        project_plane_inliers.setModelCoefficients  ( coeffs_ptr );
        project_plane_inliers.filter                ( *inliers_cloud_ptr );
    }

    int
    DepthProcessor::calculateExtremaInFrame( MyPoint                       &min_pt
                                             , MyPoint                     &max_pt
                                             , Eigen::Matrix4f       const &frame
                                             , PCloud::ConstPtr             cloud_ptr
                                             , pcl::PointIndices::ConstPtr  indices_ptr
                                             )
    {
        // move the points to the that reference frame
        PCloud cPoints;
        {
            Eigen::Matrix4f p2w( Eigen::Matrix4f::Identity() );
            p2w.block<3,3>(0,0) = frame.block<3,3>(0,0).transpose();
            p2w.block<3,1>(0,3) = -1.f * (p2w.block<3,3>(0,0) * frame.block<3,1>(0,3));

            if ( indices_ptr )
                pcl::transformPointCloud( *cloud_ptr, *indices_ptr, cPoints, p2w );
            else
                pcl::transformPointCloud( *cloud_ptr, cPoints, p2w );
        }

        // get min-max
        pcl::getMinMax3D( cPoints, min_pt, max_pt );
    }

    // transforms to local coordinates by inverting a coordinate frame
    int
    DepthProcessor::cloud2Local( PCloud::Ptr                 &cloud_ptr,
                                 Eigen::Matrix4f       const &frame,
                                 PCloud::ConstPtr             in_cloud_ptr,
                                 pcl::PointIndices::ConstPtr  indices_ptr )
    {
        if ( !cloud_ptr )   cloud_ptr = PCloud::Ptr( new PCloud() );

        Eigen::Matrix4f transform( Eigen::Matrix4f::Identity() );
        transform.block<3,3>(0,0) = frame.block<3,3>(0,0).transpose();
        transform.block<3,1>(0,3) = -1.f * (transform.block<3,3>(0,0) * frame.block<3,1>(0,3));

        if ( indices_ptr )
            pcl::transformPointCloud( *in_cloud_ptr, *indices_ptr, *cloud_ptr, transform );
        else
            pcl::transformPointCloud( *in_cloud_ptr, *cloud_ptr, transform );
    }

    // transforms to local coordinates by inverting a coordinate frame
    int
    DepthProcessor::local2World( PCloud::Ptr                 &cloud_ptr,
                                 Eigen::Matrix4f       const &frame,
                                 PCloud::ConstPtr             in_cloud_ptr,
                                 pcl::PointIndices::ConstPtr  indices_ptr )
    {
        if ( !cloud_ptr )   cloud_ptr = PCloud::Ptr( new PCloud() );

        if ( indices_ptr )
            pcl::transformPointCloud( *in_cloud_ptr, *indices_ptr, *cloud_ptr, frame);
        else
            pcl::transformPointCloud( *in_cloud_ptr, *cloud_ptr, frame );
    }

    // returns rowmajor eigen_vectors in decreasing order
    int
    DepthProcessor::PCA( Eigen::Matrix4f             &frame,
                         PCloud::ConstPtr             cloud_ptr,
                         pcl::PointIndices::ConstPtr  indices_ptr )

    {
        Eigen::Vector4f centroid;
        Eigen::Matrix3f covariance;
        if ( indices_ptr )
        {
            pcl::compute3DCentroid( /* cloud: */ *cloud_ptr, indices_ptr->indices, centroid );
            pcl::computeCovarianceMatrixNormalized( *cloud_ptr, *indices_ptr, centroid, covariance );
        }
        else
        {
            pcl::compute3DCentroid( /* cloud: */ *cloud_ptr, centroid );
            pcl::computeCovarianceMatrixNormalized( *cloud_ptr, centroid, covariance );
        }

        // eigen
        Eigen::SelfAdjointEigenSolver<Eigen::Matrix3f> eigen_solver( covariance, Eigen::ComputeEigenvectors );
        Eigen::Matrix3f eigen_vectors = eigen_solver.eigenvectors();
        am::util::eigen::sort_eigen_vectors_desc( eigen_solver.eigenvalues(),
                                                  eigen_vectors );
        eigen_vectors.col(2) = eigen_vectors.col(0).cross( eigen_vectors.col(1) );

        // debug
        if ( eigen_vectors(0,0) != eigen_vectors(0,0)
             || eigen_vectors(1,1) != eigen_vectors(1,1)
             || eigen_vectors(2,2) != eigen_vectors(2,2) )
        {
            std::cerr << "nan eigen matrix" << std::endl;
            return EXIT_FAILURE;
        }

        frame = Eigen::Matrix4f::Identity();
        frame.block<3,3>(0,0) = eigen_vectors;
        frame.block<3,1>(0,3) = centroid.head<3>();
        //min_pt = am::util::pcl::asPointXYZ( eigen_vectors * min_pt.getVector3fMap() + centroid.head<3>() );
        //max_pt = am::util::pcl::asPointXYZ( eigen_vectors * max_pt.getVector3fMap() + centroid.head<3>() );
    }


    int DepthProcessor::PCA_withPlane( Eigen::Matrix4f               & frame
                                       , Eigen::Vector4f        const& plane_coords
                                       , PCloud::ConstPtr              cloud_ptr
                                       , pcl::PointIndices::ConstPtr   indices_ptr
                                       , bool                          downsample
                                       , PCloud::Ptr                   *projected_cloud
                                       , bool                   const  debug )
    {
        // const bool debug = true;

        // Project to 2D
        am::PCloud::Ptr _2d_cloud_ptr( new PCloud() );
        {
            pcl::ModelCoefficients::Ptr plane_coeffs( new pcl::ModelCoefficients() );
            plane_coeffs->values = { plane_coords(0)
                                     , plane_coords(1)
                                     , plane_coords(2)
                                     , plane_coords(3) };

            am::DepthProcessor::projectPlaneInliers( _2d_cloud_ptr
                                                     , cloud_ptr
                                                     , indices_ptr
                                                     , plane_coeffs );

            if ( downsample )
            {
                // Create the filtering object
                pcl::VoxelGrid<pcl::PointXYZRGBL> sor;
                sor.setInputCloud( _2d_cloud_ptr );
                sor.setLeafSize  ( 0.02f, 0.02f, 0.02f );
                sor.filter       ( *_2d_cloud_ptr );
            }

            if ( projected_cloud )
                pcl::copyPointCloud( *_2d_cloud_ptr, *(*projected_cloud) );
        }

        auto vp = debug ? am::CloudViewer::showCloud( _2d_cloud_ptr, "tmpcloud", nullptr, 2., false ) : nullptr;

        // initial PCA
        am::DepthProcessor::PCA( frame, _2d_cloud_ptr, nullptr );
        // align to upvector
        {
            // second one is perp to the first one
            frame.block<3,1>(0,1) = frame.block<3,1>(0,2).cross( frame.block<3,1>(0,0) ).normalized();
            // third one is the upvector
            frame.block<3,1>(0,2) = plane_coords.head<3>();
        }

        // find best rotation, that yields the smallest 3D volume enclosed by min_pt and max_pt
        float step      = pcl::deg2rad( 5.f );
        float limits[2] = { 0.f, M_PI/2.f };

        // granularity of 2
        for ( int it = 0; it != 2; ++it, step /= 10.f )
        {
            std::pair <float,float> min_volume = { -1, FLT_MAX }; // <ang, volume>
            int i = 0;
            for ( float ang = limits[0]; ang < limits[1]; ang += step, ++i )
            {
                // rotated frame
                Eigen::Matrix4f tmp_frame = frame;

                // rotate frame around up_vector by ang
                Eigen::AngleAxisf rot( ang, tmp_frame.block<3,1>(0,2) );
                tmp_frame.block<3,1>(0,0) = rot * tmp_frame.block<3,1>(0,0);
                tmp_frame.block<3,1>(0,1) = rot * tmp_frame.block<3,1>(0,1);

                // calculate volume
                am::MyPoint min_pt, max_pt;
                am::DepthProcessor::calculateExtremaInFrame( min_pt, max_pt, tmp_frame, _2d_cloud_ptr, nullptr );

                Eigen::Vector3f diag    = max_pt.getVector3fMap() - min_pt.getVector3fMap();
                float           volume  = diag(0) * diag(1) * diag(2);
                // select min
                if ( volume < min_volume.second )
                {
                    min_volume.first = ang;
                    min_volume.second = volume;
                }
            }

            if ( debug )
                std::cout << "min_volume.first (angle): " << min_volume.first << " radians " << pcl::rad2deg(min_volume.first) << " degrees, "  << min_volume.second << " volume" << std::endl;

            // selected apply rotation
            Eigen::AngleAxisf rot( min_volume.first, frame.block<3,1>(0,2) );
            frame.block<3,1>(0,0) = rot * frame.block<3,1>(0,0);
            frame.block<3,1>(0,1) = rot * frame.block<3,1>(0,1);

            // modify lookup around chosen angle for next iteration
            limits[0] = -step/2.f;
            limits[1] = limits[0] + step;
        }

        // debug PCA
        if ( debug )
        {
            vp->VP()->addArrow(  am::util::pcl::asPointXYZ(frame.block<3,1>(0,3) + frame.block<3,1>(0,0))
                                , am::util::pcl::asPointXYZ(frame.block<3,1>(0,3))
                                , 1., 0., 0.
                                , false, "x" );
            vp->VP()->addArrow( am::util::pcl::asPointXYZ(frame.block<3,1>(0,3) + frame.block<3,1>(0,1))
                                , am::util::pcl::asPointXYZ(frame.block<3,1>(0,3))
                                , 0., 1., 0.
                                , false, "y" );
            vp->VP()->addArrow( am::util::pcl::asPointXYZ(frame.block<3,1>(0,3) + frame.block<3,1>(0,2))
                                , am::util::pcl::asPointXYZ(frame.block<3,1>(0,3))
                                , 0., 0., 1.
                                , false, "z" );
            vp->VP()->spin();
        }
    }
} // ns am

#if 0
    int
    DepthProcessor::plane_from_cloud( Plane                        &plane,
                                      PCloud::ConstPtr             cloud_ptr,
                                      pcl::PointIndices::Ptr const &indices )
    {
        int err = EXIT_SUCCESS;

        plane.CloudPtrRef()   = cloud_ptr;
        plane.IndicesPtrRef() = indices;

        pcl::PointCloud<pcl::PointXYZRGB>::Ptr plane_cloud_ptr;
        pcl::ModelCoefficients::Ptr            plane_coeffs;

        err = Segmentation::extract_plane<pcl::PointXYZRGB>( /*  out_cloud: */ plane_cloud_ptr,
                                                             /* out_coeffs: */ plane_coeffs,
                                                             /*   in_cloud: */ plane.getROICloud() );
        if ( err != EXIT_SUCCESS )
        {
            std::cerr << "Segmentation::extract_plane() returned error with " << (int)err << "\n";
            return err;
        }

        //planes.add( Plane( plane_coeffs, cloud_ptr) );
        //plane = Plane( plane_coeffs, cloud_ptr );
        plane.CoeffsPtrRef() = plane_coeffs;

        return err;
    }
#endif
