#include "MyVoxelGrid.h"

#include "Cuboid.h"
#include "Cuboids.h"

#include <queue>

namespace am
{
    /////////////////////////////////////////////////////////////////////
    /// MyVoxelGrid
    /////////////////////////////////////////////////////////////////////

    template <typename PointT>
    int
    countPointsInVoxels( std::vector<int>                         &      points_in_voxels
                         , boost::shared_ptr<const pcl::PointCloud<PointT> > cloud_ptr
                         , pcl::PointIndices::ConstPtr                   indices_ptr
                         , std::vector<int>                       &      leafs
                         , pcl::VoxelGrid<pcl::PointXYZRGB>       *const sor
                         , pcl::visualization::PCLVisualizer::Ptr        vptr )
    {
        const Eigen::Array3f  inverse_leaf_size = Eigen::Array3f::Ones() / sor->getLeafSize().array();
        const Eigen::Vector3i min_box_coords    = sor->getMinBoxCoordinates();
        const Eigen::Vector3i divb_mul          = sor->getDivisionMultiplier();

        // init
        points_in_voxels.resize( sor->getLeafLayout().size() );
        std::fill( points_in_voxels.begin(), points_in_voxels.end(), 0 );

        int ijk0, ijk1, ijk2;
        // all indices to voxels
        for ( auto it =  indices_ptr->indices.begin();
              it != indices_ptr->indices.end  ();
              ++it )
        {
            ijk0 = static_cast<int>( floor(cloud_ptr->points[*it].x * inverse_leaf_size[0])
                    - static_cast<float>( min_box_coords[0] )              );
            ijk1 = static_cast<int>( floor(cloud_ptr->points[*it].y * inverse_leaf_size[1])
                    - static_cast<float>( min_box_coords[1] )              );
            ijk2 = static_cast<int>( floor(cloud_ptr->points[*it].z * inverse_leaf_size[2])
                    - static_cast<float>( min_box_coords[2] )              );

            // Compute the centroid leaf index
            int idx = ijk0 * divb_mul[0] + ijk1 * divb_mul[1] + ijk2 * divb_mul[2];
            if ( leafs[idx] > -1 )
                ++points_in_voxels[ idx ];

#if 0
            //debug
            if ( vptr && (ijk0 == 1) && (ijk2 == 1) )
            {
                char tit[255];
                //sprintf( tit, "arrow_%d%d%d", std::distance(indices_ptr->indices.begin(), it) );
                int counter = 0;
                bool succ = false;
                do
                {
                    //sprintf( tit, "arrow_%d%d%d_%d", ijk0, ijk1, ijk2, counter++ );
                    sprintf( tit, "arrow_%d_%d", std::distance(indices_ptr->indices.begin(), it), counter++ );
                    succ = vptr->addArrow(
                               am::util::pcl::asPointXYZ( getCentroidCoord(ijk0, ijk1, ijk2, min_box_coords, sor->getLeafSize()) )
                               , cloud_ptr->points[*it]
                           , 1., 0., 1., false, tit );
                } while ( !succ && counter < 10 );
            }
#endif
        }

        return EXIT_SUCCESS;
    }

    template <typename PointT>
    int
    removeSlabFromCloud( pcl::PointIndices::Ptr                              &out_indices_ptr
                         , std::vector<int>                                  &leafs
                         , boost::shared_ptr<const pcl::PointCloud<PointT> > cloud_ptr
                         , pcl::PointIndices::ConstPtr                       indices_ptr
                         , MaxPlanes                                  const& max_planes
                         , Eigen::MatrixX2i                           const& minmax
                         , pcl::VoxelGrid<pcl::PointXYZRGB>                * sor )
    {
        const Eigen::Array3f  inverse_leaf_size = Eigen::Array3f::Ones() / sor->getLeafSize().array();
        const Eigen::Vector3i min_box_coords    = sor->getMinBoxCoordinates();
        const Eigen::Vector3i div_mul          = sor->getDivisionMultiplier();
        //    const Eigen::Vector3i divb_mul          = sor->getDivisionMultiplier();

        // init
        if ( !out_indices_ptr )     out_indices_ptr = pcl::PointIndices::Ptr( new pcl::PointIndices() );
        out_indices_ptr->indices.clear();
        out_indices_ptr->indices.reserve( indices_ptr->indices.size() );

        //std::vector<int> to_remove;
        Eigen::Vector3i ijk( 0, 0, 0 );
        // all indices to voxels
        for ( auto it =  indices_ptr->indices.begin();
              it != indices_ptr->indices.end  ();
              ++it )
        {
            ijk[0] = static_cast<int>( floor(cloud_ptr->points[*it].x * inverse_leaf_size[0])
                    - static_cast<float>( min_box_coords[0] )              );
            ijk[1] = static_cast<int>( floor(cloud_ptr->points[*it].y * inverse_leaf_size[1])
                    - static_cast<float>( min_box_coords[1] )              );
            ijk[2] = static_cast<int>( floor(cloud_ptr->points[*it].z * inverse_leaf_size[2])
                    - static_cast<float>( min_box_coords[2] )              );

            if ( leafs[ijk.dot(div_mul)] == -1 )
                std::cerr << "this shouldn't happen at " << ijk.transpose() << " == -1 " << std::endl;

            if (    (ijk[max_planes.max_dim_id] == minmax(0,0))
                    && (ijk[max_planes.dim1_id]    >= minmax(1,0))
                    && (ijk[max_planes.dim1_id]    <= minmax(1,1))
                    && (ijk[max_planes.dim2_id]    >= minmax(2,0))
                    && (ijk[max_planes.dim2_id]    <= minmax(2,1))
                    && (leafs[ijk.dot(div_mul)]   >  -1)                         )
            {
                //to_remove.push_back( ijk.dot(div_mul) );
                //leafs[ ijk.dot(divb_mul) ] = -1;
                continue;
            }

            out_indices_ptr->indices.push_back( *it );
        }

        // clean up leafs
        //    for ( auto const& idx : to_remove )
        //    {
        //        leafs[ idx ] = -2;
        //    }

        // update leafs (-1)
        for ( int c1 = minmax(1,0); c1 <= minmax(1,1); ++c1 )
            for ( int c2 = minmax(2,0); c2 <= minmax(2,1); ++c2 )
            {
                int idx = MyVoxelGrid::getLeafId( max_planes.getMaxDimCoord(), c1, c2, max_planes.dim1_id, max_planes.dim2_id, div_mul );
                leafs[ idx ] = -2;
            }

        return EXIT_SUCCESS;
    }

    int
    countPointsInPlanes( std::vector<int>                         &pointcounts
                         , std::vector<int>                 const &points_in_voxels
                         , pcl::VoxelGrid<pcl::PointXYZRGB> /*const*/ *sor          )
    {
        // cache
        const Eigen::Vector3i divs    = sor->getNrDivisions();
        const Eigen::Vector3i div_mul = sor->getDivisionMultiplier();

        // prepare
        pointcounts.resize( divs[0]+divs[1]+divs[2] );
        std::fill( pointcounts.begin(), pointcounts.end(), 0 );

        // gather
        Eigen::Vector3i ijk;
        for ( int i = 0; i < divs[0]; ++i )
            for ( int j = 0; j < divs[1]; ++j )
                for ( int k = 0; k < divs[2]; ++k )
                {
                    ijk << i,j,k;
                    int idx = ijk.dot( div_mul );

                    if ( (idx >= 0) && (idx < points_in_voxels.size()) )
                    {
                        pointcounts[                     i ] += points_in_voxels[ idx ];
                        pointcounts[ divs[0]           + j ] += points_in_voxels[ idx ];
                        pointcounts[ divs[0] + divs[1] + k ] += points_in_voxels[ idx ];
                    }
                    else
                        std::cerr<< "MyVoxelGrid::countPointInPlanes(): ijk " << i << "," << j << "," << k << " ->  idx " << idx << "out of range.." << std::endl;
                }
        return EXIT_SUCCESS;
    }

    int
    selectMaxPlanes( MaxPlanes                                    &max_planes
                     , std::vector<int>                   const   &pointcounts
                     , pcl::VoxelGrid<pcl::PointXYZRGB> /*const*/ *sor         )
    {
        const Eigen::Vector3i divs = sor->getNrDivisions();
        Eigen::Vector3i &max_point_count       = max_planes.max_point_count;
        Eigen::Vector3i &max_point_count_id    = max_planes.max_coords;
        int             &max_dim_id            = max_planes.max_dim_id;

        int offset = 0;
        // for every dimension
        for ( int dimid = 0; dimid < 3; ++dimid )
        {
            // for every plane along that axis
            for ( int pcid = 0; pcid < divs[dimid]; ++pcid )
                // check for max in dim
                if ( pointcounts[offset+pcid] > max_point_count[dimid] )
                {
                    max_point_count   [ dimid ] = pointcounts[ offset+pcid ];
                    max_point_count_id[ dimid ] = pcid;

                    // check for max over all dims
                    if ( max_point_count[ dimid ] > max_point_count[ max_dim_id ] )
                        max_dim_id = dimid;
                }
            offset += divs[ dimid ];
        }

        max_planes.dim1_id = (max_dim_id+1)%3;
        max_planes.dim2_id = (max_dim_id+2)%3;
        if ( max_planes.dim1_id > max_planes.dim2_id )
            std::swap( max_planes.dim1_id, max_planes.dim2_id );

        return EXIT_SUCCESS;
    }

    // \param[out] minmax(3x2) contains bounding box corner coords in order [ maxplane.minmax; dim1.minmax; dim2.minmax];
    int
    slabFromPlane2( Eigen::MatrixX2i                            & minmax
                    , MaxPlanes                           const& mp
                    , std::vector<int>                    const& leafs
                    , pcl::VoxelGrid<pcl::PointXYZRGB>         *const sor
                    , /*debug*/ pcl::visualization::PCLVisualizer::Ptr v_ptr )
    {
        const Eigen::Vector3i divs    = sor->getNrDivisions();
        const Eigen::Vector3i div_mul = sor->getDivisionMultiplier();

        const LeafIndexer leafIndexer( mp.max_dim_id, mp.dim1_id, mp.dim2_id, div_mul, mp.max_dim_id );

        minmax( 0, 0 ) = minmax( 0, 1 ) = mp.getMaxDimCoord();
        //int visited[ divs[mp.dim1_id] ][ divs[mp.dim2_id] ];
        Eigen::MatrixXi visited = Eigen::MatrixXi::Zero( divs[mp.dim1_id], divs[mp.dim2_id] );

        std::queue<Eigen::Vector3i> seeds;
        for ( int j = 0; j < divs[mp.dim1_id]; ++j )
            for ( int k = 0; k < divs[mp.dim2_id]; ++k )
            {
                if ( !visited(j,k) )
                {
                    int idx = leafIndexer.leafIndex( minmax(0,0), j, k );
                }
            }

        // while empty cell on level do
        {

        }

    }

    int
    slabFromPlane( Eigen::Matrix2i                            & minmax
                   , MaxPlanes                           const& mp
                   , std::vector<int>                    const& leafs
                   , pcl::VoxelGrid<pcl::PointXYZRGB>         *const sor
                   , /*debug*/ pcl::visualization::PCLVisualizer::Ptr v_ptr )
    {
        const Eigen::Vector3i divs    = sor->getNrDivisions();
        const Eigen::Vector3i div_mul = sor->getDivisionMultiplier();

        // select min max in the biggest plane

        Eigen::Vector3i ijk( 0,0,0 );
        ijk[ mp.max_dim_id ] = mp.max_coords[ mp.max_dim_id ];

        Eigen::Matrix2i limits;
        limits << 0, divs[mp.dim1_id], 0, divs[mp.dim2_id];

        bool restart;
        do
        {
            restart = false;

            minmax << limits(0,1), limits(0,0),
                    limits(1,1), limits(1,0); // mindim1,maxdim1,mindim2,maxdim2

            for ( int c1 = limits(0,0); c1 < limits(0,1) && !restart; ++c1 )
                for ( int c2 = limits(1,0); c2 < limits(1,1) && !restart; ++c2 )
                {
                    ijk[ mp.dim1_id ] = c1;
                    ijk[ mp.dim2_id ] = c2;
                    int idx = MyVoxelGrid::getLeafId( mp.getMaxDimCoord(), c1, c2, mp.dim1_id, mp.dim2_id, div_mul ); // ijk.dot( divb_mul );

                    if ( leafs[idx] > -1 )
                    {
                        // store c1 as min or max along dim1_id
                        if ( c1 < minmax(0,0) )    minmax(0,0) = c1;
                        if ( c1 > minmax(0,1) )    minmax(0,1) = c1;

                        // store c2 as min or max along dim2_id
                        if ( c2 < minmax(1,0) )    minmax(1,0) = c2;
                        if ( c2 > minmax(1,1) )    minmax(1,1) = c2;
                    }
                    else if ( leafs[idx] == -2 )
                    {
                        Eigen::Matrix2i new_limits = limits;

                        bool moved[2] = { false, false };
                        if ( c1 - limits(0,0) > limits(0,1) - c1 - 1 )
                        {
                            if ( leafs[MyVoxelGrid::getLeafId(mp.getMaxDimCoord(), c1-1, c2, mp.dim1_id, mp.dim2_id, div_mul)] > -2 )
                            {
                                new_limits(0,1) = c1;
                                {
                                    std::cout << "ax" << mp.dim1_id << ": changed upperbound " << limits(0,1) << " to " << c1 << std::endl;
                                    moved[0] = true;

                                    Eigen::Vector3i ijk2( mp.getMaxDimCoord(), mp.getMaxDimCoord(), mp.getMaxDimCoord() );
                                    ijk2[ mp.dim1_id ] = c1-1;
                                    ijk2[ mp.dim2_id ] = c2;

                                    char str[255];
                                    sprintf( str, "(%d,%d,%d)=%d", ijk2(0),ijk2(1),ijk2(2),leafs[ ijk2.dot(div_mul) ] );

                                    std::string id = MyVoxelGrid::getCellGridId(ijk2);
                                    v_ptr->removeText3D( id );
                                    v_ptr->addText3D( /*  text: */   std::string(str)
                                                      /*   pos: */ , am::util::pcl::asPointXYZ(MyVoxelGrid::getCentroidCoord(ijk2,sor))
                                                      /* scale: */ , 0.01
                                                      /*   red: */ , 0.
                                                      /* green: */ , 0.
                                                      /*  blue: */ , 1.
                                                      /* label: */ , id );
                                }
                            }
                        }
                        else
                        {
                            if ( leafs[MyVoxelGrid::getLeafId(mp.getMaxDimCoord(), c1+1, c2, mp.dim1_id, mp.dim2_id, div_mul)] > -2 )
                            {
                                new_limits(0,0) = c1+1;

                                {
                                    std::cout << "ax" << mp.dim1_id << ": changed lowerbound " << limits(0,0) << " to " << c1+1 << std::endl;
                                    moved[0] = true;

                                    Eigen::Vector3i ijk2( mp.getMaxDimCoord(), mp.getMaxDimCoord(), mp.getMaxDimCoord() );
                                    ijk2[ mp.dim1_id ] = c1+1;
                                    ijk2[ mp.dim2_id ] = c2;

                                    char str[255];
                                    sprintf( str, "(%d,%d,%d)=%d", ijk2(0),ijk2(1),ijk2(2),leafs[ ijk2.dot(div_mul) ] );

                                    std::string id = MyVoxelGrid::getCellGridId(ijk2);
                                    v_ptr->removeText3D( id );
                                    v_ptr->addText3D( /*  text: */   std::string(str)
                                                      /*   pos: */ , am::util::pcl::asPointXYZ(MyVoxelGrid::getCentroidCoord(ijk2,sor))
                                                      /* scale: */ , 0.01
                                                      /*   red: */ , 0.
                                                      /* green: */ , 0.
                                                      /*  blue: */ , 1.
                                                      /* label: */ , id );
                                }
                            }
                        }
                        if ( c2 - limits(1,0) > limits(1,1) - c2 - 1 )
                        {
                            if ( leafs[MyVoxelGrid::getLeafId(mp.getMaxDimCoord(), c1, c2-1, mp.dim1_id, mp.dim2_id, div_mul)] > -2 )
                            {
                                new_limits(1,1) = c2;
                                {
                                    std::cout << "ax" << mp.dim2_id << ": changed upperbound " << limits(1,1) << " to " << c2 << std::endl;
                                    moved[1] = true;

                                    Eigen::Vector3i ijk2( mp.getMaxDimCoord(), mp.getMaxDimCoord(), mp.getMaxDimCoord() );
                                    ijk2[ mp.dim1_id ] = c1;
                                    ijk2[ mp.dim2_id ] = c2-1;

                                    char str[255];
                                    sprintf( str, "(%d,%d,%d)=%d", ijk2(0),ijk2(1),ijk2(2),leafs[ ijk2.dot(div_mul) ] );

                                    std::string id = MyVoxelGrid::getCellGridId(ijk2);
                                    v_ptr->removeText3D( id );
                                    v_ptr->addText3D( /*  text: */   std::string(str)
                                                      /*   pos: */ , am::util::pcl::asPointXYZ(MyVoxelGrid::getCentroidCoord(ijk2,sor))
                                                      /* scale: */ , 0.01
                                                      /*   red: */ , 0.
                                                      /* green: */ , 0.
                                                      /*  blue: */ , 1.
                                                      /* label: */ , id );
                                }
                            }
                        }
                        else
                        {
                            if ( leafs[MyVoxelGrid::getLeafId(mp.getMaxDimCoord(), c1, c2+1, mp.dim1_id, mp.dim2_id, div_mul)] > -2 )
                            {
                                new_limits(1,0) = c2+1;
                                {
                                    std::cout << "ax" << mp.dim2_id << ": changed lowerbound " << limits(1,0) << " to " << c2+1 << std::endl;
                                    moved[1] = true;

                                    Eigen::Vector3i ijk2( mp.getMaxDimCoord(), mp.getMaxDimCoord(), mp.getMaxDimCoord() );
                                    ijk2[ mp.dim1_id ] = c1;
                                    ijk2[ mp.dim2_id ] = c2+1;

                                    char str[255];
                                    sprintf( str, "(%d,%d,%d)=%d", ijk2(0),ijk2(1),ijk2(2),leafs[ ijk2.dot(div_mul) ] );

                                    std::string id = MyVoxelGrid::getCellGridId(ijk2);
                                    v_ptr->removeText3D( id );
                                    v_ptr->addText3D( /*  text: */   std::string(str)
                                                      /*   pos: */ , am::util::pcl::asPointXYZ(MyVoxelGrid::getCentroidCoord(ijk2,sor))
                                                      /* scale: */ , 0.01
                                                      /*   red: */ , 0.
                                                      /* green: */ , 0.
                                                      /*  blue: */ , 1.
                                                      /* label: */ , id );
                                }
                            }
                        }

                        if ( moved[0] && moved[1] )
                            std::cerr << "moved both!! " << std::endl;
                        else
                            std::cout << "moved " << (moved[0] ? "c1" : "c2") << std::endl;

                        limits  = new_limits;
                        std::cout << "new limits: " << limits.array() << std::endl;
                        restart = true;

                        // debug
                        char str[255];
                        sprintf( str, "(%d,%d,%d)=-2", ijk(0),ijk(1),ijk(2) );

                        std::string id = MyVoxelGrid::getCellGridId(ijk);
                        v_ptr->removeText3D( id );
                        v_ptr->addText3D( /*  text: */   std::string(str)
                                          /*   pos: */ , am::util::pcl::asPointXYZ(MyVoxelGrid::getCentroidCoord(ijk,sor))
                                          /* scale: */ , 0.01
                                          /*   red: */ , 1.
                                          /* green: */ , 0.
                                          /*  blue: */ , 0.
                                          /* label: */ , id );
                    }
                }
        } while ( restart && (limits(0,0) < limits(0,1)) && (limits(1,0) < limits(1,1)) );

        std::cout << "minmax: " << minmax << std::endl;
        return EXIT_SUCCESS;
    }

    int
    slabFromPlaneRec( std::vector<Eigen::MatrixX2i>              & out_slabs // contains slab min,max coords in [ maxdim, dim1, dim2 ] order
                      , Eigen::MatrixX2i                    const& limits    // contains limits in [ maxdim,dim1,dim2] order, inclusive!
                      , MaxPlanes                           const& mp
                      , std::vector<int>                    const& leafs
                      , pcl::VoxelGrid<pcl::PointXYZRGB>         *const sor
                      , /*debug*/ pcl::visualization::PCLVisualizer::Ptr v_ptr
                      , int level = 0)
    {

        //std::cout << "L" << level << " slabFromPlaneRec() with limits\n" << limits << std::endl;
        const bool debug = false;

        if ( out_slabs.size() )
        {
            std::cerr << "slabFromPlaneRec() cleaning out_slabs, but not empty...are you sure this is ok?" << std::endl;
            out_slabs.clear();
        }

        const LeafIndexer leafIndexer( mp.max_dim_id, mp.dim1_id, mp.dim2_id, sor->getDivisionMultiplier(), mp.getMaxDimCoord() );

        if ( (level==0) || debug )
        {
            std::cout << "leafs("<<level<<"):\n";
            for ( int j = limits(1,0); j <= limits(1,1); ++j )
            {
                for ( int k = limits(2,0); k <= limits(2,1); ++k )
                {
                    std::cout << leafs[ leafIndexer.leafIndex(j,k) ] << ", ";
                }
                std::cout << "\n";
            }
        }

        bool split = false;
        Eigen::MatrixX2i slab( 3, 2 );
        slab << limits(0,0), limits(0,1), // min_max_dim = max_max_dim
                limits(1,1), limits(1,0), // mindim1,maxdim1,
                limits(2,1), limits(2,0); // mindim2,maxdim2
        if ( debug )
            std::cout << __func__ << "(" << level << "): IN\n" << slab.transpose() << std::endl;
        const std::vector<Eigen::Vector3i> dirs = { {0,1,0}, {0,-1,0}, {0,0,1}, {0,0,-1},
                                                    {0,1,1}, {0,-1,-1}, {0,-1,1}, {0,1,-1} };

        const int c0 = limits(0,0); // looking in 2D
        for ( int c1 = limits(1,0); c1 <= limits(1,1) && !split; ++c1 )
        {
            for ( int c2 = limits(2,0); c2 <= limits(2,1) && !split; ++c2 )
            {
                int idx = leafIndexer.leafIndex( c1, c2 );

                // if free cell
                if ( leafs[idx] > -1 )
                {
                    // store c1 as min or max along dim1_id
                    if ( c1 < slab(1,0) )
                        slab(1,0) = c1;
                    if ( c1 > slab(1,1) )
                        slab(1,1) = c1;

                    // store c2 as min or max along dim2_id
                    if ( c2 < slab(2,0) )
                        slab(2,0) = c2;
                    if ( c2 > slab(2,1) )
                        slab(2,1) = c2;
                }
                else if ( leafs[idx] == -2 ) // split
                {
                    split = true;
                    if ( debug )
                        std::cout << "-2 at " << c0 << "," << c1 << "," << c2 << std::endl;

                    // look for first free cell in each direction
                    for ( int dir_id = 0; dir_id < dirs.size(); ++dir_id )
                    {
                        Eigen::Vector3i new_coord( c0, c1, c2 );
                        int  idx2   = idx;
                        bool inside = true;
                        while ( inside && (leafs[idx2] == -2) )
                        {
                            // move
                            new_coord += dirs[ dir_id ];

                            // check boundaries
                            for ( int dim = 0; (dim < 3) && inside; ++dim )
                            {
                                inside &= (new_coord(dim) >= limits(dim,0));
                                inside &= (new_coord(dim) <= limits(dim,1));
                            }

                            // lookup
                            idx2 = leafIndexer.leafIndex( new_coord(1), new_coord(2) );
                        }

                        // skip, if outside
                        if ( !inside )
                            continue;

                        // found a free cell, prepare recurse
                        Eigen::MatrixX2i new_limits = limits;
                        for ( int dim = 0; dim < 3; ++dim )
                            if ( dirs[dir_id](dim) > 0 ) // looked forward, move lower limit
                                new_limits( dim, 0 ) = new_coord( dim );
                            else if ( dirs[dir_id](dim) < 0 ) // looked backward, move upper limit
                                new_limits( dim, 1 ) = new_coord( dim );

                        // check limits
                        for ( int dim = 0; dim < 3; ++dim )
                            if ( new_limits(dim,0) > new_limits(dim,1) )
                            {
                                std::cerr << __func__ << ": NOOOO, this shouldn't happen...( new_limits("<<dim<<",0) < new_limits("<<dim<<",1) )";
                                continue;
                            }
                        // recurse
                        std::vector<Eigen::MatrixX2i> slabs;
                        slabFromPlaneRec( slabs,
                                          new_limits,
                                          mp,
                                          leafs,
                                          sor,
                                          v_ptr
                                          , level+1 );
                        // output
                        out_slabs.insert( out_slabs.end(), slabs.begin(), slabs.end() );
                        // sort output decreasing
                        std::sort( out_slabs.begin(), out_slabs.end(), MyVoxelGrid::compareSlabsByVolume );
                    }
                }
            } // for c2
        } // for c1

        if ( !split )
        {
            bool positive_volume = true;
            for ( int dim = 0; dim < 3; ++dim )
                positive_volume &= (slab(dim,0) <= slab(dim,1)) ;
            if ( positive_volume )
                out_slabs.push_back( slab );
        }

        if ( debug )
        {
            std::cout << __func__ << "(" << level << "): OUT\n";
            for ( auto &slab : out_slabs )
                std::cout << slab.transpose() << std::endl;
            std::cout << std::endl;
        }

        return EXIT_SUCCESS;
    }

    int
    MyVoxelGrid::segment( am::Cuboids                 & cuboids
                          , am::PCloud::ConstPtr        cloud_ptr
                          , Eigen::Matrix4f      const& frame
                          , pcl::PointIndices::ConstPtr indices_ptr
                          , float                       resolution
                          , float                       threshold
                          , int                         max_slab_count
                          , bool                        debug
                          )
    {
        // CONSTs
        const auto colours         = am::util::nColoursEigen( 25, 1. );
        const int  slab_size_limit = indices_ptr ? (indices_ptr->indices.size() * threshold) : (cloud_ptr->size() * threshold);
        std::cout << __func__ << ": slab_size_limit: " << slab_size_limit << " points" << std::endl;

        // LOCAL
        PCloud::Ptr local_cloud_ptr;
        {
            am::DepthProcessor::cloud2Local( local_cloud_ptr
                                             , frame
                                             , cloud_ptr
                                             , indices_ptr
                                             );
        }

        // show
        boost::shared_ptr<am::CloudViewer> vptr = nullptr;
        if ( debug )
        {
            vptr = am::CloudViewer::showCloud( local_cloud_ptr, "localcloud", nullptr, 2., false );
            vptr->ViewerPtr()->setBackgroundColor( 1.f, 1.f, 1.f );
        }

        // strip
        auto xyzrgb_cloud_ptr = am::util::pcl::toXYZRGBCloud<am::PCloud::Ptr>( local_cloud_ptr );

        //am::CloudViewer::showCloud( xyzrgb_cloud_ptr, "xyzrgb_cloud_ptr", nullptr, 2., true );

        pcl::PointIndices::Ptr remainder_indices_ptr( am::util::pcl::allIndicesOf(xyzrgb_cloud_ptr) );
        //am::CloudViewer::showCloud( xyzrgb_cloud_ptr, "xyzrgb_cloud_ptr", remainder_indices_ptr, 2., true );

        // Create the filtering object
        pcl::VoxelGrid<pcl::PointXYZRGB> sor;
        std::vector<int>                 leafs;
        {
            pcl::PointCloud<pcl::PointXYZRGB>::Ptr cloud_filtered_ptr( new pcl::PointCloud<pcl::PointXYZRGB> );
            sor.setInputCloud    ( xyzrgb_cloud_ptr      );
            sor.setIndices       ( remainder_indices_ptr );
            sor.setLeafSize      ( resolution, resolution, resolution );
            sor.setSaveLeafLayout( true                  );
            sor.filter           ( *cloud_filtered_ptr   );

            // store leafs if first grid
            leafs = sor.getLeafLayout();
        }
        //MyVoxelGrid::drawGrid<pcl::PointXYZRGB>( vptr->VP()
        //                                         , leafs
        //                                         , &sor );

        std::pair<float,int> filledness_statistics = { 0, 0 }; // < sum, count >

#if 0
        // debug
        {
            std::vector<Eigen::Vector3f> pos_focal_up = {  {	1.23439f, -1.39432f, -1.37946f }
                                                          ,{ -0.127083f, 0.305916f, -0.067527f}
                                                          ,{ -0.377182f, 0.357282f, -0.854449f } };
            am::util::pcl::setCameraPosFocalUp( vptr->ViewerPtr(), pos_focal_up );
            drawRealGrid( vptr->ViewerPtr(), leafs, &sor );
            vptr->ViewerPtr()->spin();
            return 0;
        }
#endif


        std::vector<std::pair<Slab,MaxPlanes>> out_slabs;
        int added_slab_id = 0;
        int max_slab_size = 0;
        do
        {
            std::cout << "continuing, since max_slab_size > limit: " << max_slab_size << " > " << slab_size_limit << std::endl;
            max_slab_size = 0;

            // pointcount in leafs
            std::vector<int> points_in_voxels; // size: d1 * d2 * d3
            {
                countPointsInVoxels<pcl::PointXYZRGB>( points_in_voxels
                                                       , xyzrgb_cloud_ptr
                                                       , remainder_indices_ptr
                                                       , leafs
                                                       , &sor
                                                       , debug ? vptr->VP() : nullptr );
            }

            std::vector<int> points_in_planes; // size: d1 + d2 + d3
            {
                countPointsInPlanes( points_in_planes
                                     , points_in_voxels
                                     , &sor );
            }

            // select max dims
            MaxPlanes max_planes;
            {
                selectMaxPlanes( max_planes
                                 , points_in_planes
                                 , &sor );
            }

            // debug
            if ( debug )
            {
                am::util::printVector( points_in_planes, "pointcounts" );

                std::cout << "maxdim: "          << max_planes.max_dim_id
                          << " with "            << max_planes.max_point_count[ max_planes.max_dim_id ]
                          << " points at coord " << max_planes.getMaxDimCoord()
                          << std::endl;
            }




            // select min max in this plane
            std::vector<Eigen::MatrixX2i> slabs;
            {
                Eigen::MatrixX2i limits( 3, 2 );
                limits << max_planes.getMaxDimCoord(), max_planes.getMaxDimCoord(),
                          0                          , sor.getNrDivisions()(max_planes.dim1_id)-1, // inclusive!
                          0                          , sor.getNrDivisions()(max_planes.dim2_id)-1; // inclusive!
                slabFromPlaneRec( slabs
                                  , limits
                                  , max_planes
                                  , leafs
                                  , &sor
                                  /* debug: */ , debug ? vptr->VP() : nullptr );
                if ( debug )
                    for ( int i = 0; i < slabs.size(); ++i )
                        std::cout << "slabs[" << i << "]:\n" << slabs[i] << std::endl;
            }

            // split slab?
            if ( 0 )
            {
                std::vector<Eigen::MatrixX2i> tmp_slabs;
                MyVoxelGrid::splitSlabs( tmp_slabs
                                        , slabs
                                        , .1f
                                        , leafs
                                        , points_in_voxels
                                        , max_planes
                                        , &sor );
                slabs = tmp_slabs;
            }

            {
                std::vector<Eigen::MatrixX2i> tmp_slabs;
                MyVoxelGrid::splitSlabs2( tmp_slabs
                                        , slabs
                                        , .65f
                                        , leafs
                                        , points_in_voxels
                                        , max_planes
                                        , &sor
                                        , vptr?vptr->ViewerPtr():nullptr
                                        );
                slabs = tmp_slabs;
            }

            // DEBUG
            if ( 0 )
            {
                am::CloudViewer v( "splitslabs" );
                v.VP()->addPointCloud( xyzrgb_cloud_ptr, "thecloud" );
                int l_slab_id = 0;
                for ( auto const& slab : slabs )
                {
                    drawSlab( v.ViewerPtr(), slab, max_planes, &sor, am::util::sprintf<int>("slab%d", l_slab_id) );
                    ++l_slab_id;
                }

                v.ViewerPtr()->spin();
            }

            // remove slab's points
            {
                for ( int l_slab_id = 0; l_slab_id < slabs.size(); ++l_slab_id )
                {
                    auto & slab = slabs[ l_slab_id ];

                    pcl::PointIndices::Ptr tmp_indices_ptr;
                    removeSlabFromCloud<pcl::PointXYZRGB>( tmp_indices_ptr
                                                           , leafs
                                                           , xyzrgb_cloud_ptr
                                                           , remainder_indices_ptr
                                                           , max_planes
                                                           , slab
                                                           , &sor                  );
                    //am::CloudViewer::showCloud( xyzrgb_cloud_p0.00720685  0.393566 -0.919268tr, "removed", tmp_indices_ptr, false );

                    int point_count_diff = remainder_indices_ptr->indices.size() - tmp_indices_ptr->indices.size();
                    std::cout << "SLAB size: " << point_count_diff << std::endl;

                    if ( point_count_diff > max_slab_size )
                        max_slab_size = point_count_diff;

                    std::cout << "point_count_diff / (float)getVolume(slab): " << point_count_diff / (float)getVolume(slab)
                              << " vs " << filledness_statistics.first / filledness_statistics.second
                              << std::endl;

#if 0
                    c
                         || ( (point_count_diff / (float)getVolume(slab)) < (filledness_statistics.first / filledness_statistics.second)/10.f )  )
#else
                    if ( point_count_diff < std::max(20,slab_size_limit) )
#endif
                    {
                        // DEBUG
                        if ( 0 )
                        {
                            am::CloudViewer v;
                            v.VP()->addPointCloud( xyzrgb_cloud_ptr, "thecloud" );
                            drawSlab( v.ViewerPtr(), slab, max_planes, &sor, "slab" );
                            v.ViewerPtr()->spin();
                        }
                        std::cout << "skipping slab with size " << point_count_diff << std::endl;
                        continue;
                    }

                    filledness_statistics.first += point_count_diff / (float)MyVoxelGrid::getVolume(slab);
                    ++filledness_statistics.second;


                    // add slab to output
                    {
                        //Slab out_slab( 3, 2 );
                        //out_slab.row( max_planes.max_dim_id ) = slab.row( 0 );
                        //out_slab.row( max_planes.dim1_id    ) = slab.row( 1 );
                        //out_slab.row( max_planes.dim2_id    ) = slab.row( 2 );
                        out_slabs.push_back( {slab, max_planes} );
                    }

                    // shrink
                    am::Cuboid cuboid_non_shrunk, cuboid;
                    //if ( 1 )
                    int err = EXIT_SUCCESS;
                    {
                        //auto tmp_slab = slab;
                        err = MyVoxelGrid::shrinkSlabInsideGrid<pcl::PointXYZRGB>( cuboid
                                                                             , slabs[ l_slab_id ]
                                                                             , max_planes
                                                                             , leafs
                                                                             , xyzrgb_cloud_ptr
                                                                             , remainder_indices_ptr
                                                                             , &sor );
                        //slab = tmp_slab;
                    }

                    //else
                    if ( debug )
                    {
                        //MyVoxelGrid::drawSlab( vptr->VP(), slabs[l_slab_id] /*minmax*/, max_planes, &sor, slab_name, colours[slab_id] );

                        MyVoxelGrid::cuboidFromSlab( /*       out_cuboid: */ cuboid_non_shrunk
                                                     , /*        in_slab: */ slabs[ l_slab_id ]
                                                     , /* voxelgrid_info: */ max_planes
                                                     , /*  voxelgrid_pcl: */ &sor              );
                    }

                    // draw
                    if ( debug )
                    {
                        char slab_name[255];
                        sprintf( slab_name, "slab%d", added_slab_id );
//                        vptr->VP()->addText3D( slab_name
//                                               , cuboid_non_shrunk.Corners().back()
//                                               , .01
//                                               , .8, .1, .1
//                                               , std::string(slab_name) + "_title" );

                        cuboid_non_shrunk.draw( vptr->VP(), slab_name                      , colours[added_slab_id%colours.size()], 4. );
                        cuboid           .draw( vptr->VP(), slab_name+std::string("shrunk"), (Eigen::Vector3f() << 1.f, 0.f, 0.f).finished(), 2. );
                    }

                    ++added_slab_id;

                    // back to world and store
                    for ( int corner_id = 0; corner_id != cuboid.Corners().size(); ++corner_id )
                    {
                        cuboid.Corners()[ corner_id ] = DepthProcessor::local2World( cuboid.Corners()[corner_id], frame );
                    }
                    cuboids.add( cuboid );

                    remainder_indices_ptr = tmp_indices_ptr;
                }
            }

            if ( debug )
            {
                vptr->VP()->spin();
                auto posfocalups = am::util::pcl::getCamerasPosFocalUp( vptr->VP() );
                am::util::printVector( posfocalups[0], "posfocalup" );
            }

            // am::CloudViewer::showCloud( cloud_filtered_ptr, "voxelized", nullptr, 5., true );
        } while (    (added_slab_id       < max_slab_count )
                  && (max_slab_size > slab_size_limit) );

        // add slabs to cuboids
        {
//            std::vector<Slab> tmp_slabs;
//            for ( auto const& slab_with_mp : out_slabs )
//                tmp_slabs.push_back( slab_with_mp.first );

            std::vector<std::pair<Slab,MaxPlanes> > final_slabs;
            MyVoxelGrid::mergeSlabs( final_slabs
                                     , out_slabs
                                     , 1.f
                                     , leafs
                                     //, points_in_voxels
                                     //, MaxPlanes()
                                     , &sor );

            auto vptr2 = debug ? am::CloudViewer::showCloud( xyzrgb_cloud_ptr, "final_slabs", nullptr, 2., false ) : nullptr;
            int final_slab_id = 0;
            for ( auto const& slab_with_mp : final_slabs )
            {
                am::Cuboid cuboid_non_shrunk;
                if ( debug )
                {
                    //MyVoxelGrid::drawSlab( vptr->VP(), slabs[l_slab_id] /*minmax*/, max_planes, &sor, slab_name, colours[slab_id] );

                    MyVoxelGrid::cuboidFromSlab( /*       out_cuboid: */ cuboid_non_shrunk
                                                 , /*        in_slab: */ slab_with_mp.first
                                                 , /* voxelgrid_info: */ slab_with_mp.second
                                                 , /*  voxelgrid_pcl: */ &sor              );
                }

                // draw
                if ( debug )
                {
                    char slab_name[255];
                    sprintf( slab_name, "slab%d", final_slab_id );
                    vptr2->VP()->addText3D( slab_name
                                           , cuboid_non_shrunk.Corners().back()
                                           , .01
                                           , .8, .1, .1
                                           , std::string(slab_name) + "_title" );

                    cuboid_non_shrunk.draw( vptr2->VP(), slab_name                      , colours[added_slab_id%colours.size()] );
                    //cuboid           .draw( vptr->VP(), slab_name+std::string("shrunk"), (Eigen::Vector3f() << 1.f, 0.f, 0.f).finished(),1. );
                }
                final_slab_id++;
            }
            if ( debug  )
                vptr2->VP()->spin();
        }

        std::cout << __func__ << ": stopped, since slab size(" << max_slab_size << ") < slab_size_limit(" << slab_size_limit
                  << ")"
                  << " in percentage: " << (max_slab_size / (float)cloud_ptr->size()) << std::endl;

    } // segment

    int
    MyVoxelGrid::getLeafId( int c0, int c1, int c2, int d1, int d2, Eigen::Vector3i const& div_mul )
    {
        Eigen::Vector3i ijk( c0, c0, c0 );
        ijk[ d1 ] = c1;
        ijk[ d2 ] = c2;
        return ijk.dot( div_mul );
    }

    std::string
    MyVoxelGrid::getCellGridId( Eigen::Vector3i ijk )
    {
        char id[255];
        sprintf( id, "grid_%d%d%d", ijk(0), ijk(1), ijk(2) );
        return id;
    }

    int
    MyVoxelGrid::drawSlab( pcl::visualization::PCLVisualizer::Ptr          v_ptr
                           , Eigen::MatrixX2i                       const& minmax
                           , MaxPlanes                              const& mp
                           , pcl::VoxelGrid<pcl::PointXYZRGB>            * sor
                           , std::string                                   title
                           , Eigen::Vector3f                               colour
                           //, int                                         viewport = 0
                           )
    {
        const Eigen::Vector3i min_box_coords = sor->getMinBoxCoordinates();
        const Eigen::Vector3f leaf_size      = sor->getLeafSize();

        // prepare drawing targets
        am::Cuboid cuboid;
        cuboid.Corners().resize(8);
        //pcl::PointCloud<pcl::PointXYZ>::Ptr polygon_cloud_ptr( new pcl::PointCloud<pcl::PointXYZ>() );
        std::cout << "drawSlab:\n" << "{" << minmax.row(0) << ", " << minmax.row(1) << ", " << minmax.row(2) << "}" << std::endl;

        // draw slabs
        Eigen::Vector3i coord;
        coord[ mp.max_dim_id ] = mp.getMaxDimCoord();
        std::vector<int> corner_ids = { 0, 1, 3, 2 };
        pcl::PointXYZ pnt;
        for ( int corner_id = 0; corner_id < 4; ++corner_id )
        {
            coord[ mp.dim1_id ] = minmax( 1, (corner_ids[corner_id]%2) ) + (corner_ids[corner_id]%2); // upper bound is inclusive
            coord[ mp.dim2_id ] = minmax( 2, (corner_ids[corner_id]/2) ) + (corner_ids[corner_id]/2); // upper bound is inclusive

            pnt = pcl::PointXYZ(
                      (coord[0]+min_box_coords(0)) * leaf_size(0),
                    (coord[1]+min_box_coords(1)) * leaf_size(1),
                    (coord[2]+min_box_coords(2)) * leaf_size(2) );

            cuboid.Corners()[corner_id] = pnt;
            pnt.data[mp.max_dim_id] += leaf_size[mp.max_dim_id];
            cuboid.Corners()[4+corner_id] = pnt;
        }
        v_ptr->addText3D( title
                          , pnt
                          , .01
                          , .8, .1, .1
                          , title + "_title" );

        return cuboid.draw( v_ptr, title, colour );
    }

    int
    MyVoxelGrid::cuboidFromSlab( am::Cuboid &cuboid
                                 , Eigen::MatrixX2i              const& slab
                                 , MaxPlanes                     const& mp
                                 , pcl::VoxelGrid<pcl::PointXYZRGB>     *sor
                                 )
    {
        const Eigen::Vector3i min_box_coords = sor->getMinBoxCoordinates();
        const Eigen::Vector3f leaf_size      = sor->getLeafSize();

        // prepare drawing targets
        //cuboid.Corners().resize( 8 );
        std::vector<float> corners( 24 );

        // draw slabs
        Eigen::Vector3i coord;
        coord[ mp.max_dim_id ] = mp.getMaxDimCoord();
        std::vector<int> corner_ids = { 0, 1, 3, 2 };
        pcl::PointXYZ pnt;
        for ( int corner_id = 0; corner_id < 4; ++corner_id )
        {
            coord[ mp.dim1_id ] = slab( 1, (corner_ids[corner_id]%2) ) + (corner_ids[corner_id]%2); // upper bound is inclusive
            coord[ mp.dim2_id ] = slab( 2, (corner_ids[corner_id]/2) ) + (corner_ids[corner_id]/2); // upper bound is inclusive

            pnt = pcl::PointXYZ(
                      (coord[0]+min_box_coords(0)) * leaf_size(0),
                      (coord[1]+min_box_coords(1)) * leaf_size(1),
                      (coord[2]+min_box_coords(2)) * leaf_size(2) );

            //cuboid.Corners()[corner_id] = pnt;
            std::copy( pnt.data, pnt.data+3, corners.begin() + corner_id*3 );

            pnt.data[mp.max_dim_id] += leaf_size[mp.max_dim_id];

            //cuboid.Corners()[4+corner_id] = pnt;
            std::copy( pnt.data, pnt.data+3, corners.begin() + (4+corner_id)*3 );
        }

        return am::Cuboid::fromCorners( cuboid
                                        , corners
                                        , nullptr
                                        , nullptr );
    }

    int
    MyVoxelGrid::project2D( std::vector<std::vector<int> >       & projections
                            , std::vector<int>              const& points_in_voxels
                            , Eigen::MatrixX2i              const& slab
                            , MaxPlanes                     const& mp
                            , pcl::VoxelGrid<pcl::PointXYZRGB>   * sor
                            , bool                                 print)
    {

        const Eigen::Vector3i divs          = sor->getNrDivisions();
        const Eigen::Vector3i div_mul       = sor->getDivisionMultiplier();
        //const Eigen::Vector3i min_box_coords = sor->getMinBoxCoordinates();

        Eigen::MatrixX2i limits = slab;
        // minmax is inclusive
        for ( int dim = 0; dim < slab.rows(); ++dim )
            ++limits(dim,1);

        // init
        projections.resize( slab.rows() );
        for ( int c = 0; c < projections.size(); ++c )
        {
            projections[c].resize( limits(c,1) - limits(c,0) );
            std::fill( projections[c].begin(), projections[c].end(), 0 );
        }

        // voxel coordinates
        Eigen::Vector3i ijk( 0,0,0 );
        Eigen::Vector3i coord;
        for ( coord(0) = limits(0,0); coord(0) < limits(0,1); ++coord(0) )
            for ( coord(1) = limits(1,0); coord(1) < limits(1,1); ++coord(1) )
                for ( coord(2) = limits(2,0); coord(2) < limits(2,1); ++coord(2) )
                {
                    ijk[ mp.max_dim_id ] = coord(0);
                    ijk[ mp.dim1_id ]    = coord(1);
                    ijk[ mp.dim2_id ]    = coord(2);
                    int idx = ijk.dot( div_mul );
                    for ( int dim = 0; dim < 3; ++dim )
                        //projections[dim][ coord(dim) - limits(dim,0) ] += ((points_in_voxels[ idx ] > 10) ? 1 : 0); //points_in_voxels[ idx ];
                        projections[dim][ coord(dim) - limits(dim,0) ] += points_in_voxels[ idx ];
                }

        // print
        if ( print )
        {
            std::cout << "projections:\n\t";
            for ( int c = 0; c < projections.size(); ++c )
            {
                for ( int cx = 0; cx < projections[c].size(); ++cx )
                {
                    std::cout << projections[c][cx] << ", ";
                }
                std::cout << "\n\t";
            }
            std::cout << "\n";
        }

        return EXIT_SUCCESS;
    }

    // volume in cell
    inline int
    MyVoxelGrid::getVolume( Slab slab )
    {
        return   ( slab(0,1) - slab(0,0) + 1) *
                 ( slab(1,1) - slab(1,0) + 1) *
                 ( slab(2,1) - slab(2,0) + 1);
    }

    bool
    MyVoxelGrid::compareSlabsByVolume( Eigen::MatrixX2i const& a, Eigen::MatrixX2i const& b )
    {
        int volumes[2] = { getVolume(a),getVolume(b) };

        // error check
        if ( (volumes[0] < 1) || (volumes[1] < 1) )    std::cerr << "volume<1...!!!!" << std::endl;

        // decreasing sort
        return volumes[0] > volumes[1];
    }

    bool
    isSlabAll( Eigen::MatrixX2i limits
                  //, bool (*condition)(int)
                  , std::function<bool(int)> condition
                  , MaxPlanes                     const& mp
                  , std::vector<int>                   const& leafs
                  , pcl::VoxelGrid<pcl::PointXYZRGB>   * sor )
    {
        const Eigen::Vector3i div_mul       = sor->getDivisionMultiplier();

        Eigen::Vector3i ijk( 0,0,0 );
        Eigen::Vector3i coord;
        for ( coord(0) = limits(0,0); coord(0) <= limits(0,1); ++coord(0) )
            for ( coord(1) = limits(1,0); coord(1) <= limits(1,1); ++coord(1) )
                for ( coord(2) = limits(2,0); coord(2) <= limits(2,1); ++coord(2) )
                {
                    ijk[ mp.max_dim_id ] = coord(0);
                    ijk[ mp.dim1_id ]    = coord(1);
                    ijk[ mp.dim2_id ]    = coord(2);
                    int idx = ijk.dot( div_mul );

                    if ( !condition( leafs[idx] ) )
                    {
                        return false;
                    }
                }

        return true;
    }

    int
    setSlabInLeafsTo( Eigen::MatrixX2i slab
                      , int value
                      , std::vector<int> &leafs
                      , MaxPlanes const& mp
                      , Eigen::Vector3i div_mul )
    {
        for ( int c0 = slab(0,0); c0 <= slab(0,1); ++c0 )
            for ( int c1 = slab(1,0); c1 <= slab(1,1); ++c1 )
                for ( int c2 = slab(2,0); c2 <= slab(2,1); ++c2 )
                {
                    int idx = MyVoxelGrid::getLeafId( c0, c1, c2, mp.dim1_id, mp.dim2_id, div_mul );
                    leafs[ idx ] = value;
                }
    }


    int
    MyVoxelGrid::splitSlabs2( std::vector<Eigen::MatrixX2i>        & out_slabs
                              , std::vector<Eigen::MatrixX2i> const& slabs
                              , float                                threshold
                              , std::vector<int>                     leafs // need the local copy
                              , std::vector<int>              const& points_in_voxels
                              , MaxPlanes                     const& mp
                              , pcl::VoxelGrid<pcl::PointXYZRGB>   * sor
                              , pcl::visualization::PCLVisualizer::Ptr vptr )
    {
        const Eigen::Vector3i div_mul       = sor->getDivisionMultiplier();

        for ( auto const& slab : slabs )
        {
            auto limits = slab;
            std::cout << __func__ << ": starting slab " << slab.row(0) << ", " << slab.row(1) << ", " << slab.row(2) << std::endl;

            std::vector<Eigen::Vector3i> occupied = {{0,0,0}}; //bootstrap to bulk, so that the for loop starts
            for ( int it = 0; occupied.size() && (it < 20); ++it )
            {
                // select occupied cells
                occupied.clear();

                Eigen::Vector3i ijk( 0,0,0 );
                Eigen::Vector3i coord;
                for ( coord(0) = limits(0,0); coord(0) <= limits(0,1); ++coord(0) )
                    for ( coord(1) = limits(1,0); coord(1) <= limits(1,1); ++coord(1) )
                        for ( coord(2) = limits(2,0); coord(2) <= limits(2,1); ++coord(2) )
                        {
                            ijk[ mp.max_dim_id ] = coord(0);
                            ijk[ mp.dim1_id ]    = coord(1);
                            ijk[ mp.dim2_id ]    = coord(2);
                            int idx = ijk.dot( div_mul );
                            if ( leafs[idx] != -1 )
                            {
                                //std::cout << "adding " << coord.transpose() << std::endl;
                                if ( std::find( occupied.begin(), occupied.end(), coord ) != occupied.end() )
                                    std::cout << ".......\t " << coord.transpose() << " already in occupied["
                                              << std::distance( occupied.begin(), std::find( occupied.begin(), occupied.end(), coord ) ) << "]" << std::endl;
                                occupied.push_back( coord );
                            }
                            //else
                                //std::cout << "skipping " << coord.transpose() << ": " << "leafs[idx] " << leafs[idx] << std::endl;
                        }

                if ( !occupied.size() )
                    break;

                std::cout << "occupied percentage: " << occupied.size() / (float)getVolume( slab ) << std::endl;
                std::cout << "occupied.size(): " << occupied.size() << std::endl;
                std::cout << "(float)getVolume(slab): " << (float)getVolume(slab) << std::endl;
                std::cout << "coverage: " << occupied.size() / (float)getVolume( slab ) << std::endl;
                std::cout << "new coverage: " << occupied.size() / (float)getVolume( slab ) / (sor->getLeafSize().transpose()*sor->getLeafSize()) << std::endl;

                float coverage;
                if ( (coverage = occupied.size() / (float)getVolume( slab ) /*/ (sor->getLeafSize().transpose()*sor->getLeafSize())*/ ) > /*0.10*/ threshold )
                {
                    std::cout << __func__ << ": skipping split, since coverage is large enough: " << coverage << std::endl;
                    break;
                }

                // create slabs from cell pairs
                std::vector< std::pair<Eigen::MatrixX2i,int> > filled_slabs; // <slab,volume>
                for ( int cid = 0; cid != occupied.size()-1; ++cid )
                {
                    // looking forward to create a cell pair
                    for ( int cid1 = cid+1; cid1 != occupied.size(); ++cid1 )
                    {
                        Eigen::MatrixX2i tmp_slab(3,2);
                        for ( int dim = 0; dim != 3; ++dim )
                        {
                            tmp_slab(dim,0) = std::min( occupied[cid](dim), occupied[cid1](dim) ); // min
                            tmp_slab(dim,1) = std::max( occupied[cid](dim), occupied[cid1](dim) ); // max
                        }

                        if ( isSlabAll(tmp_slab,[](int leaf){ return leaf > -1; },mp,leafs,sor) ) // "is all not -1" -> filled
                        {
                            filled_slabs.push_back( {tmp_slab, getVolume(tmp_slab)} );
                        }
                    }
                }

                if ( !filled_slabs.size() )
                    break;

                // sort by volume decreasing
                std::sort( filled_slabs.begin(), filled_slabs.end(), [](std::pair<Eigen::MatrixX2i,int> a, std::pair<Eigen::MatrixX2i,int> b){
                    return a.second > b.second;
                });

                // remove biggest
                if ( filled_slabs.size() )
                {
                    setSlabInLeafsTo( filled_slabs[0].first, -2, leafs, mp, div_mul );

                    static auto colours = am::util::nColoursEigen( 5, 1.f );
                    static int drawn = 0;
                    if ( vptr )
                    {
                        //drawSlab( vptr, filled_slabs[0].first, mp, sor, am::util::sprintf( "filled%d", drawn ), colours[ drawn % colours.size()] );
                        ++drawn;
                    }

                    out_slabs.push_back( filled_slabs[0].first );
                }
            }

//            if ( vptr )
//                for ( auto const& cell : occupied )
//                {
//                    Eigen::MatrixX2i slab_from_cell(3,2);
//                    slab_from_cell.block<3,1>(0,0) = cell;
//                    slab_from_cell.block<3,1>(0,1) = cell;
//                    drawSlab( vptr, slab_from_cell, mp, sor, am::util::sprintf( "cell%d", cell(0)*10000+cell(1)*100+cell(2) ) );
//                }

        }
        if ( out_slabs.empty() )
            out_slabs = slabs;

        //out_slabs = slabs;

        return EXIT_SUCCESS;
    }

    inline void
    printSlab( std::ostream & str, MyVoxelGrid::Slab slab )
    {
        str << slab.row(0) << ", " << slab.row(1) << ", " << slab.row(2);
    }

    int MyVoxelGrid::mergeSlabs( std::vector<std::pair<Slab,MaxPlanes>>        & out_slabs
                                 , std::vector<std::pair<Slab,MaxPlanes>> const& slabs
                                 , float                                         threshold
                                 , std::vector<int>                              leafs // need the local copy
                                 //, std::vector<int>              const& points_in_voxels
                                 //, MaxPlanes                     const& mp
                                 , pcl::VoxelGrid<pcl::PointXYZRGB>            * sor
                                 , pcl::visualization::PCLVisualizer::Ptr        vptr )
    {
        for ( int slab_id = 0; slab_id != slabs.size(); ++slab_id )
        {
            auto const& slab = slabs[slab_id].first;

            for ( int neigh_id = slab_id+1; neigh_id < slabs.size(); ++neigh_id )
            {
                auto const& neigh = slabs[neigh_id].first;
                // check same size:
                if ( (slab(0,1) - slab(0,0)) == (neigh(0,1) - neigh(0,0))
                     && (slab(1,1) - slab(1,0)) == (neigh(1,1) - neigh(1,0))
                     && (slab(2,1) - slab(2,0)) == (neigh(2,1) - neigh(2,0)) )
                {

                    std::cerr << "slab" << slab_id << ": ";
                    printSlab( std::cerr, slab );
                    std::cerr << " has same size as slab" << neigh_id << ": ";
                    printSlab( std::cerr, neigh );
                    std::cerr << std::endl;
                }
            }
        }

        if ( out_slabs.empty() )
            out_slabs = slabs;

        return EXIT_SUCCESS;
#if 0
        const Eigen::Vector3i div_mul       = sor->getDivisionMultiplier();

        for ( auto const& slab : slabs )
        {
            auto limits = slab;
            std::cout << __func__ << ": starting slab " << slab.row(0) << ", " << slab.row(1) << ", " << slab.row(2) << std::endl;

            std::vector<Eigen::Vector3i> occupied = {{0,0,0}}; //bootstrap to bulk, so that the for loop starts
            for ( int it = 0; occupied.size() && (it < 20); ++it )
            {
                // select occupied cells
                occupied.clear();

                Eigen::Vector3i ijk( 0,0,0 );
                Eigen::Vector3i coord;
                for ( coord(0) = limits(0,0); coord(0) <= limits(0,1); ++coord(0) )
                    for ( coord(1) = limits(1,0); coord(1) <= limits(1,1); ++coord(1) )
                        for ( coord(2) = limits(2,0); coord(2) <= limits(2,1); ++coord(2) )
                        {
                            ijk[ mp.max_dim_id ] = coord(0);
                            ijk[ mp.dim1_id ]    = coord(1);
                            ijk[ mp.dim2_id ]    = coord(2);
                            int idx = ijk.dot( div_mul );
                            if ( leafs[idx] != -1 )
                            {
                                //std::cout << "adding " << coord.transpose() << std::endl;
                                if ( std::find( occupied.begin(), occupied.end(), coord ) != occupied.end() )
                                    std::cout << ".......\t " << coord.transpose() << " already in occupied["
                                              << std::distance( occupied.begin(), std::find( occupied.begin(), occupied.end(), coord ) ) << "]" << std::endl;
                                occupied.push_back( coord );
                            }
                            //else
                                //std::cout << "skipping " << coord.transpose() << ": " << "leafs[idx] " << leafs[idx] << std::endl;
                        }

                if ( !occupied.size() )
                    break;

                std::cout << "occupied percentage: " << occupied.size() / (float)getVolume( slab ) << std::endl;
                std::cout << "occupied.size(): " << occupied.size() << std::endl;
                std::cout << "(float)getVolume(slab): " << (float)getVolume(slab) << std::endl;

                float coverage;
                if ( (coverage = occupied.size() / (float)getVolume( slab )) > threshold )
                {
                    std::cout << __func__ << ": skipping split, since coverage is large enough: " << coverage << std::endl;
                    break;
                }




                // create slabs from cell pairs
                std::vector< std::pair<Eigen::MatrixX2i,int> > filled_slabs; // <slab,volume>
                for ( int cid = 0; cid != occupied.size()-1; ++cid )
                {
                    // looking forward to create a cell pair
                    for ( int cid1 = cid+1; cid1 != occupied.size(); ++cid1 )
                    {
                        Eigen::MatrixX2i tmp_slab(3,2);
                        for ( int dim = 0; dim != 3; ++dim )
                        {
                            tmp_slab(dim,0) = std::min( occupied[cid](dim), occupied[cid1](dim) ); // min
                            tmp_slab(dim,1) = std::max( occupied[cid](dim), occupied[cid1](dim) ); // max
                        }

                        if ( isSlabAll(tmp_slab,[](int leaf){ return leaf > -1; },mp,leafs,sor) ) // "is all not -1" -> filled
                        {
                            filled_slabs.push_back( {tmp_slab, getVolume(tmp_slab)} );
                        }
                    }
                }

                if ( !filled_slabs.size() )
                    break;

                // sort by volume decreasing
                std::sort( filled_slabs.begin(), filled_slabs.end(), [](std::pair<Eigen::MatrixX2i,int> a, std::pair<Eigen::MatrixX2i,int> b){
                    return a.second > b.second;
                });

                // remove biggest
                if ( filled_slabs.size() )
                {
                    setSlabInLeafsTo( filled_slabs[0].first, -2, leafs, mp, div_mul );

                    static auto colours = am::util::nColoursEigen( 5, 1.f );
                    static int drawn = 0;
                    if ( vptr )
                    {
                        //drawSlab( vptr, filled_slabs[0].first, mp, sor, am::util::sprintf( "filled%d", drawn ), colours[ drawn % colours.size()] );
                        ++drawn;
                    }

                    out_slabs.push_back( filled_slabs[0].first );
                }
            }

//            if ( vptr )
//                for ( auto const& cell : occupied )
//                {
//                    Eigen::MatrixX2i slab_from_cell(3,2);
//                    slab_from_cell.block<3,1>(0,0) = cell;
//                    slab_from_cell.block<3,1>(0,1) = cell;
//                    drawSlab( vptr, slab_from_cell, mp, sor, am::util::sprintf( "cell%d", cell(0)*10000+cell(1)*100+cell(2) ) );
//                }

        }
        if ( out_slabs.empty() )
            out_slabs = slabs;

        //out_slabs = slabs;

        return EXIT_SUCCESS;
#endif
    }

    int
    MyVoxelGrid::splitSlabs( std::vector<Eigen::MatrixX2i>        & out_slabs
                             , std::vector<Eigen::MatrixX2i> const& slabs
                             , float                                threshold
                             , std::vector<int>                   & leafs
                             , std::vector<int>              const& points_in_voxels
                             , MaxPlanes                     const& mp
                             , pcl::VoxelGrid<pcl::PointXYZRGB>   * sor       )
    {

        //const Eigen::Vector3i divs           = sor->getNrDivisions();
        //const Eigen::Vector3i divb_mul       = sor->getDivisionMultiplier();
        //const Eigen::Vector3i min_box_coords = sor->getMinBoxCoordinates();

        for ( auto const& slab : slabs )
        {
            std::vector<std::vector<int> > projections;
            project2D( projections
                       , points_in_voxels
                       , slab
                       , mp
                       , sor
                       , true );

            std::pair<int,float> best_dir_n_count = { -1, 0 }; // best direction and it's above_thresh count
            //std::vector<float>   avg_of_projs( projections.size(), 0 );
            std::vector<int>     max_of_projs( projections.size(), 0 );
            {
                for ( int dir = 0; dir < projections.size(); ++dir )
                {
                    if ( projections[dir].size() < 2 )  continue;

                    max_of_projs[dir] = *std::max_element( projections[dir].begin(), projections[dir].end() );
                    //avg_of_projs[dir] = std::accumulate ( projections[dir].begin(), projections[dir].end(), 0 ) / (float)projections[dir].size();

                    const float limit = std::min( 10.f, max_of_projs[dir] * threshold );
                    float below_thresh_count = std::count_if( projections[dir].begin(), projections[dir].end(), [&limit](int value) {
                        return value < limit;
                    }) * (float)projections[dir].size();

                    if ( below_thresh_count > best_dir_n_count.second )
                    {
                        best_dir_n_count = { dir, below_thresh_count };
                    }
                }
            }
            if ( best_dir_n_count.first < 0 )
            {
                std::cerr << "best_dir_n_count.first < 0, not splitting..." << std::endl;
                out_slabs = { slabs[0] };
                break;
            }

            const float limit = std::min( 10.f, max_of_projs[best_dir_n_count.first] * threshold);
            char tit[255];
            sprintf( tit, "bestProj(%f, <%f): ", best_dir_n_count.second, limit );
            am::util::printVector( projections[best_dir_n_count.first], tit );
            const int split_dim = best_dir_n_count.first;
            if ( split_dim == mp.max_dim_id )
            {
                std::cerr << "split_dim == mp.max_dim_id, not splitting..." << std::endl;
                out_slabs = { slabs[0] };
                break;
            }
            if ( projections[split_dim].size() < 2 )
            {
                out_slabs = { slabs[0] };
                break;
            }

            out_slabs.clear();
            bool looking_for_start = true;
            Eigen::MatrixX2i tmp_slab = slab;
            for ( int c = 0; c < projections[split_dim].size(); ++c )
            {
                std::cout << "inspecting " << c << ": " << projections[split_dim][c];
                if ( looking_for_start && projections[split_dim][c] > limit )
                {
                    looking_for_start = false;
                    tmp_slab( split_dim, 0 ) = slab(split_dim,0) + c;
                    tmp_slab( split_dim, 1 ) = slab(split_dim,0) + projections[split_dim].size()-1;
                    out_slabs.push_back( tmp_slab );
                    std::cout << " starting at\n";
                }
                else if ( !looking_for_start && projections[split_dim][c] < limit )
                {
                    looking_for_start = true;
                    out_slabs.back()( split_dim, 1 ) = slab(split_dim,0) + c - 1;
                    std::cout << " stopping at " << out_slabs.back() << std::endl;
                }
                else
                {
                    std::cout << " skipping\n";
                }
            }

            if ( !looking_for_start )
                std::cout << " stopped at " << out_slabs.back() << std::endl;

#if 1
            // shrink
            std::for_each( out_slabs.begin(), out_slabs.end(), [&](Eigen::MatrixX2i &slab)
            {
                Eigen::MatrixX2i out_slab;
                MyVoxelGrid::shrinkSlab( out_slab, slab, mp, leafs, sor );
                slab = out_slab;
            });
#endif
            std::sort( out_slabs.begin(), out_slabs.end(), MyVoxelGrid::compareSlabsByVolume );

            break;
            continue;


            /// compactness
            std::vector<CompactnessStruct> structs;
            // for every dir
            for ( int dir = 0; dir < projections.size(); ++dir )
            {
                // for every slab width
                for ( int width = 1; width <= projections[dir].size(); ++width )
                {
                    // for every start position
                    for ( int start = 0; start <= projections[dir].size()-width; ++start )
                    {
                        int sum = 0;
                        for ( int cx = start; cx < start+width; ++cx )
                            sum += projections[dir][cx];
                        structs.push_back( CompactnessStruct(dir,width,start,sum,(float)sum/width) );
                        structs.back().print();
                    }
                }
            }

            std::sort( structs.begin(), structs.end(), std::greater<CompactnessStruct>() ); // descending

            std::vector<float> diffsums;
            for ( int i = 0; i < 10; ++i )
            {
                int w = structs[i].width;
                diffsums.push_back( 0 );
                int count = 0;
                for ( auto const& structt : structs )
                {
                    if ( structt.width == w )
                    {
                        int diff = (structt.compactness - structs[i].compactness);
                        diffsums[i] += diff * diff;
                        ++count;
                    }
                }
                diffsums[i] /= (float)count;
                structs[i].outlierscore = diffsums[i];
            }

            std::cout << "-------- SORTED --------\n";
            std::sort( structs.begin(), structs.end(), std::greater<CompactnessStruct>() ); // descending
            for ( int i = 0; i < 20; ++i )
            {
                structs[i].print();
            }
        }

        return EXIT_SUCCESS;
    }

    int
    MyVoxelGrid::shrinkSlab( Eigen::MatrixX2i                           & out_slab // contains slab min,max coords in [ maxdim, dim1, dim2 ] order
                             , Eigen::MatrixX2i                    const& slab    // contains limits in [ maxdim,dim1,dim2] order, inclusive!
                             , MaxPlanes                           const& mp
                             , std::vector<int>                    const& leafs
                             , pcl::VoxelGrid<pcl::PointXYZRGB>         *const sor
                             , /*debug*/ pcl::visualization::PCLVisualizer::Ptr v_ptr
                             , int level )
    {
        const LeafIndexer leafIndexer( mp.max_dim_id, mp.dim1_id, mp.dim2_id, sor->getDivisionMultiplier(), mp.getMaxDimCoord() );

        out_slab = Eigen::MatrixX2i( slab.rows(), slab.cols() );
        out_slab << slab(0,0), slab(0,1), // min_max_dim = max_max_dim
                slab(1,1), slab(1,0), // mindim1,maxdim1,
                slab(2,1), slab(2,0); // mindim2,maxdim2

        for ( int c0 = slab(0,0); c0 <= slab(0,1); ++c0 )
            for ( int c1 = slab(1,0); c1 <= slab(1,1); ++c1 )
                for ( int c2 = slab(2,0); c2 <= slab(2,1); ++c2 )
                {
                    int idx = leafIndexer.leafIndex( c0, c1, c2 );

                    // if free cell
                    if ( leafs[idx] > -1 )
                    {
                        // store c1 as min or max along dim1_id
                        if ( c1 < out_slab(1,0) )
                            out_slab(1,0) = c1;
                        if ( c1 > out_slab(1,1) )
                            out_slab(1,1) = c1;

                        // store c2 as min or max along dim2_id
                        if ( c2 < out_slab(2,0) )
                            out_slab(2,0) = c2;
                        if ( c2 > out_slab(2,1) )
                            out_slab(2,1) = c2;
                    }
                }
    } // shrinkSlab

    template<typename PointT>
    int
    MyVoxelGrid::shrinkSlabInsideGrid( am::Cuboid &out_cuboid
                                       , Eigen::MatrixX2i                            const&      slab     // contains limits in [ maxdim,dim1,dim2] order, inclusive!
                                       , MaxPlanes                                   const&      mp
                                       , std::vector<int>                            const&      leafs
                                       , boost::shared_ptr<const pcl::PointCloud<PointT> >       cloud_ptr
                                       , pcl::PointIndices::ConstPtr                             indices_ptr_arg
                                       , pcl::VoxelGrid<pcl::PointXYZRGB>                 *const sor
                                       , /*debug*/ pcl::visualization::PCLVisualizer::Ptr        v_ptr
                                       , int level )
    {
        const Eigen::Array3f  inverse_leaf_size = Eigen::Array3f::Ones() / sor->getLeafSize().array();
        const Eigen::Vector3i min_box_coords    = sor->getMinBoxCoordinates();

        // init
        // gather slab planes
        am::Cuboid cuboid;
        MyVoxelGrid::cuboidFromSlab( cuboid
                                     , slab
                                     , mp
                                     , sor );

        std::vector<Eigen::Vector4f> plane_coeffs;
        for ( int plane_id = 0; plane_id != cuboid.Planes().size(); ++plane_id )
        {
            plane_coeffs.push_back( cuboid.Planes()[plane_id].Coeffs() );
        }

        std::vector<float> min_dists( plane_coeffs.size() );
        std::fill( min_dists.begin(), min_dists.end(), FLT_MAX );

        pcl::PointIndices::ConstPtr indices_ptr( indices_ptr_arg ? indices_ptr_arg
                                                                 : am::util::pcl::allIndicesOf(cloud_ptr) );

        int slab_point_count = 0;

        Eigen::Vector3i ijk;
        // all indices to voxels
        for ( auto it =  indices_ptr->indices.begin();
                   it != indices_ptr->indices.end  ();
                  ++it )
        {
            ijk[0] =    static_cast<int  >( floor(cloud_ptr->points[*it].x * inverse_leaf_size[0])
                      - static_cast<float>( min_box_coords[0] )              );
            ijk[1] =    static_cast<int  >( floor(cloud_ptr->points[*it].y * inverse_leaf_size[1])
                      - static_cast<float>( min_box_coords[1] )              );
            ijk[2] =    static_cast<int  >( floor(cloud_ptr->points[*it].z * inverse_leaf_size[2])
                      - static_cast<float>( min_box_coords[2] )              );

            if (    (ijk[mp.max_dim_id] >= slab(0,0))
                 && (ijk[mp.max_dim_id] <= slab(0,1))
                 && (ijk[mp.dim1_id   ] >= slab(1,0))
                 && (ijk[mp.dim1_id   ] <= slab(1,1))
                 && (ijk[mp.dim2_id   ] >= slab(2,0))
                 && (ijk[mp.dim2_id   ] <= slab(2,1)) )
            {
                ++slab_point_count;
                for ( int plane_id = 0; plane_id != plane_coeffs.size(); ++plane_id )
                {
                    float dist = plane_coeffs[ plane_id ].dot( (Eigen::Vector4f() <<
                                                                cloud_ptr->points[*it].x,
                                                                cloud_ptr->points[*it].y,
                                                                cloud_ptr->points[*it].z,
                                                                1.f).finished() );

                    if ( dist < min_dists[plane_id] )
                        min_dists[plane_id] = dist;
                }
            }
        }
        if ( slab_point_count < 20 )
            return EXIT_FAILURE;

        // assign corners to planes
        // get the three planes, each corner sits at
        std::vector< std::vector<int> > planes_corners( cuboid.Planes().size() ); // { plane_0: { corner_id_0, corner_id_1, corner_id_2, corner_id_3 }, ... };
        for ( int corner_id = 0; corner_id != cuboid.Corners().size(); ++corner_id )
        {
            //
            std::vector<std::pair<int,float> > dists( cuboid.Planes().size() ); // <pid,dist>
            for ( int plane_id = 0; plane_id != cuboid.Planes().size(); ++plane_id )
            {
                dists[plane_id] = { plane_id, plane_coeffs[ plane_id ].dot(cuboid.Corners()[corner_id].getVector4fMap()) };
            }
            std::sort( dists.begin(), dists.end(), [](std::pair<int,float> a, std::pair<int,float> b){
                return a.second < b.second;
            });

            for ( int plane_id = 0 ; plane_id != 3; ++plane_id )
                planes_corners[ dists[plane_id].first ].push_back( corner_id );
        }

        //out_cuboid = cuboid;
        cuboid.cloneTo( out_cuboid );
        for ( int plane_id = 0; plane_id != min_dists.size(); ++plane_id )
        {
            if  ( min_dists[plane_id] >= FLT_MAX )
            {
                std::cerr << __func__ << ":  min_dists[" << plane_id << "] >= FLT_MAX...skipping shrinking this plane" << std::endl;
                continue;
            }
            am::Plane &plane = out_cuboid.PlanesRef()[plane_id];

            // centroid
            Eigen::Vector3f centroid( Eigen::Vector3f::Zero() );
            {
                for ( int corner_id = 0; corner_id != plane.Corners().size(); ++corner_id )
                    centroid += plane.Corners()[corner_id].getVector3fMap();
                centroid /= 4.f;
            }

            // normal
            const Eigen::Vector3f normal = plane.Normal().normalized();
            const Eigen::Vector3f offset = min_dists[plane_id] * normal;
            centroid += offset;
            plane.CoeffsPtrRef()->values[3] = -normal.dot( centroid );
            plane.Frame().block<3,1>(0,3)   = centroid;

            for ( int plane_corner_id = 0; plane_corner_id != plane.Corners().size(); ++plane_corner_id )
            {
                plane.Corners()[plane_corner_id ] = am::util::pcl::asPointXYZ( plane.Corners()[plane_corner_id].getVector3fMap() + offset );

                // modify the 4 corners assigned to this plane in planes_corners
                out_cuboid.Corners()[ planes_corners[plane_id][plane_corner_id] ] =
                        am::util::pcl::asPointXYZ( out_cuboid.Corners()[ planes_corners[plane_id][plane_corner_id] ].getVector3fMap() + offset );
            }

        }

        return EXIT_SUCCESS;

    } // shrinkSlab

    Eigen::Vector3f
    MyVoxelGrid::getCentroidCoord( Eigen::Vector3i ijk, pcl::VoxelGrid<pcl::PointXYZRGB> *sor )
    {
        const auto leaf_size      = sor->getLeafSize();
        const auto min_box_coords = sor->getMinBoxCoordinates();

        return (Eigen::Vector3f() <<
                (ijk(0)+.5f+min_box_coords(0)) * leaf_size(0),
                (ijk(1)+.5f+min_box_coords(1)) * leaf_size(1),
                (ijk(2)+.5f+min_box_coords(2)) * leaf_size(2)).finished();
    }

    Eigen::Vector3f
    MyVoxelGrid::getCentroidCoord( int i, int j, int k, Eigen::Vector3i const& min_box_coords, Eigen::Vector3f const& leaf_size )
    {
        return (Eigen::Vector3f() << (i+.5f+min_box_coords(0)) * leaf_size(0),
                (j+.5f+min_box_coords(1)) * leaf_size(1),
                (k+.5f+min_box_coords(2)) * leaf_size(2)).finished();
    }

    /////////////////////////////////////////////////////////////////////
    /// MaxPlanes
    /////////////////////////////////////////////////////////////////////
    MaxPlanes::MaxPlanes()
        : max_point_count( 0, 0, 0 )
        , max_coords( 0, 0, 0 )
        , max_dim_id( 0 )
        , dim1_id( 0 )
        , dim2_id( 0 )
    {
    }

    int
    MaxPlanes::getMaxDimCoord() const
    {
        return max_coords[ max_dim_id ];
    }

    int
    MaxPlanes::getDim1Coord() const
    {
        return max_coords[dim1_id];
    }

    int
    MaxPlanes::getDim2Coord() const
    {
        return max_coords[dim2_id];
    }

    /////////////////////////////////////////////////////////////////////
    /// LeafIndexer
    /////////////////////////////////////////////////////////////////////
    LeafIndexer::LeafIndexer( int max_dim_id, int dim1_id, int dim2_id, Eigen::Vector3i div_mul, int max_dim_coord )
        : _dim_ids( max_dim_id, dim1_id, dim2_id )
        , _div_mul( div_mul )
        , _max_dim_coord( max_dim_coord )
    {}

    int
    LeafIndexer::leafIndex( int i, int j, int k ) const
    {
        Eigen::Vector3i ijk;
        ijk( _dim_ids(0) ) = i;
        ijk( _dim_ids(1) ) = j;
        ijk( _dim_ids(2) ) = k;

        return ijk.dot( _div_mul );
    }

    int
    LeafIndexer::leafIndex( int j, int k ) const
    {
        return leafIndex( _max_dim_coord, j, k );
    }



} // ns am

