#include "ui/QtCloudViewer.h"
#include "ui_QtCloudViewer.h" // autogenerated by qt from QtCloudViewer.h

#include <QDebug>
#include <QVTKWidget.h>
#include <QFileDialog>
#include <QSplashScreen>
#include <QMessageBox>
#include <QColorDialog>
#include <vtkRenderWindow.h>

#include <boost/property_tree/json_parser.hpp>

#include "cuboid_approx/common.h"

#include "ui/QtGuiUpdaterThread.h"

#include "primitives/Cuboid.h"
#include "cuboid_approx/primitives/Contact.h"

#include "pcl/io/ply_io.h" // loadPLYcloud
#include "CuboidUtil.h" // get_files_by_ext
#include "cuboid_approx/ui/DisplayedPrimitive.h"
#include "cuboid_approx/proc3d/ContactsFiltering.h"
#include "cuboid_approx/io/walls_io.h"
#include "cuboid_approx/io/primitives_io.h" // saveMoosPlane

#include <iomanip>

/// LOCALs ///
inline std::string
enforceExtension( boost::filesystem::path ext, boost::filesystem::path path )
{
    if ( path.extension() != ext )  path.replace_extension( ext );
    return path.string();
}

inline int
appendToTextFile( std::string line, std::string path )
{
    std::ofstream myfile;
    myfile.open ( path, std::ios_base::app );
    if ( !myfile.is_open() )
    {
        std::cerr << "could not open file " << path << std::endl;
        return EXIT_FAILURE;
    }

    myfile << line;
    myfile.close();

    return EXIT_SUCCESS;
}

/// Protected methods ///

int
QtCloudViewer::_refreshDisplayProperties()
{
    on_pointSizeSpinBox_valueChanged( _ui->pointSizeSpinBox->value() );
    on_colourCuboids_stateChanged( _ui->colourCuboids->isChecked() );
    on_colourEdgesCheckBox_stateChanged( _ui->colourEdgesCheckBox->isChecked() );
    on_sphereSizeSpinBox_editingFinished();
    on_showIntersections_toggled( _ui->showIntersections->isChecked() );

    return EXIT_SUCCESS;
}

std::string
QtCloudViewer::_getDepthDirAsStdString()
{
    return boost::filesystem::path(_ui->depthPathEdit->text().toStdString()).parent_path().string();
}

QString
QtCloudViewer::_getDepthDir()
{
    return QString::fromStdString(boost::filesystem::path(_ui->depthPathEdit->text().toStdString()).parent_path().string());
}

std::string
QtCloudViewer::_getFrameIdFromDepthPath()
{
    boost::filesystem::path depth_path( _ui->depthPathEdit->text().toStdString() );
    std::string depth_fname = depth_path.stem().string();
    return depth_fname.substr( depth_fname.find_last_of("_")+1);
}

int
QtCloudViewer::_saveLastPath( std::string str )
{
    boost::filesystem::path dir_path( str );
    if ( !boost::filesystem::is_directory(dir_path) )
    {
        dir_path = dir_path.parent_path();
    }

    //std::cout << "saving last_path: " << dir_path.string() << std::endl;

    _pt.put( "paths.last", dir_path.string() );
    write_json( "settings.json", _pt );

    return EXIT_SUCCESS;
}

std::string
QtCloudViewer::_getLastPath( std::string default_value )
{
    return _pt.get<std::string>("paths.last", default_value.empty() ? QApplication::applicationDirPath().toStdString() : default_value );
}

void
QtCloudViewer::_setDepthPath( QString fileName )
{
    _pt.put( "paths.depth", fileName.toStdString() );
    _ui->depthPathEdit->setText( fileName );

    write_json( "settings.json", _pt );
}

int
QtCloudViewer::_saveWalls( std::string path )
{

//    std::ofstream myfile;
//    myfile.open ( path );
//    if ( !myfile.is_open() )
//    {
//        std::cerr << "could not open file " << path << std::endl;
//        return EXIT_FAILURE;
//    }

//    std::string message;
//    for ( int i = 0; i != _ui->primitivesListWidget->count(); ++i )
//    {

//        message += am::util::sprintf("%02d",i) + ",";
//        message += _controller->getPrimitiveIsWall(i) ? "w," : "n,";
//        message += _controller->getPrimitiveIsFloor(i) ? "f," : "n,";
//        message += "\n";
//    }

//    myfile << message;
//    myfile.close();

//    message += " saved to " + path;

    am::Cuboid::TypeMap walls;
    _extractWallsMap( walls );
    std::string message = ca::io::writeWalls( walls, path, _ui->primitivesListWidget->count() );
    message += "\n to " + path;

    QMessageBox::information( this, "Walls", message.c_str(), QMessageBox::Ok );

    return EXIT_SUCCESS;
}

int
QtCloudViewer::loadWalls( std::string path )
{
    std::map<int, am::Cuboid::TYPE> walls;
    int cuboid_count = _ui->primitivesListWidget->count();
    int err          = ca::io::readWalls( walls, path, &cuboid_count );
    if ( err != EXIT_SUCCESS )
    {
        if ( err == ca::io::WALL_EXCEEDS_CUBOID_COUNT )
        {
            //QMessageBox::warning( this, "LoadWalls failed", "loadWalls: don't have this many cuboids", QMessageBox::Ok);
        }
        else
            QMessageBox::warning( this, "LoadWalls failed", "loadWalls: exception", QMessageBox::Ok);

        //return err;
    }

    for ( auto it = walls.begin(); it != walls.end(); ++it )
    {
        if ( it->first < cuboid_count )
        switch ( it->second )
        {
            case am::Cuboid::FLOOR:
                _controller->setPrimitiveIsFloor( /*  cuboid_id: */ it->first
                                                 , /*  is_floor: */ true
                                                 , /*  selected: */ false );
                break;
            case am::Cuboid::WALL:
                _controller->setPrimitiveIsWall( /*  cuboid_id: */ it->first
                                                 , /*  is_wall: */ true
                                                 , /* selected: */ false );
                break;
            case am::Cuboid::SPHERE:
                _controller->setPrimitiveIsSphere( /*    cuboid_id: */ it->first
                                                   , /*  is_sphere: */ true
                                                   , /*   selected: */ false );
                break;
            case am::Cuboid::CYLINDER:
                _controller->setPrimitiveIsCylinder( /*      cuboid_id: */ it->first
                                                     , /*  is_cylinder: */ true
                                                     , /*     selected: */ false );
                break;
            default:
                std::cerr << "[" << __func__ << "]: " << "unrecognized wall type..." << int(it->second) << std::endl;
                break;
        }
        else
        {
            std::cerr << "[" << __func__ << "]: " << "ignoring wall for cuboid " << it->first << std::endl;
            fflush(stderr);
        }
    }

    _ui->qvtkWidget->update();
    return EXIT_SUCCESS;
}

void
QtCloudViewer::_refreshPrimitivesList()
{
    auto primitives = _controller->getPrimitives();

    _ui->primitivesListWidget->clear();
    for ( auto it = primitives.begin(); it != primitives.end(); ++it )
    {
        _ui->primitivesListWidget->addItem(QString::fromStdString(
                                                  //std::string(it->second.floor           ? "F" : "")
                                                //+ std::string(it->second.wall            ? "W" : "")
                                                //+ std::string(it->second.rotation_target ? "T" : "")
                                                              it->first
                                                + std::string(it->second.tag.empty()     ? "" : (" - " + it->second.tag))
                                                )
                                            );
    }
    _ui->qvtkWidget->update();
}

std::vector<int>
QtCloudViewer::_getSelectedRows()
{
    std::vector<int> selection;

    QList<QListWidgetItem*> selected_items = _ui->primitivesListWidget->selectedItems();
    selection.reserve( selected_items.size() );

    for ( QListWidgetItem* widgetitemptr : selected_items )
        selection.push_back( _ui->primitivesListWidget->row(widgetitemptr) );

    return selection;
}

void
QtCloudViewer::_movePrimitive( char axis, float step )
{
    Eigen::Vector3f offset( Eigen::Vector3f::Zero() );
    offset( axis ) = step * _ui->stepSpinBox->value();

    std::vector<int> row_ids = _getSelectedRows();
    for ( auto const& row : row_ids )
    {
        ca::DisplayedPrimitive primitive = _controller->getPrimitive( row );
        auto cuboid_ptr = primitive.CuboidPtr();

        cuboid_ptr->moveAlongFrame( offset );

        _controller->redrawPrimitive( row, true );

        // debug
        {
            pcl::PointXYZ mn, mx;
            cuboid_ptr->getMinMaxCorner( mn, mx );
            std::cout << "new min max of row " << row << " are: " << mn << ", " << mx << std::endl;
        }
    }
    _ui->qvtkWidget->update();
}

void
QtCloudViewer::_resizePrimitive( char axis, float step )
{
    Eigen::Vector3f offset( Eigen::Vector3f::Zero() );
    offset( axis ) = step * _ui->stepSpinBox->value();

    std::vector<int> row_ids = _getSelectedRows();
    for ( auto const& row : row_ids )
    {
        ca::DisplayedPrimitive primitive = _controller->getPrimitive( row );
        auto cuboid_ptr = primitive.CuboidPtr();

        Eigen::Vector3f local_offset = offset;
//        if ( primitive.shape == am::Cuboid::TYPE::SPHERE )
//        {
//            local_offset.setConstant( offset(axis) );
//        }
        std::cout << "[" << __func__ << "]: " << "resizing by " << local_offset.transpose() << std::endl;
        cuboid_ptr->resizeAlongFrame( local_offset );
        std::cout << "[" << __func__ << "]: " << "size is " << cuboid_ptr->getExtent().transpose() << std::endl;

        _controller->redrawPrimitive( row, true );
    }
    _ui->qvtkWidget->update();
}

void
QtCloudViewer::_rotatePrimitive( char axis, float step )
{
    Eigen::Vector3f offset( Eigen::Vector3f::Zero() );
    offset( axis ) = step * _ui->stepSpinBox->value();

    std::vector<int> row_ids = _getSelectedRows();
    for ( auto const& row : row_ids )
    {
        ca::DisplayedPrimitive primitive = _controller->getPrimitive( row );
        auto cuboid_ptr = primitive.CuboidPtr();

        cuboid_ptr->rotateAroundFrameAxis( offset );

        _controller->redrawPrimitive( row, true );
    }
    _ui->qvtkWidget->update();
}

void
QtCloudViewer::_sliderReleasedSlot()
{
   // e.g. check with member variable _foobarButton
   QObject* obj = sender();

   // e.g. casting to the class you know its connected with
   QSlider* slider = dynamic_cast<QSlider*>( sender() );
   if ( slider != NULL )
   {
      slider->setValue( (slider->maximum() - slider->minimum())/2.f + 1 );
   }
}

void
QtCloudViewer::_filteringParamChanged( int id )
{
    on_clearButton_clicked();
}

void
QtCloudViewer::_set_camera_params( std::string params_str )
{
    std::vector<Eigen::Vector3f> pos_focal_up(3);

    std::vector<float> floats;
    try
    {
        std::istringstream iss( params_str );
        std::string tmp_str;
        while ( std::getline(iss, tmp_str, ',') )
        {
            floats.push_back( atof(tmp_str.c_str()) );
        }
    }
    catch (std::exception ex )
    {
        std::cerr << __func__ << ex.what() << std::endl;
    }

    // set to camera, if valid
    if ( floats.size() != 9 )
        std::cerr << __func__ << "not 9 numbers, skipping..." << std::endl;
    else
    {
        for ( int i = 0; i != 3; ++i )
            pos_focal_up[i] = (Eigen::Vector3f() << floats[i*3], floats[i*3+1], floats[i*3+2]).finished();
        _controller->setCameraPosFocalUp( pos_focal_up );
        _ui->qvtkWidget->update();
    }
}

std::pair<int,std::string>
QtCloudViewer::_rotateCamera( bool rotate )
{
    std::string message;
    if ( !rotate )
    {
        // TODO: stop rotating
        return std::pair<int,std::string>( EXIT_SUCCESS, "" );
    }

    /// start rotating
    // get floor
    auto floor_ptrs = _controller->getFloors();
    if ( floor_ptrs.empty() )
    {
        return std::pair<int,std::string>( EXIT_FAILURE, "Need a floor for upvector" );
    }
    // get cuboids centroid
    auto target_ptrs =_controller->getRotationTargets();
    if ( target_ptrs.empty() )
    {
        return std::pair<int,std::string>( EXIT_FAILURE, "Need a target for center" );
    }

    // rotationcentroid
    Eigen::Vector3f centroid;
    for ( std::shared_ptr<am::Cuboid> const cuboid_ptr : target_ptrs )
    {
        centroid += cuboid_ptr->Center();
    }
    centroid /= target_ptrs.size();

    // draw rotation axis
    Eigen::Vector3f rot_axis = floor_ptrs[0]->Planes()[2].Normal().normalized();
    //Eigen::AngleAxisf rot( ang, tmp_frame.block<3,1>(0,2) );
    //tmp_frame.block<3,1>(0,0) = rot * tmp_frame.block<3,1>(0,0);

    _controller->getViewer()->addArrow( am::util::pcl::asPointXYZ(centroid-rot_axis*1.5f) //upvector is always from plane_id=2
            ,am::util::pcl::asPointXYZ(centroid)
            ,0, 0, 0
            , false
            , "rotuparrow", 0 );

    std::vector<Eigen::Vector3f> pos_focal_up = _controller->getCameraPosFocalUp(), tmp_pos_focal_up;
    if ( pos_focal_up[0].dot(Eigen::Vector3f::Ones()) == 0.f )
        pos_focal_up[0](0) += .01f;

    tmp_pos_focal_up = pos_focal_up;
    std::cout << "firstpos is: " << pos_focal_up[0] << std::endl;

    for ( float deg = 180.f; deg < 360.f; ++deg )
    {
        Eigen::AngleAxisf rot( pcl::deg2rad(deg), rot_axis );
        tmp_pos_focal_up[0] = rot * pos_focal_up[0];

        _controller->setCameraPosFocalUp( tmp_pos_focal_up );
        std::cout << "new pos is: " << tmp_pos_focal_up[0].transpose() << std::endl;
        _ui->qvtkWidget->update();
        usleep(100000);
    }

    // start thread to rotate

    return { EXIT_SUCCESS, "Ok" };
}

////////////////////////////////////// PUBLIC //////////////////////////////////////////////////////////////////////////

QtCloudViewer::QtCloudViewer( QWidget *parent )
    : QMainWindow( parent )
    , _ui        ( new Ui::QtCloudViewer        )
    , _controller( new am::AnnotationController )
    , _last_cuboids_path( "" )
{
    _ui->setupUi( this );

    {
        connect( _ui->xMoveSlider, SIGNAL(sliderReleased()), this, SLOT(_sliderReleasedSlot()) );
        connect( _ui->yMoveSlider, SIGNAL(sliderReleased()), this, SLOT(_sliderReleasedSlot()) );
        connect( _ui->zMoveSlider, SIGNAL(sliderReleased()), this, SLOT(_sliderReleasedSlot()) );
        connect( _ui->xResizeSlider, SIGNAL(sliderReleased()), this, SLOT(_sliderReleasedSlot()) );
        connect( _ui->yResizeSlider, SIGNAL(sliderReleased()), this, SLOT(_sliderReleasedSlot()) );
        connect( _ui->zResizeSlider, SIGNAL(sliderReleased()), this, SLOT(_sliderReleasedSlot()) );
        connect( _ui->xRotateSlider, SIGNAL(sliderReleased()), this, SLOT(_sliderReleasedSlot()) );
        connect( _ui->yRotateSlider, SIGNAL(sliderReleased()), this, SLOT(_sliderReleasedSlot()) );
        connect( _ui->zRotateSlider, SIGNAL(sliderReleased()), this, SLOT(_sliderReleasedSlot()) );

        connect( _ui->loadButton            , SIGNAL(clicked()), this, SLOT(my_on_loadButton_clicked())             );
        connect( _ui->addPointsButton       , SIGNAL(clicked()), this, SLOT(my_on_addPointsButton_clicked())        );
        connect( _ui->loadContactsButton    , SIGNAL(clicked()), this, SLOT(my_on_loadContactsButton_clicked())     );
        connect( _ui->screenshotButton      , SIGNAL(clicked()), this, SLOT(my_on_screenshotButton_clicked())       );
        connect( _ui->loadCameraParamsButton, SIGNAL(clicked()), this, SLOT(my_on_loadCameraParamsButton_clicked()) );
    }

    // wire pcl visualizer to qtwidget
    vtkSmartPointer<vtkRenderWindow> renderWindow = _controller->getRenderWindow();
    _ui->qvtkWidget->SetRenderWindow( renderWindow );

    // load properties
    try
    {
        read_json( "settings.json", _pt );

        _ui->imagePathEdit->setText( QString::fromStdString(_pt.get<std::string>("paths.rgb"  )) );
        _ui->depthPathEdit->setText( QString::fromStdString(_pt.get<std::string>("paths.depth")) );
        _ui->planeInlierThresholdDoubleSpinBox->setValue( QString::fromStdString(_pt.get<std::string>("params.plane_inlier_threshold", "0.1")).toDouble() );
        _ui->depthBetaDoubleSpinBox->setValue( QString::fromStdString(_pt.get<std::string>("params.grabcut.betas.depth", "1.0")).toDouble() );
        _ui->rgbBetaDoubleSpinBox->setValue( QString::fromStdString(_pt.get<std::string>("params.grabcut.betas.rgb", "1.0")).toDouble() );
    }
    catch ( boost::property_tree::json_parser::json_parser_error err )
    {
        std::cerr << "error reading tree: " << err.what() << std::endl;

        _pt.put( "paths.rgb"  , _ui->imagePathEdit->text().toStdString() );
        _pt.put( "paths.depth", _ui->depthPathEdit->text().toStdString() );
    }

    // ensure opacity
    _controller->initViewer();
    _controller->setCuboidOpacity( -1, _ui->opacitySpinBox->value() );

    // load cloud

    if ( /**/ (_pt.find("path.points") != _pt.not_found())
         &&    _pt.get<std::string>("paths.rgb").empty()
         &&   !_pt.get<std::string>("paths.points").empty()
    )
        this->_addPoints( _pt.get<std::string>("paths.points") );
    else
        this->on_clearButton_clicked();

    _refreshDisplayProperties();

//    this->on_pointSizeSpinBox_valueChanged( _ui->pointSizeSpinBox->value() );
//    this->on_colourCuboids_stateChanged( _ui->colourCuboids->isChecked() );
    //    this->on_colourEdgesCheckBox_stateChanged( _ui->colourEdgesCheckBox->isChecked() );

    // bootstrap
    {
        //std::string cuboids_path = _pt.get<std::string>("paths.loaded_cuboids","");
        //if ( !cuboids_path.empty() )
        //{
        //   _controller->loadCuboids( cuboids_path );
        //}
        on_resetCameraButton_clicked();
        _ui->showInvalidCloud->setChecked( false );
        //on_showInvalidCloud_toggled( false );
        _refreshPrimitivesList();
    }
}

QtCloudViewer::~QtCloudViewer()
{
    {
        _pt.put( "params.plane_inlier_threshold", _ui->planeInlierThresholdDoubleSpinBox->value() );
        _pt.put( "params.grabcut.betas.depth", _ui->depthBetaDoubleSpinBox->value() );
        _pt.put( "params.grabcut.betas.rgb", _ui->rgbBetaDoubleSpinBox->value() );
         write_json( "settings.json", _pt );
    }

    SAFE_DELETE( _ui );
    _controller.reset();
}

void
QtCloudViewer::updateQvtWidget()
{
    _ui->qvtkWidget->update();
}

////////////////////////////////////// EVENTS //////////////////////////////////////////////////////////////////////////

/// CONTENT ///

void QtCloudViewer::on_imagePathButton_clicked()
{
    QString fileName = QFileDialog::getOpenFileName(this,
                                                    tr("Open rgb image file"),
                                                    _ui->imagePathEdit->text(),
                                                    tr("Image files (*.png *.bmp *.jpg);; All Files (*.*)"));
    if ( fileName.length() )
    {
        _pt.put( "paths.rgb", fileName.toStdString() );
        _ui->imagePathEdit->setText( fileName );

        boost::filesystem::path image_path( _ui->imagePathEdit->text().toStdString() );
        std::string path1 = (image_path.parent_path() / image_path.stem()).string() + ".raw";

        if ( !boost::filesystem::exists(path1) )
            path1 = (image_path.parent_path() / image_path.stem()).string() + ".dat";

        if ( !boost::filesystem::exists(path1) )
        {
            boost::filesystem::directory_iterator end_iter;
            for ( boost::filesystem::directory_iterator dir_iter( image_path.parent_path() );
                  dir_iter != end_iter;
                  ++dir_iter )
            {
                boost::filesystem::path file_path = *dir_iter;
                if (    (boost::filesystem::is_regular_file(dir_iter->status()) )
                     && (file_path.extension() == ".raw"                        ) )
                {
                    path1 = file_path.string();
                    break;
                }
            }
        }

        _setDepthPath( QString::fromStdString(path1) );
        if ( boost::filesystem::exists(path1) )
        {
            on_clearButton_clicked();
        }

        write_json( "settings.json", _pt );
    }
}

void QtCloudViewer::on_depthPathButton_clicked()
{
    QString fileName = QFileDialog::getOpenFileName(this,
                                                    tr("Open depth image file"),
                                                    _ui->depthPathEdit->text(),
                                                     tr("PGM or RAW (*.pgm *.raw);; DAT (*.dat);; All Files (*.*)") );
    if ( fileName.length() )
    {
        _setDepthPath( fileName );
        on_clearButton_clicked();
    }
}

void QtCloudViewer::on_clearButton_clicked()
{
    _ui->showIntersections->setChecked( false );
    on_showIntersections_toggled( _ui->showIntersections->isChecked() );
    //std::cout << __func__ << ": intersections now " << (_ui->showIntersections->isChecked() ? "CHECKED" : "NOT_CHECKED" ) << std::endl;

    _controller->loadNewCloudFromImagePaths( _ui->imagePathEdit->text().toStdString()
                                             , _ui->depthPathEdit->text().toStdString()
                                             , _ui->spatialSigmaSpinBox->value()
                                             , _ui->rangeSigmaSpinBox->value()
                                             , _ui->kernelRadiusSlider->value()
                                             , _ui->iterationsSpinBox->value()
                                             );

    _refreshPrimitivesList();
    on_pointSizeSpinBox_valueChanged( _ui->pointSizeSpinBox->value() );
    _saveLastPath( _ui->depthPathEdit->text().toStdString() );
}

int QtCloudViewer::_addPoints( std::string fileName )
{
    int err = EXIT_SUCCESS;

    // load cloud
    am::PCloud::Ptr cloud;
    if ( EXIT_SUCCESS == err )
    {
        // save last location
        _saveLastPath( fileName );

        // load points
        if ( fileName.find("pts2") != std::string::npos )
            err += am::AnnotationController::readAdditionalPoints<double, am::PCloud::Ptr>( cloud, fileName ); // *.pts2 - Yuoyi's format
        else if ( fileName.find("pts") != std::string::npos )
            err += am::AnnotationController::readAdditionalPoints<float,  am::PCloud::Ptr>( cloud, fileName ); // *.pts  - Tianjia's format
        else if ( fileName.find("ply") != std::string::npos )
        {
            cloud.reset( new am::PCloud() );
            err += abs( pcl::io::loadPLYFile(fileName, *cloud) );
        }
        else
        {
            std::cerr << "unkown filetype: " << fileName << std::endl;
            return EXIT_FAILURE;
        }
    }

    std::cout << "[" << __func__ << "]: " << "cloud has now " << (cloud?cloud->size():-1) << " points (-1 means no pointer...)" << std::endl;

    // add cloud
    if ( EXIT_SUCCESS == err )
    {
        err += _controller->addPoints( cloud );
    }

    // project test TODO: move away
    {
        //_controller->mapCurrentCloud();
    }

    // change point size
    if ( EXIT_SUCCESS == err )
    {
        on_pointSizeSpinBox_valueChanged( _ui->pointSizeSpinBox->value() );
//        QMessageBox::information( this
//                                  , "Info"
//                                  , ("Loaded additional points from " + fileName.toStdString()).c_str()
//                                  , QMessageBox::Ok );

        // change cached app state
        _pt.put( "paths.rgb"  , "" );
        _pt.put( "paths.depth", "" );
        _pt.put( "paths.points", fileName );
        write_json( "settings.json", _pt );
    }
    else
    {
        QMessageBox::warning( this
                              , "Warning"
                              , ("Could not load additional points from " + fileName).c_str()
                              , QMessageBox::Ok );

    }

    return err;
}

void QtCloudViewer::my_on_addPointsButton_clicked( std::string path )
{
    _addPoints( path );
}

void QtCloudViewer::my_on_addPointsButton_clicked()
{
    int     err         = EXIT_SUCCESS;
    QString fileName;

    // get file location
    if ( EXIT_SUCCESS == err )
    {
        fileName = QFileDialog::getOpenFileName( this
                                                 , tr("Open additional points file")
                                                 //QString::fromStdString(boost::filesystem::path(_ui->depthPathEdit->text().toStdString()).parent_path().string()),
                                                 , QString::fromStdString( _getLastPath(_getDepthDirAsStdString()) )
                                                 , tr("Points (*.pts *.pts2 *.ply);; All Files (*.*)")
                                                 );

        err += static_cast<int>( fileName.isEmpty() );
    }

    if ( EXIT_SUCCESS == err )
        my_on_addPointsButton_clicked( fileName.toStdString() );
}

void QtCloudViewer::on_removePointsButton_clicked()
{
    _controller->removeAllPoints();
    _ui->qvtkWidget->update();
}

/**
 * @brief QtCloudViewer::on_loadRegionsButton_clicked Loads NYU ground truth region masks
 */
void QtCloudViewer::on_loadRegionsButton_clicked()
{
    std::string frame_id = _getFrameIdFromDepthPath();
    char regions_fname[1024];
    sprintf( regions_fname, "/regions_from_labels_%06d.pgm", atoi(frame_id.c_str()) );

    QString fileName;
    fileName = QFileDialog::getOpenFileName(this,
                                            tr("Regions"),
                                            _getDepthDir()+regions_fname,
                                            tr("regions .PGM (*.pgm);; All Files (*.*)") );
    if ( fileName.isEmpty() )
        return;

    am::Params params;
    params.setDouble( "plane_distance_threshold", _ui->planeInlierThresholdDoubleSpinBox->value() );

    _controller->loadRegions( fileName.toStdString(), params );
    _refreshPrimitivesList();
}

void QtCloudViewer::on_loadMasksButton_clicked()
{
    QFileDialog dialog( this );

    std::string prev_path = _pt.get<std::string>("paths.last", QApplication::applicationDirPath().toStdString() );
    dialog.setDirectory( QString::fromStdString(prev_path) );
    dialog.setFileMode( QFileDialog::ExistingFiles );
    dialog.setNameFilter(trUtf8("masks (*.bmp *.pgm);; All Files (*.*)"));
    QStringList fileNames;
    if (dialog.exec())
        fileNames = dialog.selectedFiles();

    if ( !fileNames.empty() )
    {
        _saveLastPath( fileNames.at(0).toStdString() );

        std::vector<std::string> paths;
        for ( auto it = fileNames.begin(); it != fileNames.end(); ++it )
        {
            std::cout << "filename: " << it->toStdString() << std::endl;
            paths.push_back( it->toStdString() );
        }

//        if ( paths.size() )
//            _controller->addMasks()

    }
}

void QtCloudViewer::on_addModelButton_clicked()
{

//    QString fileName = QFileDialog::getOpenFileName(this,
//                                                    tr("Open model file"),
//                                                    _getDepthDir(),
//                                                    tr("Model files (*.obj *.stl *.ply);; All Files (*.*)") );
    QFileDialog dialog(this);
    {
        dialog.setWindowTitle( tr("Open model files") );
        dialog.setDirectory  ( _getDepthDir() );
        dialog.setFileMode   ( QFileDialog::ExistingFiles );
        dialog.setNameFilter ( trUtf8("Model files (*.obj *.stl *.ply);; All Files (*.*)") );
    }
    QStringList fileNames;
    if ( dialog.exec() )
        fileNames = dialog.selectedFiles();

    if ( !fileNames.isEmpty() )
    {
        for ( auto const& fileName : fileNames )
            _controller->addModel( fileName.toStdString() );
        _ui->qvtkWidget->update();
    }
}

int
QtCloudViewer::_setContacts( std::vector<am::Contact> const& contacts )
{
    if ( contacts.size() )
    {
        // clear prev
        on_showIntersections_clicked( false );
        // draw
        _ui->showIntersections->setChecked( true );
        _controller->drawIntersections( true, &contacts );

        return EXIT_SUCCESS;
    }
    else
    {
        std::cerr << "[" << __func__ << "]: no contacts to draw" << std::endl;
        return EXIT_FAILURE;
    }
}

void QtCloudViewer::my_on_loadContactsButton_clicked( std::string path )
{
    using namespace std;

    QString fileName;
    if ( path.empty() || !boost::filesystem::exists(boost::filesystem::path(path)) )
    {
        std::string suggested_path = _getLastPath(boost::filesystem::path(_ui->depthPathEdit->text().toStdString()).parent_path().string());
        if ( !_last_contacts_path.empty() )
            suggested_path = _last_contacts_path;
        fileName = QFileDialog::getOpenFileName(this,
                                                tr("Open graphs file"),
                                                QString::fromStdString( suggested_path ),
                                                tr("Graphs (*.txt);; All Files (*.*)") );
        if ( !fileName.isEmpty() && !path.empty() )
        {
            std::cerr << "copying " << fileName.toStdString() << " to " << path << std::endl;
            std::string command = "cp " + fileName.toStdString() + " " + path;
            int err = system( command.c_str() );
            if ( EXIT_SUCCESS != err )
            {
                std::cerr << "[" << __func__ << "]: " << "system(\"" << command << "\") exited with code " << err << std::endl;
            }
        }
    }
    else
        fileName = QString::fromStdString( path );

    int                      max_read_cuboid_id = 0;
    std::vector<am::Contact> contacts;
    int err = am::Contact::read( contacts, fileName.toStdString(), &max_read_cuboid_id );
    if ( EXIT_SUCCESS != err )
    {
        std::cerr << "[" << __func__ << "]: " << "Contact::read returned failure...exiting...\n";
        return;
    }

    int max_current_cuboid_id = _ui->primitivesListWidget->count() - 1;
    if ( max_read_cuboid_id > max_current_cuboid_id )
    {
        //int reply = QMessageBox::warning( this, "Warning", "More proxies in contacts, than cuboids...still read?", QMessageBox::Yes | QMessageBox::No );
        std::cerr << "[" << __func__ << "]: " << "[Warning] More proxies in contacts(" << fileName.toStdString() << "), than cuboids...continueing\n" << std::endl;
        int reply = QMessageBox::Yes;
        if ( reply == QMessageBox::No )
            return;
        else
        {
            // filter
            am::Contact::Vec tmp_contacts;
            for ( size_t contact_id = 0; contact_id != contacts.size(); ++contact_id )
            {
                if ( std::find_if( contacts[contact_id].ids.begin()
                                   , contacts[contact_id].ids.end()
                                   , [&max_current_cuboid_id](int const& id){ return id > max_current_cuboid_id;}) == contacts[contact_id].ids.end() )
                {
                    tmp_contacts.push_back( contacts[contact_id] );
                }
            }
            contacts = tmp_contacts;
        }
    }

    _last_contacts_path = fileName.toStdString();
    _setContacts( contacts );
}

/// SEGMENT ///

void QtCloudViewer::on_annotateButton_clicked()
{
    if ( !_controller )
    {
        std::cerr << "QtCloudViewer::on_annotateButton_clicked(): no controller to send info to...\n";
        return;
    }

    // save state
    {
        _pt.put( "paths.rgb"  , _ui->imagePathEdit->text().toStdString() );
        _pt.put( "paths.depth", _ui->depthPathEdit->text().toStdString() );
    }


    QRadioButton *p_planesRadioButton = _ui->annotationGroupBox->findChild<QRadioButton*>("planesRadioButton");

    am::Params params;
    params.setInt   ( "task_selected", p_planesRadioButton->isChecked() ? 0 : 1 );
    params.setString( "img_path", _ui->imagePathEdit->text().toStdString() );
    params.setString( "dep_path", _ui->depthPathEdit->text().toStdString() );
    params.setDouble( "plane_distance_threshold", _ui->planeInlierThresholdDoubleSpinBox->value() );
    params.setDouble( "betas[rgb]", _ui->rgbBetaDoubleSpinBox->value() );
    params.setDouble( "betas[depth]", _ui->depthBetaDoubleSpinBox->value() );
    _controller->annotate_button_clicked( params );

    _refreshPrimitivesList();

}

void QtCloudViewer::on_planeInlierThresholdSlider_sliderMoved( int position )
{
    _ui->planeInlierThresholdDoubleSpinBox->setValue( position / 100. );
    _pt.put( "params.plane_inlier_threshold", position / 100. );
}

void QtCloudViewer::on_planeInlierThresholdDoubleSpinBox_valueChanged( double arg1 )
{
    _ui->planeInlierThresholdSlider->setValue( round(arg1 * 100.) );
    _pt.put( "params.plane_inlier_threshold", arg1 );
}

void QtCloudViewer::on_segmentButton_clicked()
{
    Eigen::Vector3f up; bool use_up = false;
    if ( !_ui->upDirLineEdit->text().isEmpty() )
    {
        std::istringstream iss( _ui->upDirLineEdit->text().toStdString() );
        std::string word;

        std::vector<float> floats;
        while ( std::getline(iss, word, ',') )
            floats.push_back( atof(word.c_str()) );

        if ( floats.size() >= 3 )
        {
            up << floats[0], floats[1], floats[2];
            use_up = true;
        }
    }
    _controller->segment( /*                 min_cluster_size: */ _ui->minClusterSizeSpinBox->value()
                          , /*              cluster_tolerance: */ _ui->clusterToleranceSpinBox->value()
                          , /*           voxelgrid_resolution: */ _ui->voxelResolutionSpinBox->value()
                          , /* slab_size_threshold_percentage: */ _ui->slab_size_threshold_SpinBox->value() * 0.01f
                          , /*                 max_slab_count: */ _ui->maxSlabsSpinBox->value()
                          , /*                         up_dir: */ use_up ? &up : NULL );
    _refreshPrimitivesList();
}

void QtCloudViewer::on_clusterButton_clicked()
{
    _controller->cluster( /*                 min_cluster_size: */ _ui->minClusterSizeSpinBox->value()
                          , /*              cluster_tolerance: */ _ui->clusterToleranceSpinBox->value()
                          );
    _refreshPrimitivesList();
}

void QtCloudViewer::on_isFloorCheckBox_clicked(bool checked)
{
    QList<QListWidgetItem*> selected_items = _ui->primitivesListWidget->selectedItems();
    for ( QListWidgetItem* widgetitemptr : selected_items )
    {
        int row = _ui->primitivesListWidget->row( widgetitemptr );
        _controller->setPrimitiveIsFloor( row, checked, /* selected: */ true );

        _controller->setPrimitiveIsWall( row, false, /* selected: */ false );
        _ui->isWallCheckBox->setChecked( false );

    }
    _ui->qvtkWidget->update();
}

void QtCloudViewer::on_isWallCheckBox_clicked(bool checked)
{
    QList<QListWidgetItem*> selected_items = _ui->primitivesListWidget->selectedItems();
    for ( QListWidgetItem* widgetitemptr : selected_items )
    {
        int row = _ui->primitivesListWidget->row( widgetitemptr );
        _controller->setPrimitiveIsWall( row, checked, /* selected: */ true );

        _controller->setPrimitiveIsFloor( row, false, /* selected: */ false );
        _ui->isFloorCheckBox->setChecked( false );
    }

    _ui->qvtkWidget->update();
}

void QtCloudViewer::on_isSphereCheckBox_clicked(bool checked)
{
    QList<QListWidgetItem*> selected_items = _ui->primitivesListWidget->selectedItems();
    for ( QListWidgetItem* widgetitemptr : selected_items )
    {
        int row = _ui->primitivesListWidget->row( widgetitemptr );
        _controller->setPrimitiveIsSphere( row, checked, /* selected: */ true );
        _ui->isCylinderCheckBox->setChecked( false );
    }

    _ui->qvtkWidget->update();
}

void QtCloudViewer::on_isCylinderCheckBox_clicked( bool checked )
{
    QList<QListWidgetItem*> selected_items = _ui->primitivesListWidget->selectedItems();
    for ( QListWidgetItem* widgetitemptr : selected_items )
    {
        int row = _ui->primitivesListWidget->row( widgetitemptr );
        _controller->setPrimitiveIsSphere( row, false, /* selected: */ true );
        _controller->setPrimitiveIsCylinder( row, checked, /* selected: */ true );
        _ui->isSphereCheckBox->setChecked( false );
    }

    _ui->qvtkWidget->update();

}

void QtCloudViewer::on_isTargetCheckBox_clicked(bool checked)
{
    QList<QListWidgetItem*> selected_items = _ui->primitivesListWidget->selectedItems();
    for ( QListWidgetItem* widgetitemptr : selected_items )
    {
        int row = _ui->primitivesListWidget->row( widgetitemptr );
        _controller->setPrimitiveIsTarget( row, checked );
    }
}

void QtCloudViewer::on_saveWallsButton_clicked()
{
    QString walls_fileName;
    if ( _last_cuboids_path.empty() ) // || !boost::filesystem::exists(cuboidPathToWallsPath(_last_cuboids_path))
    {
        QString datFileName = QFileDialog::getOpenFileName( this,
                                                 tr("Save walls props file for which cuboids.dat"),
                                                 _getDepthDir(),
                                                 tr("Cuboids(*.dat);; All Files (*.*)") );
        if ( !datFileName.isEmpty() )
            walls_fileName = QString::fromStdString( am::util::cuboidPathToWallsPath(datFileName.toStdString()) );
    }
    else
    {
        walls_fileName = QString::fromStdString( am::util::cuboidPathToWallsPath(_last_cuboids_path) );
    }

    if ( !walls_fileName.isEmpty() )
    {
       _saveWalls( walls_fileName.toStdString() );
    }
}

void QtCloudViewer::on_loadWallsButton_clicked()
{
    QString fileName;
    if ( _last_cuboids_path.empty() || !boost::filesystem::exists( am::util::cuboidPathToWallsPath(_last_cuboids_path)) )
    {
        std::string last_path = _getLastPath(/* default_value: */ _getDepthDirAsStdString());
        std::string suggested_path = last_path;

        std::vector<std::string> wall_files;
        am::util::get_files_by_ext( /*    out: */ wall_files
                                    , /* dirs: */ boost::filesystem::path(last_path)
                                    , /*  ext: */ "walls" );
        if ( wall_files.size() )
            suggested_path = wall_files[0];

        if ( wall_files.size() == 1 )
        {
            fileName = QString::fromStdString( wall_files[0] );
        }
        else
        {
            fileName = QFileDialog::getOpenFileName(this,
                                                    tr("Walls"),
                                                    QString::fromStdString( suggested_path ),
                                                    tr("Walls (*.walls);; All Files (*.*)") );
        }
    }
    else
        fileName = QString::fromStdString( am::util::cuboidPathToWallsPath(_last_cuboids_path) );

    if ( fileName.isEmpty() )
        return;

    loadWalls( fileName.toStdString() );
}

/// PRIMITIVEs ///

void QtCloudViewer::on_saveButton_clicked()
{
    std::string start_path = _controller->getOutputPath();
    QFileDialog dialog( this,
                       tr("Open save depth file"),
                       QString::fromStdString(start_path),
                       tr("All Files (*.*)") );
    dialog.setFileMode( QFileDialog::Directory );
    int res = dialog.exec();
    if ( res )
    {
        QString fileName = dialog.selectedFiles().at(0);

        //QString fileName = dialog.getOpenFileName();
        std::cout << "onPushButton_clicked(): " << fileName.toStdString() << std::endl;
        //    QString fileName = QFileDialog::getOpenFileName( this,
        //                                                     tr("Open save location"),
        //                                                     _ui->depthPathEdit->text(),
        //                                                     tr("PGM (*.pgm);; PNG (*.png);; JPG(*.jpg *.jpeg);; All Files (*.*)") );

        am::Params params;
        params.setString( "img_path", _ui->imagePathEdit->text().toStdString() );
        params.setString( "dep_path", _ui->depthPathEdit->text().toStdString() );

        _controller->save( fileName.toStdString(), params );
    }
    else
    {
        std::cerr << "res is " << res << std::endl;
    }
}

void QtCloudViewer::my_on_loadButton_clicked( std::string path )
{
    QString fileName;
    if ( path.empty() || !boost::filesystem::exists(boost::filesystem::path(path)) )
    {
        fileName = QFileDialog::getOpenFileName( this
                                                 , tr("Open cuboids file")
                                                 , QString::fromStdString( _getLastPath(/* default_value: */ _getDepthDirAsStdString()) )
                                                 //QString::fromStdString(boost::filesystem::path(_ui->depthPathEdit->text().toStdString()).parent_path().string()),
                                                 //QString::fromStdString( _pt.get<std::string>("paths.loaded_cuboids",
                                                 //                                              _ui->depthPathEdit->text().toStdString() )),
                                                 , tr("Cuboids (*.dat *.box *.gt);; All Files (*.*)")
                                                 );
        if ( !fileName.isEmpty() && !path.empty() )
        {
            std::cerr << "copying " << fileName.toStdString() << " to " << path << std::endl;

            std::string command = "cp " + fileName.toStdString() + " " + path;
            int err = system( command.c_str() );
            if ( EXIT_SUCCESS != err )
            {
                std::cerr << "[" << __func__ << "]: " << "system(\"" << command << "\") exited with code " << err << std::endl;
            }
        }
    }
    else
        fileName = QString::fromStdString( path );

    if ( fileName.isEmpty() )
        return;

    am::Params params;
    params.setDouble( "plane_distance_threshold", _ui->planeInlierThresholdDoubleSpinBox->value() );
    std::vector<am::Contact> contacts;
    int ret = _controller->loadCuboids( fileName.toStdString(), &params, &contacts );
    if ( ret == EXIT_SUCCESS )
    {
        _pt.put( "paths.loaded_cuboids", fileName.toStdString() );
        _saveLastPath( fileName.toStdString() );
        write_json( "settings.json", _pt );
        _last_cuboids_path = fileName.toStdString();

        _refreshPrimitivesList();
        _refreshDisplayProperties();
//        on_pointSizeSpinBox_valueChanged( _ui->pointSizeSpinBox->value() );
//        on_showIntersections_toggled( _ui->showIntersections->isChecked() );
//        on_colourEdgesCheckBox_stateChanged( _ui->colourEdgesCheckBox->isChecked() );

        //on_loadWallsButton_clicked();

        if ( contacts.size() )
        {
            am::util::saveGTContacts( contacts, fileName.toStdString() );
//            _setContacts( contacts );
        }
    }
}

void QtCloudViewer::_clearPrimitives()
{
    _ui->primitivesListWidget->setUpdatesEnabled( false );

    const int cnt = _ui->primitivesListWidget->count();
    std::vector<int> ids_to_delete;
    for ( int row = 0; row != cnt; ++row )
        ids_to_delete.push_back( row );

    TIC;
    _ui->primitivesListWidget->clear();
    TOC("clear");

    RETIC;
    //_ui->primitivesListWidget->unblockSignals(); // optionally
    _ui->primitivesListWidget->setUpdatesEnabled(true);
    _ui->primitivesListWidget->update(); // force update
    TOC("update");

    RETIC;
    _controller->removePrimitive( ids_to_delete );
    TOC("removeprimitives");

    RETIC;
    _ui->qvtkWidget->update();
    TOC("qvtkupdate");
}

void QtCloudViewer::on_removePrimitiveButton_clicked()
{
    TIC
    QList<QListWidgetItem*> selected_items = _ui->primitivesListWidget->selectedItems();
    TOC("Selected items");
    std::vector<int> ids_to_delete;

    if ( selected_items.size() == _ui->primitivesListWidget->count() )
    {
        std::cout << "[" << __func__ << "]: "
                  << "selected_items.size() " << selected_items.size() << " == "
                  << _ui->primitivesListWidget->count() << " _ui->primitivesListWidget->count()"
                  << std::endl;

        _clearPrimitives();
        return;
    }

    _ui->primitivesListWidget->setUpdatesEnabled(false);
    for ( QListWidgetItem* widgetitemptr : selected_items )
    {
        RETIC;
        //std::cout << "would remove " << widgetitemptr->text().toStdString() << std::endl;
        int row = _ui->primitivesListWidget->row( widgetitemptr );
        std::cout << "adding " << row << "to ids_to_delete\n";
        ids_to_delete.push_back( row );
        std::cout<<"ids_to_delete:";for(size_t vi=0;vi!=ids_to_delete.size();++vi)std::cout<<ids_to_delete[vi]<<" ";std::cout << "\n";
        TOC("row");
    }

    std::sort( ids_to_delete.begin(), ids_to_delete.end() );
    for ( int i = 0; i != ids_to_delete.size(); ++i )
    {
        RETIC;
        auto ptr = _ui->primitivesListWidget->takeItem( ids_to_delete[i] - i );
        TOC("takeItem");

        if ( ptr ) delete ptr;
    }

    RETIC;
    //_ui->primitivesListWidget->unblockSignals(); // optionally
    _ui->primitivesListWidget->setUpdatesEnabled(true);
    _ui->primitivesListWidget->update(); // force update
    TOC("update");

    RETIC;
    _controller->removePrimitive( ids_to_delete );
    TOC("removeprimitives");

    RETIC;
    _ui->qvtkWidget->update();
    TOC("qvtkupdate");
}

void QtCloudViewer::on_hidePrimitiveButton_clicked()
{
    QList<QListWidgetItem*> selected_items = _ui->primitivesListWidget->selectedItems();
    for ( QListWidgetItem* widgetitemptr : selected_items )
    {
        int row = _ui->primitivesListWidget->row( widgetitemptr );
        _controller->toggleHidePrimitive( row );

        if ( _controller->getPrimitive(row).hidden )
        {
            _ui->primitivesListWidget->item( row )->setTextColor( QColor(200,200,200) );
        }
        else
        {
            _ui->primitivesListWidget->item( row )->setTextColor( QColor(0,0,0) );
        }
    }
    _ui->qvtkWidget->update();
}

void QtCloudViewer::on_primitivesListWidget_currentRowChanged( int currentRow )
{
    auto ptr = _ui->primitivesListWidget->item( currentRow );
    if ( ptr )
        std::cout << "selected " << ptr->text().toStdString() << std::endl;

    _controller->selectPrimitive( _getSelectedRows() );
    _ui->qvtkWidget->update();

    if ( currentRow > -1 )
    {
        auto primitive = _controller->getPrimitive( currentRow );
        _ui->isFloorCheckBox->setChecked( primitive.floor );
        _ui->isWallCheckBox->setChecked( primitive.wall );
        _ui->isSphereCheckBox->setChecked( primitive.shape == am::Cuboid::SPHERE );
        _ui->isCylinderCheckBox->setChecked( primitive.shape == am::Cuboid::CYLINDER );

        _ui->isTargetCheckBox->setChecked( primitive.rotation_target );
    }
    else
    {
        _ui->isFloorCheckBox->setChecked( false);
        _ui->isWallCheckBox->setChecked ( false);
        _ui->isTargetCheckBox->setChecked( false );
        _ui->isSphereCheckBox->setChecked( false );
        _ui->isCylinderCheckBox->setChecked( false );
    }
}

void QtCloudViewer::on_primitivesListWidget_itemSelectionChanged()
{
    _controller->selectPrimitive( _getSelectedRows() );
    _ui->qvtkWidget->update();
}

void QtCloudViewer::on_unselectButton_clicked()
{
    _ui->primitivesListWidget->setCurrentRow( -1 );
}

/// DISPLAY ///

void QtCloudViewer::on_sphereSizeSlider_valueChanged( int value )
{
    _ui->sphereSizeSpinBox->setValue( value / 100. );
    on_sphereSizeSpinBox_editingFinished();
}

void QtCloudViewer::on_sphereSizeSpinBox_editingFinished()
{
    _ui->sphereSizeSlider->setValue( _ui->sphereSizeSpinBox->value() * 100. );
    _controller->setSphereSize( _ui->sphereSizeSpinBox->value() );
    _ui->qvtkWidget->update();
}

void QtCloudViewer::on_colourEdgesCheckBox_stateChanged( int arg1 )
{
    _controller->setColourEdges( _ui->colourEdgesCheckBox->isChecked() );
    _ui->qvtkWidget->update();
}

void QtCloudViewer::on_opacitySlider_valueChanged(int value)
{
    std::cout << __func__ << ": " << value << std::endl;
    _ui->opacitySpinBox->setValue( value / 100. );
}

void QtCloudViewer::on_opacitySpinBox_valueChanged( double value )
{
    std::cout << __func__ << ": " << value << std::endl;

    _ui->opacitySlider->setValue( value * 100. );

    // needs only call here, slider affects this as well
    std::vector<int> row_ids = _getSelectedRows();

    if ( row_ids.size() )
        for ( auto const& row_id : row_ids )
        {
            _controller->setCuboidOpacity( row_id, value );
        }
    else
        _controller->setCuboidOpacity( -1, value ); // all

    _ui->qvtkWidget->update();
}

void QtCloudViewer::on_colourCuboids_stateChanged(int arg1)
{
    _controller->setColourCodeCuboids( _ui->colourCuboids->isChecked() );
    _ui->qvtkWidget->update();
}

void QtCloudViewer::on_colorButton_clicked()
{
    std::vector<int> row_ids = _getSelectedRows();
    for ( auto const& row_id : row_ids )
    {
        QColor color = QColorDialog::getColor( Qt::yellow );
        std::cout << "color is: "
                  << qRed   (color.rgb() ) << ", "
                  << qGreen( color.rgb() ) << ", "
                  << qBlue ( color.rgb() ) << ", "
                  << "isValid: " << (color.isValid() ? "YES" : "NO" )
                  << std::endl;

        if ( color.isValid() )
        {
            Eigen::Vector4f colour;
            colour << qRed( color.rgb() )     / 255.f
                      , qGreen( color.rgb() ) / 255.f
                      , qBlue( color.rgb() )  / 255.f
                      , _ui->opacitySpinBox->value();
            _controller->setCuboidColour( row_id, colour );
        }
    }

    _ui->qvtkWidget->update();
}

void QtCloudViewer::on_pointSizeSpinBox_valueChanged( double arg1 )
{
    _controller->setPointSize( arg1 );
    _ui->qvtkWidget->update();
}

void QtCloudViewer::on_showIntersections_toggled(bool checked)
{
    //std::cout << "[" << __func__ << "]: " << ": checked " << (checked?"CHECKED":"NOT") << std::endl;
}

void QtCloudViewer::on_showIntersections_clicked(bool checked)
{
    //std::cout << "[" << __func__ << "]: " << ": checked " << (checked?"CHECKED":"NOT") << std::endl;

    // clear
    _controller->drawIntersections( false );

    // redraw
    {
        // read prev
        am::Contact::Vec contacts( _controller->getLatestContacts() );
        if ( checked )
            _controller->drawIntersections( checked, contacts.size() ? &contacts : NULL );
    }

    // update
    _ui->qvtkWidget->update();
}

void QtCloudViewer::on_showInvalidCloud_toggled(bool checked)
{
    std::cout << __func__ << ": is now " << (checked ? "CHECKED" : "NOT CHECKED") << std::endl;

    _controller->showAuxCloud( 0, checked );
    on_pointSizeSpinBox_valueChanged( _ui->pointSizeSpinBox->value() );
    _ui->qvtkWidget->update();
}

void QtCloudViewer::on_infoButton_clicked()
{
    std::string lengths_message;
    std::string intersection_message;
    std::string colour_message;
    std::vector< std::pair<int,int> > intersections;

    std::vector<int> row_ids = _getSelectedRows();
    std::map<int,std::shared_ptr<am::Cuboid> > cuboid_ptrs;
    for ( auto const& row_id : row_ids )
    {
        ca::DisplayedPrimitive primitive = _controller->getPrimitive( row_id );
        auto cuboid_ptr                  = primitive.CuboidPtr();
        auto lengths                     = cuboid_ptr->getSideLengths();

        char chr[255];
        for ( auto const& length : lengths )
        {
            sprintf( chr, "%f", length );
            lengths_message += std::string(chr) + "\n";
        }

        sprintf( chr, "{%f,%f,%f}\n", primitive.colour(0), primitive.colour(1), primitive.colour(2) );
        colour_message += std::string( chr );

        for ( auto const& prev_pair : cuboid_ptrs )
        {
            if ( prev_pair.first < row_id )
            {
                std::cout << prev_pair.first << " - " << row_id;
                if ( am::Cuboid::intersect(*prev_pair.second, *cuboid_ptr, am::AnnotationController::INTERSECT_OFFSET) )
                {
                    if ( prev_pair.first < row_id )
                        intersections.push_back( {prev_pair.first, row_id} );
                    else
                        intersections.push_back( {row_id,prev_pair.first} );
                }
            }
        }

        std::cout << " setting " << row_id << std::endl;
        cuboid_ptrs[row_id] = cuboid_ptr;
    }

    std::sort( intersections.begin(), intersections.end(), []( std::pair<int,int> a, std::pair<int,int> b){
        if ( a.first == b.first )
            return a.second < b.second;
        else
            return a.first < b.first;
    });


    char chr[255];
    for ( auto const& intersection : intersections )
    {
        sprintf( chr, "%d\t%d\tT", intersection.first, intersection.second );
        intersection_message += std::string(chr) + "\n";
    }

    auto contacts = _controller->getLatestContacts();
    {
        std::sort( contacts.begin(), contacts.end(), []( am::Contact const& a, am::Contact const&b ){
            if ( a.ids[0] == b.ids[0] )
                return a.ids[1] < b.ids[1];
            else
                return a.ids[0] < b.ids[0];
        });

        intersection_message += "AltContacts:\n";
        for ( auto const& contact : contacts )
        {
            sprintf( chr, "%d\t%d\t%s", contact.ids[0], contact.ids[1], ((contact.type==am::Contact::FIXED)?"F":"T") );
            intersection_message += std::string(chr) + "\n";
        }
    }

    QMessageBox::StandardButton reply;
    reply = QMessageBox::question( this,
                                   "Info",
                                   (lengths_message + "\nIntersections\n" + intersection_message + "\nColours\n" + colour_message + "\nStore alt_contacts?").c_str(),
                                    QMessageBox::Yes | QMessageBox::No );
    if ( reply == QMessageBox::Yes )
    {
        QString fileName = QFileDialog::getSaveFileName(this,
                                                        tr("Save intersections"),
                                                        QString::fromStdString( _getLastPath(".") ),
                                                        tr("Graphs (*.txt);; All Files (*.*)") );
        if ( !fileName.isEmpty() )
            am::Contact::write( contacts, fileName.toStdString() );
    }
}

void QtCloudViewer::on_printButton_clicked()
{
    char chr[255];

    std::string lengths_message;

    std::vector<int> row_ids = _getSelectedRows();
    std::map<int,std::shared_ptr<am::Cuboid> > cuboid_ptrs;
    for ( auto const& row_id : row_ids )
    {
        sprintf( chr, "Cuboid%d:", row_id );
        lengths_message += std::string(chr) + "\n";

        ca::DisplayedPrimitive primitive = _controller->getPrimitive( row_id );
        auto cuboid_ptr                  = primitive.CuboidPtr();

        for ( auto const& plane : cuboid_ptr->Planes().Primitives() )
        {
            sprintf( chr, "Plane: { %f, %f, %f, %f }"
                     , plane.CoeffsPtr()->values[0]
                     , plane.CoeffsPtr()->values[1]
                     , plane.CoeffsPtr()->values[2]
                     , plane.CoeffsPtr()->values[3]
                    );
            lengths_message += std::string(chr) + "\n";
        }
    }

    QMessageBox::StandardButton reply;
    reply = QMessageBox::information( this
                                      , "Info"
                                      , lengths_message.c_str()
                                      , QMessageBox::Ok );

}

/// CAMERA ///

void QtCloudViewer::on_cameraInfoButton_clicked()
{
    std::stringstream ss;
    auto pos_focal_up = _controller->getCameraPosFocalUp();

    static const std::vector<std::string> tags = { "pos", "focal", "up" };
    for ( int vec_id = 0; vec_id != pos_focal_up.size(); ++vec_id )
    {
        ss << tags[vec_id] << vec_id << ": " << pos_focal_up[vec_id].transpose() << std::endl;
    }
    ss << "in one line: ";
    for ( int i = 0; i != pos_focal_up.size(); ++i )
        ss << pos_focal_up[i](0) << ","
           << pos_focal_up[i](1) << ","
           << pos_focal_up[i](2) << ",";

    auto proj_mat = _controller->getProjMatrix();
    auto view_mat = _controller->getViewMatrix();
    ss << "\nproj_mat^T:\n" << proj_mat.transpose() << "\nview_mat^T:\n" << view_mat.transpose();

    QMessageBox::StandardButton reply;
    reply = QMessageBox::information( this
                                      , "Info"
                                      , ss.str().c_str()
                                      , QMessageBox::Ok );
}

void QtCloudViewer::on_setCameraParamsButton_clicked()
{
    std::string params_str = _ui->setCameraParamsLineEdit->text().toStdString();
    _set_camera_params( params_str );
}

void QtCloudViewer::my_on_loadCameraParamsButton_clicked( std::string path )
{
    QString fileName;
    if ( path.empty() || !boost::filesystem::exists(boost::filesystem::path(path)) )
    {
        fileName = QFileDialog::getOpenFileName(this,
                                                tr("Camera Params"),
                                                //_getDepthDir(),
                                                QString::fromStdString( _getLastPath(".") ),
                                                tr("Cam parameters (*.cam);; All Files (*.*)") );
        if ( !fileName.isEmpty() && !path.empty() )
        {
            std::cerr << "copying " << fileName.toStdString() << " to " << path << std::endl;
            std::string command = "cp " + fileName.toStdString() + " " + path;
            int err = system( command.c_str() );
            if ( EXIT_SUCCESS != err )
            {
                std::cerr << "[" << __func__ << "]: " << "system(\"" << command << "\") exited with code " << err << std::endl;
            }
        }
    }
    else
        fileName = QString::fromStdString( path );

    std::ifstream myfile;
    myfile.open( fileName.toStdString() );
    if ( !myfile.is_open() )
    {
        std::cerr << __func__ << " could not open file " << path << std::endl;
        return;
    }

    std::string line;
    if ( getline(myfile,line) )
    {
        _set_camera_params( line );
    }
    else
    {
        std::cerr << __func__ << ": camfile has no lines..." << std::endl;
    }
}

void QtCloudViewer::on_saveCameraParamsButton_clicked()
{
    QString fileName = QFileDialog::getSaveFileName(this,
                                                    tr("Save camera pose"),
                                                    //QString::fromStdString(boost::filesystem::path(_ui->depthPathEdit->text().toStdString()).parent_path().string()+"/back_view.cam"),
                                                    QString::fromStdString( _getLastPath(".") ),
                                                    tr("Camera poses (*.cam);; All Files (*.*)") );
    if ( !fileName.isEmpty() )
    {
        int err = _controller->saveCameraPose( enforceExtension(".cam",fileName.toStdString()) );
        return;

        if ( err == EXIT_SUCCESS )
        {
            QMessageBox::information( this
                                      , "Info"
                                      , ("Saved camera pose to" + fileName.toStdString()).c_str()
                                      , QMessageBox::Ok );
        }
        else
        {
            QMessageBox::warning( this
                                  , "Info"
                                  , ("Could not save camera pose to" + fileName.toStdString()).c_str()
                                  , QMessageBox::Ok );
        }
    }
}

void QtCloudViewer::on_resetCameraButton_clicked()
{
    _controller->resetCamera();
    on_pointSizeSpinBox_valueChanged( _ui->pointSizeSpinBox->value() );
    //_ui->showInvalidCloud->setChecked( true );
    //_ui->qvtkWidget->update();
}

/// ANIMATION ///

void QtCloudViewer::on_getRotInfoButto_clicked()
{
    QMessageBox::information( this, "info", QString::fromStdString( _controller->getRotInfo() ), QMessageBox::Ok);
}

void QtCloudViewer::on_changeAnimationSavePathButton_clicked()
{
    QFileDialog dialog( this,
                        tr("Save animation"),
                        tr("/media/Data2/workspace_store/animation"),
                        tr("All Files (*.*)") );
    dialog.setFileMode( QFileDialog::Directory );
    int res = dialog.exec();
    if ( res )
    {
        boost::filesystem::path dir( _getDepthDirAsStdString() );
        std::string folder_name;
        for ( auto it = dir.begin(); it != dir.end(); ++it )
            folder_name = it->string();

        _ui->animationSavePath->setText( dialog.selectedFiles().at(0) + "/" + QString::fromStdString(folder_name) );
    }
}

void QtCloudViewer::on_rotateCameraCheckBox_clicked()
{
    bool checked = _ui->rotateCameraCheckBox->isChecked();
    std::cout << __func__ << ": " << (checked?"CHECKED":"NOT checked") << std::endl;

    if ( checked )
    {
        // Create an instance of your woker
        RotatorThread *workerThread = new RotatorThread( _controller );
        // Connect our signal and slot
        connect(workerThread, SIGNAL(progressChanged(QString)), SLOT(onProgressChanged(QString)));
        // Setup callback for cleanup when it finishes
        connect(workerThread, SIGNAL(finished()), workerThread, SLOT(deleteLater()));
        // dest directory
        {
            boost::filesystem::path dest_path( _ui->animationSavePath->text().toStdString() );
            if ( !boost::filesystem::exists(dest_path) )
            {
                boost::filesystem::create_directory( dest_path );
            }
            _animation_save_path = dest_path.string();
            _frame_count = 0;
        }

        // 282 rot: 0.0217914 0.918758  0.39422; centroid: -0.0294572 0.194195  1.59438
        // 51_opt: rot: 0.0217914 0.918758  0.39422; centroid: -0.0294572 0.194195  1.59438
        // desk: rot: -0.0305702, 0.937246, 0.347327; centroid: -0.153334, 0.0654176,   1.26939
        // heap: rot: 0.0113994, 0.890302, 0.455227; centroid: 0.0620371, 0.123995, 0.992288
        // teaser: rot: -0.0662052,  0.98009, 0.187192, -0.112986, 0.436043, 2.1205
        //  teaser with: rot: -0.0662052,  0.98009, 0.187192; centroid: -0.112986, 0.436043,   2.1205
        // 5lobby with: rot: -0.00754832, 0.968396, 0.249302; centroid: -0.0496226, 0.259554,  2.37553


        Eigen::Vector3f centroid, rot_axis;
        //rot_axis = (Eigen::Vector3f() << 0.0217914f,0.918758f,0.39422f ).finished();
        //centroid = (Eigen::Vector3f() << -0.0294572f,0.194195f,1.59438f).finished();
        std::pair<int,std::string> succ;
        if ( !_ui->rotAndCentroidText->text().isEmpty() )
        {

            std::istringstream iss( _ui->rotAndCentroidText->text().toStdString() );
            std::string word;

            std::vector<float> floats;
            while ( std::getline(iss, word, ',') )
            {
                floats.push_back( atof(word.c_str()) );
            }
            if ( floats.size() != 6 )
                std::cerr << __func__ << ": sorry...not 6" << std::endl;
            else
            {
                rot_axis = (Eigen::Vector3f() << floats[0], floats[1], floats[2]).finished();
                centroid = (Eigen::Vector3f() << floats[3], floats[4], floats[5]).finished();
                std::stringstream ss;
                ss << rot_axis.transpose() << "; " << centroid.transpose();
                QMessageBox::information( this, "asdf", QString::fromStdString(ss.str()));
            }
            succ = workerThread->prepare( 360.f, 0.f, 1.f, &centroid, &rot_axis  );
        }
        else
            succ = workerThread->prepare( 360.f, 0.f, 1.f );

        //rot_axis = (Eigen::Vector3f() << -0.00754832, 0.968396, 0.249302 ).finished();
        //centroid = (Eigen::Vector3f() << -0.0496226, 0.259554,  2.37553).finished();
        // Run, Forest, run!

        //succ = workerThread->prepare( 360.f, 0.f, 1.f, &centroid, &rot_axis  );
        //auto succ = workerThread->prepare( 360.f, 0.f, 1.f );
        if ( succ.first != EXIT_SUCCESS )
        {
            QMessageBox::warning( this, __func__, succ.second.c_str(), QMessageBox::Ok );

            // uncheck, if returned error
            if ( checked ) _ui->rotateCameraCheckBox->setChecked( false );
        }
        else
        {
            QMessageBox::information( this, "Centroidaxis", QString::fromStdString(succ.second), QMessageBox::Ok );
            if ( _thread )
            {
                std::cout << __func__ << "stopping thread for restart..." << std::endl;

                _thread->Stop();
                _thread->terminate();
                _thread = nullptr;
            }
            _thread = workerThread;

            workerThread->start(); // This invokes WorkerThread::run in a new thread
        }
    }
    else
    {
        if ( _thread )
        {
            std::cout << __func__ << "stopping thread..." << std::endl;
            _thread->Stop();
            _thread->terminate();
            _thread = nullptr;
        }
    }

    return;

    auto succ = _controller->rotateCamera( checked, _ui->qvtkWidget );
    if ( succ.first != EXIT_SUCCESS )
    {
        QMessageBox::warning( this, __func__, succ.second.c_str(), QMessageBox::Ok );

        // uncheck, if returned error
        if ( checked ) _ui->rotateCameraCheckBox->setChecked( false );
    }
}

void QtCloudViewer::onProgressChanged(QString info)
{
    // Processing code
    std::cout << __func__ << ": " << info.toStdString() << std::endl;
    _ui->qvtkWidget->update();

    char tit[255];
    sprintf( tit, "frame%06d.bmp", _frame_count );
    my_on_screenshotButton_clicked( _animation_save_path + "/" + tit );
    _thread->setDoneFrameId( _frame_count );
    ++_frame_count;
}

void QtCloudViewer::on_recordRowButton_clicked()
{
//    % 1: initial big
//    % 2.1: init boxes frontal coloured
//    % 2.2: init boxes back coloured
//    % 3.1: opti graph without points frontal
//    % 3.2: opti graph with points back

    on_resetCameraButton_clicked();
    on_clearButton_clicked();
    _ui->colourCuboids->setChecked( true );
    //_ui->pointSizeSpinBox()

    bool has_back_pose = boost::filesystem::exists( boost::filesystem::path(_getDepthDirAsStdString()+"/back_view.cam") );

    //    % 1: initial big
    if ( has_back_pose )
    {
        _ui->showInvalidCloud->setChecked( true );
        on_showInvalidCloud_toggled( _ui->showInvalidCloud->isChecked() );
        my_on_screenshotButton_clicked( _getDepthDirAsStdString() + "/input.png" );
    }

    //    % 3.2: opti graph with points back
    {
        //_ui->primitivesListWidget->setCurrentRow(-1);
        my_on_loadButton_clicked( _getDepthDirAsStdString() + "/optimized.dat" );
        if ( !has_back_pose )
        {
            my_on_loadContactsButton_clicked( _getDepthDirAsStdString() + "/graph.txt" );
            QMessageBox::information( this
                                      , "Info"
                                      , "Please save a back pose as \"back_view.cam\"!"
                                      , QMessageBox::Ok );


            return;
        }
        if ( loadWalls(  am::util::cuboidPathToWallsPath(_last_cuboids_path) ) == EXIT_FAILURE )
        {
            QMessageBox::warning( this, "No walls", ("No walls file at " + am::util::cuboidPathToWallsPath(_last_cuboids_path)).c_str(), QMessageBox::Ok );
        }

        // intense colours
        _ui->opacitySpinBox->setValue( 0.5 );
        on_showInvalidCloud_toggled(true);
        _ui->colourEdgesCheckBox->setChecked( false );
        my_on_screenshotButton_clicked( _getDepthDirAsStdString() + "/optimized_front_full.png" );
        on_showInvalidCloud_toggled(false);

        my_on_loadCameraParamsButton_clicked( _getDepthDirAsStdString()+"/back_view.cam" );
        my_on_loadContactsButton_clicked( _getDepthDirAsStdString() + "/graph.txt" );
        //_ui->primitivesListWidget->selectAll();

        // no invalid cloud
        _ui->showInvalidCloud->setChecked( false );
        on_showInvalidCloud_toggled( _ui->showInvalidCloud->isChecked() );

        // save image
        my_on_screenshotButton_clicked( _getDepthDirAsStdString() + "/optimized_back.png" );
    }

    //    % 3.1: opti graph without points frontal
    _ui->primitivesListWidget->setCurrentRow( -1 );
    on_resetCameraButton_clicked();
    on_removePointsButton_clicked();
    _ui->opacitySpinBox->setValue( 0.3 );
    my_on_screenshotButton_clicked( _getDepthDirAsStdString() + "/optimized_front.png" );


    // 2.1: init boxes frontal
    {
        on_clearButton_clicked();
        // hide intersections
        on_showIntersections_clicked( false );
        my_on_loadButton_clicked( _getDepthDirAsStdString() + "/cuboids.dat" );
        loadWalls(  am::util::cuboidPathToWallsPath(_last_cuboids_path) );

        _ui->showInvalidCloud->setChecked( true );
        on_showInvalidCloud_toggled( _ui->showInvalidCloud->isChecked() );
        _ui->colourEdgesCheckBox->setChecked( true );
        on_colourEdgesCheckBox_stateChanged( _ui->colourEdgesCheckBox->isChecked() );
        _ui->primitivesListWidget->selectAll();

        my_on_screenshotButton_clicked( _getDepthDirAsStdString() + "/initial_front.png" );
    }

    // 2.2: init boxes back
    _ui->showInvalidCloud->setChecked( false );
    on_showInvalidCloud_toggled( _ui->showInvalidCloud->isChecked() );
    my_on_loadCameraParamsButton_clicked( _getDepthDirAsStdString()+"/back_view.cam" );
    my_on_screenshotButton_clicked( _getDepthDirAsStdString() + "/initial_back.png" );

    appendToTextFile( _getDepthDirAsStdString()+"\n", "/media/Data2/Dropbox/UCL/PHD/rgbd/rgbd_structure_datasets/results_paths.txt" );

    return;

    //    % 3.1: opti graph without points frontal
    my_on_loadButton_clicked( _getDepthDirAsStdString() + "/optimized.dat" );
    _ui->opacitySpinBox->setValue( 0.3 );
    //_ui->primitivesListWidget->selectAll();
    on_removePointsButton_clicked();
    _ui->showInvalidCloud->setChecked( false );

    my_on_screenshotButton_clicked( _getDepthDirAsStdString() + "/optimized.png" );
    my_on_loadContactsButton_clicked( _getDepthDirAsStdString() + "/graph.txt" );

    return;

    //    % 2.1: init boxes frontal coloured
    my_on_loadButton_clicked( _getDepthDirAsStdString() + "/cuboids.dat" );
    _ui->opacitySpinBox->setValue( 0.3 );
    _ui->colourEdgesCheckBox->setChecked( true );
    _ui->primitivesListWidget->selectAll();
    my_on_screenshotButton_clicked( _getDepthDirAsStdString() + "/initial_front.png" );

    //    % 2.2: init boxes back coloured
    my_on_loadCameraParamsButton_clicked( _getDepthDirAsStdString() + "/optimized_back.cam" );
    my_on_screenshotButton_clicked( _getDepthDirAsStdString() + "/initial_back.png" );




    ///char line[1024];
    //sprintf( line, "printf \"%s\\n\" >> /media/Data2/Dropbox/UCL/PHD/rgbd/rgbd_structure_datasets/paths.txt", _getDepthDirAsStdString().c_str() );
    //std::cout << "executing " << line << std::endl;
    //system( line );
}

/// MANIP ///

void QtCloudViewer::on_xMoveSlider_sliderMoved( int position )
{
    _movePrimitive( 0, position - _ui->xMoveSlider->value() );
}

void QtCloudViewer::on_yMoveSlider_sliderMoved( int position )
{
    _movePrimitive( 1, position - _ui->yMoveSlider->value() );
}

void QtCloudViewer::on_zMoveSlider_sliderMoved( int position )
{
    _movePrimitive( 2, position - _ui->zMoveSlider->value() );
}

void QtCloudViewer::on_xResizeSlider_sliderMoved(int position)
{
    _resizePrimitive( 0, position - _ui->xResizeSlider->value() );
}

void QtCloudViewer::on_yResizeSlider_sliderMoved(int position)
{
    _resizePrimitive( 1, position - _ui->yResizeSlider->value() );
}

void QtCloudViewer::on_zResizeSlider_sliderMoved(int position)
{
    _resizePrimitive( 2, position - _ui->zResizeSlider->value() );
}

void QtCloudViewer::on_xRotateSlider_sliderMoved(int position)
{
    _rotatePrimitive( 0, 5 * (position - _ui->xRotateSlider->value()) );
}

void QtCloudViewer::on_yRotateSlider_sliderMoved(int position)
{
    _rotatePrimitive( 1, 5 * (position - _ui->yRotateSlider->value()) );
}

void QtCloudViewer::on_zRotateSlider_sliderMoved(int position)
{
    _rotatePrimitive( 2, 5 * (position - _ui->zRotateSlider->value()) );
}

/// FILTER ///

void QtCloudViewer::on_spatialSigmaSlider_valueChanged(int value)
{
    _ui->spatialSigmaSpinBox->setValue( value / 10.f);
    _filteringParamChanged( 0 );
}

void QtCloudViewer::on_spatialSigmaSpinBox_valueChanged(double value)
{
    _ui->spatialSigmaSlider->setValue( value * 10. );
    _filteringParamChanged( 0 );
}

void QtCloudViewer::on_rangeSigmaSlider_valueChanged(int value)
{
    _ui->rangeSigmaSpinBox->setValue( value / 10.f);
    _filteringParamChanged( 1 );
}

void QtCloudViewer::on_rangeSigmaSpinBox_valueChanged(double value)
{
    _ui->rangeSigmaSlider->setValue( value * 10. );
    _filteringParamChanged( 1 );
}

void QtCloudViewer::on_kernelRadiusSlider_valueChanged(int value)
{
    _ui->kernelRadiusSpinBox->setValue( value );
    _filteringParamChanged( 2 );
}

void QtCloudViewer::on_kernelRadiusSpinBox_valueChanged(int value)
{
    _ui->kernelRadiusSlider->setValue( value );
    _filteringParamChanged( 2 );
}

void QtCloudViewer::on_iterationsSpinBox_valueChanged( int value )
{
    _ui->iterationsSlider->setValue( value );
    _filteringParamChanged( 3 );
}

void QtCloudViewer::on_iterationsSlider_valueChanged( int value )
{
    _ui->iterationsSpinBox->setValue( value );
    _filteringParamChanged( 3 );
}

/// OTHER ///

void QtCloudViewer::on_fillButton_clicked()
{
    QList<QListWidgetItem*> selected_items = _ui->primitivesListWidget->selectedItems();
    for ( QListWidgetItem* widgetitemptr : selected_items )
    {
        int row = _ui->primitivesListWidget->row( widgetitemptr );
        _controller->fillPrimitive( row );
    }

    _ui->qvtkWidget->update();
}

int
QtCloudViewer::centerCamera()
{
    _controller->centerCamera();
}

void QtCloudViewer::my_on_screenshotButton_clicked( std::string path )
{
    int res = 0;
    QString fileName;

    if ( path.empty() )
    {
        QFileDialog dialog( this
                            , tr("Save screenshot")
                            //QString::fromStdString(boost::filesystem::path(_ui->depthPathEdit->text().toStdString()).parent_path().string()+"/screenshot.png"),
                            , QString::fromStdString( _getLastPath(".") + +"/screenshot.png" )
                            , tr("All Files (*.png)") );
        //dialog.setFileMode( QFileDialog::Directory );
        dialog.setConfirmOverwrite( true );
        res = dialog.exec();
        if ( res ) fileName = dialog.selectedFiles().at(0);
    }
    else
    {
        fileName = QString::fromStdString( path );
    }

    if ( !fileName.isEmpty() )
    {

        //std::string save_path = enforceExtension( ".bmp", fileName.toStdString() );
        std::string save_path = fileName.toStdString();
        if ( boost::filesystem::path(fileName.toStdString()).extension().empty() ) fileName += ".png";

        int err = EXIT_SUCCESS;
        // save screenshot
        err += _controller->saveScreenshot( save_path );
        // save camera pose
        std::string camera_pose_path = boost::filesystem::path(fileName.toStdString()).replace_extension(".cam").string();
        err += _controller->saveCameraPose( camera_pose_path );

        if ( path.empty() ) // if path provided, no popup
        {
            if ( err == EXIT_SUCCESS )
                QMessageBox::information( this
                                          , "Info"
                                          , ("Saved screenshot to " + save_path + ", camera pose to " + camera_pose_path).c_str()
                                          , QMessageBox::Ok );
            else
                QMessageBox::warning( this
                                      , "Warning"
                                      , ("Could not save screenshot to " + save_path+ ", camera pose to " + camera_pose_path).c_str()
                                      , QMessageBox::Ok );
        }
    }

}

void QtCloudViewer::on_regGrowButton_clicked()
{
    _controller->region_growing();
}

void QtCloudViewer::on_createCuboidButton_clicked()
{
    am::Params params;
    params.setDouble( "plane_distance_threshold", _ui->planeInlierThresholdDoubleSpinBox->value() );

    _controller->createCuboid( &params );
    _controller->clearCurrentPick();
    _refreshPrimitivesList();
    on_clearPickButton_clicked();
}

void QtCloudViewer::on_clearPickButton_clicked()
{
    _controller->clearCurrentPick();
    _ui->qvtkWidget->update();
}

void QtCloudViewer::on_save2PLYButton_clicked()
{
    _controller->saveCuboidsToPLY( "mesh.ply" );
}

void QtCloudViewer::on_clearPrimitivesButton_clicked()
{
    _clearPrimitives();
    on_showIntersections_clicked( false );
}

void QtCloudViewer::on_filterContactsButton_clicked()
{
    ca::DisplayedPrimitive::Map primitives_map = _controller->getPrimitives();

    am::Cuboids cuboids;
    for ( auto it = primitives_map.begin(); it != primitives_map.end(); ++it )
    {
        std::shared_ptr<am::Cuboid> cuboid_ptr = it->second.CuboidPtr();
        if ( cuboid_ptr )
            cuboids.add( *cuboid_ptr );
    }

    am::Cuboid::TypeMap walls;
    _extractWallsMap( walls );

    am::Contact::Vec filtered_contacts;
    int err = ca::ContactsFiltering::filterContacts( filtered_contacts
                                                     , _controller->getLatestContacts()
                                                     , cuboids
                                                     , walls );
    if ( EXIT_SUCCESS != err )
    {
        std::cerr << "[" << __func__ << "]: " << "filterContacts returned error " << err << std::endl; fflush(stderr);
    }

    boost::filesystem::path out_path(_last_contacts_path );
    std::cout << out_path.parent_path().string() + "/" + out_path.stem().string() + ".filtered.txt" << std::endl;
    QString saveFileName = QFileDialog::getSaveFileName( this
                                                         , tr("Save filtered contacts")
                                                         , QString::fromStdString(out_path.parent_path().string() + "/" + out_path.stem().string() + ".filtered.txt")
                                                         , tr("Contacts(*.txt);; All Files (*.*)") );
    if ( !saveFileName.isEmpty() )
    {
        int err = am::Contact::write( filtered_contacts, saveFileName.toStdString() );
        if ( err == EXIT_SUCCESS )
            QMessageBox::information( this, tr("Written"), tr(("Wrote filtered contacts to " + saveFileName.toStdString()).c_str())
                                      , QMessageBox::Ok );
    }
}

int
QtCloudViewer::_extractWallsMap( am::Cuboid::TypeMap &walls )
{
    int cuboid_id = 0;
    auto end_it = _controller->getPrimitives().end();
    for ( auto it = _controller->getPrimitives().begin(); it != end_it; ++it, ++cuboid_id )
    {
        if ( it->second.floor )
            walls[ cuboid_id ] = am::Cuboid::FLOOR;
        else if ( it->second.wall )
            walls[ cuboid_id ] = am::Cuboid::WALL;
        else if ( (it->second.shape == am::Cuboid::SPHERE) || (it->second.shape == am::Cuboid::CYLINDER) )
            walls[ cuboid_id ] = it->second.shape;
    }

    return EXIT_SUCCESS;
}

void QtCloudViewer::on_colourCloudCheckBox_toggled(bool checked)
{
    std::cout << "[" << __func__ << "]: " << "colourCloudCheckBox " << (checked ? "ON" : "OFF" ) << std::endl;

    std::vector<Eigen::Vector3f> pos_focal_up = _controller->getCameraPosFocalUp();
    if ( checked )  _controller->colourCloud( am::AnnotationController::COLOUR_CLOUD_MODE::DISTANCE_BASED, &(pos_focal_up[0]) );
    else            _controller->colourCloud( am::AnnotationController::COLOUR_CLOUD_MODE::NONE );

    _ui->qvtkWidget->update();
}

void QtCloudViewer::on_showIDsCheckBox_stateChanged(int arg1)
{
    std::cout << "[" << __func__ << "]: " << "showIDs " << (arg1 ? "ON" : "OFF" ) << std::endl;
    _controller->setShowCuboidIDs( arg1 );
    updateQvtWidget();
}

void QtCloudViewer::on_moosPlanesButton_clicked()
{
    QString saveFileName = QFileDialog::getSaveFileName( this
                                                         , tr("saveMoosPlanes")
                                                         , QString::fromStdString( _getLastPath() )
                                                         , tr("*.txt;; All Files (*.*)") );
    if ( !saveFileName.isEmpty() )
    {
        std::ofstream file( saveFileName.toStdString() );
        if ( !file.is_open() )
        {
            std::cerr << "[" << __func__ << "]: " << "cannot open " << saveFileName.toStdString() << std::endl;
            return;
        }

        // for cuboids
        // p=[(x1, y1, z1), (x2, y2, z2), (x3, y3, z3), (x4, y4, z4)]
        auto end_it = _controller->getPrimitives().end();
        for ( auto it = _controller->getPrimitives().begin(); it != end_it; ++it )
        {
            std::shared_ptr<am::Cuboid> cuboid_ptr = it->second.CuboidPtr();
            if ( cuboid_ptr )
            {
                std::cout << "cuboid: " << std::endl;
                for ( int c = 0; c != cuboid_ptr->Planes().size(); ++c )
                {
                    std::cout << "cuboid_plane: " << std::endl;
                    std::string plane_str = ca::io::printMoosPlane( cuboid_ptr->Planes()[c] );
                    std::cout << plane_str << std::endl;
                    file << plane_str << std::endl;
                }
            }
            else
            {
                std::shared_ptr<am::Plane> plane_ptr = it->second.PlanePtr();
                if ( plane_ptr )
                {
                    std::cout << "plane: " << std::endl;
                    std::string plane_str = ca::io::printMoosPlane( *plane_ptr );
                    std::cout << plane_str << std::endl;
                    file << plane_str << std::endl;
                }
                else
                {
                    std::cerr << "cannot parse primitive...\n"; fflush(stderr);
                }
            }
        }

        file.close();
    }
}

void QtCloudViewer::on_savePointsButton_clicked()
{
    QString saveFileName = QFileDialog::getSaveFileName( this
                                                         , tr("savePoints")
                                                         , QString::fromStdString( _getLastPath() )
                                                         , tr("*.ply;; All Files (*.*)") );
    if ( !saveFileName.isEmpty() )
    {
        _controller->savePointsToPLY( saveFileName.toStdString() );
    }
}

void QtCloudViewer::on_wshopDumpButton_clicked()
{
    QString saveFileName = QFileDialog::getSaveFileName( this
                                                         , tr("save Scene in Workshop format")
                                                         , QString::fromStdString( _getLastPath() + "/abstraction.txt" )
                                                         , tr("*.txt;; All Files (*.*)") );
    if ( !saveFileName.isEmpty() )
    {
        std::ofstream file( saveFileName.toStdString() );
        if ( !file.is_open() )
        {
            std::cerr << "[" << __func__ << "]: " << "cannot open " << saveFileName.toStdString() << std::endl;
            return;
        }
        file << "# Format:" << std::endl;
        file << "#  cuboid,v3_principal0,v3_principal1,v3_principal2,v3_position,v3_scale" << std::endl;
        file << "#  sphere,v3_position,radius" << std::endl;
        file << "#  cylinder,v3_center,v3_axis,height,radius" << std::endl;
        // for cuboids
        // p=[(x1, y1, z1), (x2, y2, z2), (x3, y3, z3), (x4, y4, z4)]
        auto end_it = _controller->getPrimitives().end();
        for ( auto it = _controller->getPrimitives().begin(); it != end_it; ++it )
        {
            std::shared_ptr<am::Cuboid> cuboid_ptr = it->second.CuboidPtr();
            if ( cuboid_ptr )
            {
                std::stringstream ss;
                if ( it->second.shape == am::Cuboid::CUBOID )
                {
                    ss << "cuboid,";
                    for ( int c = 0; c != cuboid_ptr->R().cols(); ++c )
                        for ( int d = 0; d != cuboid_ptr->R().col(c).rows(); ++d )
                            ss << cuboid_ptr->R()(d,c) << ",";
                    Eigen::Vector3f center = cuboid_ptr->Center();
                    for ( int d = 0; d != 3; ++d )
                        ss << center(d) << ",";

                    Eigen::Vector3f extent = cuboid_ptr->getExtent();
                    for ( int d = 0; d != 3; ++d )
                        ss << extent(d) << ",";

                            //ss << cuboid_ptr->R().array() << cuboid_ptr->Center().transpose() << cuboid_ptr->getExtent().transpose() << std::endl;
                }
                else if ( it->second.shape == am::Cuboid::SPHERE )
                {
                    ss << "sphere,";
                    for ( int d = 0; d != 3; ++d )
                        ss << cuboid_ptr->Center()(d) << ",";
                    ss << cuboid_ptr->getExtent()(0) << ",";
                }
                else if ( it->second.shape == am::Cuboid::CYLINDER )
                {
                    ss << "cylinder,";
                    for ( int d = 0; d != 3; ++d )
                        ss << cuboid_ptr->Center()(d) << ",";
                    for ( int d = 0; d != 3; ++d )
                        ss << cuboid_ptr->R().col(0)(d) << ",";
                    ss << cuboid_ptr->getExtent()(0) << "," << cuboid_ptr->getExtent()(1) << ",";
                }

                std::cout << ss.str().substr(0,ss.str().size()-1) << std::endl;
                file << ss.str().substr(0,ss.str().size()-1) << std::endl;
            }
        }
        file.close();

        {
            std::string colourFName = saveFileName.toStdString() + ".colour";
            std::ofstream file2( colourFName );
            if ( !file2.is_open() )
            {
                std::cerr << "[" << __func__ << "]: " << "cannot open " << colourFName << std::endl;
                return;
            }

//            char tmp[4];
            auto end_it = _controller->getPrimitives().end();
            for ( auto it = _controller->getPrimitives().begin(); it != end_it; ++it )
            {
                std::cout << " itt " << std::endl;
                Eigen::Vector4f colour = it->second.colour;

                file2 << std::hex << std::setfill ('0') << std::setw(2) << (int)round(colour(0) * 255);
                file2 << std::hex << std::setfill ('0') << std::setw(2) << (int)round(colour(1) * 255);
                file2 << std::hex << std::setfill ('0') << std::setw(2) << (int)round(colour(2) * 255)<< std::endl;
                //file2 << "#";
//                for ( int i = 0; i != 3; ++i )
//                {
//                    std::cout << "sprintf" << std::endl;
//                    sprintf(tmp, "%02X", colour(0) );
//                    file2 << tmp;
//                }
            }
            file2.close();
        }

    }
}

void QtCloudViewer::on_duplicateCuboidButton_clicked()
{
    std::vector<int> row_ids = _getSelectedRows();
    if ( row_ids.size() != 1 )
    {
        std::cerr << "[" << __func__ << "]: " << "row_ids.size() != 1, exiting" << std::endl;
        return;
    }

    ca::DisplayedPrimitive primitive = _controller->getPrimitive( row_ids[0] );
    auto cuboid_ptr = primitive.CuboidPtr();
    _controller->addCuboid( *cuboid_ptr, "dupl" );
    _refreshPrimitivesList();
}
