#include "RotatorThread.h"
#include "AnnotationController.h"
#include "Cuboid.h"

RotatorThread::RotatorThread( std::shared_ptr<am::AnnotationController> controller )
    : _controller( controller )
    , _prepared( false )
    , _degrees( 0.f )
    , _stop_degrees( 360.f )
    , _step( 1.f )
{
}

std::pair<int,std::string>
RotatorThread::prepare( float to_deg, float from_deg, float step, Eigen::Vector3f* centroid, Eigen::Vector3f* rot_axis )
{
    std::string message;

#if 1
    if ( !centroid || !rot_axis )
    {
        // get floor
        auto floor_ptrs = _controller->getFloors();
        if ( floor_ptrs.empty() )
        {
            return std::pair<int,std::string>( EXIT_FAILURE, "Need a floor for upvector" );
        }

        // get cuboids centroid
        auto target_ptrs =_controller->getRotationTargets();
        if ( target_ptrs.empty() )
        {
            return std::pair<int,std::string>( EXIT_FAILURE, "Need a target for center" );
        }

        // rotationcentroid
        _centroid = Eigen::Vector3f::Zero();
        for ( std::shared_ptr<am::Cuboid> const cuboid_ptr : target_ptrs )
        {
            _centroid += cuboid_ptr->Center();
        }
        _centroid /= target_ptrs.size();

        // draw rotation axis
        _rot_axis = floor_ptrs[0]->Planes()[2].Normal().normalized();
        // old: -1 * [0]
    }
    else
    {
        _rot_axis = *rot_axis;
        _centroid = *centroid;
    }


#else
    _rot_axis = (Eigen::Vector3f() << 0.0217914f,0.918758f,0.39422f ).finished();
    _centroid = (Eigen::Vector3f() << -0.0294572f,0.194195f,1.59438f).finished();

#endif

    //Eigen::AngleAxisf rot( ang, tmp_frame.block<3,1>(0,2) );
    //tmp_frame.block<3,1>(0,0) = rot * tmp_frame.block<3,1>(0,0);

//    _controller->getViewer()->addArrow( am::util::pcl::asPointXYZ(_centroid-_rot_axis*1.5f) //upvector is always from plane_id=2
//                                        ,am::util::pcl::asPointXYZ(_centroid)
//                                        ,0, 0, 0
//                                        , false
//                                        , "rotuparrow", 0 );

    _degrees      = from_deg;
    _stop_degrees = to_deg;
    _step         = step;
    _myframeid = 0;
    _doneframeid = -1;

    _prepared = true;
    std::stringstream ss;
    ss << "rot: " << _rot_axis.transpose() << "; centroid: " << _centroid.transpose() << std::endl;
    return { EXIT_SUCCESS, ss.str()};
}

void RotatorThread::setDoneFrameId( int frame_id )
{
    _doneframeid = frame_id;
}

void RotatorThread::one_loop()
{
    if ( !_prepared )
    {
        std::cerr << __func__ << ": call prepare before starting thread...skipping loop" << std::endl;
        return;
    }

    // wait
    while ( _myframeid > _doneframeid+1 )
    {
        msleep( 50 );
    }

    std::vector<Eigen::Vector3f> pos_focal_up = _controller->getCameraPosFocalUp();
    if ( pos_focal_up[0].dot(Eigen::Vector3f::Ones()) == 0.f )
        pos_focal_up[0](0) += .01f;

    Eigen::Vector3f pos2centroid = _centroid - pos_focal_up[0];

    Eigen::AngleAxisf rot( pcl::deg2rad(_step), _rot_axis );
    //
    pos_focal_up[0] = _centroid - (rot * pos2centroid);
    pos_focal_up[1] = _centroid;
    pos_focal_up[2] = -1.f * _rot_axis;
    // view is towards centroid
    //pos_focal_up[1] = (_centroid - pos_focal_up[0]).normalized();
    //pos_focal_up[2] = pos_focal_up[1].cross( pos_focal_up[0] );

    _controller->setCameraPosFocalUp( pos_focal_up );
    std::cout << "new pos is: " << pos_focal_up[0].transpose() << std::endl;

    // inc and ifstop
    if ( (_degrees+=_step) > _stop_degrees )
    {
        std::cout << __func__ << "degrees: " << _degrees << " > " << _stop_degrees << "...stopping from inside" << std::endl;
        _stopped = true;
    }
    else // debug
    {
        std::cout << _degrees << std::endl;
    }

    ++_myframeid;
}
