#include "cuboid_approx/util/AMUtil2.h"

#include <fstream>
#include <iostream>
#include <iomanip>
#include <time.h>
#include <stdlib.h>
#include <stdio.h>

#include "opencv2/imgproc/types_c.h"
#include "opencv2/imgproc/imgproc.hpp"
#include "opencv2/core/core.hpp"
#include "opencv2/highgui/highgui.hpp"

#include "vtkCamera.h"
#include "pcl/visualization/pcl_visualizer.h"

namespace am
{
    namespace util
    {
        namespace eigen
        {
            // assumes eigen_vectors in increasing column-major order
            int sort_eigen_vectors_desc( Eigen::Vector3f const& eigen_values, Eigen::Matrix3f &eigen_vectors )
            {
                std::vector<int> eigen_vector_ids = { 0, 1, 2 };
                am::util::algo::sort3_desc( eigen_values(0), eigen_values(1), eigen_values(2), eigen_vector_ids );

                Eigen::Matrix3f tmp( eigen_vectors );
                eigen_vectors.col(0) = tmp.col( eigen_vector_ids[0] );
                eigen_vectors.col(1) = tmp.col( eigen_vector_ids[1] );
                eigen_vectors.col(2) = tmp.col( eigen_vector_ids[2] );
            }
        }

        namespace cv
        {
            void
            unsignedIntToFloat( ::cv::Mat & img32FC1, ::cv::Mat const& img32UC1 )
            {
                img32FC1.create( img32UC1.size(), CV_32FC1 );
                for ( int y = 0; y < img32UC1.rows; ++y )
                {
                    for ( int x = 0; x < img32UC1.cols; ++x )
                    {
                        img32FC1.at<float>( y, x ) = img32UC1.at<unsigned>( y, x );
                    }
                }
            }

            int
            imread( /* out: */ ::cv::Mat &mat, /*  in: */ std::string const& path )
            {
                int ret = EXIT_SUCCESS;

                if ( path.find("pfm") != std::string::npos )
                {
                    ret += am::util::loadPFM( mat, path );
                }
                else
                {
                    mat = ::cv::imread( path.c_str(), -1 );
                }

                return ret;
            }

            int
            writePNG( std::string title, ::cv::Mat const& mat, bool silent )
            {
                // png no compression
                std::vector<int> params(2);
                params[0] = 16; params[1] = 0;

                if ( !silent ) std::cout << "saving " << title << "..." << std::endl;
                ::cv::imwrite( title.c_str(), mat, params );

                return EXIT_SUCCESS;
            }

            int
            blend( ::cv::Mat &blendedUC3, ::cv::Mat const& dep, float depMax, ::cv::Mat const& rgb, float rgbMax )
            {
                ::cv::Mat depFC1_1;
                dep.convertTo( depFC1_1, CV_32FC1, 1.f / depMax );

                std::vector< ::cv::Mat> depFC1Vector;
                depFC1Vector.push_back( depFC1_1 );
                depFC1Vector.push_back( depFC1_1 );
                depFC1Vector.push_back( depFC1_1 );
                ::cv::Mat depFC1_3;
                ::cv::merge( depFC1Vector.data(), 3, depFC1_3 );

                ::cv::Mat rgbFC1;
                rgb.convertTo( rgbFC1, CV_32FC1, 1.f / rgbMax );

                ::cv::Mat blendedFC1( depFC1_3.rows, depFC1_3.cols, CV_32FC1 );
                ::cv::addWeighted( rgbFC1  , 0.5,
                                   depFC1_3, 0.7,
                                   0.0, blendedFC1, CV_32FC1 );
                blendedFC1.convertTo( blendedUC3, CV_8UC3, 255.f );

                return EXIT_SUCCESS;
            }

            int
            replaceValue( ::cv::Mat &img, ::cv::InputArray value_to_compare, ::cv::InputArray value_to_write )
            {
                ::cv::Mat local_mask( img.size(), CV_8UC1 );

                ::cv::compare( img, value_to_compare, local_mask, ::cv::CMP_EQ );
                img.setTo( value_to_write, local_mask ); // bg pixels not copied
            }


        }

        namespace pcl
        {
            Eigen::Vector3f
            point2To3D( Eigen::Vector2f const& pnt2, Eigen::Matrix3f const& intrinsics )
            {
                Eigen::Vector3f pnt3D;
                pnt3D << (pnt2(0) - intrinsics(0,2)) / intrinsics(0,0),
                        (pnt2(1) - intrinsics(1,2)) / intrinsics(1,1),
                        1.f;
                return pnt3D;
            }

            Eigen::Vector2f
            point3To2D( Eigen::Vector3f const& pnt3, Eigen::Matrix3f const& intrinsics )
            {
                Eigen::Vector2f pnt2;
                pnt2 << (pnt3(0) / pnt3(2)) * intrinsics(0,0) + intrinsics(0,2),
                        (pnt3(1) / pnt3(2)) * intrinsics(1,1) + intrinsics(1,2);
                return pnt2;
            }

            ::pcl::PointXYZ
            asPointXYZ(Eigen::Vector3f const& vector3f )
            {
                return ::pcl::PointXYZ( vector3f.x(), vector3f.y(), vector3f.z() );
            }

            ::pcl::PointXYZ
            v4asPointXYZ( Eigen::Vector4f const& vector4f, bool dont_divide )
            {
                if ( dont_divide || (fabs(vector4f(3)) < FLT_EPSILON)  )
                    return ::pcl::PointXYZ( vector4f.x(), vector4f.y(), vector4f.z() );
                else
                    return ::pcl::PointXYZ( vector4f.x()/vector4f.w(), vector4f.y()/vector4f.w(), vector4f.z()/vector4f.w() );
            }

            ::pcl::PointXYZRGB
            PointXYZRGB( float x, float y, float z, uint8_t r, uint8_t g, uint8_t b )
            {
                ::pcl::PointXYZRGB p;
                p.x = x;
                p.y = y;
                p.z = z;
                uint32_t rgb = ((uint32_t)r << 16 | (uint32_t)g << 8 | (uint32_t)b);
                p.rgb = *reinterpret_cast<float*>(&rgb);

                return p;
            }

            std::vector< std::vector<Eigen::Vector3f> >
            getCamerasPosFocalUp( ::pcl::visualization::PCLVisualizer::Ptr viewer )
            {
                std::vector< std::vector<Eigen::Vector3f> > ret;

                auto rens = viewer->getRendererCollection();
                rens->InitTraversal ();
                vtkRenderer* renderer = NULL;
                while ((renderer = rens->GetNextItem ()) != NULL)
                {
                    ret.push_back( std::vector<Eigen::Vector3f>() );

                    auto pos         = renderer->GetActiveCamera()->GetPosition();
                    ret.back().push_back( (Eigen::Vector3f() <<         pos[0],         pos[1],         pos[2]).finished() );

                    auto focal_point = renderer->GetActiveCamera()->GetFocalPoint();
                    ret.back().push_back( (Eigen::Vector3f() << focal_point[0], focal_point[1], focal_point[2]).finished() );

                    auto view_up     = renderer->GetActiveCamera()->GetViewUp();
                    ret.back().push_back( (Eigen::Vector3f() <<     view_up[0],     view_up[1],     view_up[2]).finished() );

                    /*std::cout << renderer->GetActiveCamera ()->GetPosition ()[0] << ", ";
                    std::cout << renderer->GetActiveCamera ()->GetPosition ()[1]<< ", ";
                    std::cout << renderer->GetActiveCamera ()->GetPosition ()[2]<< "\n";
                    std::cout << renderer->GetActiveCamera ()->GetFocalPoint ()[0]<< ", ";
                    std::cout << renderer->GetActiveCamera ()->GetFocalPoint ()[1]<< ", ";
                    std::cout << renderer->GetActiveCamera ()->GetFocalPoint ()[2]<<"\n";
                    std::cout << renderer->GetActiveCamera ()->GetViewUp ()[0]<< ", ";
                    std::cout << renderer->GetActiveCamera ()->GetViewUp ()[1]<< ", ";
                    std::cout << renderer->GetActiveCamera ()->GetViewUp ()[2]<<"\n";*/
                }

                return ret;
            }

            // should be moved to amutilPCL later
            int
            setCameraPosFocalUp( ::pcl::visualization::PCLVisualizer::Ptr viewer
                                 , std::vector<Eigen::Vector3f>           pos_focal_up
                                 , int                                    viewport_id  )
            {
                int vid = 0;

                auto rens = viewer->getRendererCollection();
                rens->InitTraversal ();
                vtkRenderer* renderer = NULL;
                while ((renderer = rens->GetNextItem ()) != NULL)
                {
                    if ( (viewport_id == 0) || (vid == viewport_id) )
                    {
                        renderer->GetActiveCamera()->SetPosition  ( pos_focal_up[0][0], pos_focal_up[0][1], pos_focal_up[0][2] );
                        renderer->GetActiveCamera()->SetFocalPoint( pos_focal_up[1][0], pos_focal_up[1][1], pos_focal_up[1][2] );
                        renderer->GetActiveCamera()->SetViewUp    ( pos_focal_up[2][0], pos_focal_up[2][1], pos_focal_up[2][2] );
                    }

                    ++vid;
                }

                return EXIT_SUCCESS;
            }

        }

        int
        savePFM( ::cv::Mat const& imgF, std::string path, float scale, bool silent )
        {
            if ( imgF.empty() || imgF.type() != CV_32FC1 )
            {
                std::cerr << "AMUtil::savePFM: expects non-empty image of type CV_32FC1" << std::endl;
                return EXIT_FAILURE;
            }

            if ( !silent ) std::cout << "saving " << path << "..." << std::endl;

            std::fstream file;
            file.open( path.c_str(), std::ios_base::out | std::ios_base::binary | std::ios_base::trunc );
            if ( !file.is_open() )
            {
                std::cerr << "AMUtil::savePFM: could not open file at path " << path << std::endl;
                return EXIT_FAILURE;
            }

            // header
            file << "Pf" << std::endl;
            file << imgF.cols << " " << imgF.rows << std::endl;
            file << scale << std::endl;

            // transposed image
            for ( int x = 0; x < imgF.cols; ++x )
            {
                for ( int y = 0; y < imgF.rows; ++y )
                {
                    file << std::setprecision(23) << imgF.at<float>(y,x) << " ";
                }
                file << std::endl;
            }
            file.close();

            return EXIT_SUCCESS;
        }

        int
        loadPFM( ::cv::Mat & imgF, std::string path )
        {
            std::ifstream file( path.c_str() );
            //file.open( path.c_str(), std::ios_base::in | std::ios_base::binary );
            if ( !file.is_open() )
            {
                std::cerr << "AMUtil::loadPFM: could not open file at path " << path << std::endl;
                return EXIT_FAILURE;
            }

            int lineid = 0;
            int rows, cols, row = 0, col = 0;
            float scale;

            std::string line;
            for( std::string line; getline( file, line ); )
            {
                if ( lineid < 3 )
                {
                    if ( (lineid == 0) && (line.find("Pf") == std::string::npos) )
                    {
                        std::cerr << "Not PFM file..." << std::endl;
                        file.close();
                        return EXIT_FAILURE;
                    }
                    else if ( lineid == 1 )
                    {
                        std::stringstream ss( line );
                        std::string tmp;
                        getline( ss, tmp, ' ' );
                        cols = atoi( tmp.c_str() );
                        getline( ss, tmp, ' ' );
                        rows = atoi( tmp.c_str() );

                        imgF.create( rows, cols, CV_32FC1 );
                    }
                    else if ( lineid == 2 )
                    {
                        scale = atof( line.c_str() );
                    }
                    ++lineid;
                    continue;
                }

                std::stringstream ss( line );
                std::string word;

                row = 0;
                while ( getline(ss, word, ' ') )
                {
                    imgF.at<float>( row, col ) = atof( word.c_str() );
                    ++row;
                }

                ++lineid;
                ++col;
            }
            file.close();

            return EXIT_SUCCESS;
        }

        // depth to ::cv::MAT
        int PclDepthImageToCvMat( ::cv::Mat* pCvDepth16, boost::shared_ptr<openni_wrapper::DepthImage> const& depth_img, ::cv::Mat* pCvDepth8 )
        {
            static const double conversionRatio = 255.0/2047.0;

            // CHECK
            if ( (pCvDepth8 == NULL) && (pCvDepth16 == NULL) ) return -1;

            // INIT
            xn::DepthMetaData depthMD;
            {
                depthMD.InitFrom( depth_img->getDepthMetaData() );
                if ( (depthMD.FullXRes() == 0) || (depthMD.FullYRes() == 0) )
                {
                    std::cerr << "PclDepthImageToCvMat: depthMD.size is zero...: " << depthMD.FullXRes() << "x" << depthMD.FullYRes() << std::endl;
                    throw new std::runtime_error("PclDepthImageToCvMat: depthMD.size is zero");
                }
            }

            ::cv::Mat cvDepth16;
            {
                cvDepth16.create( depthMD.FullYRes(), depthMD.FullXRes(), CV_16UC1 );
            }

            // COPY
            {
                const XnDepthPixel *pDepthPixels = depthMD.Data();
                int offset = 0;
                for ( XnUInt y = 0; y < depthMD.YRes(); ++y, offset += depthMD.XRes() )
                    memcpy( cvDepth16.data + cvDepth16.step * y, pDepthPixels + offset, depthMD.XRes() * sizeof(XnDepthPixel) );
            }

            // OUTPUT
            {
                if ( pCvDepth8 )
                {
                    pCvDepth8->create( depthMD.FullYRes(), depthMD.FullXRes(), CV_8UC1 );
                    cvDepth16.convertTo( *pCvDepth8, CV_8UC1, conversionRatio );
                }
                if ( pCvDepth16 )
                {
                    pCvDepth16->create( depthMD.FullYRes(), depthMD.FullXRes(), CV_16UC1 );
                    cvDepth16.copyTo( *pCvDepth16 );
                }
            }

            return 0;
        }

        // image to ::cv::MAT
        int PclRgbImageToCvMat( ::cv::Mat *pImgRGB, boost::shared_ptr<openni_wrapper::Image> const& img )
        {
            // update
            xn::ImageMetaData imageMD;
            ::cv::Mat imgBGR;

            // INIT
            {
                imageMD.InitFrom( img->getMetaData() );
                if ( (imageMD.FullXRes() == 0) || (imageMD.FullYRes() == 0) )
                {
                    std::cerr << "PclDepthImageToCvMat: depthMD.size is zero...: " << imageMD.FullXRes() << "x" << imageMD.FullYRes() << std::endl;
                    throw new std::runtime_error("PclRgbImageToCvMat: imageMD.size is zero");
                }
                imgBGR.create( imageMD.FullYRes(), imageMD.FullXRes(), imageMD.BytesPerPixel() == 1 ? CV_8UC1 : CV_8UC3  );
            }

            // copy
            const XnUInt8* pImageRow = imageMD.Data();
            //            std::cout << "sizeof(*pImageRow): " << sizeof(*pImageRow) << std::endl;
            //            std::cout << "imgBGR.step: " << imgBGR.step << std::endl;
            //            std::cout << "imgBGR.step1()" << imgBGR.step1() << std::endl;
            //            std::cout << "BytesPerPixel: " << imageMD.BytesPerPixel() << std::endl;
            //            std::cout << "imageMD.pixelFormat()" << imageMD.PixelFormat() << std::endl;
            for ( unsigned y = 0; y < imgBGR.rows; ++y  )
            {
                //std::cout << "y: " << y << std::endl;
                for ( unsigned x = 0; x < imgBGR.cols; ++x )
                {
                    for ( uchar c = 0; c < imageMD.BytesPerPixel(); ++c )
                    {
                        imgBGR.at<uchar>(y,x*imageMD.BytesPerPixel()+c) = pImageRow[y*imageMD.FullXRes()*imageMD.BytesPerPixel() + x*imageMD.BytesPerPixel() + c];

                    }
                }
            }

            //            for ( XnUInt y = 0; y < imageMD.YRes(); ++y, pImageRow += imageMD.XRes() )
            //            {
            //                memcpy( imgBGR.data + imgBGR.step*y, pImageRow, imageMD.XRes() * sizeof(*pImageRow) );
            //            }

            // OUT
            if ( pImgRGB )
            {
                pImgRGB->create( imgBGR.size(), imgBGR.type() );
                if ( imageMD.BytesPerPixel() > 1 ) cvtColor( imgBGR, *pImgRGB, CV_BGR2RGB );
                else                               *pImgRGB = imgBGR.clone();

            }

            return 0;
        }

        std::string timestamp2Str()
        {
            time_t rawtime;
            struct tm * timeinfo;
            char buffer [80];

            time ( &rawtime );
            timeinfo = localtime (&rawtime);

            strftime ( buffer, 80, "_%Y%m%d_%H%M", timeinfo );

            return std::string( buffer );
        }

        ::cv::Point3f hsv2rgb( ::cv::Point3f const& in )
        {
            double      hh, p, q, t, ff;
            long        i;
            ::cv::Point3f out;

            if ( in.y <= 0.f ) // < is bogus, just shuts up warnings
            {
                out.x = in.z;
                out.y = in.z;
                out.z = in.z;
                return out;
            }

            hh = in.x;
            if ( hh >= 360.f ) hh = 0.f;
            hh /= 60.f;
            i = (long)hh;
            ff = hh - i;
            p = in.z * (1.f - in.y);
            q = in.z * (1.f - (in.y * ff));
            t = in.z * (1.f - (in.y * (1.f - ff)));

            switch(i) {
                case 0:
                    out.x = in.z;
                    out.y = t;
                    out.z = p;
                    break;
                case 1:
                    out.x = q;
                    out.y = in.z;
                    out.z = p;
                    break;
                case 2:
                    out.x = p;
                    out.y = in.z;
                    out.z = t;
                    break;

                case 3:
                    out.x = p;
                    out.y = q;
                    out.z = in.z;
                    break;
                case 4:
                    out.x = t;
                    out.y = p;
                    out.z = in.z;
                    break;

                case 5:
                default:
                    out.x = in.z;
                    out.y = p;
                    out.z = q;
                    break;
            }

            return out;
        }

        // generates n different colours
        // assumes hue [0, 360), saturation [0, 100), lightness [0, 100)
        std::vector<::cv::Point3f>
        nColoursCv(int n, float scale, bool random_shuffle )
        {
            std::vector<::cv::Point3f> out;
            //srand(time(NULL));

            float step = 360. / n;
            for ( int i = 0; i < n; ++i )
            {
                ::cv::Point3f c;
                c.x = i * step; // hue
                c.y = (90.f + rand()/(float)RAND_MAX * 10.f) / 100.f; // saturation
                c.z = (50.f + rand()/(float)RAND_MAX * 50.f) / 100.f; // value

                // convert and store
                out.push_back( hsv2rgb(c) * scale );
            }

            if ( random_shuffle )
                std::random_shuffle( out.begin(), out.end() );

            return out;
        }

        template <class PointT> inline typename PointT::value_type&
        _at( PointT &pnt, int i )
        {
            switch ( i )
            {
                case 0:
                    return pnt.x;
                    break;
                case 1:
                    return pnt.y;
                    break;
                case 2:
                    return pnt.z;
                    break;
                default:
                    std::cerr << "[" << __func__ << "]: " << "overindexed..." << std::endl;
                    return pnt.x;
                    break;
            }

            return pnt.x;
        }

        // generates n different colours
        // assumes hue [0, 360), saturation [0, 100), lightness [0, 100)
        // example call: am::util::nColoursCv2( 32, 1.f, .3f, .8 );
        std::vector<::cv::Point3f>
        nColoursCv2( int n, float scale, float value_random_part, double start, float brightness_limit )
        {
            std::vector<::cv::Point3f> out( n );
            srand( 123456 );

            const double golden_ratio_conjugate = 1. - 0.618033988749895;
            double hue = start;

            ::cv::Point3f c;
            for ( int i = 0; i < n; ++i )
            {
                hue += golden_ratio_conjugate;
                hue  = hue - static_cast<double>( static_cast<int>(hue) );

                c.x  = hue * 360.f;                                       // hue
                c.y  = .85f + RANDF() * .15f;                               // saturation
                c.z  = 1.f - value_random_part + hue * value_random_part; // value

                // store
                out[i] = am::util::hsv2rgb( c ) * scale;

                // no white colours
                float   dist    = 0.f, tmp;
                int     ch_id   = 0;
                for ( int j = 0; j != 3; ++j )
                    if ( (tmp=1.f-_at(out[i],j)) > dist )
                        dist = tmp;
                if ( dist < 1. - brightness_limit )
                    _at(out[i],ch_id) = brightness_limit;

                //if ( (out[i].x > 0.95f) )
            }

            return out;
        }

        std::vector<::Eigen::Vector3f>
        nColoursEigen( int n, float scale, bool random_shuffle )
        {
            std::vector<::cv::Point3f>     colours = nColoursCv( n, scale, random_shuffle );
            std::vector<::Eigen::Vector3f> colours_eigen;
            for ( auto const& colour : colours )
            {
                colours_eigen.push_back( (Eigen::Vector3f() << colour.x, colour.y, colour.z).finished() );
            }

            return colours_eigen;
        }

        std::string
        toHexString( int number )
        {
            char buffer[255];
            std::sprintf( buffer, "%02X", number );

            return buffer;
        }
    } // end ns util

    namespace geom
    {
        Eigen::Vector3f
        lineToPlaneIntersection( Eigen::VectorXf line, Eigen::Vector4f plane )
        {
            // line_coeffs.head<3>() = plane.getCentroid();
            // line_coeffs.segment<3>(3) = plane.Normal();

            // t = -(l0 . N + d) / (lv . N );
            double t = -1.f * (line.head<3>().dot(plane.head<3>()) + plane(3)) / line.segment<3>(3).dot( plane.head<3>());
            // l0 + t * lv;
            return line.head<3>() + t * line.segment<3>(3);
        }
    }
} // end ns am
